Type: Package
Package: accessibility
Title: Transport Accessibility Measures
Version: 1.4.0
Authors@R: c(
    person("Rafael H. M.", "Pereira", , "rafa.pereira.br@gmail.com", role = "aut",
           comment = c(ORCID = "0000-0003-2125-7465")),
    person("Daniel", "Herszenhut", , "dhersz@gmail.com", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0001-8066-1105")),
    person("Joey", "Reid", role = "ctb"),
    person("Ipea - Institute for Applied Economic Research", role = c("cph", "fnd"))
  )
Description: A set of fast and convenient functions to help conducting
    accessibility analyses. Given a pre-computed travel cost matrix and a
    land use dataset (containing the location of jobs, healthcare and
    population, for example), the package allows one to calculate
    accessibility levels and accessibility poverty and inequality. The
    package covers the majority of the most commonly used accessibility
    measures (such as cumulative opportunities, gravity-based and floating
    catchment areas methods), as well as the most frequently used
    inequality and poverty metrics (such as the Palma ratio, the
    concentration and Theil indices and the FGT family of measures).
License: MIT + file LICENSE
URL: https://github.com/ipeaGIT/accessibility,
        https://ipeagit.github.io/accessibility/
BugReports: https://github.com/ipeaGIT/accessibility/issues
Depends: R (>= 3.5.0)
Imports: checkmate, data.table, Hmisc, Rdpack (>= 0.7), stats, utils
Suggests: covr, ggplot2, knitr, rmarkdown, sf, testthat, tibble
VignetteBuilder: knitr
RdMacros: Rdpack
Encoding: UTF-8
NeedsCompilation: no
RoxygenNote: 7.2.3
Packaged: 2024-03-06 20:44:17 UTC; dhersz
Author: Rafael H. M. Pereira [aut] (<https://orcid.org/0000-0003-2125-7465>),
  Daniel Herszenhut [aut, cre] (<https://orcid.org/0000-0001-8066-1105>),
  Joey Reid [ctb],
  Ipea - Institute for Applied Economic Research [cph, fnd]
Maintainer: Daniel Herszenhut <dhersz@gmail.com>
Repository: CRAN
Date/Publication: 2024-03-06 23:40:06 UTC
Built: R 4.2.3; ; 2024-03-07 01:47:31 UTC; unix

Type: Package
Package: accessibility
Title: Transport Accessibility Measures
Version: 1.5.0
Authors@R: c(
    person("Rafael H. M.", "Pereira", , "rafa.pereira.br@gmail.com", role = "aut",
           comment = c(ORCID = "0000-0003-2125-7465")),
    person("Daniel", "Herszenhut", , "dhersz@gmail.com", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0001-8066-1105")),
    person("Anastasia", "Soukhov", role = "aut"),
    person("Christopher", "Higgins", role = "ctb"),
    person("Joey", "Reid", role = "ctb"),
    person("Ipea - Institute for Applied Economic Research", role = c("cph", "fnd"))
  )
Description: A set of fast and convenient functions to help conducting 
    accessibility analyses. Given a pre-computed travel cost matrix and a land 
    use dataset (containing for example the location of jobs, healthcare and 
    population), the package allows one to calculate accessibility levels, and 
    accessibility poverty and inequality. The package covers the majority of the 
    most commonly used accessibility measures (such as cumulative opportunities, 
    gravity-based and floating catchment areas methods), some cutting edge 
    measures proposed in the literature (e.g. balancing cost and constrained 
    accessibility) as well as the most frequently used inequality and poverty 
    metrics (such as the Palma ratio, the concentration and Theil indices and 
    the FGT family of measures).
License: MIT + file LICENSE
URL: https://github.com/ipeaGIT/accessibility,
        https://ipeagit.github.io/accessibility/
BugReports: https://github.com/ipeaGIT/accessibility/issues
Depends: R (>= 3.5.0)
Imports: checkmate, data.table, Hmisc, Rdpack (>= 0.7), stats, utils
Suggests: covr, ggplot2, knitr, rmarkdown, sf, testthat, tibble
VignetteBuilder: knitr
RdMacros: Rdpack
Encoding: UTF-8
NeedsCompilation: no
RoxygenNote: 7.3.3
Packaged: 2026-05-12 00:49:29 UTC; rafap
Author: Rafael H. M. Pereira [aut] (ORCID:
    <https://orcid.org/0000-0003-2125-7465>),
  Daniel Herszenhut [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-8066-1105>),
  Anastasia Soukhov [aut],
  Christopher Higgins [ctb],
  Joey Reid [ctb],
  Ipea - Institute for Applied Economic Research [cph, fnd]
Maintainer: Daniel Herszenhut <dhersz@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-12 05:10:21 UTC
Built: R 4.6.0; ; 2026-05-12 07:08:21 UTC; unix

# Generated by roxygen2: do not edit by hand

export(balancing_cost)
export(concentration_index)
export(constrained_accessibility)
export(cost_to_closest)
export(cumulative_cutoff)
export(cumulative_interval)
export(decay_binary)
export(decay_exponential)
export(decay_linear)
export(decay_logistic)
export(decay_power)
export(decay_stepped)
export(fgt_poverty)
export(floating_catchment_area)
export(gini_index)
export(gravity)
export(palma_ratio)
export(spatial_availability)
export(theil_t)
importFrom(Rdpack,reprompt)
importFrom(data.table,"%chin%")
importFrom(data.table,":=")
importFrom(data.table,.GRP)
importFrom(data.table,.I)
importFrom(data.table,.N)
importFrom(data.table,.SD)
importFrom(utils,globalVariables)

# accessibility <img align="right" src="man/figures/logo.png" alt="logo" width="180">

[![CRAN
status](https://www.r-pkg.org/badges/version/accessibility)](https://CRAN.R-project.org/package=accessibility)
[![rcmdcheck](https://github.com/ipeaGIT/accessibility/workflows/rcmdcheck/badge.svg)](https://github.com/ipeaGIT/accessibility/actions)
[![CRAN/METACRAN Total
downloads](https://cranlogs.r-pkg.org/badges/grand-total/accessibility?color=yellow)](https://CRAN.R-project.org/package=accessibility)
[![Codecov test
coverage](https://codecov.io/gh/ipeaGIT/accessibility/branch/main/graph/badge.svg)](https://app.codecov.io/gh/ipeaGIT/accessibility?branch=main)
[![Lifecycle:
stable](https://lifecycle.r-lib.org/articles/figures/lifecycle-stable.svg)](https://lifecycle.r-lib.org/articles/stages.html)

**accessibility** offers a set of fast and convenient functions to help
conducting accessibility analyses. Given a pre-computed travel cost
matrix and a land use dataset (containing the location of jobs,
healthcare and population, for example), the package allows one to
calculate accessibility levels and accessibility poverty and inequality.
The package covers the majority of the most commonly used accessibility
measures (such as cumulative opportunities, gravity-based and floating
catchment areas methods), some cutting edge measures proposed in the
literature (e.g. balancing cost and constrained accessibility), as well
as the most frequently used inequality and poverty metrics (such as the
Palma ratio, the concentration and Theil indices and the FGT family of
measures).

## Installation

Stable version:

``` r
install.packages("accessibility")
```

Development version:

``` r
# install.packages("remotes")
remotes::install_github("ipeaGIT/accessibility")
```

## Usage

This section aims to present a very brief overview of some of the
packages’ features. Fore more details please read the vignettes:

- [Introduction to **accessibility**: calculating accessibility
  measures](https://ipeagit.github.io/accessibility/articles/accessibility.html)
- [Decay
  functions](https://ipeagit.github.io/accessibility/articles/decay_functions.html)
- [Calculating accessibility inequality and
  poverty](https://ipeagit.github.io/accessibility/articles/inequality_and_poverty.html)

To calculate accessibility levels, one simply needs a pre-calculated
travel matrix and some land use data. Below we showcase some of the
available functions:

``` r
library(accessibility)

data_dir <- system.file("extdata", package = "accessibility")
travel_matrix <- readRDS(file.path(data_dir, "travel_matrix.rds"))
land_use_data <- readRDS(file.path(data_dir, "land_use_data.rds"))

cum_cutoff <- cumulative_cutoff(
  travel_matrix,
  land_use_data,
  opportunity = "jobs",
  travel_cost = "travel_time",
  cutoff = 30
)
head(cum_cutoff)
#>                 id  jobs
#>             <char> <int>
#> 1: 89a881a5a2bffff 14561
#> 2: 89a881a5a2fffff 29452
#> 3: 89a881a5a67ffff 16647
#> 4: 89a881a5a6bffff 10700
#> 5: 89a881a5a6fffff  6669
#> 6: 89a881a5b03ffff 37029

grav <- gravity(
  travel_matrix,
  land_use_data,
  opportunity = "schools",
  travel_cost = "travel_time",
  decay_function = decay_exponential(decay_value = 0.2)
)
head(grav)
#>                 id    schools
#>             <char>      <num>
#> 1: 89a88cdb57bffff 0.03041853
#> 2: 89a88cdb597ffff 1.15549493
#> 3: 89a88cdb5b3ffff 0.56519126
#> 4: 89a88cdb5cfffff 0.19852152
#> 5: 89a88cd909bffff 0.41378042
#> 6: 89a88cd90b7ffff 0.95737555

fca <- floating_catchment_area(
  travel_matrix,
  land_use_data,
  opportunity = "jobs",
  travel_cost = "travel_time",
  demand = "population",
  method = "2sfca",
  decay_function = decay_binary(cutoff = 50)
)
head(fca)
#>                 id      jobs
#>             <char>     <num>
#> 1: 89a88cdb57bffff 0.4278111
#> 2: 89a88cdb597ffff 0.3863614
#> 3: 89a88cdb5b3ffff 0.4501725
#> 4: 89a88cdb5cfffff 0.5366707
#> 5: 89a88cd909bffff 0.4280401
#> 6: 89a88cd90b7ffff 0.5176583
```

Calculating inequality and poverty levels is equally easy. Below we use
the previously calculated cumulative accessibility dataset to show some
of the available inequality and poverty functions:

``` r
palma <- palma_ratio(
  cum_cutoff,
  sociodemographic_data = land_use_data,
  opportunity = "jobs",
  population = "population",
  income = "income_per_capita"
)
palma
#>    palma_ratio
#>          <num>
#> 1:    3.800465

poverty <- fgt_poverty(
  cum_cutoff,
  sociodemographic_data = land_use_data,
  opportunity = "jobs",
  population = "population",
  poverty_line = 95368
)
poverty
#>         FGT0      FGT1      FGT2
#>        <num>     <num>     <num>
#> 1: 0.5745378 0.3277383 0.2218769
```

## Related work:

- [r5r](https://github.com/ipeaGIT/r5r): Rapid Realistic Routing with R5
  in R
- [tracc](https://github.com/jamaps/tracc): Transport accessibility
  measures in Python
- [access](https://access.readthedocs.io/en/latest/): Spatial Access for
  PySAL
- [aceso](https://github.com/tetraptych/aceso): a lightweight Python
  package for measuring spatial accessibility

## Acknowledgement <a href="https://www.ipea.gov.br"><img src="man/figures/ipea_logo.png" alt="IPEA" align="right" width="300"/></a>

**accessibility** is developed by a team at the Institute for Applied
Economic Research (Ipea), Brazil.

## ----include = FALSE----------------------------------------------------------
Sys.setenv(OMP_THREAD_LIMIT = 2)

knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## -----------------------------------------------------------------------------
library(accessibility)

output_fn <- decay_exponential(c(0.2, 0.3))

output_fn(c(10, 15, 20))

## -----------------------------------------------------------------------------
stepped_output <- decay_stepped(
  steps = list(c(10, 20, 30), c(10, 20, 30, 40)),
  weights = list(c(0.67, 0.33, 0), c(0.75, 0.5, 0.25, 0))
)

stepped_output(c(15, 25, 35, 45))

## ----eval = requireNamespace("ggplot2", quietly = TRUE), out.width = "80%", fig.width = 6, fig.height = 6----
library(data.table)
library(ggplot2)

binary <- decay_binary(cutoff = 50)
linear <- decay_linear(cutoff = 50)
negative_exp <- decay_exponential(decay_value = 0.2)
inverse_power <- decay_power(decay_value = 0.2)
stepped <- decay_stepped(steps = c(30, 60, 90), weights = c(0.67, 0.33, 0))
logistic <- decay_logistic(cutoff = 50, sd = 10)

travel_costs <- seq(1, 100, 0.1)

weights <- data.table(
  minutes = travel_costs,
  binary = as.numeric(binary(travel_costs)[["50"]]),
  linear = linear(travel_costs)[["50"]],
  negative_exp = negative_exp(travel_costs)[["0.2"]],
  inverse_power = inverse_power(travel_costs)[["0.2"]],
  stepped = stepped(travel_costs)[["s(30,60,90);w(0.67,0.33,0)"]],
  logistic = logistic(travel_costs)[["c50;sd10"]]
)

# reshape data to long format
weights <- melt(
  weights,
  id.vars = "minutes",
  variable.name = "decay_function",
  value.name = "weights"
)

ggplot(weights) +
  geom_line(
    aes(minutes, weights, color = decay_function),
    show.legend = FALSE
  ) +
  facet_wrap(. ~ decay_function, ncol = 2) +
  theme_minimal()

## -----------------------------------------------------------------------------
my_decay <- function(travel_cost) {
  weights <- 1 / travel_cost
  weights[weights > 1] <- 1
  return(weights)
}

## -----------------------------------------------------------------------------
my_decay(c(0, 0.5, 1, 2, 5, 10))

## -----------------------------------------------------------------------------
data_dir <- system.file("extdata", package = "accessibility")

travel_matrix <- readRDS(file.path(data_dir, "travel_matrix.rds"))
land_use_data <- readRDS(file.path(data_dir, "land_use_data.rds"))

custom_gravity <- gravity(
  travel_matrix,
  land_use_data,
  opportunity = "jobs",
  travel_cost = "travel_time",
  decay_function = my_decay
)
head(custom_gravity)

## -----------------------------------------------------------------------------
my_second_decay <- function(decay_parameter) {
  function(travel_cost) {
    weights <- 1 / (decay_parameter * travel_cost)
    weights[weights > 1] <- 1
    return(weights)
  }
}

output_fn <- my_second_decay(2)
output_fn(c(0, 0.5, 1, 2, 5, 10))

# compare to the first custom decay function
my_decay(c(0, 0.5, 1, 2, 5, 10))

## -----------------------------------------------------------------------------
second_custom_gravity <- gravity(
  travel_matrix,
  land_use_data,
  opportunity = "jobs",
  travel_cost = "travel_time",
  decay_function = my_second_decay(1)
)
head(second_custom_gravity)

## -----------------------------------------------------------------------------
decay_power(1)

## -----------------------------------------------------------------------------
my_third_decay <- function(decay_parameter) {
  function(travel_cost) {
    weighting_list <- lapply(
      decay_parameter,
      function(x) {
        weights <- 1 / (x * travel_cost)
        weights[weights > 1] <- 1
        return(weights)
      }
    )

    names(weighting_list) <- decay_parameter
    weighting_list
  }
}

output_fn <- my_third_decay(c(1, 2))
output_fn(c(0, 0.5, 1, 2, 5, 10))

# compare to the first and second custom decay functions

my_decay(c(0, 0.5, 1, 2, 5, 10))

my_second_decay(2)(c(0, 0.5, 1, 2, 5, 10))

## -----------------------------------------------------------------------------
third_custom_gravity <- gravity(
  travel_matrix,
  land_use_data,
  opportunity = "jobs",
  travel_cost = "travel_time",
  decay_function = my_third_decay(c(1, 2))
)
third_custom_gravity

## ----include = FALSE----------------------------------------------------------
Sys.setenv(OMP_THREAD_LIMIT = 2)

knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----eval = FALSE-------------------------------------------------------------
# install.packages("accessibility")

## ----eval = FALSE-------------------------------------------------------------
# # install.packages("remotes")
# remotes::install_github("ipeaGIT/accessibility")

## ----message = FALSE, warning = FALSE-----------------------------------------
library(accessibility)
library(data.table)
library(ggplot2)
library(sf)

## -----------------------------------------------------------------------------
data_dir <- system.file("extdata", package = "accessibility")

travel_matrix <- readRDS(file.path(data_dir, "travel_matrix.rds"))
head(travel_matrix)

## -----------------------------------------------------------------------------
land_use_data <- readRDS(file.path(data_dir, "land_use_data.rds"))
head(land_use_data)

## ----message = FALSE----------------------------------------------------------
mtc <- cost_to_closest(
  travel_matrix,
  land_use_data,
  opportunity = "schools",
  travel_cost = "travel_time",
  n = 1
)
head(mtc)

## ----message = FALSE----------------------------------------------------------
cum_cutoff <- cumulative_cutoff(
  travel_matrix,
  land_use_data,
  opportunity = "jobs",
  travel_cost = "travel_time",
  cutoff = 30
)
head(cum_cutoff)

## ----message = FALSE----------------------------------------------------------
passive_cum_cutoff <- cumulative_cutoff(
  travel_matrix,
  land_use_data,
  opportunity = "population",
  travel_cost = "travel_time",
  cutoff = 30,
  active = FALSE
)
head(passive_cum_cutoff)

## ----message = FALSE----------------------------------------------------------
cum_interval <- cumulative_interval(
  travel_matrix = travel_matrix,
  land_use_data = land_use_data,
  opportunity = "jobs",
  travel_cost = "travel_time",
  interval = c(40, 60),
  summary_function = base::mean
)
head(cum_interval)

## ----message = FALSE----------------------------------------------------------
negative_exp <- gravity(
  travel_matrix,
  land_use_data,
  opportunity = "schools",
  travel_cost = "travel_time",
  decay_function = decay_exponential(decay_value = 0.2)
)
head(negative_exp)

## ----message = FALSE----------------------------------------------------------
bfca <- floating_catchment_area(
  travel_matrix,
  land_use_data,
  opportunity = "jobs",
  travel_cost = "travel_time",
  demand = "population",
  method = "bfca",
  decay_function = decay_exponential(decay_value = 0.5)
)
head(bfca)

## -----------------------------------------------------------------------------
spatial_avlblt <- spatial_availability(
  travel_matrix,
  land_use_data,
  opportunity = "jobs",
  travel_cost = "travel_time",
  demand = "population",
  decay_function = decay_exponential(decay_value = 0.1)
)
head(spatial_avlblt)

## -----------------------------------------------------------------------------
bal_cost <- balancing_cost(
  travel_matrix,
  land_use_data,
  opportunity = "jobs",
  travel_cost = "travel_time",
  demand = "population"
)
head(bal_cost)

## ----eval = requireNamespace(c("sf", "ggplot2"), quietly = TRUE), out.width = "80%", fig.width = 6, fig.height = 6----
grid <- system.file("extdata/grid_bho.rds", package = "accessibility")
grid <- readRDS(grid)

spatial_data <- merge(grid, cum_cutoff, by = "id")

ggplot() +
  geom_sf(data = spatial_data, aes(fill = jobs), color = NA) +
  labs(
    title = "Job accessibility by transit in under 30 min.",
    fill = "Accessible jobs"
  ) +
  scale_fill_viridis_c() +
  theme_void()

## ----include = FALSE----------------------------------------------------------
Sys.setenv(OMP_THREAD_LIMIT = 2)

knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## -----------------------------------------------------------------------------
library(accessibility)

data_dir <- system.file("extdata", package = "accessibility")
travel_matrix <- readRDS(file.path(data_dir, "travel_matrix.rds"))
land_use_data <- readRDS(file.path(data_dir, "land_use_data.rds"))

access <- cumulative_cutoff(
  travel_matrix,
  land_use_data,
  opportunity = "jobs",
  travel_cost = "travel_time",
  cutoff = 30
)
head(access)

## -----------------------------------------------------------------------------
palma <- palma_ratio(
  access,
  sociodemographic_data = land_use_data,
  opportunity = "jobs",
  population = "population",
  income = "income_per_capita"
)
palma

## -----------------------------------------------------------------------------
gini <- gini_index(
  access,
  sociodemographic_data = land_use_data,
  opportunity = "jobs",
  population = "population"
)
gini

## -----------------------------------------------------------------------------
ci <- concentration_index(
  access,
  sociodemographic_data = land_use_data,
  opportunity = "jobs",
  population = "population",
  income = "income_per_capita",
  type = "corrected"
)
ci

## -----------------------------------------------------------------------------
theil_without_groups <- theil_t(
  access,
  sociodemographic_data = land_use_data,
  opportunity = "jobs",
  population = "population"
)
theil_without_groups

# some cells are classified as in the decile NA because their income per capita
# is NaN, as they don't have any population. we filter these cells from our
# accessibility data, otherwise the output would include NA values (note that
# subsetting the data like this doesn't affect the assumption that groups are
# completely exhaustive, because cells with NA income decile don't have any
# population)

na_decile_ids <- land_use_data[is.na(land_use_data$income_decile), ]$id
no_na_access <- access[! access$id %in% na_decile_ids, ]
sociodem_data <- land_use_data[! land_use_data$id %in% na_decile_ids, ]

theil_with_groups <- theil_t(
  no_na_access,
  sociodemographic_data = sociodem_data,
  opportunity = "jobs",
  population = "population",
  socioeconomic_groups = "income_decile"
)
theil_with_groups

## -----------------------------------------------------------------------------
poverty <- fgt_poverty(
  access,
  sociodemographic_data = land_use_data,
  opportunity = "jobs",
  population = "population",
  poverty_line = 50000
)

poverty

#' @keywords internal
merge_by_reference <- function(data,
                               land_use_data,
                               opportunity,
                               left_df_idcol) {
  # when calculating active accessibility, we want the number of opportunities
  # in the destination. if passive, the number of opportunities in the origin

  right_df_idcol <- "id"

  join_vector <- right_df_idcol
  names(join_vector) <- left_df_idcol

  data[
    land_use_data,
    on = join_vector,
    eval(opportunity) := get(paste0("i.", ..opportunity))
  ]
}

#' accessibility: Transport accessibility measures
#'
#' @section Usage:
#' Please check the vignettes for more on the package usage:
#' - Introduction to accessibility: calculating accessibility measures. Run
#' `vignette("accessibility")` or check it on the [website](
#' https://ipeagit.github.io/accessibility/articles/accessibility.html).
#' - Decay functions. Run `vignette("decay_functions", package =
#' "accessibility")` or check it on the [website](
#' https://ipeagit.github.io/accessibility/articles/decay_functions.html).
#' - Calculating accessibility inequality and poverty. Run
#' `vignette("inequality_and_poverty", package = "accessibility")` or check it
#' on the
#' [website](https://ipeagit.github.io/accessibility/articles/inequality_and_poverty.html).
#'
#' @docType package
#' @name accessibility
#' @aliases accessibility-package
#'
#' @importFrom data.table := .I .SD %chin% .GRP .N
#' @importFrom utils globalVariables
#' @importFrom Rdpack reprompt
#'
#' @keywords internal
"_PACKAGE"

utils::globalVariables(
  c(
    ".",
    "from_id",
    "schools",
    "df",
    "travel_time",
    "to_id",
    "cum_opport",
    "impedance",
    "pop_served",
    "ppr",
    "access",
    "min_cost",
    "destination",
    "opp_weight",
    "balanced_opp_weight_i",
    "balanced_opp_weight_j",
    "balanced_pop_served",
    "balanced_ppr",
    "..opportunity",
    "demand_bal_fac",
    "impedance_bal_fac",
    "combined_bal_fac",
    "decay_function_arg",
    "cost",
    "did_balance",
    "avg_access",
    "i.avg_access",
    ".palma_group",
    "poorest_avg_access",
    "p",
    "FGT0",
    "FGT1",
    "FGT2",
    ".area_under_curve",
    ".big_edge",
    ".cum_total_access",
    ".prop_pop",
    ".small_edge",
    ".fgt0",
    ".fgt1",
    ".fgt2",
    ".norm_opp_shortfall",
    "pop_share",
    "upper",
    "lower",
    "cont_to_total",
    "correction_factor",
    "diff_from_avg",
    "fractional_rank",
    "..population",
    "access_share",
    "component",
    "group_avg_access",
    "group_theil_t",
    "theil_share",
    "total_access",
    "group_total_access",
    "contribution",
    "value",
    "within_ineq",
    "between_share",
    "group_tot_access",
    "share_of_component",
    "group_tot_pop",
    "total",
    "share_of_total"
  )
)

#' Total constrained accessibility
#'
#' Allocates total opportunities in the region proportionally based on travel
#' impedance. Uses the logic of a total (or unconstrained by Wilson's terms)
#' constraint. Returns values in units of 'supply' (i.e., opportunities) if
#' `active = TRUE` and returns values in units of demand' (i.e., population)
#' if `active = FALSE`.
#'
#' Internal helper used by [constrained_accessibility()] when `constraint = "total"`.
#'
#' @name total_constrained
#' @keywords internal
#' @return A `data.table`/`data.frame` with results (structure mirrors the wrapper).
#' @examples NULL
#'
#' @importFrom utils globalVariables
#'
utils::globalVariables(c(
  # common ids / columns
  "from_id", "to_id", "id", "opp_weight", "supply", "demand", "weighted_demand", "weighted_supply",
  # total-constrained
  "kappa_total", "hatkappa_total", "K_total", "hatK_total", "constrained_opportunity",
  # singly-constrained
  "denom_i", "denom_j", "kappa_singly", "hatkappa_singly", "A_i", "B_j", "singly_access",
  # doubly-constrained
  "kappa_doubly", "flow", "error"
))
total_constrained <- function(travel_matrix,
                              land_use_data,
                              travel_cost,
                              decay_function,
                              group_by = character(0),
                              fill_missing_ids = TRUE,
                              detailed_results = FALSE,
                              active,
                              demand = NULL,   # population
                              supply = NULL) { # jobs

  # Validate inputs
  checkmate::assert_string(travel_cost)
  checkmate::assert_logical(detailed_results, len = 1, any.missing = FALSE)
  checkmate::assert_logical(active, len = 1, any.missing = FALSE)

  if (active) {
    #active accessibility
    if (is.null(supply) || !is.null(demand)) {
      stop("For active = TRUE, supply must be specified and demand must be NULL.")
    }
    merge_id <- "to_id"
    group_id <- "from_id"
    weighted_col <- "weighted_supply"
    kappa_col <- "kappa_total"
    total_col <- "K_total"
    result_col <- "supply"
    opportunity_col <- supply
  } else {
    #passive accessibility
    if (is.null(demand) || !is.null(supply)) {
      stop("For active = FALSE, demand must be specified and supply must be NULL.")
    }
    merge_id <- "from_id"
    group_id <- "to_id"
    weighted_col <- "weighted_demand"
    kappa_col <- "hatkappa_total"
    total_col <- "hatK_total"
    result_col <- "demand"
    opportunity_col <- demand
  }

  assert_decay_function(decay_function)
  assert_group_by(group_by)
  assert_detailed_fill_missing_ids(fill_missing_ids, detailed_results)
  assert_travel_matrix(travel_matrix, travel_cost, group_by)
  assert_land_use_data(land_use_data, travel_matrix, opportunity_col)

  # Convert to data.table
  if (!inherits(travel_matrix, "data.table")) {
    original_class <- class(travel_matrix)
    data <- data.table::as.data.table(travel_matrix)
  } else {
    data <- data.table::copy(travel_matrix)
  }
  if (!inherits(land_use_data, "data.table")) {
    land_use_data <- data.table::as.data.table(land_use_data)
  }

  # Merge land use data
  merge_by_reference(data, land_use_data, opportunity_col, left_df_idcol = merge_id)

  # Apply decay
  data <- apply_gravity_measure(data, decay_function, travel_cost)

  # Core calculation
  data[, (weighted_col) := get(opportunity_col) * opp_weight]
  total_weighted_system <- data[, sum(get(weighted_col))]
  data[, (kappa_col) := get(weighted_col) / total_weighted_system]

  ids_for_total <- unique(data[[merge_id]])                   # e.g., A,B,C when active=TRUE
  total_opportunities_region <- land_use_data[id %in% ids_for_total, sum(get(opportunity_col), na.rm = TRUE)]

  data[, constrained_opportunity := get(kappa_col) * total_opportunities_region]

  if (detailed_results) {
    if (!active) {
      access <- data[, .(
        from_id,
        to_id,
        weighted_demand,
        hatkappa_total,
        demand = constrained_opportunity,
        hatK_total = total_opportunities_region / total_weighted_system
      )]
    } else {
      # supply-side details (active accessibility)
      access <- data[, .(
        from_id,
        to_id,
        weighted_supply,
        kappa_total,
        supply = constrained_opportunity,
        K_total = total_opportunities_region / total_weighted_system
      )]
    }
  } else {
    if (active) {
      # supply aggregated by origin (from_id)
      access <- data[, .(supply = sum(constrained_opportunity)), by = c("from_id", group_by)]
      if (fill_missing_ids) access <- fill_missing_ids(access, travel_matrix, c("from_id", group_by))
    } else {
      # demand aggregated by destination (to_id)
      access <- data[, .(demand = sum(constrained_opportunity)), by = c("to_id", group_by)]
      if (fill_missing_ids) access <- fill_missing_ids(access, travel_matrix, c("to_id", group_by))
    }
  }

  if (exists("original_class")) class(access) <- original_class
  return(access[])
}

#' Constrained accessibility
#'
#' Calculates accessibility using constraints, as proposed in
#' \insertCite{soukhov2025family;textual}{accessibility}. Accessibility is
#' conceptualised as potential spatial interaction. This function covers three
#' constraint cases. Please see the Details section for more information.
#'
#' @template travel_matrix
#' @template land_use_data
#' @template travel_cost
#' @template demand
#' @template supply
#' @param constraint A string. One of `"total"`, `"singly"`, or `"doubly"`. See
#'        Details section for more information.
#' @template decay_function
#' @param active A logical. When `TRUE`, the function calculates active
#'        accessibility (the quantity of opportunities that can be reached from
#'        a given origin). When `FALSE`, it calculates passive accessibility (by
#'        how many people each destination can be reached), which is equivalent
#'        to the notion of market potential. This parameter only works for
#'        `constraint` types `"total"` and `"singly"`. Ignored for
#'        `constraint = "doubly"`.
#' @param error_threshold Numeric. Convergence criterion used only for
#' calibration in the doubly-constrained case (`constraint = "doubly"`).
#' @param improvement_threshold Numeric. Convergence criterion for improvement
#' used only for calibration in the doubly-constrained case
#' (`constraint = "doubly"`).
#' @param max_iterations Integer. Maximum iterations used only for calibration
#' in the doubly-constrained case (`constraint = "doubly"`).
#' @template group_by
#' @template fill_missing_ids_combinations
#' @param detailed_results Logical. Whether to return detailed OD-level results.
#'
#'
#' @section Details:
#' This function covers the family of constrained accessibility measures
#' proposed in \insertCite{soukhov2025family;textual}{accessibility}.
#'
#' ## Total Constrained Accessibility
#'
#' Allocates the system-wide total proportionally based on travel impedance
#' between origins and destinations. This measure uses the logic of a
#' total ~(or 'unconstrained' by Wilson's terms)~ constraint.
#'
#' Use this measure when the total quantity of **supply** OR **demand** in the
#' system is known and representing accessibility as a proportion of this total
#'  is meaningful.
#'
#' ### Requirement:
#' - Either `demand` or `supply` must be provided (cannot provide both).
#'
#' ### Interpretation:
#' - `active = TRUE` (*active accessibility*): Results represent the total
#' number of  **opportunities** (supply) accessible from each origin based on
#' region-relative travel impedance. The units are in 'supply' (e.g., jobs,
#' school seats).
#'   - If `detailed_results = FALSE`, outputs are aggregated and returned by
#'   origin.
#'   - If `detailed_results = TRUE`, OD-level flows are returned. Summing flows
#'   by origin equals the aggregated result.
#'
#' - `active = FALSE` (*passive accessibility*, the notion of market potential):
#' Results represent the total number of **population** (demand) that can reach
#' each destination based on region-relative travel impedance. The units are in
#' 'demand' (e.g., population).
#'   - If `detailed_results = FALSE`, outputs are aggregated by destination.
#'   - If `detailed_results = TRUE`, OD-level flows are returned. Summing flows
#'   by destination equals the aggregated result.
#'
#' ### Use cases:
#' - Active accessibility (aggregated):
#'   "How many jobs can be reached from origin zone A given its region-relative
#'   travel impedance?"
#'
#' - Active accessibility (flow-level):
#'   "How many jobs can be reached by flow A->1 given A->1's region-relative
#'   travel impedance?"
#'
#' - Passive accessibility (aggregated):
#'   "How many people can reach destination zone 1 given its region-relative
#'   travel impedance?"
#'
#' - Passive accessibility (flow-level):
#'   "How many people are reached by flow 1->A given 1->A's region-relative
#'   travel impedance?"
#'
#'
#' ## Singly Constrained Accessibility
#'
#' Allocates opportunities at each destination (or population at each origin)
#' proportionally based on travel impedance and the opposite marginal. This
#'  measure uses the logic of single constraint from
#'  \insertCite{wilson1971family;textual}{accessibility}.
#'
#' Use this measure when modeling **competition**, where both demand and supply
#' are conceptualised to influence accessibility but only one side is fixed.
#' The measure distributes flows so that totals match the constrained side
#' while weighting by travel impedance and the unconstrained side.
#'
#' ### Requirements:
#' - Both `demand` and `supply` must be provided (the logical for `active`
#' determines if either demand or supply is constrained).
#'
#' ### Interpretation:
#' - `active = TRUE` (*active accessibility*): constrains supply. Results
#' represent the total number of **opportunities** (supply) accessible from each
#'  origin based on region-relative travel impedance and population at the origin.
#'    The units are in 'supply' (e.g., jobs, school seats).
#'   - If `detailed_results = FALSE`, outputs are aggregated and returned by origin.
#'   - If `detailed_results = TRUE`, OD-level flows are returned. Summing flows
#'    by origin equals the aggregated result.
#'
#' - `active = FALSE` (*passive accessibility*, the notion of market potential):
#' constrains demand. Results represent the total number of
#' **population** (demand) that can reach each destination based on
#' region-relative travel impedance and opportunities at the destination. The
#' units are in 'demand' (e.g., population).
#'   - If `detailed_results = FALSE`, outputs are aggregated by destination.
#'   - If `detailed_results = TRUE`, OD-level flows are returned. Summing flows
#'   by destination equals the aggregated result.
#'
#' ### Use cases:
#' - Active accessibility (aggregated):
#'   "How many jobs can be reached from origin zone A given its region-relative
#'   travel impedance and demand?"
#'
#' - Active accessibility (flow-level):
#'   "How many jobs can be reached by flow A->1 given A->1's region-relative
#'   travel impedance and demand?"
#'
#' - Passive accessibility (aggregated):
#'   "How many people can reach destination zone 1 given its region-relative
#'   travel impedance and supply?"
#'
#' - Passive accessibility (flow-level):
#'   "How many people are reached by flow 1->A given 1->A's region-relative
#'   travel impedance and supply?"
#'
#' **NOTE:** the active form of this measure yields equivalent results to the
#'  `spatial_availability()` function, through different logic.
#'
#' ## Doubly Constrained Accessibility
#'
#' Allocates flows so supply at each destination matches demand at each origin.
#' It uses Wilson's *doubly-constrained* gravity model \insertCite{wilson1971family;textual}{accessibility}.
#'
#' The model uses iterative proportional fitting to update balancing factors in
#' order to calibrate OD flows on both margins (`A_i` for origins and `B_j` for
#' destinations) until convergence (i.e. the sum of `demand` and `supply` match).
#' This guarantees that flows satisfy both marginals while being weighted by
#' travel  impedance.
#'
#' ### Requirements:
#' - Both `demand` and `supply` must be provided.
#' - Unlike `total` and `singly`, `doubly` requires the sum of demand and supply
#' to match; otherwise, the model will not converge.
#'  - `active` must be `NULL`. Since supply must match demand, their units are
#'  the same and there is no distinction between 'active' and 'passive' notions.
#' - Only accepts`detailed_results = TRUE`.
#'
#' ### Interpretation:
#' - Results include OD-level flows (`flow`) along with balancing factors (`A_i`,
#' `B_j`) and travel impedance weights. The resulting flows represent the
#' distribution of demand and supply across all origin-destination pairs. NOTE:
#' OD flows are in flow units (jointly determined by demand and supply).
#'
#' ### Use cases:
#' - flow-level:
#' "What is the count of A->1 flows given A->1's region-relative travel impedance,
#'  demand and supply?"
#'
#' @references
#' \insertAllCited{}

#' @family Constrained accessibility
#'
#' @examples
#' # Load demo data shipped with the package
#' data_dir <- system.file("extdata", package = "accessibility")
#' travel_matrix <- readRDS(file.path(data_dir, "travel_matrix.rds"))
#' land_use_data <- readRDS(file.path(data_dir, "land_use_data.rds"))
#'
#' # Total-constrained (active accessibility, aggregated): returns units of
#' # accessible supply by origin (requires supply)
#' constrained_accessibility(
#'   travel_matrix   = travel_matrix,
#'   land_use_data   = land_use_data,
#'   travel_cost     = "travel_time",
#'   constraint      = "total",
#'   decay_function  = decay_exponential(0.1),
#'   demand          = NULL,
#'   supply          = "jobs",
#'   active          = TRUE,
#'   detailed_results = FALSE
#' )
#'
#' # Total-constrained (passive accessibility, aggregated): returns units of
#' # accessible demand by destination  (requires demand)
#' constrained_accessibility(
#'   travel_matrix  = travel_matrix,
#'   land_use_data  = land_use_data,
#'   travel_cost    = "travel_time",
#'   constraint     = "total",
#'   decay_function = decay_exponential(0.1),
#'   demand         = "population",
#'   supply         = NULL,
#'   active         = FALSE,
#'   detailed_results = FALSE
#' )
#'
#' # Singly-constrained (active accessibility, aggregated): returns units of
#' # accessible supply by origin (requires supply and demand)
#' constrained_accessibility(
#'   travel_matrix   = travel_matrix,
#'   land_use_data   = land_use_data,
#'   travel_cost     = "travel_time",
#'   constraint      = "singly",
#'   decay_function  = decay_exponential(0.1),
#'   demand          = "population",
#'   supply          = "jobs",
#'   active          = TRUE,
#'   detailed_results = FALSE
#' )
#'
#' # Doubly-constrained: returns units of flow (requires both demand and supply
#' # (totals that match) and `detailed_results = TRUE`)
#'
#' # Using a small toy dataset with matching totals.
#' tm_small <- data.table::data.table(
#'   expand.grid(from_id = c("1","2","3"), to_id = c("1","2","3"))
#' )
#' tm_small[, travel_time := c(10, 30, 15, 30, 10, 25, 15, 25, 10)]
#' lu_small <- data.table::data.table(
#'   id         = c("1","2","3"),
#'   population = c(4, 10, 6),   # sum = 20
#'   jobs       = c(7,  5,  8)   # sum = 20
#' )
#'
#' constrained_accessibility(
#'   travel_matrix   = tm_small,
#'   land_use_data   = lu_small,
#'   travel_cost     = "travel_time",
#'   constraint      = "doubly",
#'   decay_function  = decay_exponential(0.1),
#'   demand          = "population",
#'   supply          = "jobs",
#'   detailed_results = TRUE
#' )
#'
#' @export
constrained_accessibility <- function(travel_matrix,
                                      land_use_data,
                                      travel_cost,
                                      demand = NULL,
                                      supply = NULL,
                                      constraint,
                                      decay_function,
                                      active = NULL,
                                      error_threshold = 0.001,
                                      improvement_threshold = 1e-6,
                                      max_iterations = 1000,
                                      group_by = character(0),
                                      fill_missing_ids = TRUE,
                                      detailed_results = FALSE) {

  checkmate::assert_choice(constraint, c("total", "singly", "doubly"))
  checkmate::assert_string(travel_cost)
  assert_decay_function(decay_function)
  assert_group_by(group_by)
  assert_travel_matrix(travel_matrix, travel_cost, group_by)
  assert_land_use_data(land_use_data, travel_matrix, opportunity = supply, demand = demand)
  checkmate::assert_logical(detailed_results, len = 1, any.missing = FALSE)
  assert_detailed_fill_missing_ids(fill_missing_ids, detailed_results)


  if (constraint == "doubly") {
    return(doubly_constrained(
      travel_matrix = travel_matrix,
      land_use_data = land_use_data,
      travel_cost = travel_cost,
      decay_function = decay_function,
      demand = demand,
      supply = supply,
      error_threshold = error_threshold,
      improvement_threshold = improvement_threshold,
      max_iterations = max_iterations,
      group_by = group_by,
      fill_missing_ids = fill_missing_ids,
      detailed_results = detailed_results
    ))
  }
  # For 'total' and 'singly' we require active to be TRUE or FALSE
  if (is.null(active)) {
    stop(sprintf("For '%s', active must be TRUE or FALSE.", constraint))
  }

  # Validate the relevant side (opportunity column) only
  # 'active = TRUE' -> supply-side is returned; 'active = FALSE' -> demand-side is returned
  opportunity_col <- if (active) supply else demand
  assert_land_use_data(land_use_data, travel_matrix, opportunity = opportunity_col)


  if (constraint == "singly") {
    return(singly_constrained(
      travel_matrix = travel_matrix,
      land_use_data = land_use_data,
      travel_cost = travel_cost,
      decay_function = decay_function,
      demand = demand,
      supply = supply,
      active = active,
      group_by = group_by,
      fill_missing_ids = fill_missing_ids,
      detailed_results = detailed_results
    ))
  }

  if (constraint == "total") {
    return(total_constrained(
      travel_matrix = travel_matrix,
      land_use_data = land_use_data,
      travel_cost = travel_cost,
      decay_function = decay_function,
      group_by = group_by,
      fill_missing_ids = fill_missing_ids,
      detailed_results = detailed_results,
      active = active,
      demand = if (!active) demand else NULL,
      supply = if (active) supply else NULL
    ))
  }


}

# testthat::local_edition(3)

tm <- data.table::data.table(
  expand.grid(from_id = c("1","2"), to_id = c("1","2"))
)
tm[, travel_time := c(10, 20, 15, 10)]
lu <- data.table::data.table(
  id         = c("1","2"),
  population = c(5, 7),
  jobs       = c(100, 120)
)
decay <- decay_exponential(decay_value = 0.1)


tester <- function(
    constraint = "total",
    travel_matrix = tm,
    land_use_data = lu,
    travel_cost = "travel_time",
    decay_function = decay,
    demand = "population",
    supply = "jobs",
    active = FALSE,
    error_threshold = 0.001,
    improvement_threshold = 1e-6,
    max_iterations = 1000,
    group_by = character(0),
    fill_missing_ids = TRUE,
    detailed_results = FALSE
) {
  constrained_accessibility(
    constraint,
    travel_matrix,
    land_use_data,
    travel_cost,
    decay_function,
    demand,
    supply,
    return_demand_side,
    error_threshold,
    improvement_threshold,
    max_iterations,
    group_by,
    fill_missing_ids,
    detailed_results
  )
}

test_that("constrained_accessibility: wrapper argument rules are enforced", {

  # invalid constraint -> any error is fine


  # total/singly require TRUE or FALSE (not NULL)
  expect_error(
    tester(constraint = "total", return_demand_side = NULL)
  )

  expect_error(
    tester(constraint = "singly", return_demand_side = NULL)
  )

  # doubly requires NULL (error if TRUE)
  expect_error(
    tester(constraint = "doubly", active = FALSE)
  )
})





test_that("raises errors due to incorrect input", {

  expect_error(tester(constraint = "banana"))

  expect_error(tester(decay_function = "a"))
  expect_error(tester(decay_function = mean))
  expect_error(tester(decay_function = get))

  expect_error(tester(opportunity = 1))
  expect_error(tester(opportunity = c("schools", "jobs")))

  expect_error(tester(travel_cost = 1))
  expect_error(tester(travel_cost = c("travel_time", "monetary_cost")))

  expect_error(tester(demand = 1))
  expect_error(tester(demand = c("population", "population")))

  expect_error(tester(supply = 1))
 # expect_error(tester(supply = c("population", "population"))) 666666666

  expect_error(tester(group_by = 1))
  expect_error(tester(group_by = NA))
  expect_error(tester(group_by = "from_id"))
  expect_error(tester(group_by = c("mode", "mode")))

  expect_error(tester(fill_missing_ids = 1))
  expect_error(tester(fill_missing_ids = c(TRUE, TRUE)))
  expect_error(tester(fill_missing_ids = NA))

  expect_error(tester(as.list(tm)))
  expect_error(tester(tm[, .(oi = from_id, to_id, travel_time)]))
  expect_error(tester(tm[, .(from_id, oi = to_id, travel_time)]))
  expect_error(
    tester(
      tm[, .(from_id, to_id, oi = travel_time)],
      travel_cost = "travel_time"
    )
  )
  expect_error(
    tester(
      tm[, .(from_id, to_id, travel_time, oi = mode)],
      group_by = "mode"
    )
  )

  expect_error(tester(as.list(lu)))
  expect_error(
    tester(lu = lu[, .(oi = id, jobs, population)])
  )
  expect_error(
    tester(
      lu = lu[, .(id, oi = jobs, population)],
      opportunity = "jobs"
    )
  )
  expect_error(
    tester(
      lu = lu[, .(id, jobs, oi = population)],
      demand = "population"
    )
  )
})

# Small internal helpers
make_tm_3x3 <- function() {
  tm <- data.table::data.table(
    expand.grid(from_id = c("1","2","3"), to_id = c("1","2","3"))
  )
  tm[, travel_time := c(10, 30, 15,  30, 10, 25,  15, 25, 10)]
  tm
}
make_lu_large <- function() {
  data.table::data.table(
    id         = c("1","2","3"),
    population = c(4, 10, 6),
    jobs       = c(160, 150, 180)
  )
}
make_lu_match20 <- function() {
  data.table::data.table(
    id         = c("1","2","3"),
    population = c(4, 10, 6),  # sum = 20
    jobs       = c(7,  5,  8)  # sum = 20
  )
}

exp_decay_01 <- decay_exponential(decay_value = 0.1)
pow_decay_3  <- decay_power(decay_value = 3)

# ------------------------------- TOTAL -------------------------------
test_that("total-constrained: preserves totals on supply/demand sides", {
  tm <- make_tm_3x3()
  lu <- make_lu_large()

  # Supply-side output (returns accessible opportunities)
  det_supply <- constrained_accessibility(
    constraint = "total",
    travel_matrix = tm,
    land_use_data = lu,
    travel_cost = "travel_time",
    decay_function = exp_decay_01,
    demand = NULL,
    supply = "jobs",
    detailed_results = TRUE,
    active = TRUE
  )
  agg_supply <- constrained_accessibility(
    constraint = "total",
    travel_matrix = tm,
    land_use_data = lu,
    travel_cost = "travel_time",
    decay_function = exp_decay_01,
    demand = NULL,
    supply = "jobs",
    detailed_results = FALSE,
    active = TRUE
  )

  # Demand-side output (returns accessible population)
  det_demand <- constrained_accessibility(
    constraint = "total",
    travel_matrix = tm,
    land_use_data = lu,
    travel_cost = "travel_time",
    decay_function = exp_decay_01,
    demand = "population",
    supply = NULL,
    detailed_results = TRUE,
    active = FALSE
  )
  agg_demand <- constrained_accessibility(
    constraint = "total",
    travel_matrix = tm,
    land_use_data = lu,
    travel_cost = "travel_time",
    decay_function = exp_decay_01,
    demand = "population",
    supply = NULL,
    detailed_results = FALSE,
    active = FALSE
  )

  expect_equal(sum(det_supply$supply), sum(lu$jobs))
  expect_equal(sum(agg_supply$supply), sum(lu$jobs))
  expect_equal(sum(det_demand$demand), sum(lu$population))
  expect_equal(sum(agg_demand$demand), sum(lu$population))
})

# ------------------------------ SINGLY -------------------------------
test_that("singly-constrained: preserves totals and kappa properties", {
  tm <- make_tm_3x3()
  lu <- make_lu_large()

  # Supply-constrained (returns accessible opportunities)
  det_supply <- constrained_accessibility(
    travel_matrix = tm,
    land_use_data = lu,
    constraint = "singly",
    travel_cost = "travel_time",
    decay_function = exp_decay_01,
    demand = "population",
    supply = "jobs",
    detailed_results = TRUE,
    active = TRUE
  )
  agg_supply <- constrained_accessibility(
    constraint = "singly",
    travel_matrix = tm,
    land_use_data = lu,
    travel_cost = "travel_time",
    decay_function = exp_decay_01,
    demand = "population",
    supply = "jobs",
    detailed_results = FALSE,
    active = TRUE
  )

  # Demand-constrained (returns accessible population)
  det_demand <- constrained_accessibility(
    constraint = "singly",
    travel_matrix = tm,
    land_use_data = lu,
    travel_cost = "travel_time",
    decay_function = exp_decay_01,
    demand = "population",
    supply = "jobs",
    detailed_results = TRUE,
    active = FALSE
  )
  agg_demand <- constrained_accessibility(
    constraint = "singly",
    travel_matrix = tm,
    land_use_data = lu,
    travel_cost = "travel_time",
    decay_function = exp_decay_01,
    demand = "population",
    supply = "jobs",
    detailed_results = FALSE,
    active = FALSE
  )

  # Totals
  expect_equal(sum(det_supply$supply), sum(lu$jobs))
  expect_equal(sum(agg_supply$supply), sum(lu$jobs))
  expect_equal(sum(det_demand$demand), sum(lu$population))
  expect_equal(sum(agg_demand$demand), sum(lu$population))

  # κ normalization properties (these DO hold for singly)
  by_to   <- det_supply[, .(sum_k = sum(kappa_singly)),    by = to_id]
  by_from <- det_demand[, .(sum_k = sum(hatkappa_singly)), by = from_id]
  expect_true(all(abs(by_to$sum_k   - 1) < 1e-8))  # per-destination sums == 1
  expect_true(all(abs(by_from$sum_k - 1) < 1e-8))  # per-origin sums == 1
  expect_equal(sum(det_supply$kappa_singly),    length(unique(det_supply$to_id)))
  expect_equal(sum(det_demand$hatkappa_singly), length(unique(det_demand$from_id)))
})

# ------------------------------ DOUBLY -------------------------------
test_that("doubly-constrained: errors when totals mismatch", {
  tm <- make_tm_3x3()
  lu_bad <- data.table::copy(make_lu_match20())
  lu_bad[id == "3", jobs := jobs + 1]  # break equality

  expect_error(
    constrained_accessibility(
      constraint = "doubly",
      travel_matrix = tm,
      land_use_data = lu_bad,
      travel_cost = "travel_time",
      decay_function = exp_decay_01,
      demand = "population",
      supply = "jobs",
      detailed_results = TRUE,
      active = NULL
    ),     "sum of origins must equal the sum of destinations"
  )
})



test_that("doubly-constrained: OD flows match marginals", {
  tm    <- make_tm_3x3()
  lu    <- make_lu_match20()
  decay <- pow_decay_3

  det <- constrained_accessibility(
    constraint = "doubly",
    travel_matrix = tm,
    land_use_data = lu,
    travel_cost = "travel_time",
    decay_function = decay,
    demand = "population",
    supply = "jobs",
    detailed_results = TRUE,
    active = NULL
  )

  # Robustly pick the numeric flow column
  candidate_cols <- c("flow", "supply", "demand")
  value_col <- intersect(candidate_cols, names(det))[1]
  expect_true(!is.null(value_col), info = "No flow/supply/demand column in detailed output")
  expect_true(is.numeric(det[[value_col]]), info = "Flow column must be numeric")

  # Sum flows by origin and destination
  by_origin <- det[, .(sum_flow = sum(get(value_col))), by = from_id]
  by_dest   <- det[, .(sum_flow = sum(get(value_col))), by = to_id]

  # Build expected marginals from land use
  O <- lu[match(by_origin$from_id, lu$id), population]
  D <- lu[match(by_dest$to_id,   lu$id), jobs]

  expect_equal(by_origin$sum_flow, O, tolerance = 1e-3)
  expect_equal(by_dest$sum_flow,   D, tolerance = 1e-3)
})

#' Singly constrained accessibility
#'
#' Allocates opportunities at each destination proportionally based on travel
#' impedance and population at the origin. Uses the logic of Wilson's single
#' constraint. Returns values in the unit of 'demand'.
#' Internal helper function used by [constrained_accessibility()] when `constraint = "singly"`.
#'
#' @name singly_constrained
#' @keywords internal
#' @return A `data.table`/`data.frame` with results (structure mirrors the wrapper).
#' @examples NULL
#' @importFrom utils globalVariables
utils::globalVariables(c(
  # common ids / columns
  "from_id", "to_id", "id", "opp_weight", "supply", "demand", "weighted_demand", "weighted_supply",
  # total-constrained
  "kappa_total", "hatkappa_total", "K_total", "hatK_total", "constrained_opportunity",
  # singly-constrained
  "denom_i", "denom_j", "kappa_singly", "hatkappa_singly", "A_i", "B_j", "singly_access",
  # doubly-constrained
  "kappa_doubly", "flow", "error"
))
singly_constrained <- function(travel_matrix,
                               land_use_data,
                               travel_cost,
                               decay_function,
                               demand,
                               supply,
                               active,
                               group_by = character(0),
                               fill_missing_ids = TRUE,
                               detailed_results = FALSE) {
  # Validate
  checkmate::assert_string(demand)
  checkmate::assert_string(supply)
  checkmate::assert_logical(active, len = 1)

  # Convert to data.table
  if (!inherits(travel_matrix, "data.table")) {
    original_class <- class(travel_matrix)
    data <- data.table::as.data.table(travel_matrix)
  } else {
    data <- data.table::copy(travel_matrix)
  }
  if (!inherits(land_use_data, "data.table")) {
    land_use_data <- data.table::as.data.table(land_use_data)
  }

  # Merge demand and supply
  merge_by_reference(data, land_use_data, demand, left_df_idcol = "from_id")
  merge_by_reference(data, land_use_data, supply, left_df_idcol = "to_id")

  # Apply decay
  data <- apply_gravity_measure(data, decay_function, travel_cost)

  if (active) {
    # Supply-constrained (returns units of supply).. V_ij = (O_i f(c_ij) / sum_i O_i f(c_ij)) * D_j
    data[, weighted_demand := get(demand) * opp_weight]
    data[, denom_j := sum(weighted_demand), by = c("to_id", group_by)]
    data[, kappa_singly := data.table::fifelse(denom_j > 0, weighted_demand / denom_j, 0)]
    data[, singly_access := kappa_singly * get(supply)]

    if (detailed_results) {
      access <- data[, .(
        from_id,
        to_id,
        kappa_singly,
        supply = singly_access,
        B_j = data.table::fifelse(denom_j > 0, 1 / denom_j, 0)
      )]
    } else {
      access <- data[, .(supply = sum(singly_access)), by = c("from_id", group_by)]
      if (fill_missing_ids) access <- fill_missing_ids(access, travel_matrix, c("from_id", group_by))
    }

  } else {
    # Demand-constrained (returns units of demand).... M_ij = (D_j f(c_ij) / sum_j D_j f(c_ij)) * O_i
    data[, weighted_supply := get(supply) * opp_weight]
    data[, denom_i := sum(weighted_supply), by = c("from_id", group_by)]
    data[, hatkappa_singly := data.table::fifelse(denom_i > 0, weighted_supply / denom_i, 0)]
    data[, singly_access := hatkappa_singly * get(demand)]

    if (detailed_results) {
      access <- data[, .(
        from_id,
        to_id,
        hatkappa_singly,
        demand = singly_access,
        A_i = data.table::fifelse(denom_i > 0, 1 / denom_i, 0)
      )]
    } else {
      access <- data[, .(demand = sum(singly_access)), by = c("to_id", group_by)]
      if (fill_missing_ids) access <- fill_missing_ids(access, travel_matrix, c("to_id", group_by))
    }
  }

  if (exists("original_class")) class(access) <- original_class
  return(access[])
}

#' @param travel_matrix A data frame. The travel matrix describing the costs
#'   (i.e. travel time, distance, monetary cost, etc.) between the origins and
#'   destinations in the study area. Must contain the columns `from_id`, `to_id`
#'   and any others specified in `travel_cost`.

#' Foster-Greer-Thorbecke (FGT) poverty measures
#'
#' Calculates the FGT metrics, a family of poverty measures originally proposed
#' by \insertCite{foster1984class;textual}{accessibility} that capture the
#' extent and severity of poverty within an accessibility distribution. The FGT
#' family is composed of three measures that differ based on the \eqn{\alpha}
#' parameter used to calculate them (either 0, 1 or 2) and which also changes
#' their interpretation. Please see the details section for more information on
#' the interpretation of the measures.
#'
#' @param accessibility_data A data frame. The accessibility levels whose
#'   poverty levels should be calculated. Must contain the columns `id` and any
#'   others specified in `opportunity`.
#' @template sociodem_data_without_income
#' @param opportunity A string. The name of the column in `accessibility_data`
#'   with the accessibility levels to be considerend when calculating
#'   accessibility poverty.
#' @param population A string. The name of the column in `sociodemographic_data`
#'   with the number of people in each cell. Used to weigh accessibility levels
#'   when calculating poverty.
#' @param poverty_line A `numeric`. The poverty line below which individuals are
#'   considered to be in accessibility poverty.
#' @param poor_below_threshold Logic. If `TRUE` (default,) the observations below
#'       the poverty line are considered to be poor. This is the correct approach
#'       for primal accessibility measures (e.g.cumulative accessibility). If
#'       `FALSE`, then observations above the poverty line are considered to be
#'       poor. This is the correct approach for dual accessibility measures (e.g.
#'       travel time to the closest facility). When set to `FALSE`, FGT 1 and 2
#'       do not have an upper bound.
#' @param group_by A `character` vector. When not `character(0)` (the default),
#'   indicates the `accessibility_data` columns that should be used to group the
#'   poverty estimates by. For example, if `accessibility_data` includes a
#'   `race` column that specifies the racial category of the population (e.g.
#'   `"black"` and `"white"`) that each entry refers to, passing `"race"` to
#'   this parameter results in poverty estimates grouped by race.
#'
#' @return A data frame containing the three poverty estimates (FGT0, FGT1 and
#'   FGT2) for the study area.
#'
#' @section Interpretation of FGT measures:
#' The interpretation of each FGT measure depends on the \eqn{\alpha} parameter
#' used to calculate it:
#'
#' - with \eqn{\alpha = 0} (FGT0) the measure captures the *extent* of poverty
#' as a simple headcount - i.e. the proportion of people below the poverty line;
#'
#' - with \eqn{\alpha = 1} (FGT1) the measure, also know as the "poverty gap
#' index", captures the *severity* of poverty as the average percentage distance
#' between the poverty line and the accessibility of individuals below the
#' poverty line;
#'
#' - with \eqn{\alpha = 2} (FGT2) the measure simultaneously captures the
#' *extent* and the *severity* of poverty by calculating the number of people
#' below the poverty line weighted by the size of the accessibility shortfall
#' relative to the poverty line.
#'
#' FGT values range from 0 to 1. A value of 0 indicates that every individual is
#' above the poverty line. When every individual is below the poverty line,
#' however, FGT0 value is 1 and FGT1 and FGT2 values approach 1.
#'
#' @references
#' \insertAllCited{}
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' data_dir <- system.file("extdata", package = "accessibility")
#' travel_matrix <- readRDS(file.path(data_dir, "travel_matrix.rds"))
#' land_use_data <- readRDS(file.path(data_dir, "land_use_data.rds"))
#'
#' access <- cumulative_cutoff(
#'   travel_matrix,
#'   land_use_data,
#'   cutoff = 30,
#'   opportunity = "jobs",
#'   travel_cost = "travel_time"
#' )
#'
#' poverty <- fgt_poverty(
#'   accessibility_data = access,
#'   opportunity = "jobs",
#'   sociodemographic_data = land_use_data,
#'   population = "population",
#'   poverty_line = 50000
#' )
#' poverty
#'
#' @export
fgt_poverty <- function(accessibility_data,
                         sociodemographic_data,
                         opportunity,
                         population,
                         poverty_line,
                         poor_below_threshold = TRUE,
                         group_by = character(0)) {

  checkmate::assert_string(opportunity)
  checkmate::assert_string(population)
  checkmate::assert_numeric(poverty_line, lower = 0)
  checkmate::assert_logical(poor_below_threshold)
  assert_access_group_by(group_by)
  assert_accessibility_data(accessibility_data, opportunity, group_by)
  assert_sociodemographic_data(
    sociodemographic_data,
    accessibility_data,
    population = population
  )

  if (!inherits(accessibility_data, "data.table")) {
    original_class <- class(accessibility_data)
    data <- data.table::as.data.table(accessibility_data)
  } else {
    data <- data.table::copy(accessibility_data)
  }

  if (!inherits(sociodemographic_data, "data.table")) {
    sociodemographic_data <- data.table::as.data.table(sociodemographic_data)
  }

  merge_by_reference(
    data,
    sociodemographic_data,
    population,
    left_df_idcol = "id"
  )

  warn_extra_access_cols(accessibility_data, opportunity, group_by)

  .opp_colname <- opportunity
  .pop_colname <- population
  .groups <- group_by

  # primal accessibility ---------------------------------------------------------
  if (isTRUE(poor_below_threshold)) {

    data[get(.opp_colname) >= poverty_line, c(".fgt0", ".fgt1", ".fgt2") := 0]

    data[
      get(.opp_colname) < poverty_line,
      .norm_opp_shortfall := (poverty_line - get(.opp_colname)) / poverty_line
    ]
    data[
      get(.opp_colname) < poverty_line,
      `:=`(
        .fgt0 = .norm_opp_shortfall ^ 0,
        .fgt1 = .norm_opp_shortfall ^ 1,
        .fgt2 = .norm_opp_shortfall ^ 2
      )
    ]
  }

  # dual accessibility ---------------------------------------------------------
  if (isFALSE(poor_below_threshold)) {
    data[get(.opp_colname) <= poverty_line, c(".fgt0", ".fgt1", ".fgt2") := 0]

    data[
      get(.opp_colname) > poverty_line,
      .norm_opp_shortfall := (get(.opp_colname) - poverty_line) / poverty_line
      # .norm_opp_shortfall := (get(.opp_colname)  / poverty_line)
    ]

    data[
      get(.opp_colname) > poverty_line,
      `:=`(
        .fgt0 = .norm_opp_shortfall ^ 0,
        .fgt1 = .norm_opp_shortfall ^ 1,
        .fgt2 = .norm_opp_shortfall ^ 2
      )
    ]
  }


  if (nrow(data) == 0 && identical(group_by, character(0))) {
    fgt <- data.table::data.table(
      FGT0 = numeric(0),
      FGT1 = numeric(0),
      FGT2 = numeric(0)
    )
  } else {
    fgt <- data[
      ,
      .(
        FGT0 = stats::weighted.mean(x = .fgt0, w = get(.pop_colname)),
        FGT1 = stats::weighted.mean(x = .fgt1, w = get(.pop_colname)),
        FGT2 = stats::weighted.mean(x = .fgt2, w = get(.pop_colname))
      ),
      by = .groups
    ]
  }

  if (exists("original_class")) class(fgt) <- original_class

  return(fgt[])
}

# if running manually, please run the following line first:
# source("tests/testthat/setup.R")

tester <- function(accessibility_data = small_access,
                   sociodemographic_data = land_use_data,
                   opportunity = "jobs",
                   population = "population",
                   poverty_line = 10000,
                   poor_below_threshold = TRUE,
                   group_by = "mode") {
  fgt_poverty(
    accessibility_data,
    sociodemographic_data,
    opportunity,
    population,
    poverty_line,
    poor_below_threshold,
    group_by
  )
}

test_that("raises errors due to incorrect input", {
  expect_error(tester(opportunity = 1))
  expect_error(tester(opportunity = c("schools", "jobs")))

  expect_error(tester(population = 1))
  expect_error(tester(population = c("schools", "jobs")))

  expect_error(tester(poverty_line = "a"))
  expect_error(tester(poverty_line = -1))

  expect_error(tester(group_by = 1))
  expect_error(tester(group_by = NA))
  expect_error(tester(group_by = "id"))
  expect_error(tester(group_by = c("mode", "mode")))

  expect_error(tester(as.list(small_access)))
  expect_error(tester(small_access[, .(oi = id, jobs, mode)]))
  expect_error(tester(small_access[, .(id, oi = jobs, mode)]))
  expect_error(tester(small_access[, .(id, jobs, oi = mode)]))

  expect_error(tester(sociodemographic_data = as.list(land_use_data)))
  expect_error(
    tester(sociodemographic_data = land_use_data[, .(oi = id, population)])
  )
  expect_error(
    tester(sociodemographic_data = land_use_data[, .(id, oi = population)])
  )

  expect_error(
    tester(poor_below_threshold = "banana")
  )
})

test_that("throws warning if accessibility_data has an extra col", {
  expect_warning(tester(group_by = character(0)))
})

test_that("returns a dataframe with same class as accessibility_data's", {
  result <- tester()
  expect_is(result, "data.table")
  result <- tester(sociodemographic_data = as.data.frame(land_use_data))
  expect_is(result, "data.table")

  result <- tester(as.data.frame(small_access))
  expect_false(inherits(result, "data.table"))
  expect_is(result, "data.frame")
  result <- tester(
    as.data.frame(small_access),
    sociodemographic_data = as.data.frame(land_use_data)
  )
  expect_false(inherits(result, "data.table"))
  expect_is(result, "data.frame")
})

test_that("result has correct structure", {
  result <- tester()
  expect_true(ncol(result) == 4)
  expect_is(result$mode, "character")
  expect_is(result$FGT0, "numeric")
  expect_is(result$FGT1, "numeric")
  expect_is(result$FGT2, "numeric")

  suppressWarnings(result <- tester(group_by = character(0)))
  expect_true(ncol(result) == 3)
  expect_is(result$FGT0, "numeric")
  expect_is(result$FGT1, "numeric")
  expect_is(result$FGT2, "numeric")

  result <- tester(small_access[0])
  expect_true(nrow(result) == 0)
  expect_true(ncol(result) == 4)
  expect_is(result$mode, "character")
  expect_is(result$FGT0, "numeric")
  expect_is(result$FGT1, "numeric")
  expect_is(result$FGT2, "numeric")

  suppressWarnings(result <- tester(small_access[0], group_by = character(0)))
  expect_true(nrow(result) == 0)
  expect_true(ncol(result) == 3)
  expect_is(result$FGT0, "numeric")
  expect_is(result$FGT1, "numeric")
  expect_is(result$FGT2, "numeric")
})

test_that("input data sets remain unchanged", {
  original_access_data <- cumulative_cutoff(
    smaller_matrix,
    land_use_data,
    opportunity = "jobs",
    travel_cost = "travel_time",
    cutoff = 30,
    group_by = "mode"
  )
  original_sociodem_data <- readRDS(file.path(data_dir, "land_use_data.rds"))

  result <- tester()

  expect_equal(original_access_data, small_access)
  expect_equal(original_sociodem_data, land_use_data)
})

test_that("poverty levels are correctly calculated", {
  selected_ids <- c(
    "89a88cd909bffff",
    "89a88cdb57bffff",
    "89a88cdb597ffff",
    "89a88cdb5b3ffff",
    "89a88cdb5cfffff"
  )
  access_data <- cumulative_cutoff(
    travel_matrix[from_id %in% selected_ids],
    land_use_data,
    opportunity = "jobs",
    travel_cost = "travel_time",
    cutoff = 30,
    group_by = "mode"
  )

  # everything is 0 when all access levels are above the poverty line (10000)

  result <- tester(access_data)
  expected_result <- data.table::data.table(
    mode = c("transit", "transit2"),
    FGT0 = 0,
    FGT1 = 0,
    FGT2 = 0
  )
  expect_identical(result, expected_result)

  # FGT0 is 1 when all access levels are below poverty lines

  result <- tester(access_data, poverty_line = 100000)
  result[, `:=`(FGT1 = round(FGT1, 4), FGT2 = round(FGT2, 4))]
  expected_result <- data.table::data.table(
    mode = c("transit", "transit2"),
    FGT0 = 1,
    FGT1 = 0.5588,
    FGT2 = 0.3256
  )
  expect_identical(result, expected_result)

  access_data <- access_data[!(id == "89a88cdb5cfffff" & mode == "transit2")]
  result <- tester(access_data, poverty_line = 100000)
  result[, `:=`(FGT1 = round(FGT1, 4), FGT2 = round(FGT2, 4))]
  expected_result <- data.table::data.table(
    mode = c("transit", "transit2"),
    FGT0 = 1,
    FGT1 = c(0.5588, 0.6367),
    FGT2 = c(0.3256, 0.4129)
  )
  expect_identical(result, expected_result)
})

test_that("works even if access_data and sociodem_data has specific colnames", {
  selected_ids <- c(
    "89a88cdb57bffff",
    "89a88cdb5b3ffff"
  )
  access_data <- cumulative_cutoff(
    travel_matrix[from_id %in% selected_ids],
    land_use_data,
    opportunity = "jobs",
    travel_cost = "travel_time",
    cutoff = 30,
    group_by = "mode"
  )
  expected_result <- tester(access_data)

  access_data[, opportunity := "oi"]
  result <- suppressWarnings(tester(access_data))
  expect_identical(expected_result, result)

  access_data[, opportunity := NULL]
  land_use_data[, population_temp := population]
  land_use_data[, population := 1]
  result <- tester(access_data, population = "population_temp")
  expect_identical(expected_result, result)

  land_use_data[, population := population_temp]
  land_use_data[, population_temp := NULL]
  access_data[, group_by := "oi"]
  result <- suppressWarnings(tester(access_data))
  expect_identical(expected_result, result)

  access_data[, group_by := NULL]

  # poor above the threshold
  poor_above <- tester(poor_below_threshold = F, poverty_line = 41)
  expect_identical(poor_above$FGT0, 1)
  expect_identical(poor_above$FGT1, 1)
  expect_identical(poor_above$FGT2, 1)
})

#' Floating catchment area accessibility
#'
#' Calculates accessibility accounting for the competition of resources using a
#' measure from the floating catchment area (FCA) family. Please see the details
#' for the available FCA measures.
#' @template description_generic_cost
#'
#' @template travel_matrix
#' @template land_use_data
#' @template opportunity
#' @template travel_cost
#' @template demand
#' @param method A string. Which floating catchment area measure to use.
#'   Current available options are `"2sfca"` and `"bfca"`. More info in the
#'   details.
#' @template decay_function
#' @template group_by
#' @template fill_missing_ids_combinations
#'
#' @template return_accessibility
#'
#' @section Details:
#' The package currently includes two built-in FCA measures:
#'
#' - 2SFCA - the 2-Step Floating Catchment Area measure was the first
#' accessibility metric in the FCA family. It was originally proposed by
#' \insertCite{luo2003measures;textual}{accessibility}.
#'
#' - BFCA - the Balanced Floating Catchment Area measure calculates
#' accessibility accounting for competition effects while simultaneously
#' correcting for issues of inflation of demand and service levels that are
#' present in other FCA measures. It was originally proposed by
#' \insertCite{paez2019demand;textual}{accessibility} and named in
#' \insertCite{pereira2021geographic;textual}{accessibility}.
#'
#' @references
#' \insertAllCited{}
#'
#' @family Floating catchment area
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' data_dir <- system.file("extdata", package = "accessibility")
#' travel_matrix <- readRDS(file.path(data_dir, "travel_matrix.rds"))
#' land_use_data <- readRDS(file.path(data_dir, "land_use_data.rds"))
#'
#' # 2SFCA with a step decay function
#' df <- floating_catchment_area(
#'   travel_matrix,
#'   land_use_data,
#'   method = "2sfca",
#'   decay_function = decay_binary(cutoff = 50),
#'   opportunity = "jobs",
#'   travel_cost = "travel_time",
#'   demand = "population"
#' )
#' head(df)
#'
#'
#' # BFCA with an exponential decay function
#' df <- floating_catchment_area(
#'   travel_matrix,
#'   land_use_data,
#'   method = "bfca",
#'   decay_function = decay_exponential(decay_value = 0.5),
#'   opportunity = "jobs",
#'   travel_cost = "travel_time",
#'   demand = "population"
#' )
#' head(df)
#'
#' @export
floating_catchment_area <- function(travel_matrix,
                                    land_use_data,
                                    opportunity,
                                    travel_cost,
                                    demand,
                                    method,
                                    decay_function,
                                    group_by = character(0),
                                    fill_missing_ids = TRUE) {
  checkmate::assert(
    checkmate::check_string(method),
    checkmate::check_names(method, subset.of = c("2sfca", "bfca")),
    combine = "and"
  )
  checkmate::assert_string(opportunity)
  checkmate::assert_string(travel_cost)
  checkmate::assert_string(demand)
  checkmate::assert_logical(fill_missing_ids, len = 1, any.missing = FALSE)
  assert_decay_function(decay_function)
  assert_group_by(group_by)
  assert_travel_matrix(travel_matrix, travel_cost, group_by)
  assert_land_use_data(
    land_use_data,
    travel_matrix,
    opportunity,
    demand = demand
  )

  # if not a dt, keep original class to assign later when returning result

  if (!inherits(travel_matrix, "data.table")) {
    original_class <- class(travel_matrix)
    data <- data.table::as.data.table(travel_matrix)
  } else {
    data <- data.table::copy(travel_matrix)
  }

  if (!inherits(land_use_data, "data.table")) {
    land_use_data <- data.table::as.data.table(land_use_data)
  }

  merge_by_reference(data, land_use_data, opportunity, left_df_idcol = "to_id")
  merge_by_reference(data, land_use_data, demand, left_df_idcol = "from_id")

  data <- apply_gravity_measure(data, decay_function, travel_cost)

  groups <- c("from_id", group_by)
  if ("decay_function_arg" %in% names(data)) {
    groups <- c(groups, "decay_function_arg")
  }

  warn_extra_cols(
    travel_matrix,
    travel_cost,
    group_id = "from_id",
    groups = groups
  )

  fca_function <- if (method == "2sfca") {
    fca_2sfca
  } else {
    fca_bfca
  }

  access <- fca_function(data, opportunity, demand, setdiff(groups, "from_id"))

  if (fill_missing_ids) {
    access <- fill_missing_ids(access, travel_matrix, groups)
  }

  data.table::setnames(access, c("from_id", "access"), c("id", opportunity))

  if (exists("original_class")) class(access) <- original_class

  return(access)
}


#' @keywords internal
fca_2sfca <- function(data, opportunity, demand, group_by) {
  .opportunity_colname <- opportunity
  .demand_colname <- demand

  # step 1a - calculate the demand to each destination as the weight between
  # each od pair multiplied by the number of people at the origin

  data[
    ,
    pop_served := sum(get(.demand_colname) * opp_weight, na.rm = TRUE),
    by = c("to_id", group_by)
  ]

  # step 1b - calculate the provider-to-population ratio (ppr) of a destination
  # as number of opportunities/resources located at it divided by the
  # population it serves

  data[, ppr := get(.opportunity_colname) / pop_served]

  # step 2 - calculate accessibility at each origin i as the sum of the
  # multiplication between the ppr of destination j and the weight between od
  # pair ij

  access <- data[
    ,
    .(access = sum(ppr * opp_weight, na.rm = TRUE)),
    by = c("from_id", group_by)
  ]

  return(access)
}


#' @keywords internal
fca_bfca <- function(data, opportunity, demand, group_by) {
  .opportunity_colname <- opportunity
  .demand_colname <- demand

  # calculate balanced (normalized) opp_weight by origin and by destination

  data[
    ,
    balanced_opp_weight_i := opp_weight / sum(opp_weight),
    by = c("from_id", group_by)
  ]
  data[
    ,
    balanced_opp_weight_j := opp_weight / sum(opp_weight),
    by = c("to_id", group_by)
  ]

  # step 1a - calculate the balanced demand to each destination as the balanced
  # weight by origin multiplied by the number of people at the origin

  data[
    ,
    balanced_pop_served := sum(
      get(.demand_colname) * balanced_opp_weight_i,
      na.rm = TRUE
    ),
    by = c("to_id", group_by)
  ]

  # step 1b - calculate the balanced provider-to-population ratio
  # (balanced_ppr) of a destination as number of opportunities/resources
  # located at it divided by the balanced demand

  data[, balanced_ppr := get(.opportunity_colname) / balanced_pop_served]

  # step 2 - calculate accessibility at each origin i as the sum of the
  # multiplication between the balanced_ppr of destination j and the balanced
  # weight by destination between od pair ij

  access <- data[
    ,
    .(access = sum(balanced_ppr * balanced_opp_weight_j, na.rm = TRUE)),
    by = c("from_id", group_by)
  ]

  return(access)
}

# if running manually, please run the following line first:
# source("tests/testthat/setup.R")

tester <- function(
  travel_matrix = smaller_matrix,
  land_use_data = get("land_use_data", envir = parent.frame()),
  opportunity = "jobs",
  travel_cost = "travel_time",
  demand = "population",
  method = "2sfca",
  decay_function = decay_binary(45),
  group_by = "mode",
  fill_missing_ids = TRUE
) {
  floating_catchment_area(
    travel_matrix,
    land_use_data,
    opportunity,
    travel_cost,
    demand,
    method,
    decay_function,
    group_by,
    fill_missing_ids
  )
}


test_that("raises errors due to incorrect input", {
  expect_error(tester(method = 1))
  expect_error(tester(method = "a"))
  expect_error(tester(method = c("bfca", "bfca")))

  expect_error(tester(decay_function = "a"))
  expect_error(tester(decay_function = mean))
  expect_error(tester(decay_function = get))

  expect_error(tester(opportunity = 1))
  expect_error(tester(opportunity = c("schools", "jobs")))

  expect_error(tester(travel_cost = 1))
  expect_error(tester(travel_cost = c("travel_time", "monetary_cost")))

  expect_error(tester(demand = 1))
  expect_error(tester(demand = c("population", "population")))

  expect_error(tester(group_by = 1))
  expect_error(tester(group_by = NA))
  expect_error(tester(group_by = "from_id"))
  expect_error(tester(group_by = c("mode", "mode")))

  expect_error(tester(fill_missing_ids = 1))
  expect_error(tester(fill_missing_ids = c(TRUE, TRUE)))
  expect_error(tester(fill_missing_ids = NA))

  expect_error(tester(as.list(travel_matrix)))
  expect_error(tester(travel_matrix[, .(oi = from_id, to_id, travel_time)]))
  expect_error(tester(travel_matrix[, .(from_id, oi = to_id, travel_time)]))
  expect_error(
    tester(
      travel_matrix[, .(from_id, to_id, oi = travel_time)],
      travel_cost = "travel_time"
    )
  )
  expect_error(
    tester(
      travel_matrix[, .(from_id, to_id, travel_time, oi = mode)],
      group_by = "mode"
    )
  )

  expect_error(tester(as.list(land_use_data)))
  expect_error(
    tester(land_use_data = land_use_data[, .(oi = id, jobs, population)])
  )
  expect_error(
    tester(
      land_use_data = land_use_data[, .(id, oi = jobs, population)],
      opportunity = "jobs"
    )
  )
  expect_error(
    tester(
      land_use_data = land_use_data[, .(id, jobs, oi = population)],
      demand = "population"
    )
  )
})

test_that("throws warning if travel_matrix extra col", {
  # i.e. col not listed in travel_cost and by_col
  expect_warning(tester(group_by = character(0)))
})

test_that("returns a dataframe whose class is the same as travel_matrix's", {
  result <- tester()
  expect_is(result, "data.table")
  result <- tester(land_use_data = as.data.frame(land_use_data))
  expect_is(result, "data.table")

  result <- tester(as.data.frame(smaller_matrix))
  expect_false(inherits(result, "data.table"))
  expect_is(result, "data.frame")
  result <- tester(
    as.data.frame(smaller_matrix),
    land_use_data = as.data.frame(land_use_data)
  )
  expect_false(inherits(result, "data.table"))
  expect_is(result, "data.frame")
})

test_that("result has correct structure", {
  result <- tester()
  expect_true(ncol(result) == 3)
  expect_is(result$id, "character")
  expect_is(result$mode, "character")
  expect_is(result$jobs, "numeric")

  result <- tester(opportunity = "schools")
  expect_true(ncol(result) == 3)
  expect_is(result$id, "character")
  expect_is(result$mode, "character")
  expect_is(result$schools, "numeric")

  suppressWarnings(result <- tester(group_by = character(0)))
  expect_true(ncol(result) == 2)
  expect_is(result$id, "character")
  expect_is(result$jobs, "numeric")

  result <- tester(
    data.table::data.table(
      mode = character(),
      from_id = character(),
      to_id = character(),
      travel_time = integer()
    )
  )
  expect_true(ncol(result) == 3)
  expect_true(nrow(result) == 0)
  expect_is(result$id, "character")
  expect_is(result$mode, "character")
  expect_is(result$jobs, "numeric")
})

test_that("input data sets remain unchanged", {
  original_smaller_matrix <- travel_matrix[1:10]
  original_land_use_data <- readRDS(file.path(data_dir, "land_use_data.rds"))

  result <- tester()

  # subsets in other functions tests set smaller_matrix index
  data.table::setindex(smaller_matrix, NULL)

  expect_equal(original_smaller_matrix, smaller_matrix)
  expect_equal(original_land_use_data, land_use_data)
})

test_that("fill_missing_ids arg works correctly", {
  small_travel_matrix <- travel_matrix[
    from_id %in% c("89a88cdb57bffff", "89a88cdb597ffff")
  ]
  small_travel_matrix <- small_travel_matrix[
    !(from_id == "89a88cdb57bffff" & mode == "transit2")
  ]

  result <- tester(small_travel_matrix, fill_missing_ids = TRUE)
  result[, jobs := as.integer(jobs)]
  data.table::setkey(result, NULL)
  expect_identical(
    result,
    data.table::data.table(
      id = rep(c("89a88cdb57bffff", "89a88cdb597ffff"), each = 2),
      mode = rep(c("transit", "transit2"), times = 2),
      jobs = c(233L, 0L, 195L, 372L)
    )
  )

  result <- tester(small_travel_matrix, fill_missing_ids = FALSE)
  result[, jobs := as.integer(jobs)]
  expect_identical(
    result,
    data.table::data.table(
      id = c("89a88cdb57bffff", "89a88cdb597ffff", "89a88cdb597ffff"),
      mode = c("transit", "transit", "transit2"),
      jobs = c(233L, 195L, 372L)
    )
  )
})

test_that("accepts custom decay function", {
  selected_ids <- c(
    "89a88cdb57bffff",
    "89a88cdb597ffff",
    "89a88cdb5b3ffff",
    "89a88cdb5cfffff",
    "89a88cd909bffff"
  )
  smaller_travel_matrix <- travel_matrix[
    from_id %in% selected_ids & to_id %in% selected_ids
  ]

  custom_function <- function(travel_cost) rep(1L, length(travel_cost))

  result <- tester(smaller_travel_matrix, decay_function = custom_function)
  result[, jobs := round(jobs, digits = 4)]
  expect_identical(
    result,
    data.table::data.table(
      id = rep(selected_ids, 2),
      mode = rep(c("transit", "transit2"), each = 5),
      jobs = 0.1108
    )
  )
})

test_that("calculates 2sfca correctly", {
  selected_ids <- c(
    "89a88cdb57bffff",
    "89a88cdb597ffff",
    "89a88cdb5b3ffff",
    "89a88cdb5cfffff",
    "89a88cd909bffff"
  )
  smaller_travel_matrix <- travel_matrix[
    from_id %in% selected_ids & to_id %in% selected_ids
  ]

  result <- tester(smaller_travel_matrix, method = "2sfca")
  result[, jobs := round(jobs, digits = 4)]
  expect_identical(
    result,
    data.table::data.table(
      id = rep(selected_ids, 2),
      mode = rep(c("transit", "transit2"), each = 5),
      jobs = c(0.0152, 0.1491, 0.1491, 0.1068, 0)
    )
  )
})

test_that("calculates bfca correctly", {
  selected_ids <- c(
    "89a88cdb57bffff",
    "89a88cdb597ffff",
    "89a88cdb5b3ffff",
    "89a88cdb5cfffff",
    "89a88cd909bffff"
  )
  smaller_travel_matrix <- travel_matrix[
    from_id %in% selected_ids & to_id %in% selected_ids
  ]

  result <- tester(smaller_travel_matrix, method = "bfca")
  result[, jobs := round(jobs, digits = 4)]
  expect_identical(
    result,
    data.table::data.table(
      id = rep(selected_ids, 2),
      mode = rep(c("transit", "transit2"), each = 5),
      jobs = c(0.0094, 0.1993, 0.1993, 0.1147, 0)
    )
  )
})

test_that("works even if travel_matrix and land_use has specific colnames", {
  expected_result <- tester()

  smaller_matrix[, opportunity := "oi"]
  result <- suppressWarnings(tester(smaller_matrix))
  expect_identical(expected_result, result)

  smaller_matrix[, opportunity := NULL]
  smaller_matrix[, travel_cost := "oi"]
  result <- suppressWarnings(tester(smaller_matrix))
  expect_identical(expected_result, result)

  smaller_matrix[, travel_cost := NULL]
  smaller_matrix[, groups := "oi"]
  result <- suppressWarnings(tester(smaller_matrix))
  expect_identical(expected_result, result)

  smaller_matrix[, groups := NULL]
  smaller_matrix[, demand := "oi"]
  result <- suppressWarnings(tester(smaller_matrix))
  expect_identical(expected_result, result)

  smaller_matrix[, demand := NULL]
  land_use_data[, opportunity := "oi"]
  result <- suppressWarnings(tester(land_use_data = land_use_data))
  expect_identical(expected_result, result)

  land_use_data[, opportunity := NULL]
  land_use_data[, demand := "oi"]
  result <- suppressWarnings(tester(land_use_data = land_use_data))
  expect_identical(expected_result, result)

  land_use_data[, demand := NULL]
})

test_that("results are grouped by decay_function_arg when needed", {
  small_travel_matrix <- travel_matrix[
    from_id %in% c("89a88cdb57bffff", "89a88cdb597ffff") &
      mode != "transit2"
  ]

  result <- tester(
    small_travel_matrix,
    decay_function = decay_exponential(c(0.5, 0.6))
  )
  data.table::setkey(result, NULL)
  result[, jobs := round(jobs, 2)]

  expect_identical(
    result,
    data.table::data.table(
      id = rep(c("89a88cdb57bffff", "89a88cdb597ffff"), times = 2),
      mode = rep("transit", 4),
      decay_function_arg = rep(c(0.5, 0.6), each = 2),
      jobs = c(456.32, 272.26, 456.85, 271.10)
    )
  )
})

https://maxcovr.njtierney.com/index.html
https://prioritizr.net/articles/publication_record.html
https://dirkschumacher.github.io/ompr/articles/problem-warehouse-location.html
https://github.com/njtierney/maxcovr

https://pysal.org/spopt/notebooks/p-median.html

# location-allocation brute force

library(accessibility)
library(data.table)
library(furrr)




data_dir <- system.file("extdata", package = "accessibility")
travel_matrix <- readRDS(file.path(data_dir, "travel_matrix.rds"))
land_use_data <- readRDS(file.path(data_dir, "land_use_data.rds"))


df <- accessibility:::merge_by_reference(data = travel_matrix,
                                         land_use_data = land_use_data,
                                         left_df_idcol = 'from_id',
                                         opportunity = 'population')


df <- accessibility:::merge_by_reference(data = df,
                                         land_use_data = land_use_data,
                                         left_df_idcol = 'to_id'
                                         , opportunity = 'schools'
                                         )
head(df)




#' references
#'
#' Open-source approaches for location cover models: capabilities and efficiency
#' https://link.springer.com/article/10.1007/s10109-021-00350-w
#'
#' Mark green
#' https://github.com/markagreen/mapping_test_accessibility
#'
#' Pysal spopt
#' https://github.com/pysal/spopt


optimal_location <- function(df, candidates='all'){

  # candidates <- c('89a88cd90b7ffff', '89a88cdb607ffff')



  # ## drop origin areas with no population
  # df <- df[population>0,]

  ## get baseline time to access closest facility

  # find time closest
  # dfclosest <- df[,.(population=population[1L],
  #                    travel_to_closest=min(travel_time[which(schools>0)]),na.rm=TRUE),
  #                 by=from_id]

 # internal function to find travel time to closest facility
  get_closest <-function(ttmfull){
    temp_closest <- ttmfull[schools>0,.(population=population[1L],
                                travel_to_closest=min(travel_time,na.rm=TRUE)),
                    by=from_id]
    return(temp_closest)
  }

  # find travel time to closest facility
  dfclosest <- get_closest(ttmfull=df)

  # internal function to find average time to closest
  avg_time2closest <- function(df_closest){
    df_avg_time <- df_closest[population>0,.(wmean_time= weighted.mean(travel_to_closest, w=population, na.rm=TRUE))]
    return(df_avg_time)
    }

  # get baseline average time to closest
  baseline <- avg_time2closest(df_closest=dfclosest)$wmean_time

  # find all location candidates which do not have any facility
  # in a future version, the user can pass a vector of location candidates
  if(candidates=='all'){
    all_cadidates <- unique(df$to_id[which(df$schools==0)])
  }

  if (all(candidates!='all')) {
  all_cadidates = candidates
  }


  # function to simulate impact of allocation at an specific destination
  sim_one_location <- function(candn,
                               df = parent.frame()$df){

    # candn <- '89a88cd968fffff'

    # add 1 school to candn
    df[to_id == candn,]
    df[to_id == candn, schools := schools + 1]

    # find travel time to closest facility
    step1_closest <- get_closest(ttmfull = df)

    # get baseline average time to closest
    step1_avg_time <- avg_time2closest(df_closest = step1_closest)

    # add simulation id
    step1_avg_time[, candn := candn]
    return(step1_avg_time)
  }

  ## apply function to all candidate locations
  # df2 <- pbapply::pblapply(X=all_cadidates, FUN=sim_one_location, df=df) |>
  #        data.table::rbindlist()
  # parallel
  future::plan(strategy = 'multisession')
  df2 <- furrr::future_map(.x = all_cadidates,
                    .f = sim_one_location, .progress = TRUE, df=df) |>
    data.table::rbindlist()


  # location which minimizes average distance
  step2 <- df2[ wmean_time== min(wmean_time)]

  # organize output
  step2[, baseline_time := baseline]
  data.table::setnames(step2, old = 'candn', new = 'optimal_locations')
  data.table::setcolorder(step2, neworder = c('optimal_locations',
                                              'wmean_time',
                                              'baseline_time'))


  return(step2)
}


t <- optimal_location(df = df, candidates = 'all')




#' P-Median Problem
#' location which minimizes total distance
#' https://pysal.org/spopt/notebooks/p-median.html
df[, .(d = to_id[which.min(sum(travel_time * population))] )]

##### Decay functions ------------------------

library(accessibility)
library(data.table)
library(ggplot2)

vec <- 0:100
decay_value <- 0.2
cutoff <- 50

step        <- decay_binary(cutoff=cutoff)
linear      <- decay_linear(cutoff=cutoff)
exponential <- decay_exponential(decay_value=decay_value)
power       <- decay_power(decay_value = decay_value)

df <- data.table(
  minutes = vec,
  binary =          step(vec),
  linear =          linear(vec),
  exponential =     exponential(vec),
  inverse_power =   power(vec)
)

df2 <- data.table::melt.data.table(data = df, id.vars = 'minutes', variable.name = 'decay_function', value.name = 'impedance_factor')

ggplot() +
  geom_line(data=df2, aes(x=minutes, y=impedance_factor, color=decay_function), show.legend = FALSE) +
  facet_wrap(.~decay_function, ncol = 2) +
  theme_minimal()


ggsave('./man/figures/decay_functions.png', width = 12, height = 10, units = 'cm')

#' @keywords internal
assert_cutoff <- function(cutoff, travel_cost) {
  if (length(travel_cost) == 1) {
    checkmate::assert_numeric(
      cutoff,
      lower = 0,
      finite = TRUE,
      any.missing = FALSE,
      min.len = 1,
      unique = TRUE
    )
  } else {
    checkmate::assert_list(
      cutoff,
      any.missing = FALSE,
      len = length(travel_cost)
    )

    cutoff_collection <- checkmate::makeAssertCollection()

    for (i in seq.int(1, length(travel_cost))) {
      checkmate::assert_numeric(
        cutoff[[i]],
        lower = 0,
        finite = TRUE,
        any.missing = FALSE,
        min.len = 1,
        unique = TRUE,
        .var.name = paste0("cutoff[[", i, "]]"),
        add = cutoff_collection
      )
    }

    checkmate::reportAssertions(cutoff_collection)
  }

  return(invisible(TRUE))
}


#' @keywords internal
assert_group_by <- function(group_by) {
  checkmate::assert_character(group_by, any.missing = FALSE, unique = TRUE)

  checkmate::assert_names(group_by, disjunct.from = c("from_id", "to_id"))

  return(invisible(TRUE))
}


#' @keywords internal
assert_access_group_by <- function(group_by) {
  checkmate::assert_character(group_by, any.missing = FALSE, unique = TRUE)

  checkmate::assert_names(group_by, disjunct.from = "id")

  return(invisible(TRUE))
}


#' @keywords internal
assert_travel_matrix <- function(travel_matrix, travel_cost, group_by) {
  travel_matrix_req_names <- c("from_id", "to_id", travel_cost, group_by)
  checkmate::assert_data_frame(travel_matrix)
  checkmate::assert_names(
    names(travel_matrix),
    must.include = travel_matrix_req_names,
    .var.name = "travel_matrix"
  )

  return(invisible(TRUE))
}


#' @keywords internal
assert_accessibility_data <- function(accessibility_data,
                                      opportunity,
                                      group_by) {
  required_names <- c("id", opportunity, group_by)

  checkmate::assert_data_frame(accessibility_data)
  checkmate::assert_names(
    names(accessibility_data),
    must.include = required_names,
    .var.name = "accessibility_data"
  )

  return(invisible(TRUE))
}


#' @keywords internal
assert_land_use_data <- function(land_use_data,
                                 travel_matrix,
                                 opportunity,
                                 active = NULL,
                                 demand = NULL) {
  land_use_data_req_names <- c("id", opportunity)
  if (!is.null(demand)) {
    land_use_data_req_names <- c(land_use_data_req_names, demand)
  }

  checkmate::assert_data_frame(land_use_data)
  checkmate::assert_names(
    names(land_use_data),
    must.include = land_use_data_req_names,
    .var.name = "land_use_data"
  )

  # if an id is present in travel_matrix but not in land_use_data, merging the
  # datasets will produce NAs, so we warn users about that

  travel_matrix_ids <- if (is.null(active)) {
    c(travel_matrix$from_id, travel_matrix$to_id)
  } else if (active) {
    travel_matrix$to_id
  } else {
    travel_matrix$from_id
  }

  if (!all(travel_matrix_ids %in% land_use_data$id)) {
    warning(
      "'land_use_data' is missing ids listed in 'travel_matrix', which may ",
      "produce NAs in the final output.",
      call. = FALSE
    )
  }

  # if either land_use_data$opportunity or $demand contains NA values, the final
  # result may contain NAs as well, so we warn about that too

  cols_to_check <- c(
    opportunity,
    if (is.null(demand)) character(0) else demand
  )

  for (col in cols_to_check) {
    if (any(is.na(land_use_data[[col]]))) {
      warning(
        "'land_use_data$", col, "' contains NA values, which may produce NAs ",
        "in the final output.",
        call. = FALSE
      )
    }
  }

  return(invisible(TRUE))
}


#' @keywords internal
assert_sociodemographic_data <- function(sociodemographic_data,
                                         accessibility_data,
                                         population = NULL,
                                         income = NULL,
                                         extra_cols = NULL) {
  population <- if (is.null(population)) character() else population
  income <- if (is.null(income)) character() else income
  extra_cols <- if (is.null(extra_cols)) character() else extra_cols
  required_names <- c("id", population, income, extra_cols)

  checkmate::assert_data_frame(sociodemographic_data)
  checkmate::assert_names(
    names(sociodemographic_data),
    must.include = required_names,
    .var.name = "sociodemographic_data"
  )

  # if an id is present in accessibility_data but not in sociodemographic_data,
  # merging the datasets will produce NAs, so we warn users about that

  if (!all(accessibility_data$id %in% sociodemographic_data$id)) {
    warning(
      "'sociodemographic_data' is missing ids listed in 'accessibility_data', ",
      "which may produce NAs in the final output.",
      call. = FALSE
    )
  }

  # if any of the columns in sociodemographic_data required to calculate
  # inequality contains NA values, the final result may contain NAs as well, so
  # we warn about that too
  # if NAs are found in income column, we check if the correspondent population
  # entries are 0. if not, we throw the warning, otherwise we ignore it. related
  # to https://github.com/ipeaGIT/accessibility/issues/42

  non_income_cols <- c(population, extra_cols)

  for (col in non_income_cols) {
    if (any(is.na(sociodemographic_data[[col]]))) {
      warning(
        "'sociodemographic_data$", col, "' contains NA values, which may ",
        "produce NAs in the final output.",
        call. = FALSE
      )
    }
  }

  if (!identical(income, character())) {
    if (any(is.na(sociodemographic_data[[income]]))) {
      .inc_col <- income
      na_entries <- sociodemographic_data[
        is.na(sociodemographic_data[[.inc_col]]),
      ]
      pop_na_entries <- na_entries[[population]]

      should_warn <- any(pop_na_entries > 0)
      should_warn <- ifelse(is.na(should_warn), TRUE, should_warn)

      if (should_warn) {
        warning(
          "'sociodemographic_data$", income, "' contains NA values whose ",
          "correspodent 'sociodemographic_data$population' is not 0, which ",
          "may produce NAs in the final output.",
          call. = FALSE
        )
      }
    }
  }

  return(invisible(TRUE))
}


#' @keywords internal
assert_summary_function <- function(summary_function) {
  checkmate::assert_function(summary_function)

  result <- tryCatch(
    summary_function(1:100),
    error = function(cnd) cnd
  )

  if (inherits(result, "error")) {
    stop(
      "Assertion on 'summary_function' failed: Must be a function that ",
      "takes a numeric vector as input. Instead, failed with message: ",
      "'", result$message, "'",
      call. = FALSE
    )
  }

  if (!is.numeric(result) || length(result) != 1) {
    stop(
      "Assertion on 'summary_function' failed: Must be a function that takes ",
      "a numeric vector as input and returns a single numeric as output. ",
      "Instead, returned: ",
      "{", paste0(result, collapse = ","), "}",
      call. = FALSE
    )
  }

  return(invisible(TRUE))
}


#' @keywords internal
assert_decay_function <- function(decay_function) {
  checkmate::assert_function(decay_function)

  result <- tryCatch(
    decay_function(1:100),
    error = function(cnd) cnd
  )

  if (inherits(result, "error")) {
    stop(
      "Assertion on 'decay_function' failed: Must be a function that ",
      "takes a numeric vector as input. Instead, failed with message: ",
      "'", result$message, "'",
      call. = FALSE
    )
  }

  # result must be either a numeric vector with same length as input or a list
  # of numeric vectors whose lenghts are the same as input

  is_numeric_like <- function(x) is.numeric(x) || is.integer(x)

  right_numeric <- is_numeric_like(result) && length(result) == 100

  numeric_like_elements <- vapply(result, is_numeric_like, logical(1))
  right_classes <- length(result) > 0 && all(numeric_like_elements)

  elements_length <- vapply(result, length, integer(1))
  right_lengths <- length(result) > 0 && all(elements_length == 100)

  if (!(right_numeric || (is.list(result) && right_classes && right_lengths))) {
    stop(
      "Assertion on 'decay_function' failed: Must be a function that takes ",
      "a numeric vector as input and returns either:\n",
      " - a numeric vector with the same length of input, or;\n",
      " - a list of numeric vectors whose lengths are the same of input.\n",
      "Instead, with input 1:100 it returned: ",
      "{", paste0(result, collapse = ","), "}",
      call. = FALSE
    )
  }

  return(invisible(TRUE))
}


#' @keywords internal
assert_interval_increment <- function(interval_increment) {
  checkmate::assert_number(interval_increment, finite = TRUE)

  if (interval_increment <= 0) {
    stop(
      "Assertion on 'interval_increment' failed: ",
      "Value must be greater than 0."
    )
  }

  return(invisible(TRUE))
}


#' @keywords internal
assert_and_assign_interval <- function(interval) {
  is_numeric_like <- function(x) is.numeric(x) || is.integer(x)

  build_interval_name <- function(pair) {
    paste0("[", pair[1], ",", pair[2], "]")
  }

  if (is.list(interval)) {
    checkmate::assert_list(
      interval,
      types = "numeric",
      any.missing = FALSE,
      min.len = 1,
      unique = TRUE
    )

    for (i in seq.int(length(interval))) {
      checkmate::assert_numeric(
        interval[[i]],
        lower = 0,
        any.missing = FALSE,
        len = 2,
        unique = TRUE,
        sorted = TRUE,
        finite = TRUE,
        .var.name = paste0("interval[[", i, "]]")
      )
    }

    names(interval) <- vapply(interval, build_interval_name, character(1))
  } else if (is_numeric_like(interval)) {
    checkmate::assert_numeric(
      interval,
      lower = 0,
      any.missing = FALSE,
      len = 2,
      unique = TRUE,
      sorted = TRUE,
      finite = TRUE
    )

    interval <- list(interval)
    names(interval) <- build_interval_name(interval[[1]])
  } else {
    stop(
      "Assertion on 'interval' failed: Must be either:\n",
      " - a numeric vector with 2 elements, indicating the start and end ",
      "points of the travel cost interval;\n",
      " - a list of numeric vectors with 2 elements, each one of them ",
      "indicating the start and end points of a travel cost interval."
    )
  }

  return(interval)
}


#' @keywords internal
assert_detailed_fill_missing_ids <- function(fill_missing_ids,
                                             detailed_results) {
  checkmate::assert_logical(fill_missing_ids, len = 1, any.missing = FALSE)

  if (detailed_results && !fill_missing_ids) {
    warning(
      "Found non-default 'fill_missing_ids' (FALSE) with 'detailed_results = ",
      "TRUE'. Please note that 'fill_missing_ids' does not affect the ",
      "detailed spatial availability output.",
      call. = FALSE
    )
  }

  return(invisible(TRUE))
}

# if running manually, please run the following line first:
# source("tests/testthat/setup.R")

tester <- function(land_use_data = get("land_use_data", envir = parent.frame()),
                   travel_matrix = smaller_matrix,
                   opportunity = "jobs",
                   active = TRUE,
                   demand = NULL) {
  assert_land_use_data(
    land_use_data,
    travel_matrix,
    opportunity,
    active,
    demand
  )
}

test_that("warns if travel_matrix contains ids that land_use_data does not", {
  expect_warning(
    tester(
      travel_matrix = rbind(
        smaller_matrix,
        data.table::data.table(
          mode = "transit",
          from_id = "89a88cdb57bffff",
          to_id = "hehe",
          travel_time = 50
        )
      )
    )
  )

  expect_warning(
    tester(
      travel_matrix = rbind(
        smaller_matrix,
        data.table::data.table(
          mode = "transit",
          from_id = "hehe",
          to_id = "89a88cdb57bffff",
          travel_time = 50
        )
      ),
      active = FALSE
    )
  )

  expect_warning(
    tester(
      travel_matrix = rbind(
        smaller_matrix,
        data.table::data.table(
          mode = "transit",
          from_id = "89a88cdb57bffff",
          to_id = "hehe",
          travel_time = 50
        )
      ),
      active = NULL
    )
  )

  expect_warning(
    tester(
      travel_matrix = rbind(
        smaller_matrix,
        data.table::data.table(
          mode = "transit",
          from_id = "hehe",
          to_id = "89a88cdb57bffff",
          travel_time = 50
        )
      ),
      active = NULL
    )
  )
})

test_that("warns if either opportunity or demand cols contain NAs", {
  expect_warning(
    tester(
      rbind(land_use_data, data.table::data.table(id = "hehe"), fill = TRUE),
      demand = "population"
    )
  )
})

test_that("assert_land_use_data() doesn't error with tibbles", {
  # issue 49 (https://github.com/ipeaGIT/accessibility/issues/49)
  # using land_use_data as tibbles would throw errors because this format has
  # stricter subsetting rules
  expect_no_error(tester(tibble::as_tibble(land_use_data)))
})

# if running manually, please run the following line first:
# source("tests/testthat/setup.R")

tester <- function(sociodemographic_data = land_use_data,
                   accessibility_data = small_access,
                   population = NULL,
                   income = NULL,
                   extra_cols = NULL) {
  assert_sociodemographic_data(
    sociodemographic_data,
    accessibility_data,
    population,
    income,
    extra_cols
  )
}

test_that("warns if access_data contains ids not in sociodemographic_data", {
  expect_warning(
    tester(
      accessibility_data = rbind(
        small_access,
        data.table::data.table(id = "hehe", mode = "transit", jobs = 100)
      )
    )
  )
})

test_that("warns if cols contain NAs", {
  expect_warning(
    tester(
      rbind(
        land_use_data,
        data.table::data.table(id = "hehe"),
        fill = TRUE
      ),
      population = "population"
    )
  )

  expect_warning(
    tester(
      rbind(
        land_use_data,
        data.table::data.table(id = "hehe"),
        fill = TRUE
      ),
      extra_cols = "income_decile"
    )
  )

  expect_warning(
    tester(
      rbind(
        land_use_data,
        data.table::data.table(id = "hehe", population = 1),
        fill = TRUE
      ),
      income = "income_per_capita",
      population = "population"
    )
  )

  expect_warning(
    tester(
      rbind(
        land_use_data,
        data.table::data.table(id = "hehe", population = NA),
        fill = TRUE
      ),
      income = "income_per_capita",
      population = "population"
    )
  )

  # should not warn because all NA income entries have respective population
  # entries equal to 0
  expect_silent(
    tester(
      land_use_data,
      income = "income_per_capita",
      population = "population"
    )
  )
})

library(data.table)
library(sf)
library(ggplot2)


sigla <- 'bho'

########## select radius  --------------------------------------------------
# acces <- aopdata::read_access(city=sigla, year=2019, geometry = T)
# acces <- readRDS('L:/Proj_acess_oport/data/acesso_oport/output_base_final/2019/dados2019_AcessOport_access_tpcar_v1.0.rds')
acces_for <- subset(acces, sigla_muni ==sigla & pico==1 & modo =='tp')

center <- acces_for$id_hex[which.max(acces_for$CMATT30)]

buff <- subset(acces_for, id_hex == center) |>
          st_centroid()  |>
           st_transform(crs = '+proj=utm +zone23 +datum=WGS84 +units=m +ellps=WGS84 +towgs84=0,0,0') |>
          st_buffer(dist = 7000) |>
          st_transform(crs = st_crs(acces_for))

plot(buff['id_hex'])

ggplot() +
  geom_sf(data=acces_for, aes(fill=CMATT30), color=NA) +
  geom_sf(data=buff, color='red', fill='red', alpha=.5)

buff$CMATT30 <- NULL
inter <- st_intersects(acces_for, buff, sparse = FALSE )

inter <- acces_for[inter, ]

ggplot() +
  geom_sf(data=inter, aes(fill=CMATT30), color=NA) # +geom_sf(data=buff, color='red', fill='red', alpha=.5)




########## travel time matrix --------------------------------------------------
# read ttm
ttm_files <- list.files('E:/data/ttmatrix_fixed/2019', full.names = T)
ttm_for <- ttm_files[ ttm_files %like% sigla]


dt <- readRDS(ttm_for)
head(dt)

ttm <- subset(dt,
              mode =='transit' &
              pico == 1 &
              ano == 2019 &
              city  == sigla &
              travel_time <= 120)


ttm[, c('pico', 'city', 'ano', 'mode') := NULL]
names(ttm) <- c('from_id', 'to_id', 'travel_time')
head(ttm)

ttm <- subset(ttm, from_id %in% unique(inter$id_hex) )
ttm <- subset(ttm, to_id %in% unique(inter$id_hex) )



########## land use  --------------------------------------------------
land_for <- aopdata::read_landuse(city=sigla, year=2019, geometry = F)
land_for <- land_for[, .(id_hex, P001, T001, E001)]
names(land_for) <- c('id', 'population', 'jobs', 'schools')


# merge opportunities data
ttm[land_for, on=c('from_id'='id'), population   := i.population  ]
ttm[land_for, on=c('to_id'='id'), jobs  := i.jobs ]
ttm[land_for, on=c('to_id'='id'), schools  := i.schools ]


# ttm2 <- subset(ttm, population > 0 | jobs >0 | schools >0 )
head(ttm)
summary(ttm)



########## grid  --------------------------------------------------
grid <- aopdata::read_grid(city = sigla)

grid$abbrev_muni <- NULL
grid$name_muni  <- NULL
grid$code_muni <- NULL
names(grid)[1] <- 'id'

grid2 <- subset(grid, id %in% unique(inter$id_hex) )
grid2 <- unique(grid2)
class(grid2)
plot(grid2)
head(grid2)

# check
setdiff(ttm$from_id, grid2$id)
setdiff(ttm$to_id, grid2$id)
setdiff(grid2$id, ttm$to_id)

unique(ttm$to_id) |> length()
unique(ttm$from_id) |> length()
unique(grid2$id) |> length()


# save files
saveRDS(ttm, './inst/extdata/ttm_bho.rds', compress = TRUE)
saveRDS(grid2, './inst/extdata/grid_bho.rds', compress = T)

#' @param active A logical. When `TRUE`, the function calculates active
#'        accessibility (the quantity of opportunities that can be reached from
#'        a given origin). when `FALSE`, it calculates passive accessibility (by
#'        how many people each destination can be reached), which is equivalent
#'        to the notion of market potential.

#' Binary (a.k.a. step) decay function
#'
#' Returns a binary weighting function (frequently used to calculate cumulative
#' opportunities measures) to be used inside accessibility calculating
#' functions.
#' @template description_generic_cost
#'
#' @param cutoff A `numeric` vector. The numbers indicating the travel cost
#'   cutoffs.
#'
#' @template return_decay_function
#'
#' @family decay functions
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' weighting_function <- decay_binary(cutoff = 30)
#'
#' weighting_function(c(20, 35))
#'
#' weighting_function <- decay_binary(cutoff = c(30, 45))
#'
#' weighting_function(c(20, 35))
#'
#' @export
decay_binary <- function(cutoff) {
  checkmate::assert_numeric(
    cutoff,
    lower = 0,
    finite = TRUE,
    any.missing = FALSE,
    min.len = 1,
    unique = TRUE
  )

  weighting_function <- function(travel_cost) {
    weights_list <- lapply(
      cutoff,
      function(x) {
        weights <- as.integer(travel_cost <= x)
        weights
      }
    )
    names(weights_list) <- cutoff

    return(weights_list)
  }

  return(weighting_function)
}

tester <- function(cutoff = 20) decay_binary(cutoff)

test_that("adequately raises errors", {
  expect_error(tester("banana"))
  expect_error(tester(-1))
  expect_error(tester(Inf))
  expect_error(tester(NA))
  expect_error(tester(numeric(0)))
  expect_error(tester(c(1, 1)))
})

test_that("output is a decay function that returns a numeric", {
  expect_is(tester(), "function")

  output_fn <- tester()
  output_list <- output_fn(0)
  expect_is(output_list, "list")
  expect_length(output_list, 1)
  expect_named(output_list, "20")

  expect_equal(
    output_fn(c(10, 20, 40, 60))[["20"]],
    c(1, 1, 0, 0)
  )

  output_fn <- tester(40)
  output_list <- output_fn(0)
  expect_is(output_list, "list")
  expect_length(output_list, 1)
  expect_named(output_list, "40")

  expect_equal(
    output_fn(c(10, 20, 40, 60))[["40"]],
    c(1, 1, 1, 0)
  )

  output_fn <- tester(c(20, 40))
  output_list <- output_fn(0)
  expect_is(output_list, "list")
  expect_length(output_list, 2)
  expect_named(output_list, c("20", "40"))

  expect_equal(
    output_fn(c(10, 20, 40, 60))[["20"]],
    c(1, 1, 0, 0)
  )
  expect_equal(
    output_fn(c(10, 20, 40, 60))[["40"]],
    c(1, 1, 1, 0)
  )
})

test_that("output fn returns empty integer if receives empty numeric/integer", {
  output_fn <- tester(c(20, 40))

  expect_identical(output_fn(integer())[["20"]], integer())
  expect_identical(output_fn(integer())[["40"]], integer())

  expect_identical(output_fn(numeric())[["20"]], integer())
  expect_identical(output_fn(numeric())[["40"]], integer())
})

apply_gravity_measure <- function(data, decay_function, travel_cost) {
  .cost_colname <- travel_cost

  # the decay function can come in two "formats"
  # - a function that returns a numeric vector
  # - a function that returns a list of numeric vectors
  # we have first to figure out which function we're dealing with, and then use
  # it accordingly

  decay_fn_output <- decay_function(1)

  is_list_length_1 <- is.list(decay_fn_output) && length(decay_fn_output) == 1
  is_numeric_like <- is.numeric(decay_fn_output) || is.integer(decay_fn_output)

  if (is_numeric_like || is_list_length_1) {
    data[, opp_weight := decay_function(get(.cost_colname))]

    access <- data
  } else {
    opp_weights <- decay_function(data[[travel_cost]])

    access <- lapply(
      opp_weights,
      function(opp_weight) cbind(data, opp_weight)
    )
    access <- data.table::rbindlist(access, idcol = "decay_function_arg")
    access[, decay_function_arg := try_to_convert_to_num(decay_function_arg)]
  }

  return(access[])
}


try_to_convert_to_num <- function(x) {
  result <- tryCatch(
    as.numeric(x),
    warning = function(cnd) cnd
  )

  if (inherits(result, "warning")) return(x)

  return(result)
}

# if running manually, please run the following line first:
# source("tests/testthat/setup.R")

# creating this copy of travel_matrix because apply_gravity_measure() can modify
# the original object by reference
data <- data.table::as.data.table(travel_matrix)
data[, mode := NULL]
data <- unique(data)

tester <- function(data = get("data", envir = parent.frame()),
                   decay_function = decay_exponential(0.5),
                   travel_cost = "travel_time") {
  apply_gravity_measure(data, decay_function, travel_cost)
}

legacy_decay_exponential <- function(decay_value) {
  function(travel_cost) {
    weights <- exp(-decay_value * travel_cost)
    return(weights)
  }
}

test_that("mod by ref when input fn returns num vector or list with length 1", {
  expect_named(data, c("from_id", "to_id", "travel_time"))

  result <- tester()
  expect_identical(result, data)
  expect_named(data, c("from_id", "to_id", "travel_time", "opp_weight"))
  data[, opp_weight := NULL]

  result <- tester(decay_function = legacy_decay_exponential(0.5))
  expect_identical(result, data)
  expect_named(data, c("from_id", "to_id", "travel_time", "opp_weight"))
  data[, opp_weight := NULL]
})

test_that("doesnt mod by ref when input returns list length > 1 and add col", {
  expect_named(data, c("from_id", "to_id", "travel_time"))

  result <- tester(decay_function = decay_exponential(c(0.5, 0.6)))
  expected_result <- cbind(
    decay_function_arg = rep(c(0.5, 0.6), each = nrow(data)),
    rbind(data, data),
    opp_weight = c(
      legacy_decay_exponential(0.5)(data$travel_time),
      legacy_decay_exponential(0.6)(data$travel_time)
    )
  )
  expect_identical(result, expected_result)

  expect_named(data, c("from_id", "to_id", "travel_time"))
})

#' @param fill_missing_ids A `logical`. When calculating grouped accessibility
#'   estimates (i.e. when `by_col` is not `NULL`), some combinations of groups
#'   and origins may be missing. For example, if a single trip can depart from
#'   origin `A` at 7:15am and reach destination `B` within 55 minutes, but no
#'   trips departing from `A` at 7:30am can be completed at all, this second
#'   combination will not be included in the output. When `TRUE` (the default),
#'   the function identifies which combinations would be left out and fills
#'   their respective accessibility values with 0, which incurs in a
#'   performance penalty.

#' Negative exponential decay function
#'
#' Returns a negative exponential weighting function to be used inside
#' accessibility calculating functions.
#' @template description_generic_cost
#'
#' @param decay_value A `numeric` vector. The calibration parameters that, when
#'   multiplied by the travel cost, are used as the exponent of `e` in the
#'   negative exponential function.
#'
#' @template return_decay_function
#'
#' @family decay functions
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' weighting_function <- decay_exponential(decay_value = 0.1)
#'
#' weighting_function(c(20, 30))
#'
#' weighting_function <- decay_exponential(decay_value = c(0.1, 0.2))
#'
#' weighting_function(c(20, 30))
#'
#' @export
decay_exponential <- function(decay_value) {
  checkmate::assert_numeric(
    decay_value,
    lower = 0,
    finite = TRUE,
    any.missing = FALSE,
    min.len = 1,
    unique = TRUE
  )

  weighting_function <- function(travel_cost) {
    weights_list <- lapply(
      decay_value,
      function(x) exp(-x * travel_cost)
    )
    names(weights_list) <- decay_value

    return(weights_list)
  }

  return(weighting_function)
}

### setup ------------------------

library(accessibility)
library(data.table)
library(ggplot2)
library(sf)
library(interp)
library(sfheaders)

library(dplyr)

### prepate input data ---------------------------------
ttm_path <- system.file("extdata/travel_matrix.rds", package = "accessibility")
grid_path <- system.file("extdata/grid_bho.rds", package = "accessibility")
landuse_path <- system.file("extdata/land_use_data.rds", package = "accessibility")
ttm <- readRDS(ttm_path)
grid <- readRDS(grid_path)
landuse <- readRDS(landuse_path)


a <- left_join(landuse, grid)
a <- st_sf(a)
class(a)
ggplot() +
  geom_sf(data=a, aes(fill=schools), color=NA, show.legend = F) +
  scale_fill_viridis_c(direction = -1) +
  theme_void() +
  theme(panel.grid.major=element_line(colour="transparent"))

### prepate decay functions ---------------------------------

binary <- decay_binary(cutoff = 30)
linear <- decay_linear(cutoff = 60)
inverse_power <- decay_power(decay_value = 0.8)




# Active accessibility: number of schools accessible from each origin

df_tmi <- accessibility::cost_to_closest(travel_matrix = ttm,
                                 land_use_data = landuse,
                                 travel_cost = 'travel_time',
                                 opportunity = 'schools')

df_cum <- accessibility::gravity(travel_matrix = ttm,
                                 land_use_data = landuse,
                                 travel_cost = 'travel_time',
                                 opportunity = 'schools',
                                 decay_function = binary)

df_lin <- accessibility::gravity(travel_matrix = ttm,
                                 land_use_data = landuse,
                                 travel_cost = 'travel_time',
                                 opportunity = 'schools',
                                 decay_function = linear)

df_gra <- accessibility::gravity(travel_matrix = ttm,
                                           land_use_data = landuse,
                                           travel_cost = 'travel_time',
                                           opportunity = 'schools',
                                            decay_function = inverse_power)
df_cum$decay <- 'binary_30'
df_lin$decay <- 'linear_60'
df_gra$decay <- 'negexp_.1'

df <- rbind(df_cum, df_lin, df_gra)

# access
df2 <- df_lin[setDT(grid), on=c('id'='id'), geom := i.geom]

df2 <- st_sf(df2)



ggplot() +
  geom_sf(data=df2, aes(fill=schools), color=NA, show.legend = F) +
  scale_fill_viridis_c(option = 'inferno') +
  # facet_wrap(.~decay) +
  theme_void() +
  theme(panel.grid.major=element_line(colour="transparent"))
ggsave('aaa.png', dpi=300, width = 10, height = 10, units = 'cm')

#' Spatial availability
#'
#' Calculates spatial availability, an accessibility measured proposed by
#' \insertCite{soukhov2023introducing;textual}{accessibility} that takes into
#' account competition effects. The accessibility levels that result from using
#' this measure are proportional both to the demand in each origin and to the
#' travel cost it takes to reach the destinations. The spatial availability is
#' a particular case of `constrained_accessibility(constraint = "singly")`.
#' @template description_generic_cost
#'
#' @template travel_matrix
#' @template land_use_data
#' @template opportunity
#' @template travel_cost
#' @template demand
#' @template decay_function
#' @param alpha A `numeric`. A parameter used to modulate the effect of demand
#'   by population. When less than 1, opportunities are allocated more rapidly
#'   to smaller centers relative to larger ones; values higher than 1 achieve
#'   the opposite effect.
#' @template group_by
#' @template fill_missing_ids_combinations
#' @param detailed_results A `logical`. Whether to return spatial availability
#'   results aggregated by origin-destination pair (`TRUE`) or by origin
#'   (`FALSE`, the default). When `TRUE`, the output also includes the demand,
#'   impedance and combined balancing factors used to calculate spatial
#'   availability. Please note that the argument `fill_missing_ids`
#'   does not affect the output when `detailed_results` is `TRUE`.
#'
#' @template return_accessibility
#'
#' @references
#' \insertAllCited{}
#'
#' @family Constrained accessibility
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' # the example below is based on Soukhov et al. (2023) paper
#'
#' travel_matrix <- data.table::data.table(
#'   from_id = rep(c("A", "B", "C"), each = 3),
#'   to_id = as.character(rep(1:3, 3)),
#'   travel_time = c(15, 30, 100, 30, 15, 100, 100, 100, 15)
#' )
#' land_use_data <- data.table::data.table(
#'   id = c("A", "B", "C", "1", "2", "3"),
#'   population = c(50000, 150000, 10000, 0, 0, 0),
#'   jobs = c(0, 0, 0, 100000, 100000, 10000)
#' )
#'
#' df <- spatial_availability(
#'   travel_matrix,
#'   land_use_data,
#'   opportunity = "jobs",
#'   travel_cost = "travel_time",
#'   demand = "population",
#'   decay_function = decay_exponential(decay_value = 0.1)
#' )
#' df
#'
#' detailed_df <- spatial_availability(
#'   travel_matrix,
#'   land_use_data,
#'   opportunity = "jobs",
#'   travel_cost = "travel_time",
#'   demand = "population",
#'   decay_function = decay_exponential(decay_value = 0.1),
#'   detailed_results = TRUE
#' )
#' detailed_df
#'
#' @export
spatial_availability <- function(travel_matrix,
                                 land_use_data,
                                 opportunity,
                                 travel_cost,
                                 demand,
                                 decay_function,
                                 alpha = 1,
                                 group_by = character(0),
                                 fill_missing_ids = TRUE,
                                 detailed_results = FALSE) {
  checkmate::assert_string(opportunity)
  checkmate::assert_string(travel_cost)
  checkmate::assert_string(demand)
  checkmate::assert_number(alpha, lower = 0, finite = TRUE)
  checkmate::assert_logical(detailed_results, len = 1, any.missing = FALSE)
  assert_decay_function(decay_function)
  assert_group_by(group_by)
  assert_detailed_fill_missing_ids(fill_missing_ids, detailed_results)
  assert_travel_matrix(travel_matrix, travel_cost, group_by)
  assert_land_use_data(
    land_use_data,
    travel_matrix,
    opportunity,
    demand = demand
  )

  if (!inherits(travel_matrix, "data.table")) {
    original_class <- class(travel_matrix)
    data <- data.table::as.data.table(travel_matrix)
  } else {
    data <- data.table::copy(travel_matrix)
  }

  if (!inherits(land_use_data, "data.table")) {
    land_use_data <- data.table::as.data.table(land_use_data)
  }

  merge_by_reference(data, land_use_data, opportunity, left_df_idcol = "to_id")
  merge_by_reference(data, land_use_data, demand, left_df_idcol = "from_id")

  data <- apply_gravity_measure(data, decay_function, travel_cost)

  groups <- c("from_id", group_by)
  if ("decay_function_arg" %in% names(data)) {
    groups <- c(groups, "decay_function_arg")
    group_by <- c(group_by, "decay_function_arg")
  }

  warn_extra_cols(
    travel_matrix,
    travel_cost,
    group_id = "from_id",
    groups = groups
  )

  .demand_colname <- demand
  total_demand <- sum(land_use_data[[.demand_colname]] ^ alpha)
  data[, demand_bal_fac := (get(.demand_colname) ^ alpha) / total_demand]

  data[
    ,
    impedance_bal_fac := opp_weight / sum(opp_weight),
    by = c("to_id", group_by)
  ]

  data[
    ,
    combined_bal_fac := demand_bal_fac * impedance_bal_fac /
      sum(demand_bal_fac * impedance_bal_fac),
    by = c("to_id", group_by)
  ]

  .opportunity_colname <- opportunity
  data[
    ,
    spatial_availability := combined_bal_fac * get(.opportunity_colname),
    by = c("to_id", group_by)
  ]

  if (detailed_results) {
    cols_to_drop <- c(travel_cost, opportunity, demand, "opp_weight")
    access <- data[, (cols_to_drop) := NULL]

    data.table::setnames(access, "spatial_availability", opportunity)
  } else {
    access <- data[
      ,
      .(access = sum(spatial_availability)),
      by = c("from_id", group_by)
    ]

    if (fill_missing_ids) {
      access <- fill_missing_ids(access, travel_matrix, groups)
    }

    data.table::setnames(access, c("from_id", "access"), c("id", opportunity))
  }

  if (exists("original_class")) class(access) <- original_class

  return(access[])
}

library(accessibility)

# load a travel time matrix data in long format
data_path <- system.file("extdata/ttm_bho.rds", package = "accessibility")
ttm <- readRDS(data_path)
ttm <- rbind(ttm, ttm, ttm, ttm, ttm)
ttm <- rbind(ttm, ttm, ttm, ttm, ttm)
ttm <- rbind(ttm, ttm, ttm, ttm, ttm)
nrow(ttm)

system.time(df <- gravity_access(data = ttm,
                                 opportunity_col = 'schools',
                                 decay_function = decay_exponential(decay_value = 0.2),
                                 travel_cost_col='travel_time',
                                 by_col = 'from_id'
))

system.time(
  df <- floating_catchment_area(
    data = ttm,
    fca_metric = '2SFCA',
    orig_col = 'from_id',
    dest_col = 'to_id',
    opportunity_col = 'jobs',
    population_col = 'population',
    decay_function = decay_binary(cutoff = 50),
    travel_cost_col = 'travel_time'
  )
  )
26.12



##### example map ------------------------
library(accessibility)
library(data.table)
library(ggplot2)
library(sf)


ttm_path <- system.file("extdata/ttm_bho.rds", package = "accessibility")
grid_path <- system.file("extdata/grid_bho.rds", package = "accessibility")
ttm <- readRDS(ttm_path)
grid <- readRDS(grid_path)

setdiff(ttm$from_id, grid$id)

unique(ttm$from_id) |> length()
unique(grid$id) |> length()

# Active accessibility: number of schools accessible from each origin
df <- cumulative_time_cutoff(data = ttm,
                             opportunity_colname = 'jobs',
                             cutoff = 30,
                             by_colname = 'from_id')

df <- accessibility::gravity_access(data = ttm,
                             opportunity_colname = 'jobs',
                             decay_function = 'inverse_power',
                             decay_value = .5,
                             by_colname = 'from_id')

# access
df2 <- df[setDT(grid), on=c('from_id'='id'), geom := i.geom]

df2 <- st_sf(df2)
ggplot() +
  geom_sf(data=df2, aes(fill=access), color=NA) +
  scale_fill_viridis_c()



# isochone
id <- ttm[, .(t_min = mean(travel_time)), by = 'from_id']
id <- id[min(t_min)]$from_id
df <- ttm[ from_id ==id, ]
df2 <- df[setDT(grid), on=c('to_id'='id'), geom := i.geom]


df2 <- st_sf(df2)
ggplot() +
  geom_sf(data=df2, aes(fill=travel_time), color=NA) +
  scale_fill_viridis_c()




##### Coverage ------------------------
library(covr)
library(testthat)
Sys.setenv(NOT_CRAN = "true")


# the whole package
Sys.setenv(NOT_CRAN = "true")
cov <- covr::package_coverage(path = ".", type = "tests")
cov


x <- as.data.frame(cov)



# each function separately

a <- covr::function_coverage(fun=accessibility::cumulative_time_cutoff, test_file("tests/testthat/test-cumulative_cutoff.R"))
a <- covr::function_coverage(fun=accessibility::cumulative_time_interval, test_file("tests/testthat/test-cumulative_interval.R"))
a <- covr::function_coverage(fun=accessibility::time_to_closest, test_file("tests/testthat/test-time_to_closest.R"))
a <- covr::function_coverage(fun=accessibility::gravity_access, test_file("tests/testthat/test-gravity_access.R"))
a <- covr::function_coverage(fun=accessibility::floating_catchment_area, test_file("tests/testthat/test-floating_catchment_area.R"))
a <- covr::function_coverage(fun=accessibility::decay_binary, test_file("tests/testthat/test-decay_binary.R"))
a <- covr::function_coverage(fun=accessibility::decay_power, test_file("tests/testthat/test-decay_power.R"))
a <- covr::function_coverage(fun=accessibility::decay_exponential, test_file("tests/testthat/test-decay_exponential.R"))
a <- covr::function_coverage(fun=accessibility::decay_linear, test_file("tests/testthat/test-decay_linear.R"))
a <- covr::function_coverage(fun=accessibility::constrained_accessibility, test_file("tests/testthat/test-constrained-wrapper.R"))
a <- covr::function_coverage(fun=accessibility::constrained_accessibility, test_file("tests/testthat/test-constrained-essentials.R"))
a

zeroCov <- covr::zero_coverage(a)




##### Profiling function ------------------------
# p <-   profvis( update_newstoptimes("T2-1@1#2146") )
#
# p <-   profvis( b <- corefun("T2-1") )





# checks spelling
library(spelling)
devtools::spell_check(pkg = ".", vignettes = TRUE, use_wordlist = TRUE)

# Update documentation
devtools::document(pkg = ".")


# Write package manual.pdf
system("R CMD Rd2pdf --title=Package gtfs2gps --output=./gtfs2gps/manual.pdf")
# system("R CMD Rd2pdf gtfs2gps")






### CMD Check ----------------
# Check package errors
library(tictoc)


# LOCAL
tictoc::tic()
Sys.setenv(NOT_CRAN = "true")
devtools::check(pkg = ".",  cran = FALSE, env_vars = c(NOT_CRAN = "true"))
tictoc::toc()

#
#
# # CRAN
# tictoc::tic()
# Sys.setenv(NOT_CRAN = "false")
# devtools::check(pkg = ".",  cran = TRUE, env_vars = c(NOT_CRAN = "false"))
# tictoc::toc()


devtools::check_win_release(pkg = ".")

# devtools::check_win_oldrelease()
# devtools::check_win_devel()


beepr::beep()



tictoc::tic()
devtools::check(pkg = ".",  cran = TRUE, env_vars = c(NOT_CRAN = "false"))
tictoc::toc()





# build binary -----------------
system("R CMD build . --resave-data") # build tar.gz




# submit to CRAN -----------------
# usethis::use_cran_comments('teste 2222, , asdadsad')


devtools::submit_cran()

fill_missing_ids <- function(access_df, travel_matrix, groups) {
  unique_values <- lapply(groups, function(x) unique(travel_matrix[[x]]))
  names(unique_values) <- groups

  if ("decay_function_arg" %in% groups) {
    unique_values$decay_function_arg <- unique(access_df$decay_function_arg)
  }
  possible_combinations <- do.call(data.table::CJ, unique_values)

  if (nrow(access_df) < nrow(possible_combinations)) {
    access_df <- do_fill_missing_ids(access_df, possible_combinations, groups)
  }

  return(access_df)
}


do_fill_missing_ids <- function(access_df,
                                possible_combinations,
                                groups,
                                access_col = "access",
                                fill_value = 0) {
  filled_access_df <- merge(
    possible_combinations,
    access_df,
    by = groups,
    all.x = TRUE
  )

  filled_access_df[
    is.na(get(access_col)),
    eval(access_col, envir = parent.frame()) := fill_value
  ]

  return(filled_access_df[])
}

tester <- function(access_df,
                   possible_combinations,
                   groups,
                   access_col = "access",
                   fill_value = 0) {
  do_fill_missing_ids(
    access_df,
    possible_combinations,
    groups,
    access_col,
    fill_value
  )
}

test_that("fill correctly with one id column", {
  access_df <- data.table::data.table(
    from_id = c("1", "2", "5"),
    access = c(10, 10, 10)
  )
  possible_combinations <- data.table::data.table(from_id = as.character(1:5))
  groups <- "from_id"

  result <- tester(access_df, possible_combinations, groups)
  data.table::setkey(result, NULL)
  expect_equal(
    result,
    data.table::data.table(
      from_id = as.character(1:5),
      access = c(10, 10, 0, 0, 10)
    )
  )
})

test_that("fill correctly with two id columns", {
  access_df <- data.table::data.table(
    from_id = c("1", "2", "5", "1"),
    mode = c("mode1", "mode1", "mode1", "mode2"),
    access = c(10, 10, 10, 20)
  )
  possible_combinations <- data.table::CJ(
    from_id = as.character(1:5),
    mode = c("mode1", "mode2")
  )
  groups <- c("from_id", "mode")

  result <- tester(access_df, possible_combinations, groups)
  data.table::setkey(result, NULL)
  expect_equal(
    result,
    data.table::data.table(
      from_id = rep(as.character(1:5), each = 2),
      mode = rep(c("mode1", "mode2"), times = 5),
      access = c(10, 20, 10, 0, 0, 0, 0, 0, 10, 0)
    )
  )
})

test_that("fills correct access_col", {
  access_df <- data.table::data.table(
    from_id = c("1", "2", "5"),
    oi = c(10, 10, 10)
  )
  possible_combinations <- data.table::data.table(from_id = as.character(1:5))
  groups <- "from_id"

  result <- tester(access_df, possible_combinations, groups, access_col = "oi")
  data.table::setkey(result, NULL)
  expect_equal(
    result,
    data.table::data.table(
      from_id = as.character(1:5),
      oi = c(10, 10, 0, 0, 10)
    )
  )
})

test_that("fills with correct fill_value", {
  access_df <- data.table::data.table(
    from_id = c("1", "2", "5"),
    access = c(10, 10, 10)
  )
  possible_combinations <- data.table::data.table(from_id = as.character(1:5))
  groups <- "from_id"

  result <- tester(access_df, possible_combinations, groups, fill_value = NA)
  data.table::setkey(result, NULL)
  expect_equal(
    result,
    data.table::data.table(
      from_id = as.character(1:5),
      access = c(10, 10, NA, NA, 10)
    )
  )
})

library(data.table)
library(lpSolve)
library(accessibility)

data_dir <- system.file("extdata", package = "accessibility")
travel_matrix <- readRDS(file.path(data_dir, "travel_matrix.rds"))
land_use_data <- readRDS(file.path(data_dir, "land_use_data.rds"))


df <- accessibility:::merge_by_reference(data = travel_matrix,
                                         land_use_data = land_use_data,
                                         left_df_idcol = 'from_id',
                                         opportunity = 'population')


df <- accessibility:::merge_by_reference(data = df,
                                         land_use_data = land_use_data,
                                         left_df_idcol = 'to_id'
                                         , opportunity = 'schools'
)
head(df)


names(df) <- c('origin_id',
               'destination_id',
               'travel_time',
               'origin_population',
               'destination_facilities')

rowpos <- sample(1:nrow(df), size = 500, replace = T) |> sort()
data <- df[rowpos, ]

result <- solve_p_median(df, p = 2)

######################
data <- data.table(
  origin_id = rep(1:3, each = 3),
  destination_id = rep(1:3, times = 3),
  travel_time = c(10, 20, 30, 15, 25, 35, 20, 30, 40),
  origin_population = rep(c(100, 150, 200), each = 3),
  destination_facilities = rep(0, 9)  # Assuming no facilities are open yet
)

# Solve the p-Median Problem for p = 2
result <- solve_p_median(data, p = 2)

# View the results
print(result$total_cost)
print(result$assignment)
print(result$facilities_opened)


solve_p_median <- function(data, p) {
  # Ensure data is a data.table
  if (!is.data.table(data)) {
    stop("Input data must be a data.table.")
  }

  # Check required columns
  required_cols <- c("origin_id", "destination_id", "travel_time",
                     "origin_population", "destination_facilities")
  if (!all(required_cols %in% names(data))) {
    stop("Data must contain the following columns: ", paste(required_cols, collapse = ", "))
  }

  # Get unique origins and destinations
  origins <- unique(data$origin_id)
  destinations <- unique(data$destination_id)

  # Number of origins and destinations
  n_origins <- length(origins)
  n_destinations <- length(destinations)

  # Create index mappings
  origin_index <- setNames(1:n_origins, origins)
  destination_index <- setNames(1:n_destinations, destinations)

  # Demand at each origin
  demand <- data[, .(origin_population = mean(origin_population)), by = origin_id]
  demand <- demand[order(origin_id)]
  demand_vector <- demand$origin_population

  # Travel time matrix (origins x destinations)
  travel_time_matrix <- matrix(0, nrow = n_origins, ncol = n_destinations)
  for (i in 1:nrow(data)) {
    o <- origin_index[as.character(data$origin_id[i])]
    d <- destination_index[as.character(data$destination_id[i])]
    travel_time_matrix[o, d] <- data$travel_time[i]
  }

  # Decision variables:
  # x[i, j] = 1 if origin i is assigned to facility at destination j
  # y[j] = 1 if a facility is located at destination j

  # Objective: Minimize total weighted travel time
  # Minimize sum_{i in origins} sum_{j in destinations} demand[i] * travel_time[i, j] * x[i, j]

  # Constraints:
  # 1. Each origin is assigned to exactly one facility: sum_{j} x[i, j] = 1 for all i
  # 2. Assignment only to open facilities: x[i, j] <= y[j] for all i, j
  # 3. Number of facilities to open: sum_{j} y[j] = p
  # 4. y[j] are binary variables
  # 5. x[i, j] are binary variables

  # Number of variables:
  # Total x[i, j]: n_origins * n_destinations
  # Total y[j]: n_destinations
  total_vars <- n_origins * n_destinations + n_destinations

  # Objective coefficients
  obj_coeffs <- c(as.vector(travel_time_matrix * demand_vector), rep(0, n_destinations))

  # Variable types: binary
  var_types <- rep("binary", total_vars)

  # Constraints
  constraints <- list()
  constr_rhs <- c()
  constr_dir <- c()

  # Constraint 1: Each origin is assigned to exactly one facility
  for (i in 1:n_origins) {
    constr <- rep(0, total_vars)
    constr[((i - 1) * n_destinations + 1):(i * n_destinations)] <- 1
    constraints <- c(constraints, list(constr))
    constr_rhs <- c(constr_rhs, 1)
    constr_dir <- c(constr_dir, "==")
  }

  # Constraint 2: Assignment only to open facilities x[i, j] - y[j] <= 0
  for (i in 1:n_origins) {
    for (j in 1:n_destinations) {
      constr <- rep(0, total_vars)
      constr[(i - 1) * n_destinations + j] <- 1  # x[i, j]
      constr[n_origins * n_destinations + j] <- -1  # -y[j]
      constraints <- c(constraints, list(constr))
      constr_rhs <- c(constr_rhs, 0)
      constr_dir <- c(constr_dir, "<=")
    }
  }

  # Constraint 3: Number of facilities to open
  constr <- rep(0, total_vars)
  constr[(n_origins * n_destinations + 1):total_vars] <- 1  # y[j]
  constraints <- c(constraints, list(constr))
  constr_rhs <- c(constr_rhs, p)
  constr_dir <- c(constr_dir, "==")

  # Combine constraints
  constraint_matrix <- do.call(rbind, constraints)

  # Solve the problem
  result <- lp("min",
               objective.in = obj_coeffs,
               const.mat = constraint_matrix,
               const.dir = constr_dir,
               const.rhs = constr_rhs,
               all.bin = TRUE)

  if (result$status != 0) {
    stop("Optimal solution not found.")
  }

  # Extract the solution
  solution <- result$solution
  x_vars <- matrix(solution[1:(n_origins * n_destinations)], nrow = n_origins, byrow = TRUE)
  y_vars <- solution[(n_origins * n_destinations + 1):total_vars]

  # Prepare assignment results
  assignment <- data.table(
    origin_id = origins,
    assigned_destination_id = destinations[max.col(x_vars, ties.method = "first")]
  )

  # Facilities opened
  facilities_opened <- data.table(
    destination_id = destinations,
    facility_opened = as.logical(y_vars)
  )

  # Return results
  list(
    total_cost = result$objval,
    assignment = assignment,
    facilities_opened = facilities_opened
  )
}

#' Theil T Index
#'
#' Calculates the Theil T Index of a given accessibility distribution. Values
#' range from 0 (when all individuals have exactly the same accessibility
#' levels) to the natural log of *n*, in which *n* is the number of individuals
#' in the accessibility dataset. If the individuals can be classified into
#' mutually exclusive and completely exhaustive groups, the index can be
#' decomposed into a between-groups inequaliy component and a within-groups
#' component.
#'
#' @template accessibility_data
#' @param sociodemographic_data A data frame. The distribution of
#'   sociodemographic characteristics of the population in the study area cells.
#'   Must contain the columns `id` and any others specified in `population` and
#'   `socioeconomic_groups`.
#' @template opportunity_access
#' @template population
#' @param socioeconomic_groups A string. The name of the column in
#'   `sociodemographic_data` whose values identify the socioeconomic groups that
#'   should be used to calculate the between- and within-groups inequality
#'   levels. If `NULL` (the default), between- and within-groups components are
#'   not calculated and only the total aggregate inequality is returned.
#' @template group_by_access
#'
#' @return If `socioeconomic_groups` is `NULL`, a data frame containing the
#'   total Theil T estimates for the study area. If not, a list containing three
#'   dataframes: one summarizing the total inequality and the between- and
#'   within-groups components, one listing the contribution of each group to the
#'   between-groups component and another listing the contribution of each group
#'   to the within-groups component.
#'
#' @family inequality
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' data_dir <- system.file("extdata", package = "accessibility")
#' travel_matrix <- readRDS(file.path(data_dir, "travel_matrix.rds"))
#' land_use_data <- readRDS(file.path(data_dir, "land_use_data.rds"))
#'
#' access <- cumulative_cutoff(
#'   travel_matrix,
#'   land_use_data,
#'   cutoff = 30,
#'   opportunity = "jobs",
#'   travel_cost = "travel_time"
#' )
#'
#' ti <- theil_t(
#'   access,
#'   sociodemographic_data = land_use_data,
#'   opportunity = "jobs",
#'   population = "population"
#' )
#' ti
#'
#' # to calculate inequality between and within income deciles, we pass
#' # "income_decile" to socioeconomic_groups.
#' # some cells, however, are classified as in the decile NA because their
#' # income per capita is NaN, as they don't have any population. we filter
#' # these cells from our accessibility data, otherwise the output would include
#' # NA values (note that subsetting the data like this doesn't affect the
#' # assumption that groups are completely exhaustive, because cells with NA
#' # income decile don't have any population)
#'
#' na_decile_ids <- land_use_data[is.na(land_use_data$income_decile), ]$id
#' access <- access[! access$id %in% na_decile_ids, ]
#' sociodem_data <- land_use_data[! land_use_data$id %in% na_decile_ids, ]
#'
#' ti <- theil_t(
#'   access,
#'   sociodemographic_data = sociodem_data,
#'   opportunity = "jobs",
#'   population = "population",
#'   socioeconomic_groups = "income_decile"
#' )
#' ti
#'
#' @export
theil_t <- function(accessibility_data,
                    sociodemographic_data,
                    opportunity,
                    population,
                    socioeconomic_groups = NULL,
                    group_by = character(0)) {
  checkmate::assert_string(opportunity)
  checkmate::assert_string(population)
  checkmate::assert_string(socioeconomic_groups, null.ok = TRUE)
  assert_access_group_by(group_by)
  assert_accessibility_data(accessibility_data, opportunity, group_by)
  assert_sociodemographic_data(
    sociodemographic_data,
    accessibility_data,
    population = population,
    extra_cols = socioeconomic_groups
  )

  if (!inherits(accessibility_data, "data.table")) {
    original_class <- class(accessibility_data)
    data <- data.table::as.data.table(accessibility_data)
  } else {
    data <- data.table::copy(accessibility_data)
  }

  if (!inherits(sociodemographic_data, "data.table")) {
    sociodemographic_data <- data.table::as.data.table(sociodemographic_data)
  }

  merge_by_reference(
    data,
    sociodemographic_data,
    population,
    left_df_idcol = "id"
  )

  warn_extra_access_cols(accessibility_data, opportunity, group_by)

  if (is.null(socioeconomic_groups)) {
    result <- theil_without_groups(data, opportunity, population, group_by)
  } else {
    result <- theil_with_groups(
      data,
      sociodemographic_data,
      opportunity,
      population,
      socioeconomic_groups,
      group_by
    )
  }

  if (exists("original_class")) {
    if (inherits(result, "data.frame")) {
      class(result) <- original_class
    } else {
      class(result$summary) <- original_class
      class(result$within_group_component) <- original_class
      class(result$between_group_component) <- original_class
    }
  }

  return(result)
}

theil_without_groups <- function(data, opportunity, population, group_by) {
  # we have to filter by opportunity because opportunity = 0 results in
  # theil_t being NaN (because log(0) is NaN)
  # TODO: throw warning

  .opp_colname <- opportunity
  .groups <- group_by

  data[
    ,
    avg_access := stats::weighted.mean(
      get(..opportunity),
      w = get(..population)
    ),
    by = .groups
  ]

  theil_index <- data[
    get(.opp_colname) > 0,
    .(
      theil_t = calc_theil_t(get(..opportunity), avg_access, get(..population))
    ),
    by = .groups
  ]

  return(theil_index)
}

theil_with_groups <- function(data,
                              sociodemographic_data,
                              opportunity,
                              population,
                              socioeconomic_groups,
                              group_by) {
  merge_by_reference(
    data,
    sociodemographic_data,
    socioeconomic_groups,
    left_df_idcol = "id"
  )

  .groups <- group_by
  .socioecon_groups <- c(group_by, socioeconomic_groups)

  data[
    ,
    avg_access := stats::weighted.mean(
      get(..opportunity),
      w = get(..population)
    ),
    by = .groups
  ]
  data[
    ,
    group_avg_access := stats::weighted.mean(
      get(..opportunity),
      w = get(..population)
    ),
    by = .socioecon_groups
  ]

  summarized_data <- data[
    ,
    .(
      group_tot_access = as.numeric(
        sum(get(..opportunity) * get(..population))
      ),
      group_tot_pop = sum(get(..population)),
      group_theil_t = calc_theil_t(
        get(..opportunity),
        group_avg_access,
        get(..population)
      ),
      avg_access = mean(avg_access),
      group_avg_access = mean(group_avg_access)
    ),
    by = .socioecon_groups
  ]
  summarized_data[
    ,
    `:=`(
      group_avg_access = group_tot_access / group_tot_pop,
      access_share = group_tot_access / sum(group_tot_access),
      pop_share = group_tot_pop / sum(group_tot_pop)
    ),
    by = .groups
  ]

  # summary, containing total value and total within- and between-groups
  # inequality

  summary <- summarized_data[
    ,
    .(
      between_group = calc_theil_t(
        group_avg_access,
        avg_access,
        group_tot_pop
      ),
      within_group = sum(
        group_theil_t * group_tot_access / sum(group_tot_access)
      )
    ),
    by = .groups
  ]
  summary[, total := between_group + within_group]

  summary <- data.table::melt(
    summary,
    id.vars = .groups,
    variable.name = "component",
    value.name = "value"
  )
  summary[, component := as.character(component)]
  summary[, share_of_total := 2 * value / sum(value), by = .groups]

  # decomposed within group

  within_group <- data.table::copy(summarized_data)
  within_group[
    ,
    value := group_theil_t * group_tot_access / sum(group_tot_access),
    by = .groups
  ]
  within_group[
    ,
    share_of_component := value / sum(value),
    by = .groups
  ]

  cols_to_drop <- setdiff(
    names(within_group),
    c(.socioecon_groups, "value", "share_of_component")
  )
  within_group[, eval(cols_to_drop) := NULL]
  data.table::setorderv(within_group, .socioecon_groups)

  # decomposed between group

  between_group <- data.table::copy(summarized_data)
  between_group[, value := access_share * log(access_share / pop_share)]
  between_group[, eval(cols_to_drop) := NULL]
  data.table::setorderv(between_group, .socioecon_groups)

  output_list <- list(
    summary = summary,
    within_group_component = within_group,
    between_group_component = between_group
  )

  return(output_list)
}

calc_theil_t <- function(x, avg_x, weight) {
  stats::weighted.mean(x / avg_x * log(x / avg_x), w = weight)
}

# if running manually, please run the following line first:
# source("tests/testthat/setup.R")

na_decile_ids <- land_use_data[is.na(land_use_data$income_decile), ]$id
sociodem_data <- land_use_data[! id %in% na_decile_ids]

tester <- function(accessibility_data = small_access,
                   sociodemographic_data = sociodem_data,
                   opportunity = "jobs",
                   population = "population",
                   socioeconomic_groups = "income_decile",
                   group_by = "mode") {
  theil_t(
    accessibility_data,
    sociodemographic_data,
    opportunity,
    population,
    socioeconomic_groups,
    group_by
  )
}

test_that("raises errors due to incorrect input", {
  expect_error(tester(opportunity = 1))
  expect_error(tester(opportunity = c("schools", "jobs")))

  expect_error(tester(population = 1))
  expect_error(tester(population = c("schools", "jobs")))

  expect_error(tester(socioeconomic_groups = 1))
  expect_error(tester(socioeconomic_groups = c("income_decile", "jobs")))

  expect_error(tester(group_by = 1))
  expect_error(tester(group_by = NA))
  expect_error(tester(group_by = "id"))
  expect_error(tester(group_by = c("mode", "mode")))

  expect_error(tester(as.list(small_access)))
  expect_error(tester(small_access[, .(oi = id, jobs, mode)]))
  expect_error(tester(small_access[, .(id, oi = jobs, mode)]))
  expect_error(tester(small_access[, .(id, jobs, oi = mode)]))

  expect_error(tester(sociodemographic_data = as.list(sociodem_data)))
  expect_error(
    tester(
      sociodemographic_data = sociodem_data[
        ,
        .(oi = id, population, income_decile)
      ]
    )
  )
  expect_error(
    tester(
      sociodemographic_data = sociodem_data[
        ,
        .(id, oi = population, income_decile)
      ]
    )
  )
  expect_error(
    tester(
      sociodemographic_data = sociodem_data[
        ,
        .(id, population, oi = income_decile)
      ]
    )
  )
})

test_that("throws warning if accessibility_data has an extra col", {
  expect_warning(tester(group_by = character(0)))
})

# tests when socioeconomic_groups is not NULL

test_that("returns dataframes with same class as accessibility_data's", {
  result <- tester()
  expect_is(result, "list")
  expect_is(result$summary, "data.table")
  expect_is(result$within_group_component, "data.table")
  expect_is(result$between_group_component, "data.table")

  result <- tester(sociodemographic_data = as.data.frame(sociodem_data))
  expect_is(result, "list")
  expect_is(result$summary, "data.table")
  expect_is(result$within_group_component, "data.table")
  expect_is(result$between_group_component, "data.table")

  result <- tester(as.data.frame(small_access))
  expect_is(result, "list")
  expect_false(inherits(result$summary, "data.table"))
  expect_is(result$summary, "data.frame")
  expect_false(inherits(result$within_group_component, "data.table"))
  expect_is(result$within_group_component, "data.frame")
  expect_false(inherits(result$between_group_component, "data.table"))
  expect_is(result$between_group_component, "data.frame")

  result <- tester(as.data.frame(small_access), as.data.frame(sociodem_data))
  expect_is(result, "list")
  expect_false(inherits(result$summary, "data.table"))
  expect_is(result$summary, "data.frame")
  expect_false(inherits(result$within_group_component, "data.table"))
  expect_is(result$within_group_component, "data.frame")
  expect_false(inherits(result$between_group_component, "data.table"))
  expect_is(result$between_group_component, "data.frame")
})

test_that("result has correct structure", {
  result <- tester()

  expect_true(ncol(result$summary) == 4)
  expect_is(result$summary$mode, "character")
  expect_is(result$summary$component, "character")
  expect_is(result$summary$value, "numeric")
  expect_is(result$summary$share_of_total, "numeric")

  expect_true(ncol(result$within_group_component) == 4)
  expect_is(result$within_group_component$mode, "character")
  expect_is(result$within_group_component$income_decile, "factor")
  expect_is(result$within_group_component$value, "numeric")
  expect_is(result$within_group_component$share_of_component, "numeric")

  expect_true(ncol(result$between_group_component) == 3)
  expect_is(result$between_group_component$mode, "character")
  expect_is(result$between_group_component$income_decile, "factor")
  expect_is(result$between_group_component$value, "numeric")

  suppressWarnings(result <- tester(group_by = character(0)))

  expect_true(ncol(result$summary) == 3)
  expect_is(result$summary$component, "character")
  expect_is(result$summary$value, "numeric")
  expect_is(result$summary$share_of_total, "numeric")

  expect_true(ncol(result$within_group_component) == 3)
  expect_is(result$within_group_component$income_decile, "factor")
  expect_is(result$within_group_component$value, "numeric")
  expect_is(result$within_group_component$share_of_component, "numeric")

  expect_true(ncol(result$between_group_component) == 2)
  expect_is(result$between_group_component$income_decile, "factor")
  expect_is(result$between_group_component$value, "numeric")

  result <- tester(small_access[0])

  expect_true(nrow(result$summary) == 0)
  expect_true(ncol(result$summary) == 4)
  expect_is(result$summary$mode, "character")
  expect_is(result$summary$component, "character")
  expect_is(result$summary$value, "numeric")
  expect_is(result$summary$share_of_total, "numeric")

  expect_true(nrow(result$within_group_component) == 0)
  expect_true(ncol(result$within_group_component) == 4)
  expect_is(result$within_group_component$mode, "character")
  expect_is(result$within_group_component$income_decile, "factor")
  expect_is(result$within_group_component$value, "numeric")
  expect_is(result$within_group_component$share_of_component, "numeric")

  expect_true(nrow(result$between_group_component) == 0)
  expect_true(ncol(result$between_group_component) == 3)
  expect_is(result$between_group_component$mode, "character")
  expect_is(result$between_group_component$income_decile, "factor")
  expect_is(result$between_group_component$value, "numeric")

  suppressWarnings(result <- tester(small_access[0], group_by = character(0)))

  expect_true(ncol(result$summary) == 3)
  expect_is(result$summary$component, "character")
  expect_is(result$summary$value, "numeric")
  expect_is(result$summary$share_of_total, "numeric")

  expect_true(nrow(result$within_group_component) == 0)
  expect_true(ncol(result$within_group_component) == 3)
  expect_is(result$within_group_component$income_decile, "factor")
  expect_is(result$within_group_component$value, "numeric")
  expect_is(result$within_group_component$share_of_component, "numeric")

  expect_true(nrow(result$between_group_component) == 0)
  expect_true(ncol(result$between_group_component) == 2)
  expect_is(result$between_group_component$income_decile, "factor")
  expect_is(result$between_group_component$value, "numeric")
})

test_that("input data sets remain unchanged", {
  # with the exception of some indexes

  original_access_data <- cumulative_cutoff(
    smaller_matrix,
    land_use_data,
    opportunity = "jobs",
    travel_cost = "travel_time",
    cutoff = 30,
    group_by = "mode"
  )
  original_sociodem_data <- land_use_data[! id %in% na_decile_ids]

  result <- tester()

  data.table::setindexv(small_access, NULL)
  expect_equal(original_access_data, small_access)
  expect_equal(original_sociodem_data, original_sociodem_data)
})

test_that("theil t is correctly calculated", {
  selected_ids <- c(
    "89a88cd900bffff",
    "89a88cdb57bffff",
    "89a88cdb597ffff",
    "89a88cdb5b3ffff",
    "89a88cdb5cfffff"
  )
  access_data <- cumulative_cutoff(
    travel_matrix[from_id %in% selected_ids],
    land_use_data,
    opportunity = "jobs",
    travel_cost = "travel_time",
    cutoff = 30,
    group_by = "mode"
  )

  result <- tester(access_data)

  result$summary[, value := round(value, 4)]
  result$summary[, share_of_total := round(share_of_total, 4)]
  expected_result <- data.table::data.table(
    mode = rep(c("transit", "transit2"), 3),
    component = rep(c("between_group", "within_group", "total"), each = 2),
    value = rep(c(0.0515, 0.0337, 0.0852), each = 2),
    share_of_total = rep(c(0.6044, 0.3956, 1), each = 2)
  )
  expect_identical(result$summary, expected_result)

  result$within_group_component[, value := round(value, 4)]
  expected_result <- data.table::data.table(
    mode = rep(c("transit", "transit2"), each = 4),
    income_decile = rep(factor(c(1, 3, 6, 10), levels = 1:10), 2),
    value = rep(c(0, 0, 0, 0.0337), 2),
    share_of_component = rep(c(0, 0, 0, 1), 2)
  )
  expect_identical(result$within_group_component, expected_result)

  result$between_group_component[, value := round(value, 4)]
  expected_result <- data.table::data.table(
    mode = rep(c("transit", "transit2"), each = 4),
    income_decile = rep(factor(c(1, 3, 6, 10), levels = 1:10), 2),
    value = rep(c(0.0243, -0.0579, -0.0042, 0.0894), 2)
  )
  expect_identical(result$between_group_component, expected_result)

  access_data <- access_data[!(id == "89a88cd900bffff" & mode == "transit2")]
  result <- tester(access_data)

  result$summary[, value := round(value, 4)]
  result$summary[, share_of_total := round(share_of_total, 4)]
  expected_result <- data.table::data.table(
    mode = rep(c("transit", "transit2"), 3),
    component = rep(c("between_group", "within_group", "total"), each = 2),
    value = c(0.0515, 0.0021, 0.0337, 0.0357, 0.0852, 0.0378),
    share_of_total = c(0.6044, 0.0545, 0.3956, 0.9455, 1, 1)
  )
  expect_identical(result$summary, expected_result)

  result$within_group_component[, value := round(value, 4)]
  expected_result <- data.table::data.table(
    mode = c(rep("transit", 4), rep("transit2", 3)),
    income_decile = factor(c(1, 3, 6, 10, 1, 6, 10), levels = 1:10),
    value = c(0, 0, 0, 0.0337, 0, 0, 0.0357),
    share_of_component = c(0, 0, 0, 1, 0, 0, 1)
  )
  expect_identical(result$within_group_component, expected_result)

  result$between_group_component[, value := round(value, 4)]
  expected_result <- data.table::data.table(
    mode = c(rep("transit", 4), rep("transit2", 3)),
    income_decile = factor(c(1, 3, 6, 10, 1, 6, 10), levels = 1:10),
    value = c(0.0243, -0.0579, -0.0042, 0.0894, -0.0142, -0.0113, 0.0276)
  )
  expect_identical(result$between_group_component, expected_result)
})

test_that("works even if access_data and sociodem_data has specific colnames", {
  selected_ids <- c(
    "89a88cdb57bffff",
    "89a88cdb5b3ffff"
  )
  access_data <- cumulative_cutoff(
    travel_matrix[from_id %in% selected_ids],
    land_use_data,
    opportunity = "jobs",
    travel_cost = "travel_time",
    cutoff = 30,
    group_by = "mode"
  )
  expected_result <- tester(access_data)

  access_data[, opportunity := "oi"]
  result <- suppressWarnings(tester(access_data))
  expect_identical(expected_result, result)

  access_data[, opportunity := NULL]
  sociodem_data[, population_temp := population]
  sociodem_data[, population := 1]
  result <- tester(access_data, population = "population_temp")
  expect_identical(expected_result, result)

  sociodem_data[, population := population_temp]
  sociodem_data[, population_temp := NULL]
  sociodem_data[, socioeconomic_groups := "oi"]
  result <- tester(access_data)
  expect_identical(expected_result, result)

  sociodem_data[, socioeconomic_groups := NULL]
  access_data[, group_by := "oi"]
  result <- suppressWarnings(tester(access_data))
  expect_identical(expected_result, result)

  access_data[, group_by := NULL]
})

# tests when socioeconomic_groups is NULL

null_tester <- function(...) tester(..., socioeconomic_groups = NULL)

test_that("returns dataframes with same class as accessibility_data's", {
  result <- null_tester()
  expect_is(result, "data.table")

  result <- null_tester(sociodemographic_data = as.data.frame(sociodem_data))
  expect_is(result, "data.table")

  result <- null_tester(as.data.frame(small_access))
  expect_false(inherits(result, "data.table"))
  expect_is(result, "data.frame")

  result <- null_tester(
    as.data.frame(small_access),
    as.data.frame(sociodem_data)
  )
  expect_false(inherits(result, "data.table"))
  expect_is(result, "data.frame")
})

test_that("result has correct structure", {
  result <- null_tester()
  expect_true(ncol(result) == 2)
  expect_is(result$mode, "character")
  expect_is(result$theil_t, "numeric")

  suppressWarnings(result <- null_tester(group_by = character(0)))
  expect_true(ncol(result) == 1)
  expect_is(result$theil_t, "numeric")

  result <- null_tester(small_access[0])
  expect_true(nrow(result) == 0)
  expect_true(ncol(result) == 2)
  expect_is(result$mode, "character")
  expect_is(result$theil_t, "numeric")

  suppressWarnings(
    result <- null_tester(small_access[0], group_by = character(0))
  )
  expect_true(nrow(result) == 0)
  expect_true(ncol(result) == 1)
  expect_is(result$theil_t, "numeric")
})

test_that("input data sets remain unchanged", {
  original_access_data <- cumulative_cutoff(
    smaller_matrix,
    land_use_data,
    opportunity = "jobs",
    travel_cost = "travel_time",
    cutoff = 30,
    group_by = "mode"
  )
  original_sociodem_data <- land_use_data[! id %in% na_decile_ids]

  result <- null_tester()

  expect_equal(original_access_data, small_access)
  expect_equal(original_sociodem_data, original_sociodem_data)
})

test_that("theil t is correctly calculated", {
  selected_ids <- c(
    "89a88cd900bffff",
    "89a88cdb57bffff",
    "89a88cdb597ffff",
    "89a88cdb5b3ffff",
    "89a88cdb5cfffff"
  )
  access_data <- cumulative_cutoff(
    travel_matrix[from_id %in% selected_ids],
    land_use_data,
    opportunity = "jobs",
    travel_cost = "travel_time",
    cutoff = 30,
    group_by = "mode"
  )

  result <- null_tester(access_data)
  result[, theil_t := round(theil_t, 4)]

  expected_result <- data.table::data.table(
    mode = c("transit", "transit2"),
    theil_t = 0.0852
  )
  expect_identical(result, expected_result)

  access_data <- access_data[!(id == "89a88cdb5cfffff" & mode == "transit2")]
  result <- null_tester(access_data)
  result[, theil_t := round(theil_t, 4)]

  expected_result <- data.table::data.table(
    mode = c("transit", "transit2"),
    theil_t = c(0.0852, 0.0839)
  )
  expect_identical(result, expected_result)
})

test_that("works even if access_data and sociodem_data has specific colnames", {
  selected_ids <- c(
    "89a88cdb57bffff",
    "89a88cdb5b3ffff"
  )
  access_data <- cumulative_cutoff(
    travel_matrix[from_id %in% selected_ids],
    land_use_data,
    opportunity = "jobs",
    travel_cost = "travel_time",
    cutoff = 30,
    group_by = "mode"
  )
  expected_result <- null_tester(access_data)

  access_data[, opportunity := "oi"]
  result <- suppressWarnings(null_tester(access_data))
  expect_identical(expected_result, result)

  access_data[, opportunity := NULL]
  sociodem_data[, population_temp := population]
  sociodem_data[, population := 1]
  result <- null_tester(access_data, population = "population_temp")
  expect_identical(expected_result, result)

  sociodem_data[, population := population_temp]
  sociodem_data[, population_temp := NULL]
  access_data[, group_by := "oi"]
  result <- suppressWarnings(null_tester(access_data))
  expect_identical(expected_result, result)

  access_data[, group_by := NULL]
})

#' Balancing cost accessibility measure
#'
#' Calculates the balancing cost measure, which is defined as the travel cost
#' required to reach as many opportunities as the number of people in a given
#' origin. Originally proposed by
#'  \insertCite{barboza2021balancing;textual}{accessibility}, under the name
#' "balancing time".
#' @template description_generic_cost
#'
#' @template travel_matrix
#' @template land_use_data
#' @template opportunity
#' @template travel_cost
#' @template demand
#' @param cost_increment A number. The cost increment that should be used when
#'   defining the travel cost distribution from which the potential balancing
#'   costs will be picked. For example, an increment of 1 tends to suitable for
#'   travel time distributions, meaning that the function will first check if
#'   any origins reach their balancing cost with a travel time of 0 minutes,
#'   then 1 minute, 2 minutes, 3, 4, ..., etc. A increment of 1 might be too big
#'   for a distribution of monetary costs, on the other hand, which could
#'   possibly benefit from a smaller increment of 0.05, for example, resulting
#'   in the function looking for balancing costs first at a cost of 0, then
#'   0.05, 0.10, ..., etc. Defaults to 1.
#' @template group_by
#' @template fill_missing_ids_combinations
#'
#' @template return_accessibility
#' @return A data frame containing the accessibility estimates for each origin
#'   in the travel matrix. Origins marked with a `NA` balancing cost never reach
#'   as many opportunities as there is people residing in them, given the
#'   specified travel matrix.
#'
#' @references
#' \insertAllCited{}
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' data_dir <- system.file("extdata", package = "accessibility")
#' travel_matrix <- readRDS(file.path(data_dir, "travel_matrix.rds"))
#' land_use_data <- readRDS(file.path(data_dir, "land_use_data.rds"))
#'
#' bc <- balancing_cost(
#'   travel_matrix,
#'   land_use_data,
#'   opportunity = "jobs",
#'   travel_cost = "travel_time",
#'   demand = "population"
#' )
#' head(bc)
#' @export
balancing_cost <- function(travel_matrix,
                           land_use_data,
                           opportunity,
                           travel_cost,
                           demand,
                           cost_increment = 1,
                           group_by = character(0),
                           fill_missing_ids = TRUE) {
  checkmate::assert_string(opportunity)
  checkmate::assert_string(travel_cost)
  checkmate::assert_string(demand)
  checkmate::assert_number(cost_increment, lower = 0, finite = TRUE)
  checkmate::assert_logical(fill_missing_ids, len = 1, any.missing = FALSE)
  assert_group_by(group_by)
  assert_travel_matrix(travel_matrix, travel_cost, group_by)
  assert_land_use_data(
    land_use_data,
    travel_matrix,
    opportunity,
    demand = demand
  )

  variable <- 3

  # if not a dt, keep original class to assign later when returning result

  if (!inherits(travel_matrix, "data.table")) {
    original_class <- class(travel_matrix)
    data <- data.table::as.data.table(travel_matrix)
  } else {
    data <- data.table::copy(travel_matrix)
  }

  if (!inherits(land_use_data, "data.table")) {
    land_use_data <- data.table::as.data.table(land_use_data)
  }

  merge_by_reference(data, land_use_data, opportunity, left_df_idcol = "to_id")

  groups <- c("from_id", group_by)
  warn_extra_cols(travel_matrix, travel_cost, "from_id", groups)

  lookup_vec <- land_use_data[[demand]]
  names(lookup_vec) <- land_use_data$id

  max_lookup_cost <- if (length(data[[travel_cost]]) > 0) {
    max(data[[travel_cost]], na.rm = TRUE)
  } else {
    0
  }
  lookup_costs <- seq.int(0, max_lookup_cost, by = cost_increment)

  # origins with 0 people should always have a balancing time of 0

  .demand_colname <- demand
  no_pop_origins <- land_use_data[get(.demand_colname) == 0]$id
  no_pop_origins_in_ttm <- no_pop_origins[no_pop_origins %in% data$from_id]

  grouped_no_pop_origins <- lapply(
    groups,
    function(group) {
      if (group == "from_id") return(no_pop_origins_in_ttm)
      unique(data[[group]])
    }
  )
  names(grouped_no_pop_origins) <- groups

  access <- do.call(data.table::CJ, grouped_no_pop_origins)
  access[, cost := 0]

  data <- data[!(from_id %in% no_pop_origins_in_ttm)]

  for (i in lookup_costs) {
    cum_opps <- calc_cum_cutoff(data, groups, opportunity, travel_cost, i)

    cum_opps[, did_balance := access > lookup_vec[from_id]]
    balanced <- cum_opps[did_balance == TRUE]

    if (nrow(balanced) > 0) {
      balanced[, c("access", "did_balance") := NULL]
      balanced[, cost := i]

      access <- rbind(access, balanced)

      data <- data[!balanced, on = groups]

      if (nrow(data) == 0) break
    }
  }

  if (fill_missing_ids) {
    unique_values <- lapply(groups, function(x) unique(travel_matrix[[x]]))
    names(unique_values) <- groups

    possible_combinations <- do.call(data.table::CJ, unique_values)

    if (nrow(access) < nrow(possible_combinations)) {
      access <- do_fill_missing_ids(
        access,
        possible_combinations,
        groups = groups,
        access_col = "cost",
        fill_value = NA
      )
    }
  }

  data.table::setnames(access, c("from_id", "cost"), c("id", travel_cost))
  data.table::setcolorder(
    access,
    c("id", setdiff(groups, "from_id"), travel_cost)
  )
  data.table::setorderv(access, c("id", setdiff(groups, "from_id")))

  if (exists("original_class")) class(access) <- original_class

  return(access[])
}


calc_cum_cutoff <- function(data, groups, opportunity, travel_cost, .cutoff) {
  env <- environment()
  .opportunity_colname <- opportunity
  .cost_colname <- travel_cost

  data <- data[get(.cost_colname) <= .cutoff]

  cum_cutoff_access <- data[
    ,
    .(access = sum(get(.opportunity_colname))),
    by = eval(groups, envir = env)
  ]

  return(cum_cutoff_access)
}

# if running manually, please run the following line first:
# source("tests/testthat/setup.R")

fast_smaller_matrix <- data.table::copy(smaller_matrix)
fast_smaller_matrix[, travel_time := travel_time / 4]

tester <- function(travel_matrix = fast_smaller_matrix,
                   land_use_data = get("land_use_data", envir = parent.frame()),
                   opportunity = "jobs",
                   travel_cost = "travel_time",
                   demand = "population",
                   cost_increment = 1,
                   group_by = "mode",
                   fill_missing_ids = TRUE) {
  balancing_cost(
    travel_matrix,
    land_use_data,
    opportunity,
    travel_cost,
    demand,
    cost_increment,
    group_by,
    fill_missing_ids
  )
}

test_that("raises errors due to incorrect input", {
  expect_error(tester(opportunity = 1))
  expect_error(tester(opportunity = c("schools", "jobs")))

  expect_error(tester(travel_cost = 1))
  expect_error(tester(travel_cost = c("travel_time", "monetary_cost")))

  expect_error(tester(demand = 1))
  expect_error(tester(demand = c("population", "population")))

  expect_error(tester(cost_increment = "a"))
  expect_error(tester(cost_increment = -1))
  expect_error(tester(cost_increment = Inf))

  expect_error(tester(group_by = 1))
  expect_error(tester(group_by = NA))
  expect_error(tester(group_by = "from_id"))
  expect_error(tester(group_by = c("mode", "mode")))

  expect_error(tester(fill_missing_ids = 1))
  expect_error(tester(fill_missing_ids = c(TRUE, TRUE)))
  expect_error(tester(fill_missing_ids = NA))

  expect_error(tester(as.list(travel_matrix)))
  expect_error(
    tester(travel_matrix[, .(oi = from_id, to_id, travel_time, mode)])
  )
  expect_error(
    tester(travel_matrix[, .(from_id, oi = to_id, travel_time, mode)])
  )
  expect_error(
    tester(travel_matrix[, .(from_id, to_id, oi = travel_time, mode)])
  )
  expect_error(
    tester(travel_matrix[, .(from_id, to_id, travel_time, oi = mode)])
  )

  expect_error(tester(as.list(land_use_data)))
  expect_error(
    tester(land_use_data = land_use_data[, .(oi = id, jobs, population)])
  )
  expect_error(
    tester(land_use_data = land_use_data[, .(id, oi = jobs, population)])
  )
  expect_error(
    tester(land_use_data = land_use_data[, .(id, jobs, oi = population)])
  )
})

test_that("throws warning if travel_matrix extra col", {
  # i.e. col not listed in travel_cost and by_col
  expect_warning(tester(group_by = character(0)))
})

test_that("returns a dataframe whose class is the same as travel_matrix's", {
  result <- tester()
  expect_is(result, "data.table")
  result <- tester(land_use_data = as.data.frame(land_use_data))
  expect_is(result, "data.table")

  result <- tester(as.data.frame(fast_smaller_matrix))
  expect_false(inherits(result, "data.table"))
  expect_is(result, "data.frame")
  result <- tester(
    as.data.frame(fast_smaller_matrix),
    land_use_data = as.data.frame(land_use_data)
  )
  expect_false(inherits(result, "data.table"))
  expect_is(result, "data.frame")
})

test_that("result has correct structure", {
  result <- tester()
  expect_true(ncol(result) == 3)
  expect_is(result$id, "character")
  expect_is(result$mode, "character")
  expect_is(result$travel_time, "numeric")

  suppressWarnings(result <- tester(group_by = character(0)))
  expect_true(ncol(result) == 2)
  expect_is(result$id, "character")
  expect_is(result$travel_time, "numeric")

  result <- tester(
    data.table::data.table(
      mode = character(),
      from_id = character(),
      to_id = character(),
      travel_time = integer()
    )
  )
  expect_true(ncol(result) == 3)
  expect_true(nrow(result) == 0)
  expect_is(result$id, "character")
  expect_is(result$mode, "character")
  expect_is(result$travel_time, "numeric")
})

test_that("input data sets remain unchanged", {
  original_fast_smaller_matrix <- travel_matrix[1:10]
  original_fast_smaller_matrix[, travel_time := travel_time / 4]
  original_land_use_data <- readRDS(file.path(data_dir, "land_use_data.rds"))

  result <- tester()

  expect_identical(original_fast_smaller_matrix, fast_smaller_matrix)
  expect_identical(original_land_use_data, land_use_data)
})

test_that("calculates balancing cost correctly", {
  selected_ids <- c(
    "89a88cda64fffff",
    "89a88cdb027ffff",
    "89a88cdb12bffff",
    "89a88cdb287ffff",
    "89a88cdb67bffff"
  )
  smaller_ttm <- travel_matrix[
    from_id %in% selected_ids & to_id %in% selected_ids
  ]
  smaller_ttm[mode == "transit2", travel_time := travel_time + 10]

  result <- tester(smaller_ttm)
  expected_result <- data.table::data.table(
    id = rep(selected_ids, each = 2),
    mode = rep(c("transit", "transit2"), 5),
    travel_time = c(42, 52, 6, 16, 27, 37, 6, 16, 20, 30)
  )
  expect_identical(result, expected_result)
})

test_that("fill_missing_ids arg works correctly", {
  test_matrix <- rbind(
    fast_smaller_matrix,
    fast_smaller_matrix[1][, mode := "transit2"]
  )

  result <- tester(test_matrix)
  data.table::setkeyv(result, NULL)
  expect_identical(
    result,
    data.table::data.table(
      id = rep("89a88cdb57bffff", 2),
      mode = c("transit", "transit2"),
      travel_time = c(12, NA)
    )
  )

  result <- tester(test_matrix, fill_missing_ids = FALSE)
  expect_identical(
    result,
    data.table::data.table(
      id = "89a88cdb57bffff",
      mode = "transit",
      travel_time = 12
    )
  )
})

test_that("cost_increment arg works correctly", {
  selected_ids <- c(
    "89a88cda64fffff",
    "89a88cdb027ffff",
    "89a88cdb12bffff",
    "89a88cdb287ffff",
    "89a88cdb67bffff"
  )
  smaller_ttm <- travel_matrix[
    from_id %in% selected_ids & to_id %in% selected_ids
  ]

  result <- tester(smaller_ttm)
  expect_false(all(result$travel_time %% 2 == 0))

  result <- tester(smaller_ttm, cost_increment = 2)
  expect_true(all(result$travel_time %% 2 == 0))
})

test_that("works even if travel_matrix and land_use has specific colnames", {
  expected_result <- tester()

  fast_smaller_matrix[, opportunity := "oi"]
  result <- suppressWarnings(tester(fast_smaller_matrix))
  expect_identical(expected_result, result)

  fast_smaller_matrix[, opportunity := NULL]
  fast_smaller_matrix[, travel_cost := "oi"]
  result <- suppressWarnings(tester(fast_smaller_matrix))
  expect_identical(expected_result, result)

  fast_smaller_matrix[, travel_cost := NULL]
  fast_smaller_matrix[, groups := "oi"]
  result <- suppressWarnings(tester(fast_smaller_matrix))
  expect_identical(expected_result, result)

  fast_smaller_matrix[, groups := NULL]
  fast_smaller_matrix[, demand := "oi"]
  result <- suppressWarnings(tester(fast_smaller_matrix))
  expect_identical(expected_result, result)

  fast_smaller_matrix[, demand := NULL]
  land_use_data[, opportunity := "oi"]
  result <- suppressWarnings(tester(land_use_data = land_use_data))
  expect_identical(expected_result, result)

  land_use_data[, opportunity := NULL]
  land_use_data[, demand := "oi"]
  result <- suppressWarnings(tester(land_use_data = land_use_data))
  expect_identical(expected_result, result)

  land_use_data[, demand := NULL]
})

#' @param decay_function A `fuction` that converts travel cost into an
#'   impedance factor used to weight opportunities. This function should take a
#'   `numeric` vector and also return a `numeric` vector as output, with the
#'   same length as the input. For convenience, the package currently includes
#'   the following functions: [decay_binary()], [decay_exponential()],
#'   [decay_power()] and [decay_stepped()]. See the documentation of each decay
#'   function for more details.

#' Doubly constrained accessibility
#'
#' Calculates accessibility using Wilson's doubly-constrained gravity model.
#' This measure allocates flows between origins and destinations such that
#' origin totals equal demand and destination totals equal supply. Note: this is
#' an internal helper function used by [constrained_accessibility()] when
#' `constraint = "doubly"`.
#'
#' @name doubly_constrained
#' @keywords internal
#' @return A `data.table`/`data.frame` with either OD-level flows (`detailed_results = TRUE`)
#'   or marginals; see Details.
#' @examples NULL
#' @importFrom utils globalVariables
utils::globalVariables(c(
  # common ids / columns
  "from_id", "to_id", "id", "opp_weight", "supply", "demand", "weighted_demand", "weighted_supply",
  # total-constrained
  "kappa_total", "hatkappa_total", "K_total", "hatK_total", "constrained_opportunity",
  # singly-constrained
  "denom_i", "denom_j", "kappa_singly", "hatkappa_singly", "A_i", "B_j", "singly_access",
  # doubly-constrained
  "kappa_doubly", "flow", "error"
))
doubly_constrained <- function(travel_matrix,
                               land_use_data,
                               travel_cost,
                               decay_function,
                               demand,
                               supply,
                               error_threshold = 0.001,
                               improvement_threshold = 1e-6,
                               max_iterations = 1000,
                               active,
                               group_by = character(0),
                               fill_missing_ids = TRUE,
                               detailed_results = FALSE
                               ) {
  # Validate inputs
  checkmate::assert_string(demand)
  checkmate::assert_string(supply)
  checkmate::assert_number(error_threshold, lower = 0)
  checkmate::assert_number(improvement_threshold, lower = 0)
  checkmate::assert_int(max_iterations, lower = 1)

  # Convert to data.table
  if (!inherits(travel_matrix, "data.table")) {
    original_class <- class(travel_matrix)
    data <- data.table::as.data.table(travel_matrix)
  } else {
    data <- data.table::copy(travel_matrix)
  }
  if (!inherits(land_use_data, "data.table")) {
    land_use_data <- data.table::as.data.table(land_use_data)
  }

  # Merge demand and supply
  merge_by_reference(data, land_use_data, demand, left_df_idcol = "from_id")
  merge_by_reference(data, land_use_data, supply, left_df_idcol = "to_id")

  # Apply decay
  data <- apply_gravity_measure(data, decay_function, travel_cost)

  # Prepare vectors and impedance matrix
  origins <- unique(data$from_id)
  destinations <- unique(data$to_id)
  O <- land_use_data[id %in% origins, get(demand)]
  D <- land_use_data[id %in% destinations, get(supply)]

  # Validate totals match
  if (abs(sum(O) - sum(D)) > 1e-6) {
    stop("For doubly-constrained, the sum of origins must equal the sum of destinations.")
  }

  # Build impedance matrix
  impedance_matrix <- matrix(data$opp_weight, nrow = length(origins), ncol = length(destinations), byrow = FALSE)

  # Iterative proportional fitting
  Ai <- rep(1, length(O))
  Bj <- rep(1, length(D))
  previous_error <- Inf
  iteration_count <- 0
  stop_reason <- ""

  repeat {
    iteration_count <- iteration_count + 1

    # Update Ai
    for (i in seq_along(O)) {
      Ai[i] <- 1 / (sum(Bj * D * impedance_matrix[i, ]) + 1e-9)
    }

    # Update Bj
    Bj_new <- numeric(length(D))
    for (j in seq_along(D)) {
      Bj_new[j] <- 1 / (sum(Ai * O * impedance_matrix[, j]) + 1e-9)
    }

    # Compute flows
    Tij <- outer(Ai * O, Bj_new * D) * impedance_matrix

    # Compute error
    error <- (sum(abs(O - rowSums(Tij))) + sum(abs(D - colSums(Tij)))) / sum(O)
    error_change <- abs(previous_error - error)

    if (error < error_threshold || error_change < improvement_threshold || iteration_count >= max_iterations) {
      Bj <- Bj_new
      stop_reason <- if (iteration_count >= max_iterations) "Max iterations reached" else if (error < error_threshold) "Error threshold met" else "Slow improvement"
      break
    }

    previous_error <- error
    Bj <- Bj_new
  }

  # Compute kappa_doubly
# kappa_ij^D = A_i * B_j * O_i * f(c_ij)
  kappa_matrix <- outer(Ai * O, Bj) * impedance_matrix

  # Prepare output
  detailed_dt <- data.table::data.table(
    from_id = rep(origins, each = length(destinations)),
    to_id = rep(destinations, times = length(origins)),
    flow = as.vector(t(Tij)),
    A_i = rep(Ai, each = length(destinations)),
    B_j = rep(Bj, times = length(origins)),
    kappa_doubly = as.vector(t(kappa_matrix)),
    error = error
  )

  if (detailed_results) {
    return(detailed_dt[])
  } else {
    warning("Aggregated results equal marginals (O_i or D_j). Interpret with caution.")
    agg <- data.table::rbindlist(list(
      data.table::data.table(id = origins,      flow = O),
      data.table::data.table(id = destinations, flow = D)
    ))
    return(agg[])
  }

}

#' @param travel_cost A string. The name of the column in `travel_matrix` with
#'        the travel cost between origins and destinations. The notion of cost
#'        here is generic over any kind of numeric travel cost, such as distance,
#'        time and money.

#' @param data A `data.frame` with a travel cost matrix in long format,
#'   containing the columns with origin ids, destination ids, the travel cost
#'   between each origin-destination pair (this can be either in  terms of
#'   travel time, distance or monetary costs), and number of population and
#'   opportunities located in each origin and destination.

#' Cumulative access based on maximum travel time interval
#'
#' Calculates the average or median number of opportunities that can be reached
#' considering multiple maximum travel cost thresholds within a given travel
#' cost interval specified by the user. The time interval cumulative
#' accessibility measures was originally proposed by
#' \insertCite{tomasiello2023time;textual}{accessibility}.
#' @template description_generic_cost
#'
#' @template travel_matrix
#' @template land_use_data
#' @template opportunity
#' @template travel_cost
#' @param interval A `numeric` vector of length 2. Indicates the start and end
#'   points of the interval of travel cost thresholds to be used. The first
#'   entry must be lower than the second.
#' @param interval_increment A `numeric`. How many travel cost units separate
#'   the cutoffs used to calculate the accessibility estimates which will be
#'   used to calculate the summary estimate within the specified interval.
#'   Should be thought as the resolution of the distribution of travel costs
#'   within the interval. Defaults to 1.
#' @param summary_function A function. This function is used to summarize a
#'   distribution of accessibility estimates within a travel cost interval as a
#'   single value. Can be any function that takes an arbitrary number of
#'   numeric values as as input and returns a single number as output. Defaults
#'   to [stats::median()].
#' @template group_by
#' @template active
#'
#' @template return_accessibility
#'
#' @family cumulative access
#'
#' @references
#' \insertAllCited{}
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' data_dir <- system.file("extdata", package = "accessibility")
#' travel_matrix <- readRDS(file.path(data_dir, "travel_matrix.rds"))
#' land_use_data <- readRDS(file.path(data_dir, "land_use_data.rds"))
#'
#' df <- cumulative_interval(
#'   travel_matrix = travel_matrix,
#'   land_use_data = land_use_data,
#'   interval = c(20, 30),
#'   opportunity = "schools",
#'   travel_cost = "travel_time"
#' )
#' head(df)
#'
#' df <- cumulative_interval(
#'   travel_matrix = travel_matrix,
#'   land_use_data = land_use_data,
#'   interval = c(40, 80),
#'   opportunity = "jobs",
#'   travel_cost = "travel_time"
#' )
#' head(df)
#'
#' @export
cumulative_interval <- function(travel_matrix,
                                land_use_data,
                                opportunity,
                                travel_cost,
                                interval,
                                interval_increment = 1,
                                summary_function = stats::median,
                                group_by = character(0),
                                active = TRUE) {
  checkmate::assert_string(opportunity)
  checkmate::assert_string(travel_cost)
  checkmate::assert_logical(active, len = 1, any.missing = FALSE)
  assert_interval_increment(interval_increment)
  assert_summary_function(summary_function)
  assert_group_by(group_by)
  assert_travel_matrix(travel_matrix, travel_cost, group_by)
  assert_land_use_data(
    land_use_data,
    travel_matrix,
    opportunity,
    active = active
  )

  interval <- assert_and_assign_interval(interval)

  # if not a dt, keep original class to assign later when returning result

  if (!inherits(travel_matrix, "data.table")) {
    original_class <- class(travel_matrix)
    data <- data.table::as.data.table(travel_matrix)
  } else {
    data <- data.table::copy(travel_matrix)
  }

  if (!inherits(land_use_data, "data.table")) {
    land_use_data <- data.table::as.data.table(land_use_data)
  }

  cutoffs <- lapply(
    interval,
    function(pair) seq.int(pair[1], pair[2], by = interval_increment)
  )
  cutoffs <- unique(unlist(cutoffs))
  cutoffs <- cutoffs[order(cutoffs)]
  names(cutoffs) <- as.character(cutoffs)

  # small optimization: we can ditch anything that costs more than the highest
  # cost cutoff value, since it won't affect the results anyway
  .cost_colname <- travel_cost
  data <- data[get(.cost_colname) <= max(cutoffs)]

  merge_by_reference(
    data,
    land_use_data,
    opportunity,
    left_df_idcol = ifelse(active, "to_id", "from_id")
  )

  group_id <- ifelse(active, "from_id", "to_id")
  groups <- c(group_id, group_by)
  env <- environment()

  warn_extra_cols(travel_matrix, travel_cost, group_id, groups)

  .opportunity_colname <- opportunity
  access_list <- lapply(
    cutoffs,
    function(x) {
      data[
        get(.cost_colname) <= x,
        .(access = sum(get(.opportunity_colname))),
        by = eval(groups, envir = env)
      ]
    }
  )

  # some of the estimates for lower thresholds may have less number of rows
  # because some origins may not reach any destinations when the travel cost
  # restrictions are harder. in such case, we have to fill these datasets with
  # the missing combinations having accessibility 0, otherwise the summary
  # measure would be miscalculated (e.g. think of median of 1 3 5 and median of
  # 0 1 3 5)

  nrow_more_restrictive <- nrow(access_list[[1]])
  nrow_less_restrictive <- nrow(access_list[[length(access_list)]])

  if (nrow_more_restrictive < nrow_less_restrictive) {
    access_list <- fill_access_list(access_list, travel_matrix, groups)
  }
  access_list <- data.table::rbindlist(access_list, idcol = "cutoffs")
  access_list[, cutoffs := as.numeric(cutoffs)]

  # the as.integer() call below makes sure that, even though the summary
  # function may return doubles, the result is an integer (afterall when using
  # a cumulative measure you cannot reach half of an opportunity) and that the
  # float is rounded down to the integer (if you reach 2.8 opportunities you do
  # reach 2, but you can't reach 3)

  access_list <- lapply(
    interval,
    function(pair) {
      access_list[
        cutoffs >= pair[1] & cutoffs <= pair[2],
        .(access = as.integer(summary_function(access))),
        by = eval(groups, envir = env)
      ]
    }
  )
  access <- data.table::rbindlist(access_list, idcol = "interval")
  data.table::setcolorder(access, c(groups, "interval", "access"))
  data.table::setnames(access, c(group_id, "access"), c("id", opportunity))

  if (length(interval) == 1) access[, interval := NULL]

  if (exists("original_class")) class(access) <- original_class

  return(access[])
}


#' @keywords internal
fill_access_list <- function(access, travel_matrix, groups) {
  unique_values <- lapply(groups, function(x) unique(travel_matrix[[x]]))
  names(unique_values) <- groups
  possible_combinations <- do.call(data.table::CJ, unique_values)

  access_nrows <- vapply(access, nrow, integer(1))
  possible_nrows <- nrow(possible_combinations)
  should_fill <- access_nrows < possible_nrows

  filled_access <- mapply(
    do_fill = should_fill,
    access_df = access,
    SIMPLIFY = FALSE,
    FUN = function(do_fill, access_df) {
      if (!do_fill) return(access_df)

      filled_access_df <- do_fill_missing_ids(
        access_df,
        possible_combinations,
        groups
      )

      return(filled_access_df[])
    }
  )

  return(filled_access)
}

# if running manually, please run the following line first:
# source("tests/testthat/setup.R")

tester <- function(travel_matrix = smaller_matrix,
                   land_use_data = get("land_use_data", envir = parent.frame()),
                   opportunity = "jobs",
                   travel_cost = "travel_time",
                   interval = c(10, 30),
                   interval_increment = 1,
                   summary_function = stats::median,
                   group_by = "mode",
                   active = TRUE) {
  cumulative_interval(
    travel_matrix,
    land_use_data,
    opportunity,
    travel_cost,
    interval,
    interval_increment,
    summary_function,
    group_by,
    active
  )
}

test_that("raises errors due to incorrect input", {
  expect_error(tester(interval = "banana"))

  expect_error(tester(interval = 3))
  expect_error(tester(interval = c(-1, 10)))
  expect_error(tester(interval = c(11, 10)))
  expect_error(tester(interval = c(10, 10)))
  expect_error(tester(interval = c(1, Inf)))
  expect_error(tester(interval = c(1, NA)))

  expect_error(tester(interval = list("a")))
  expect_error(tester(interval = list()))
  expect_error(tester(interval = list(c(10, 20), c(10, 20))))
  expect_error(tester(interval = list(3)))
  expect_error(tester(interval = list(c(-1, 10))))
  expect_error(tester(interval = list(c(11, 10))))
  expect_error(tester(interval = list(c(10, 10))))
  expect_error(tester(interval = list(c(1, Inf))))
  expect_error(tester(interval = list(c(1, NA))))

  expect_error(tester(interval_increment = "a"))
  expect_error(tester(interval_increment = Inf))
  expect_error(tester(interval_increment = c(1, 1)))
  expect_error(tester(interval_increment = -1))

  expect_error(tester(summary_function = "test"))
  expect_error(tester(summary_function = readRDS))
  expect_error(tester(summary_function = identity))

  expect_error(tester(opportunity = 1))
  expect_error(tester(opportunity = c("schools", "jobs")))

  expect_error(tester(travel_cost = 1))
  expect_error(tester(travel_cost = c("travel_time", "monetary_cost")))

  expect_error(tester(group_by = 1))
  expect_error(tester(group_by = NA))
  expect_error(tester(group_by = "from_id"))
  expect_error(tester(group_by = c("mode", "mode")))

  expect_error(tester(active = 1))
  expect_error(tester(active = c(TRUE, TRUE)))
  expect_error(tester(active = NA))

  expect_error(tester(as.list(travel_matrix)))
  expect_error(tester(travel_matrix[, .(oi = from_id, to_id, travel_time)]))
  expect_error(tester(travel_matrix[, .(from_id, oi = to_id, travel_time)]))
  expect_error(
    tester(
      travel_matrix[, .(from_id, to_id, oi = travel_time)],
      travel_cost = "travel_time"
    )
  )
  expect_error(
    tester(
      travel_matrix[, .(from_id, to_id, travel_time, oi = mode)],
      group_by = "mode"
    )
  )

  expect_error(tester(as.list(land_use_data)))
  expect_error(tester(land_use_data = land_use_data[, .(oi = id, jobs)]))
  expect_error(
    tester(
      land_use_data = land_use_data[, .(id, oi = jobs)],
      opportunity = "jobs"
    )
  )
})

test_that("throws warning if travel_matrix extra col", {
  # i.e. col not listed in travel_cost and by_col
  expect_warning(tester(group_by = character(0)))
})

test_that("returns a dataframe whose class is the same as travel_matrix's", {
  result <- tester()
  expect_is(result, "data.table")
  result <- tester(land_use_data = as.data.frame(land_use_data))
  expect_is(result, "data.table")

  result <- tester(as.data.frame(smaller_matrix))
  expect_false(inherits(result, "data.table"))
  expect_is(result, "data.frame")
  result <- tester(
    as.data.frame(smaller_matrix),
    land_use_data = as.data.frame(land_use_data)
  )
  expect_false(inherits(result, "data.table"))
  expect_is(result, "data.frame")
})

test_that("result has correct structure", {
  result <- tester()
  expect_true(ncol(result) == 3)
  expect_is(result$id, "character")
  expect_is(result$mode, "character")
  expect_is(result$jobs, "integer")

  result <- tester(opportunity = "schools")
  expect_true(ncol(result) == 3)
  expect_is(result$id, "character")
  expect_is(result$mode, "character")
  expect_is(result$schools, "integer")

  suppressWarnings(result <- tester(group_by = character(0)))
  expect_true(ncol(result) == 2)
  expect_is(result$id, "character")
  expect_is(result$jobs, "integer")

  result <- tester(
    data.table::data.table(
      mode = character(),
      from_id = character(),
      to_id = character(),
      travel_time = integer()
    )
  )
  expect_true(ncol(result) == 3)
  expect_true(nrow(result) == 0)
  expect_is(result$id, "character")
  expect_is(result$mode, "character")
  expect_is(result$jobs, "integer")

  result <- tester(interval = list(c(10, 30), c(20, 30)))
  expect_true(ncol(result) == 4)
  expect_is(result$id, "character")
  expect_is(result$mode, "character")
  expect_is(result$interval, "character")
  expect_is(result$jobs, "integer")
})

test_that("input data sets remain unchanged", {
  original_smaller_matrix <- travel_matrix[1:10]
  original_land_use_data <- readRDS(file.path(data_dir, "land_use_data.rds"))

  result <- tester()

  # subsets in other functions tests set smaller_matrix index
  data.table::setindex(smaller_matrix, NULL)

  expect_equal(original_smaller_matrix, smaller_matrix)
  expect_equal(original_land_use_data, land_use_data)
})

test_that("active and passive accessibility is correctly calculated", {
  selected_ids <- c(
    "89a88cdb57bffff",
    "89a88cdb597ffff",
    "89a88cdb5b3ffff",
    "89a88cdb5cfffff",
    "89a88cd909bffff"
  )
  smaller_travel_matrix <- travel_matrix[
    from_id %in% selected_ids & to_id %in% selected_ids
  ]

  result <- tester(
    smaller_travel_matrix,
    interval = c(40, 45),
    group_by = "mode"
  )
  expected_result <- data.table::data.table(
    id = rep(selected_ids, 2),
    mode = rep(c("transit", "transit2"), each = 5),
    jobs = rep(as.integer(c(82, 517, 517, 304, 0)), 2)
  )
  expect_identical(result, expected_result)

  result <- tester(
    smaller_travel_matrix,
    interval = c(40, 45),
    opportunity = "population",
    group_by = "mode",
    active = FALSE
  )
  expected_result <- data.table::data.table(
    id = rep(selected_ids, 2),
    mode = rep(c("transit", "transit2"), each = 5),
    population = rep(as.integer(c(1946, 3457, 2363, 4552, 0)), 2)
  )
  expect_identical(result, expected_result)

  # with more than one interval

  result <- tester(
    smaller_travel_matrix,
    interval = list(c(40, 45), c(40, 50)),
    group_by = "mode"
  )
  expected_result <- data.table::data.table(
    id = rep(selected_ids, 4),
    mode = rep(rep(c("transit", "transit2"), each = 5), 2),
    interval = rep(c("[40,45]", "[40,50]"), each = 10),
    jobs = c(
      rep(as.integer(c(82, 517, 517, 304, 0)), 2),
      rep(as.integer(c(82, 599, 599, 499, 0)), 2)
    )
  )
  expect_identical(result, expected_result)

  result <- tester(
    smaller_travel_matrix,
    interval = list(c(40, 45), c(40, 50)),
    opportunity = "population",
    group_by = "mode",
    active = FALSE
  )
  expected_result <- data.table::data.table(
    id = rep(selected_ids, 4),
    mode = rep(rep(c("transit", "transit2"), each = 5), 2),
    interval = rep(c("[40,45]", "[40,50]"), each = 10),
    population = c(
      rep(as.integer(c(1946, 3457, 2363, 4552, 0)), 2),
      rep(as.integer(c(5404, 4552, 2363, 4552, 0)), 2)
    )
  )
  expect_identical(result, expected_result)
})

test_that("summarizes the result according to summary function", {
  selected_ids <- c(
    "89a88cdb57bffff",
    "89a88cdb597ffff",
    "89a88cdb5b3ffff",
    "89a88cdb5cfffff",
    "89a88cd909bffff"
  )
  smaller_travel_matrix <- travel_matrix[
    from_id %in% selected_ids & to_id %in% selected_ids
  ]

  result <- tester(
    smaller_travel_matrix,
    interval = c(40, 45),
    summary_function = min,
    group_by = "mode"
  )
  expected_result <- data.table::data.table(
    id = rep(selected_ids, 2),
    mode = rep(c("transit", "transit2"), each = 5),
    jobs = rep(as.integer(c(82, 408, 408, 109, 0)), 2)
  )
  expect_identical(result, expected_result)

  result <- tester(
    smaller_travel_matrix,
    interval = c(40, 45),
    summary_function = mean,
    group_by = "mode"
  )
  expected_result <- data.table::data.table(
    id = rep(selected_ids, 2),
    mode = rep(c("transit", "transit2"), each = 5),
    jobs = rep(as.integer(c(82, 512, 508, 304, 0)), 2)
  )
  expect_identical(result, expected_result)
})

test_that("accessibility is correctly calculated when ids are missing", {
  selected_ids <- c(
    "89a88cdb57bffff",
    "89a88cdb597ffff",
    "89a88cdb5b3ffff",
    "89a88cdb5cfffff",
    "89a88cd909bffff"
  )
  smaller_travel_matrix <- travel_matrix[
    from_id %in% selected_ids & to_id %in% selected_ids
  ]

  result <- tester(
    smaller_travel_matrix,
    interval = c(5, 6),
    summary_function = min,
    group_by = "mode"
  )
  expected_result <- data.table::data.table(
    id = rep(selected_ids[order(selected_ids)], each = 2),
    mode = rep(c("transit", "transit2"), 5),
    jobs = rep(as.integer(c(0, 0, 0, 0, 0)), 2)
  )
  expect_identical(result, expected_result)
})

test_that("interval_increment arg is being used", {
  # hard to "perfectly" assess if the interval_increment value is being applied
  # correctly, but the code is trivial enough to let us say that it's correctly
  # applied if the results are different when using different increments

  selected_ids <- c(
    "89a88cdb57bffff",
    "89a88cdb597ffff",
    "89a88cdb5b3ffff",
    "89a88cdb5cfffff",
    "89a88cd909bffff"
  )
  smaller_travel_matrix <- travel_matrix[
    from_id %in% selected_ids & to_id %in% selected_ids
  ]

  result_increment_1 <- tester(
    smaller_travel_matrix,
    interval_increment = 1,
    interval = c(5, 7)
  )
  result_increment_2 <- tester(
    smaller_travel_matrix,
    interval_increment = 2,
    interval = c(5, 7)
  )

  expect_false(identical(result_increment_1, result_increment_2))
})

test_that("works even if travel_matrix and land_use has specific colnames", {
  expected_result <- tester()

  smaller_matrix[, opportunity := "oi"]
  result <- suppressWarnings(tester(smaller_matrix))
  expect_identical(expected_result, result)

  smaller_matrix[, opportunity := NULL]
  smaller_matrix[, travel_cost := "oi"]
  result <- suppressWarnings(tester(smaller_matrix))
  expect_identical(expected_result, result)

  smaller_matrix[, travel_cost := NULL]
  smaller_matrix[, groups := "oi"]
  result <- suppressWarnings(tester(smaller_matrix))
  expect_identical(expected_result, result)

  smaller_matrix[, groups := NULL]
  land_use_data[, opportunity := "oi"]
  result <- suppressWarnings(tester(land_use_data = land_use_data))
  expect_identical(expected_result, result)

  land_use_data[, opportunity := NULL]
})

#' Stepped decay function
#'
#' Returns a stepped weighting function to be used inside accessibility
#' calculating functions.
#' @template description_generic_cost
#'
#' @param steps A `numeric` vector or a list of `numeric` vectors. The travel
#'   cost steps, in ascending order. Please do not include travel cost 0 as a
#'   step: this is already handled by the function.
#' @param weights A `numeric` vector with same length as `steps` or a list of
#'   `numeric` vectors whose lengths are equal to the lengths of the elements of
#'   same index in `steps`. The values, between 0 and 1, that the function
#'   assumes at each step. Please do not include weight 1 as the first value:
#'   this is already handled by the function. The function considers the steps'
#'   intervals "open on the right", meaning that the function assumes the step
#'   value at the actual step, not right after it. Please see the illustrative
#'   examples for effects of this assumption on the results.
#'
#' @return A `function` that takes a generic travel cost vector (`numeric`) as
#'   an input and returns a vector of weights (`numeric`).
#'
#' @details When both `steps` and `weights` parameters are given `list`s, their
#'   content are matched element-wise to define each stepped weighting function
#'   - i.e. the first element of `steps` is matched to the first element of
#'   `weights`, the second element of `steps` is matched to the second of
#'   `weights`, etc. When using a function created with `decay_stepped()`, the
#'   output is named after the combination of steps (`"s"`) and weights (`"w"`)
#'   - e.g. given the steps `c(10, 20, 30)` and the weights `c(0.66, 0.33, 0)`,
#'   the output will be named `"s(10,20,30);w(0.66,0.33,0)"`.
#'
#' @family decay functions
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' weighting_function <- decay_stepped(
#'   c(10, 20, 30, 40),
#'   weights = c(0.75, 0.5, 0.25, 0)
#' )
#'
#' weighting_function(c(5, 25, 35, 45))
#'
#' weighting_function <- decay_stepped(
#'   list(c(10, 20, 30, 40), c(10, 20, 30, 40)),
#'   weights = list(c(0.75, 0.5, 0.25, 0), c(0.8, 0.6, 0.4, 0.2))
#' )
#'
#' weighting_function(c(5, 25, 35, 45))
#'
#' # intervals are open on the right, so the values change exactly at each step
#' weighting_function(c(0, 10, 20, 30, 40))
#'
#' @export
decay_stepped <- function(steps, weights) {
  assert_steps(steps)
  assert_weights(weights, steps)

  if (!is.list(steps)) steps <- list(steps)
  if (!is.list(weights)) weights <- list(weights)

  steps <- lapply(steps, function(stps) c(0, stps))
  weights <- lapply(weights, function(wghts) c(1, wghts))

  weighting_function <- function(travel_cost) {
    weights_list <- mapply(
      stps = steps,
      wghts = weights,
      FUN = function(stps, wghts) {
        vapply(
          travel_cost,
          function(x) wghts[max(which(x >= stps))],
          numeric(1)
        )
      },
      SIMPLIFY = FALSE
    )

    list_names <- mapply(
      stps = steps,
      wghts = weights,
      FUN = function(stps, wghts) {
        paste0(
          "s(", paste(stps[-1], collapse = ","), ");",
          "w(", paste(wghts[-1], collapse = ","), ")"
        )
      },
      SIMPLIFY = TRUE
    )

    names(weights_list) <- list_names

    return(weights_list)
  }

  return(weighting_function)
}


assert_steps <- function(steps) {
  is_numeric_like <- function(x) is.numeric(x) || is.integer(x)

  if (!(is_numeric_like(steps) || is.list(steps))) {
    stop(
      "Assertion on 'steps' failed: Must be either a numeric vector or a list ",
      "of numeric vectors."
    )
  }

  if (is_numeric_like(steps)) {
    checkmate::assert_numeric(
      steps,
      lower = 1,
      finite = TRUE,
      any.missing = FALSE,
      min.len = 1,
      unique = TRUE,
      sorted = TRUE
    )
  } else {
    for (i in seq.int(length(steps))) {
      checkmate::assert_numeric(
        steps[[i]],
        lower = 1,
        finite = TRUE,
        any.missing = FALSE,
        min.len = 1,
        unique = TRUE,
        sorted = TRUE,
        .var.name = paste0("steps[[", i, "]]")
      )
    }
  }

  return(invisible(TRUE))
}


assert_weights <- function(weights, steps) {
  is_numeric_like <- function(x) is.numeric(x) || is.integer(x)

  if (is_numeric_like(steps)) {
    checkmate::assert_numeric(
      weights,
      lower = 0,
      upper = 1,
      any.missing = FALSE,
      len = length(steps)
    )
  } else {
    checkmate::assert_list(
      weights,
      types = "numeric",
      len = length(steps)
    )

    for (i in seq.int(length(weights))) {
      checkmate::assert_numeric(
        weights[[i]],
        lower = 0,
        upper = 1,
        any.missing = FALSE,
        len = length(steps[[i]]),
        .var.name = paste0("weights[[", i, "]]")
      )
    }
  }

  return(invisible(TRUE))
}

tester <- function(steps = c(25, 50), weights = c(0.5, 0)) {
  decay_stepped(steps, weights)
}

test_that("adequately raises errors", {
  expect_error(tester("a"))
  expect_error(tester(0))
  expect_error(tester(c(25, Inf)))
  expect_error(tester(c(25, NA)))
  expect_error(tester(integer()))
  expect_error(tester(c(25, 25)))
  expect_error(tester(c(25, 20)))

  expect_error(tester(list(c(25, 50), "a")))
  expect_error(tester(list(c(25, 50), 0)))
  expect_error(tester(list(c(25, 50), c(25, Inf))))
  expect_error(tester(list(c(25, 50), c(25, NA))))
  expect_error(tester(list(c(25, 50), integer())))
  expect_error(tester(list(c(25, 50), c(25, 25))))
  expect_error(tester(list(c(25, 50), c(25, 20))))

  expect_error(tester(weights = "a"))
  expect_error(tester(weights = c(0.7, -1)))
  expect_error(tester(weights = c(1.5, 0)))
  expect_error(tester(weights = c(0.7, NA)))
  expect_error(tester(weights = 1))

  expect_error(tester(list(c(20, 50), c(30, 60)), weights = "a"))
  expect_error(tester(list(c(20, 50), c(30, 60)), weights = list("a", "b")))
  expect_error(tester(list(c(20, 50), c(30, 60)), weights = list(1)))
  expect_error(
    tester(list(c(20, 50), c(30, 60)), weights = list(c(0.5, 0), c(0.7, -1)))
  )
  expect_error(
    tester(list(c(20, 50), c(30, 60)), weights = list(c(0.5, 0), c(1.5, 0)))
  )
  expect_error(
    tester(list(c(20, 50), c(30, 60)), weights = list(c(0.5, 0), c(0.7, NA)))
  )
  expect_error(
    tester(list(c(20, 50), c(30, 60)), weights = list(c(0.5, 0), c(1)))
  )
})

test_that("output is a function that returns a list of numeric vectors", {
  expect_is(tester(), "function")

  output_fn <- tester()
  output_list <- output_fn(0)
  expect_is(output_list, "list")
  expect_length(output_list, 1)
  expect_named(output_list, "s(25,50);w(0.5,0)")

  expect_equal(output_fn(c(5, 30, 55))[[1]], c(1, 0.5, 0))

  output_fn <- tester(weights = c(0.6, 0))
  output_list <- output_fn(0)
  expect_is(output_list, "list")
  expect_length(output_list, 1)
  expect_named(output_list, "s(25,50);w(0.6,0)")

  expect_equal(output_fn(c(5, 30, 55))[[1]], c(1, 0.6, 0))

  output_fn <- tester(
    steps = list(c(25, 50), c(25, 50)),
    weights = list(c(0.5, 0), c(0.6, 0))
  )
  output_list <- output_fn(0)
  expect_is(output_list, "list")
  expect_length(output_list, 2)
  expect_named(output_list, c("s(25,50);w(0.5,0)", "s(25,50);w(0.6,0)"))

  expect_equal(output_fn(c(5, 30, 55))[[1]], c(1, 0.5, 0))
  expect_equal(output_fn(c(5, 30, 55))[[2]], c(1, 0.6, 0))
})

test_that("steps intervals are open on the right", {
  output_fn <- tester(c(10, 20, 30, 40), c(0.75, 0.5, 0.25, 0))

  expect_identical(
    output_fn(c(0, 10, 20, 30, 40))[[1]],
    c(1, 0.75, 0.5, 0.25, 0)
  )
})

test_that("output fn returns empty numeric if receives empty numeric/integer", {
  output_fn <- tester()

  expect_identical(output_fn(integer())[[1]], numeric())
  expect_identical(output_fn(numeric())[[1]], numeric())
})

#' @keywords internal
warn_extra_cols <- function(travel_matrix, travel_cost, group_id, groups) {
  travel_matrix_extra_cols <- setdiff(names(travel_matrix), groups)
  travel_matrix_extra_cols <- setdiff(travel_matrix_extra_cols, travel_cost)
  if (group_id == "from_id") {
    travel_matrix_extra_cols <- setdiff(travel_matrix_extra_cols, "to_id")
  } else {
    travel_matrix_extra_cols <- setdiff(travel_matrix_extra_cols, "from_id")
  }

  if (!identical(travel_matrix_extra_cols, character(0))) {
    warning(
      "Found columns in 'travel_matrix' not listed in either ",
      "'travel_cost' or 'group_by': {",
      paste0("'", travel_matrix_extra_cols, "'", collapse = ","),
      "}.\n",
      "This can result in excessive aggregation when calculating ",
      "accessibility, if these columns serve as group ids in 'travel_matrix'.",
      call. = FALSE
    )
  }

  invisible(TRUE)
}

#' @keywords internal
warn_extra_access_cols <- function(accessibility_data, opportunity, group_id) {
  extra_cols <- setdiff(
    names(accessibility_data),
    c("id", opportunity, group_id)
  )

  if (!identical(extra_cols, character(0))) {
    warning(
      "Found columns in 'accessibility_data' not listed in either ",
      "'opportunity' or 'group_by': {",
      paste0("'", extra_cols, "'", collapse = ","),
      "}.\n",
      "This can result in excessive aggregation when calculating ",
      "inequality measures, if these columns serve as group ids in ",
      "'accessibility_data'.",
      call. = FALSE
    )
  }

  invisible(TRUE)
}

#' @param group_by A `character` vector. When not `character(0)` (the default),
#'   indicates the `accessibility_data` columns that should be used to group the
#'   inequality estimates by. For example, if `accessibility_data` includes a
#'   `scenario` column that identifies distinct scenarios that each
#'   accessibility estimates refer to (e.g. before and after a transport policy
#'   intervention), passing `"scenario"` to this parameter results in inequality
#'   estimates grouped by scenario.

#' Gini Index
#'
#' Calculates the Gini Index of a given accessibility distribution.
#'
#' @template accessibility_data
#' @template sociodem_data_without_income
#' @template opportunity_access
#' @template population
#' @template group_by_access
#'
#' @template return_inequality
#'
#' @family inequality
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' data_dir <- system.file("extdata", package = "accessibility")
#' travel_matrix <- readRDS(file.path(data_dir, "travel_matrix.rds"))
#' land_use_data <- readRDS(file.path(data_dir, "land_use_data.rds"))
#'
#' access <- cumulative_cutoff(
#'   travel_matrix,
#'   land_use_data,
#'   cutoff = 30,
#'   opportunity = "jobs",
#'   travel_cost = "travel_time"
#' )
#'
#' gini <- gini_index(
#'   access,
#'   sociodemographic_data = land_use_data,
#'   opportunity = "jobs",
#'   population = "population"
#' )
#' gini
#'
#' @export
gini_index <- function(accessibility_data,
                       sociodemographic_data,
                       opportunity,
                       population,
                       group_by = character(0)) {
  checkmate::assert_string(opportunity)
  checkmate::assert_string(population)
  assert_access_group_by(group_by)
  assert_accessibility_data(accessibility_data, opportunity, group_by)
  assert_sociodemographic_data(
    sociodemographic_data,
    accessibility_data,
    population = population
  )

  if (!inherits(accessibility_data, "data.table")) {
    original_class <- class(accessibility_data)
    data <- data.table::as.data.table(accessibility_data)
  } else {
    data <- data.table::copy(accessibility_data)
  }

  if (!inherits(sociodemographic_data, "data.table")) {
    sociodemographic_data <- data.table::as.data.table(sociodemographic_data)
  }

  merge_by_reference(
    data,
    sociodemographic_data,
    population,
    left_df_idcol = "id"
  )

  warn_extra_access_cols(accessibility_data, opportunity, group_by)

  .opp_colname <- opportunity
  .pop_colname <- population
  .groups <- group_by

  data.table::setorderv(data, cols = c(opportunity, group_by))

  data[
    ,
    .prop_pop := get(.pop_colname) / sum(get(.pop_colname)),
    by = .groups
  ]
  data[
    ,
    .cum_total_access := cumsum(
      as.numeric(get(.pop_colname) * get(.opp_colname))
    ) / sum(get(.pop_colname) * get(.opp_colname)),
    by = .groups
  ]

  # to calculate the area under the lorenz curve we simply have to calculate the
  # area of the trapezoids formed by the population on the x axis and the
  # cumulative total access in the y axis.
  # since the lorenz curve starts from the (0,0) point, we have to "add" a 0
  # value to to the beginning of the cumulative total access distribution
  # (otherwise we would have a trapezoid in the first segment, whereas it should
  # always be a triangle)

  data[
    ,
    .small_edge := data.table::shift(.cum_total_access, fill = 0),
    by = .groups
  ]
  data[, .big_edge := .cum_total_access]

  data[, .area_under_curve := ((.small_edge + .big_edge) * .prop_pop) / 2]

  if (nrow(data) == 0 && identical(group_by, character(0))) {
    gini <- data.table::data.table(gini_index = numeric(0))
  } else {
    gini <- data[
      ,
      .(gini_index = 1 - 2 * sum(.area_under_curve)), by = .groups
    ]
  }

  if (exists("original_class")) class(gini) <- original_class

  return(gini)
}

#' Cumulative access based on a travel cost cutoff
#'
#' Calculates the number of opportunities accessible under a given specified
#' travel cost cutoff.
#' @template description_generic_cost
#'
#' @template travel_matrix
#' @template land_use_data
#' @template opportunity
#' @param travel_cost A `character` vector. The name of the columns in
#'   `travel_matrix` with the travel costs between origins and destinations to
#'   be considered in the calculation.
#' @param cutoff Either a `numeric` vector or a list of `numeric` vectors, one
#'   for each cost specified in `travel_cost`. The travel cost cutoffs to
#'   consider when calculating accessibility levels. If a list, the function
#'   finds every single possible cutoff combination and use them to calculate
#'   accessibility (e.g. if one specifies that travel time cutoffs should be 30
#'   and 60 minutes and that monetary cost cutoffs should be 5 and 10 dollars,
#'   the output includes accessibility estimates limited at 30 min & 5 dollars,
#'   30 min & 10 dollars, 60 min & 5 dollars and 60 min & 10 dollars). In these
#'   cases, cost constraints are considered simultaneously - i.e. only trips
#'   that take 30 minutes or less AND 5 dollars or less to be completed, for
#'   example, are included in the accessibility output. The cutoff parameter is
#'   not included in the final output if the input includes only a single cutoff
#'   for a single travel cost.
#' @template group_by
#' @template active
#' @param fill_missing_ids A `logical`. Calculating cumulative accessibility may
#'   result in missing ids if the they cannot reach any of the destinations
#'   within the specified travel cost cutoff. For example, using a travel time
#'   cutoff of 20 minutes, when estimating the accessibility of origin `A` that
#'   can only reach destinations with more than 40 minutes results in id `A`
#'   not being included in the output. When `TRUE` (the default), the function
#'   identifies which origins would be left out and fills their respective
#'   accessibility values with 0, which incurs in a performance penalty.
#'
#' @template return_accessibility
#'
#' @family cumulative access
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' data_dir <- system.file("extdata", package = "accessibility")
#' travel_matrix <- readRDS(file.path(data_dir, "travel_matrix.rds"))
#' land_use_data <- readRDS(file.path(data_dir, "land_use_data.rds"))
#'
#' # active accessibility: number of schools accessible from each origin
#' df <- cumulative_cutoff(
#'   travel_matrix = travel_matrix,
#'   land_use_data = land_use_data,
#'   cutoff = 30,
#'   opportunity = "schools",
#'   travel_cost = "travel_time"
#' )
#' head(df)
#'
#' df <- cumulative_cutoff(
#'   travel_matrix = travel_matrix,
#'   land_use_data = land_use_data,
#'   cutoff = c(30, 60),
#'   opportunity = "schools",
#'   travel_cost = "travel_time"
#' )
#' head(df)
#'
#' # passive accessibility: number of people that can reach each destination
#' df <- cumulative_cutoff(
#'   travel_matrix = travel_matrix,
#'   land_use_data = land_use_data,
#'   cutoff = 30,
#'   opportunity = "population",
#'   travel_cost = "travel_time",
#'   active = FALSE
#' )
#' head(df)
#'
#' # using multiple travel costs
#' pareto_frontier <- readRDS(file.path(data_dir, "pareto_frontier.rds"))
#'
#' df <- cumulative_cutoff(
#'   pareto_frontier,
#'   land_use_data = land_use_data,
#'   opportunity = "jobs",
#'   travel_cost = c("travel_time", "monetary_cost"),
#'   cutoff = list(c(20, 30), c(0, 5, 10))
#' )
#' head(df)
#'
#' @export
cumulative_cutoff <- function(travel_matrix,
                              land_use_data,
                              opportunity,
                              travel_cost,
                              cutoff,
                              group_by = character(0),
                              active = TRUE,
                              fill_missing_ids = TRUE) {

  checkmate::assert_string(opportunity)
  checkmate::assert_character(
    travel_cost,
    any.missing = FALSE,
    min.len = 1,
    unique = TRUE
  )
  checkmate::assert_logical(active, len = 1, any.missing = FALSE)
  checkmate::assert_logical(fill_missing_ids, len = 1, any.missing = FALSE)
  assert_cutoff(cutoff, travel_cost)
  assert_group_by(group_by)
  assert_travel_matrix(travel_matrix, travel_cost, group_by)
  assert_land_use_data(
    land_use_data,
    travel_matrix,
    opportunity,
    active = active
  )

  # if not a dt, keep original class to assign later when returning result

  if (!inherits(travel_matrix, "data.table")) {
    original_class <- class(travel_matrix)
    data <- data.table::as.data.table(travel_matrix)
  } else {
    data <- data.table::copy(travel_matrix)
  }

  if (!inherits(land_use_data, "data.table")) {
    land_use_data <- data.table::as.data.table(land_use_data)
  }

  merge_by_reference(
    data,
    land_use_data,
    opportunity,
    left_df_idcol = ifelse(active, "to_id", "from_id")
  )

  group_id <- ifelse(active, "from_id", "to_id")
  groups <- c(group_id, group_by)
  warn_extra_cols(travel_matrix, travel_cost, group_id, groups)

  cutoff_list <- if (is.list(cutoff)) {
    cutoff
  } else {
    list(cutoff)
  }
  names(cutoff_list) <- travel_cost
  cutoff_df <- do.call(data.table::CJ, cutoff_list)

  env <- environment()
  .opportunity_colname <- opportunity

  access <- do.call(
    mapply,
    args = c(
      as.list(cutoff_df),
      SIMPLIFY = FALSE,
      FUN = function(...) {
        cutoff_args <- unlist(list(...))
        subset_expr <- paste(
          names(cutoff_args),
          cutoff_args,
          sep = " <= ",
          collapse = " & "
        )
        subset_expr <- parse(text = subset_expr)

        cum_opps <- data[eval(subset_expr)]

        if (length(travel_cost) > 1) {
          .od_group <- c(setdiff(groups, group_id), "from_id", "to_id")
          cum_opps <- cum_opps[cum_opps[, .I[1], by = .od_group]$V1]
        }

        cum_opps <- cum_opps[
          ,
          .(access = sum(get(.opportunity_colname))),
          by = eval(groups, envir = env)
        ]

        cutoff_id_expr <- paste0(
          "`:=`(",
          paste(
            names(cutoff_args),
            cutoff_args,
            sep = " = ",
            collapse = ", "
          ),
          ")"
        )
        cutoff_id_expr <- parse(text = cutoff_id_expr)

        cum_opps[, eval(cutoff_id_expr)]
      }
    )
  )

  access <- data.table::rbindlist(access)

  if (fill_missing_ids) {
    unique_values <- lapply(groups, function(x) unique(travel_matrix[[x]]))
    names(unique_values) <- groups
    unique_values <- append(unique_values, cutoff_list)

    possible_combinations <- do.call(data.table::CJ, unique_values)

    if (nrow(access) < nrow(possible_combinations)) {
      access <- do_fill_missing_ids(
        access,
        possible_combinations,
        groups = c(groups, travel_cost)
      )
    }
  }

  data.table::setnames(access, c(group_id, "access"), c("id", opportunity))
  data.table::setcolorder(
    access,
    c("id", setdiff(groups, group_id), travel_cost, opportunity)
  )
  data.table::setorderv(access, c("id", setdiff(groups, group_id), travel_cost))

  env <- environment()
  if (length(travel_cost) == 1 && length(cutoff) == 1) {
    access[, get("travel_cost", envir = env) := NULL]
  }

  if (exists("original_class")) class(access) <- original_class

  return(access[])
}

# if running manually, please run the following line first:
# source("tests/testthat/setup.R")

tester <- function(travel_matrix = smaller_matrix,
                   land_use_data = get("land_use_data", envir = parent.frame()),
                   opportunity = "jobs",
                   travel_cost = "travel_time",
                   cutoff = 30,
                   group_by = "mode",
                   active = TRUE,
                   fill_missing_ids = TRUE) {
  cumulative_cutoff(
    travel_matrix,
    land_use_data,
    opportunity,
    travel_cost,
    cutoff,
    group_by,
    active,
    fill_missing_ids
  )
}

tester_with_cost <- function(travel_matrix = small_frontier,
                             travel_cost = c("travel_time", "monetary_cost"),
                             cutoff = list(30, 10),
                             ...) {
  tester(travel_matrix, travel_cost = travel_cost, cutoff = cutoff, ...)
}

test_that("raises errors due to incorrect input", {
  expect_error(tester(opportunity = 1))
  expect_error(tester(opportunity = c("schools", "jobs")))

  expect_error(tester(travel_cost = 1))
  expect_error(tester(travel_cost = NA_character_))
  expect_error(tester(travel_cost = character()))
  expect_error(tester(travel_cost = c("travel_time", "travel_time")))

  expect_error(tester(cutoff = "banana"))
  expect_error(tester(cutoff = -3))
  expect_error(tester(cutoff = Inf))
  expect_error(tester(cutoff = NA_real_))
  expect_error(tester(cutoff = numeric(0)))
  expect_error(tester(cutoff = c(1, 1)))
  expect_error(tester_with_cost(cutoff = 3))
  expect_error(tester_with_cost(cutoff = list(3)))
  expect_error(tester_with_cost(cutoff = list(3, "a")))
  expect_error(tester_with_cost(cutoff = list(3, -3)))
  expect_error(tester_with_cost(cutoff = list(3, Inf)))
  expect_error(tester_with_cost(cutoff = list(3, NA_real_)))
  expect_error(tester_with_cost(cutoff = list(3, numeric())))
  expect_error(tester_with_cost(cutoff = list(3, c(1, 1))))

  expect_error(tester(group_by = 1))
  expect_error(tester(group_by = NA))
  expect_error(tester(group_by = "from_id"))
  expect_error(tester(group_by = c("mode", "mode")))

  expect_error(tester(active = 1))
  expect_error(tester(active = c(TRUE, TRUE)))
  expect_error(tester(active = NA))

  expect_error(tester(fill_missing_ids = 1))
  expect_error(tester(fill_missing_ids = c(TRUE, TRUE)))
  expect_error(tester(fill_missing_ids = NA))

  expect_error(tester(as.list(travel_matrix)))
  expect_error(tester(travel_matrix[, .(oi = from_id, to_id, travel_time)]))
  expect_error(tester(travel_matrix[, .(from_id, oi = to_id, travel_time)]))
  expect_error(
    tester(
      travel_matrix[, .(from_id, to_id, oi = travel_time)],
      travel_cost = "travel_time"
    )
  )
  expect_error(
    tester(
      travel_matrix[, .(from_id, to_id, travel_time, oi = mode)],
      group_by = "mode"
    )
  )

  expect_error(tester(as.list(land_use_data)))
  expect_error(tester(land_use_data = land_use_data[, .(oi = id, jobs)]))
  expect_error(
    tester(
      land_use_data = land_use_data[, .(id, oi = jobs)],
      opportunity = "jobs"
    )
  )
})

test_that("throws warning if travel_matrix has an extra col", {
  # i.e. col not listed in travel_cost and by_col
  expect_warning(tester(group_by = character(0)))
})

test_that("returns a dataframe whose class is the same as travel_matrix's", {
  result <- tester()
  expect_is(result, "data.table")
  result <- tester(land_use_data = as.data.frame(land_use_data))
  expect_is(result, "data.table")

  result <- tester(as.data.frame(smaller_matrix))
  expect_false(inherits(result, "data.table"))
  expect_is(result, "data.frame")
  result <- tester(
    as.data.frame(smaller_matrix),
    land_use_data = as.data.frame(land_use_data)
  )
  expect_false(inherits(result, "data.table"))
  expect_is(result, "data.frame")
})

test_that("result has correct structure", {
  result <- tester()
  expect_true(ncol(result) == 3)
  expect_is(result$id, "character")
  expect_is(result$mode, "character")
  expect_is(result$jobs, "integer")

  result <- tester(opportunity = "schools")
  expect_true(ncol(result) == 3)
  expect_is(result$id, "character")
  expect_is(result$mode, "character")
  expect_is(result$schools, "integer")

  result <- tester(cutoff = c(15, 30))
  expect_true(ncol(result) == 4)
  expect_is(result$id, "character")
  expect_is(result$mode, "character")
  expect_is(result$travel_time, "numeric")
  expect_is(result$jobs, "integer")

  suppressWarnings(result <- tester(group_by = character(0)))
  expect_true(ncol(result) == 2)
  expect_is(result$id, "character")
  expect_is(result$jobs, "integer")

  result <- tester(
    data.table::data.table(
      mode = character(),
      from_id = character(),
      to_id = character(),
      travel_time = integer()
    )
  )
  expect_true(ncol(result) == 3)
  expect_true(nrow(result) == 0)
  expect_is(result$id, "character")
  expect_is(result$mode, "character")
  expect_is(result$jobs, "integer")

  result <- tester(
    data.table::data.table(
      mode = character(),
      from_id = character(),
      to_id = character(),
      travel_time = integer()
    ),
    cutoff = c(15, 30)
  )
  expect_true(ncol(result) == 4)
  expect_true(nrow(result) == 0)
  expect_is(result$id, "character")
  expect_is(result$mode, "character")
  expect_is(result$travel_time, "numeric")
  expect_is(result$jobs, "integer")

  result <- tester_with_cost()
  expect_true(ncol(result) == 5)
  expect_is(result$id, "character")
  expect_is(result$mode, "character")
  expect_is(result$travel_time, "numeric")
  expect_is(result$monetary_cost, "numeric")
  expect_is(result$jobs, "integer")

  suppressWarnings(result <- tester_with_cost(group_by = character()))
  expect_true(ncol(result) == 4)
  expect_is(result$id, "character")
  expect_is(result$travel_time, "numeric")
  expect_is(result$monetary_cost, "numeric")
  expect_is(result$jobs, "integer")

  result <- tester_with_cost(small_frontier[0])
  expect_true(ncol(result) == 5)
  expect_is(result$id, "character")
  expect_is(result$mode, "character")
  expect_is(result$travel_time, "numeric")
  expect_is(result$monetary_cost, "numeric")
  expect_is(result$jobs, "integer")
})

test_that("input data sets remain unchanged", {
  original_smaller_matrix <- travel_matrix[1:10]
  original_land_use_data <- readRDS(file.path(data_dir, "land_use_data.rds"))

  result <- tester()

  # subsets in other functions tests set smaller_matrix index
  data.table::setindex(smaller_matrix, NULL)

  expect_equal(original_smaller_matrix, smaller_matrix)
  expect_equal(original_land_use_data, land_use_data)
})

test_that("active and passive accessibility is correctly calculated", {
  selected_ids <- c(
    "89a88cd909bffff",
    "89a88cdb57bffff",
    "89a88cdb597ffff",
    "89a88cdb5b3ffff",
    "89a88cdb5cfffff"
  )
  smaller_travel_matrix <- travel_matrix[
    from_id %in% selected_ids & to_id %in% selected_ids
  ]

  result <- tester(smaller_travel_matrix, group_by = "mode")
  expected_result <- data.table::data.table(
    id = rep(selected_ids, each = 2),
    mode = rep(c("transit", "transit2"), 5),
    jobs = rep(as.integer(c(0, 82, 408, 408, 109)), each = 2)
  )
  expect_identical(result, expected_result)

  result <- tester(
    smaller_travel_matrix,
    cutoff = 45,
    opportunity = "population",
    active = FALSE
  )
  expected_result <- data.table::data.table(
    id = rep(selected_ids, each = 2),
    mode = rep(c("transit", "transit2"), 5),
    population = rep(as.integer(c(0, 5404, 4552, 2363, 4552)), each = 2)
  )
  expect_identical(result, expected_result)

  # when the provided matrix/frontier includes more than 1 trip per od pair -
  # i.e. not double counting opportunities when more than 1 trip can be used
  # between origins and destinations

  test_frontier <- pareto_frontier[
    from_id %in% selected_ids & to_id %in% selected_ids
  ]

  result <- tester_with_cost(test_frontier, cutoff = list(120, c(5, 15)))
  expected_result <- data.table::data.table(
    id = rep(selected_ids, each = 4),
    mode = rep(rep(c("transit", "transit2"), each = 2), times = 5),
    travel_time = 120,
    monetary_cost = rep(c(5, 15), 10),
    jobs = rep(as.integer(c(499, 599, 499, 599, rep(599, 16))))
  )
  expect_identical(result, expected_result)

  result <- tester_with_cost(
    test_frontier,
    cutoff = list(120, c(5, 15)),
    opportunity = "population",
    active = FALSE
  )
  expected_result <- data.table::data.table(
    id = rep(selected_ids, each = 4),
    mode = rep(rep(c("transit", "transit2"), each = 2), times = 5),
    travel_time = 120,
    monetary_cost = rep(c(5, 15), 10),
    population = rep(as.integer(c(3435, 5404, 3435, 5404, rep(5404, 16))))
  )
  expect_identical(result, expected_result)
})

test_that("fill_missing_ids arg works correctly", {
  # with one travel cost and one cutoff

  small_travel_matrix <- travel_matrix[
    from_id %in% c("89a88cdb57bffff", "89a88cdb597ffff")
  ]
  small_travel_matrix <- small_travel_matrix[
    from_id != "89a88cdb57bffff" | travel_time > 40
  ]

  result <- tester(small_travel_matrix)
  data.table::setkey(result, NULL)
  expect_identical(
    result,
    data.table::data.table(
      id = rep(c("89a88cdb57bffff", "89a88cdb597ffff"), each = 2),
      mode = rep(c("transit", "transit2"), times = 2),
      jobs = rep(as.integer(c(0, 36567)), each = 2)
    )
  )

  result <- tester(
    small_travel_matrix,
    land_use_data,
    fill_missing_ids = FALSE
  )
  data.table::setkey(result, NULL)
  expect_identical(
    result,
    data.table::data.table(
      id = rep("89a88cdb597ffff", each = 2),
      mode = c("transit", "transit2"),
      jobs = rep(36567L, each = 2)
    )
  )

  # with one travel cost and more than one cutoff

  result <- tester(small_travel_matrix, cutoff = c(15, 50))
  data.table::setkey(result, NULL)
  expected_result <- data.table::data.table(
    id = rep(c("89a88cdb57bffff", "89a88cdb597ffff"), each = 4),
    mode = rep(rep(c("transit", "transit2"), each = 2), times = 2),
    travel_time = rep(c(15, 50), times = 4),
    jobs = as.integer(c(rep(c(0, 187799), 2), rep(c(3008, 257648), 2)))
  )

  expect_identical(result, expected_result)

  result <- tester(
    small_travel_matrix,
    cutoff = c(15, 50),
    fill_missing_ids = FALSE
  )
  data.table::setkey(result, NULL)

  expect_identical(result, expected_result[jobs != 0])

  # with more than one travel cost

  test_frontier <- rbind(small_frontier, small_frontier[1])
  test_frontier[11, mode := "transit2"]
  test_frontier[11, travel_time := 100]

  result <- tester_with_cost(test_frontier, cutoff = list(10, 10))
  expected_result <- data.table::data.table(
    id = "89a881a5a2bffff",
    mode = c("transit", "transit2"),
    travel_time = 10,
    monetary_cost = 10,
    jobs = as.integer(c(323, 0))
  )
  data.table::setkey(result, NULL)
  expect_identical(result, expected_result)

  result <- tester_with_cost(
    test_frontier,
    cutoff = list(10, 10),
    fill_missing_ids = FALSE
  )
  expect_identical(result, expected_result[1])
})

test_that("works even if travel_matrix and land_use has specific colnames", {
  expected_result <- tester()

  smaller_matrix[, opportunity := "oi"]
  result <- suppressWarnings(tester(smaller_matrix))
  expect_identical(expected_result, result)

  smaller_matrix[, opportunity := NULL]
  smaller_matrix[, travel_cost := "oi"]
  result <- suppressWarnings(tester(smaller_matrix))
  expect_identical(expected_result, result)

  smaller_matrix[, travel_cost := NULL]
  smaller_matrix[, groups := "oi"]
  result <- suppressWarnings(tester(smaller_matrix))
  expect_identical(expected_result, result)

  smaller_matrix[, groups := NULL]
  land_use_data[, opportunity := "oi"]
  result <- suppressWarnings(tester(land_use_data = land_use_data))
  expect_identical(expected_result, result)

  land_use_data[, opportunity := NULL]
})

# if running manually, please run the following line first:
# source("tests/testthat/setup.R")

tester <- function(accessibility_data = small_access,
                   sociodemographic_data = land_use_data,
                   opportunity = "jobs",
                   population = "population",
                   income = "income_per_capita",
                   group_by = "mode") {
  palma_ratio(
    accessibility_data,
    sociodemographic_data,
    opportunity,
    population,
    income,
    group_by
  )
}

test_that("raises errors due to incorrect input", {
  expect_error(tester(opportunity = 1))
  expect_error(tester(opportunity = c("schools", "jobs")))

  expect_error(tester(population = 1))
  expect_error(tester(population = c("schools", "jobs")))

  expect_error(tester(income = 1))
  expect_error(tester(income = c("schools", "jobs")))

  expect_error(tester(group_by = 1))
  expect_error(tester(group_by = NA))
  expect_error(tester(group_by = "id"))
  expect_error(tester(group_by = c("mode", "mode")))

  expect_error(tester(as.list(small_access)))
  expect_error(tester(small_access[, .(oi = id, jobs, mode)]))
  expect_error(tester(small_access[, .(id, oi = jobs, mode)]))
  expect_error(tester(small_access[, .(id, jobs, oi = mode)]))

  expect_error(tester(sociodemographic_data = as.list(land_use_data)))
  expect_error(
    tester(
      sociodemographic_data = land_use_data[
        ,
        .(oi = id, population, income_per_capita)
      ]
    )
  )
  expect_error(
    tester(
      sociodemographic_data = land_use_data[
        ,
        .(id, oi = population, income_per_capita)
      ]
    )
  )
  expect_error(
    tester(
      sociodemographic_data = land_use_data[
        ,
        .(id, population, oi = income_per_capita)
      ]
    )
  )
})

test_that("throws warning if accessibility_data has an extra col", {
  expect_warning(tester(group_by = character(0)))
})

test_that("returns a dataframe with same class as accessibility_data's", {
  result <- tester()
  expect_is(result, "data.table")
  result <- tester(sociodemographic_data = as.data.frame(land_use_data))
  expect_is(result, "data.table")

  result <- tester(as.data.frame(small_access))
  expect_false(inherits(result, "data.table"))
  expect_is(result, "data.frame")
  result <- tester(
    as.data.frame(small_access),
    sociodemographic_data = as.data.frame(land_use_data)
  )
  expect_false(inherits(result, "data.table"))
  expect_is(result, "data.frame")
})

test_that("result has correct structure", {
  result <- tester()
  expect_true(ncol(result) == 2)
  expect_is(result$mode, "character")
  expect_is(result$palma_ratio, "numeric")

  suppressWarnings(result <- tester(group_by = character(0)))
  expect_true(ncol(result) == 1)
  expect_is(result$palma_ratio, "numeric")

  result <- tester(
    data.table::data.table(
      mode = character(),
      id = character(),
      jobs = integer()
    )
  )
  expect_true(nrow(result) == 0)
  expect_true(ncol(result) == 2)
  expect_is(result$mode, "character")
  expect_is(result$palma_ratio, "numeric")

  suppressWarnings(
    result <- tester(
      data.table::data.table(
        mode = character(),
        id = character(),
        jobs = integer()
      ),
      group_by = character(0)
    )
  )
  expect_true(nrow(result) == 0)
  expect_true(ncol(result) == 1)
  expect_is(result$palma_ratio, "numeric")
})

test_that("input data sets remain unchanged", {
  original_access_data <- cumulative_cutoff(
    smaller_matrix,
    land_use_data,
    opportunity = "jobs",
    travel_cost = "travel_time",
    cutoff = 30,
    group_by = "mode"
  )
  original_sociodem_data <- readRDS(file.path(data_dir, "land_use_data.rds"))

  result <- tester()

  expect_equal(original_access_data, small_access)
  expect_equal(original_sociodem_data, land_use_data)
})

test_that("palma ratio is correctly calculated", {
  selected_ids <- c(
    "89a88cd909bffff",
    "89a88cdb57bffff",
    "89a88cdb597ffff",
    "89a88cdb5b3ffff",
    "89a88cdb5cfffff"
  )
  access_data <- cumulative_cutoff(
    travel_matrix[from_id %in% selected_ids],
    land_use_data,
    opportunity = "jobs",
    travel_cost = "travel_time",
    cutoff = 30,
    group_by = "mode"
  )

  result <- tester(access_data)
  result[, palma_ratio := round(palma_ratio, 4)]

  expected_result <- data.table::data.table(
    mode = c("transit", "transit2"),
    palma_ratio = rep(1.0911, 2)
  )
  expect_identical(result, expected_result)

  access_data <- access_data[!(id == "89a88cdb5cfffff" & mode == "transit2")]
  result <- tester(access_data)
  result[, palma_ratio := round(palma_ratio, 4)]

  expected_result <- data.table::data.table(
    mode = c("transit", "transit2"),
    palma_ratio = c(1.0911, 0.5249)
  )
  expect_identical(result, expected_result)
})

test_that("works even if access_data and sociodem_data has specific colnames", {
  selected_ids <- c(
    "89a88cdb57bffff",
    "89a88cdb5b3ffff"
  )
  access_data <- cumulative_cutoff(
    travel_matrix[from_id %in% selected_ids],
    land_use_data,
    opportunity = "jobs",
    travel_cost = "travel_time",
    cutoff = 30,
    group_by = "mode"
  )
  expected_result <- tester(access_data)

  access_data[, opportunity := "oi"]
  result <- suppressWarnings(tester(access_data))
  expect_identical(expected_result, result)

  access_data[, opportunity := NULL]
  land_use_data[, population_temp := population]
  land_use_data[, population := 1]
  result <- tester(access_data, population = "population_temp")
  expect_identical(expected_result, result)

  land_use_data[, population := population_temp]
  land_use_data[, population_temp := NULL]
  land_use_data[, income := "oi"]
  result <- tester(access_data)
  expect_identical(expected_result, result)

  land_use_data[, income := NULL]
  access_data[, group_by := "oi"]
  result <- suppressWarnings(tester(access_data))
  expect_identical(expected_result, result)

  access_data[, group_by := NULL]
})

test_that("handles missing data correctly", {
  # when access data is not null and it's missing either wealthiest or poorest
  # cells, palma should be NA, both with group_by = character(0) or something.
  # when access data is null, function should return a dataframe with 0 rows in
  # both cases

  selected_ids <- c(
    "89a88cdb57bffff",
    "89a88cdb5b3ffff"
  )
  custom_access <- cumulative_cutoff(
    travel_matrix[from_id %in% selected_ids],
    land_use_data,
    opportunity = "jobs",
    travel_cost = "travel_time",
    cutoff = 30,
    group_by = "mode"
  )

  result <- tester(
    custom_access[!(id == "89a88cdb5b3ffff" & mode == "transit2")]
  )
  result[, palma_ratio := round(palma_ratio, 4)]
  expect_identical(
    result,
    data.table::data.table(
      mode = c("transit", "transit2"),
      palma_ratio = c(0.5249, NA)
    )
  )

  result <- tester(
    custom_access[!(id == "89a88cdb57bffff" & mode == "transit2")]
  )
  result[, palma_ratio := round(palma_ratio, 4)]
  data.table::setkeyv(result, NULL)
  expect_identical(
    result,
    data.table::data.table(
      mode = c("transit", "transit2"),
      palma_ratio = c(0.5249, NA)
    )
  )

  custom_access <- custom_access[mode != "transit2"]

  # one row of custom_access is a wealthy cell and the other is a poor one

  result <- tester(custom_access[1])
  expect_identical(
    result,
    data.table::data.table(mode = "transit", palma_ratio = NA_real_)
  )

  result <- tester(custom_access[2])
  data.table::setkeyv(result, NULL)
  expect_identical(
    result,
    data.table::data.table(mode = "transit", palma_ratio = NA_real_)
  )

  suppressWarnings(result <- tester(custom_access[1], group_by = character()))
  expect_identical(result, data.table::data.table(palma_ratio = NA_real_))

  suppressWarnings(result <- tester(custom_access[2], group_by = character()))
  expect_identical(result, data.table::data.table(palma_ratio = NA_real_))

  no_access <- data.table::data.table(
    id = character(),
    mode = character(),
    jobs = integer()
  )

  result <- tester(no_access)
  expect_identical(
    result,
    data.table::data.table(mode = character(), palma_ratio = numeric())
  )

  suppressWarnings(result <- tester(no_access, group_by = character()))
  expect_identical(result, data.table::data.table(palma_ratio = numeric()))
})

#' Logistic decay function
#'
#' Returns a logistic weighting function (in which the weights
#' follow the distribution of a reversed cumulative logistic curve) to be
#' used inside accessibility calculating functions. The logistic curve is
#' parameterized with the cutoff that sets its inflection point and the standard
#' deviation that sets its steepness.
#' @template description_generic_cost
#'
#' @param cutoff A `numeric` vector. The cost value that serves as the
#'   inflection point of the cumulative logistic curve.
#' @param sd A `numeric` vector with same length as `cutoff`. The standard
#'   deviation of the logistic curve. Values near 0 result in weighting curves
#'   that approximate binary decay, while higher values tend to linearize the
#'   decay.
#'
#' @return A `function` that takes a generic travel cost vector (`numeric`) as
#'   input and returns a vector of weights (`numeric`).
#'
#' @details When using a function created with `decay_logistic()`, the output is
#'   named after the combination of cutoffs (`"c"`) and standard deviations
#'   (`"sd"`) - e.g. given the cutoffs `c(30, 40)` and the standard deviations
#'   `c(10, 20)`, the first element of the output will be named `"c30;sd10"` and
#'   the second will be named `"c40;sd20"`. This function uses the adjusted
#'   logistic decay curve proposed by
#'   \insertCite{bauer2016measuring;textual}{accessibility}, in which the
#'   condition f(0) = 1 is met (i.e. the weight of an opportunity whose cost to
#'   reach is 0 is 1).
#'
#' @references \insertAllCited{}
#'
#' @family decay functions
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' weighting_function <- decay_logistic(cutoff = 30, sd = 5)
#'
#' weighting_function(c(0, 30, 45, 60))
#'
#' weighting_function <- decay_logistic(cutoff = c(30, 45), sd = c(5, 10))
#'
#' weighting_function(c(0, 30, 45, 60))
#'
#' @export
decay_logistic <- function(cutoff, sd) {
  assert_logistic_cutoff(cutoff)
  assert_standard_deviation(sd, cutoff)

  weighting_function <- function(travel_cost) {
    g <- function(cost, median, std_dev) {
      1 + exp(((cost - median) * base::pi) / (std_dev * sqrt(3)))
    }

    sd_list <- mapply(
      median = cutoff,
      std_dev = sd,
      FUN = function(median, std_dev) {
        vapply(
          travel_cost,
          function(x) g(0, median, std_dev) / g(x, median, std_dev),
          numeric(1)
        )
      },
      SIMPLIFY = FALSE
    )

    list_names <- paste0("c", cutoff, ";sd", sd)
    names(sd_list) <- list_names

    return(sd_list)
  }

  return(weighting_function)
}

assert_logistic_cutoff <- function(cutoff) {
  checkmate::assert(
    checkmate::check_numeric(
      cutoff,
      finite = TRUE,
      any.missing = FALSE,
      min.len = 1
    ),
    check_higher_than_zero(cutoff),
    combine = "and"
  )
}

assert_standard_deviation <- function(sd, cutoff) {
  checkmate::assert(
    checkmate::check_numeric(
      sd,
      finite = TRUE,
      any.missing = FALSE,
      len = length(cutoff)
    ),
    check_higher_than_zero(sd),
    combine = "and"
  )
}

check_higher_than_zero <- function(x) {
  if (any(x <= 0)) return("Must be higher than 0")

  return(TRUE)
}

tester <- function(cutoff = c(20, 30), sd = c(2, 10)) {
  decay_logistic(cutoff, sd)
}

test_that("adequately raises errors", {
  expect_error(tester("a"))
  expect_error(tester(0))
  expect_error(tester(c(25, Inf)))
  expect_error(tester(c(25, NA)))
  expect_error(tester(integer()))

  expect_error(tester(cutoff = "a"))
  expect_error(tester(cutoff = 0))
  expect_error(tester(cutoff = 120))
  expect_error(tester(cutoff = c(25, Inf)))
  expect_error(tester(cutoff = c(25, NA)))
  expect_error(tester(cutoff = integer()))
})

test_that("output is a decay function that returns a list of numeric vctrs", {
  expect_is(tester(), "function")

  output_fn <- tester(c(20, 30), c(2, 10))
  output_list <- output_fn(0)
  expect_is(output_list, "list")
  expect_length(output_list, 2L)
  expect_named(output_list, c("c20;sd2", "c30;sd10"))

  expect_equal(
    output_fn(c(0, 20))[["c20;sd2"]],
    c(1, 0.5)
  )

  expect_equal(
    output_fn(c(0, 30))[["c30;sd10"]],
    c(1, 0.502),
    tolerance = 0.001
  )

  output_fn <- tester(20, 2)
  output_list <- output_fn(0)
  expect_is(output_list, "list")
  expect_length(output_list, 1L)
  expect_named(output_list, "c20;sd2")

  expect_equal(
    output_fn(c(0, 20))[["c20;sd2"]],
    c(1, 0.5)
  )
})

test_that("output fn returns empty numeric if receives empty numeric/integer", {
  output_fn <- tester(20, 2)

  expect_identical(output_fn(integer())[["c20;sd2"]], numeric())
  expect_identical(output_fn(numeric())[["c20;sd2"]], numeric())
})

#' @param group_by A `character` vector. When not `character(0)` (the default),
#'   indicates the `travel_matrix` columns that should be used to group the
#'   accessibility estimates by. For example, if `travel_matrix` includes a
#'   departure time column, that specifies the departure time of each entry in
#'   the data frame, passing `"departure_time"` to this parameter results in
#'   accessibility estimates grouped by origin and by departure time.

#' @return A `function` that takes a generic travel cost vector (`numeric`) as
#'   an input and returns a `list` of weight vectors (a list of `numeric`
#'   vectors, named after the arguments passed to the decay function).

data_dir <- system.file("extdata", package = "accessibility")

travel_matrix <- readRDS(file.path(data_dir, "travel_matrix.rds"))
travel_matrix_list <- list(transit = travel_matrix, transit2 = travel_matrix)
travel_matrix <- data.table::rbindlist(travel_matrix_list, idcol = "mode")

smaller_matrix <- travel_matrix[1:10]

land_use_data <- readRDS(file.path(data_dir, "land_use_data.rds"))

small_access <- cumulative_cutoff(
  smaller_matrix,
  land_use_data,
  opportunity = "jobs",
  travel_cost = "travel_time",
  cutoff = 30,
  group_by = "mode"
)

pareto_frontier <- readRDS(file.path(data_dir, "pareto_frontier.rds"))
frontier_list <- list(transit = pareto_frontier, transit2 = pareto_frontier)
pareto_frontier <- data.table::rbindlist(frontier_list, idcol = "mode")

small_frontier <- pareto_frontier[1:10]

#' @param orig_col A `string` with the name of the column of origin ids.
#' @param dest_col A `string` with the name of the column of destination ids.
#' @param population_col A `string` with the name of the column of origin with
#'       population count.

library(hexSticker) # https://github.com/GuangchuangYu/hexSticker
library(ggplot2)
library(sf)
library(accessibility)
library(sysfonts)
library(data.table)

# add special text font
sysfonts::font_add_google(name = "Roboto", family = "Roboto")

### setup ------------------------

library(accessibility)
library(data.table)
library(ggplot2)
library(sf)
library(interp)
library(sfheaders)



ttm_path <- system.file("extdata/travel_matrix.rds", package = "accessibility")
grid_path <- system.file("extdata/grid_bho.rds", package = "accessibility")
landuse_path <- system.file("extdata/land_use_data.rds", package = "accessibility")
ttm <- readRDS(ttm_path)
grid <- readRDS(grid_path)
landuse <- readRDS(landuse_path)

setdiff(ttm$from_id, grid$id)

unique(ttm$from_id) |> length()
unique(grid$id) |> length()

# Active accessibility: number of schools accessible from each origin
df <- accessibility::cumulative_cutoff(travel_matrix = ttm,
                                       land_use_data =landuse,
                                       travel_cost = 'travel_time',
                                       opportunity = 'jobs',
                                       cutoff = 30)

# df <- accessibility::gravity_access(data = ttm,
#                                     opportunity_colname = 'jobs',
#                                     decay_function = 'inverse_power',
#                                     decay_value = .5,
#                                     by_colname = 'from_id')

# access
df2 <- df[setDT(grid), on=c('id'='id'), geom := i.geom]

df2 <- st_sf(df2)



### network plot  ------------------------

# plot results
fig <- ggplot() +
  geom_sf(data=df2, aes(fill=jobs), color=NA, show.legend = F) +
  scale_fill_viridis_c() +
  theme_void() +
  theme(panel.grid.major=element_line(colour="transparent"))

fig

########### spatial interpolation --------------------


# interpolate estimates to get spatially smooth result
temp_xy <-  df2 |>
            st_centroid() |>
            sfheaders::sf_to_df(fill = T)

accss_interp <- interp(temp_xy$x, temp_xy$y, temp_xy$access) |>
                with(cbind(access=as.vector(z),  # Column-major order
                           x=rep(x, times=length(y)),
                           y=rep(y, each=length(x)))) |>
                as.data.frame() %>% na.omit()

# find results' bounding box to crop the map
bb_x <- c(min(accss_interp$x), max(accss_interp$x))
bb_y <- c(min(accss_interp$y), max(accss_interp$y))

fig_interp <- ggplot(data=accss_interp) +
              geom_contour_filled(aes(x=x, y=y, z=access), alpha=.8, show.legend = F) +
              # scale_fill_viridis_c() +
              theme_void() +
              theme(panel.grid.major=element_line(colour="transparent")) +
              coord_sf(xlim = bb_x, ylim = bb_y, datum = NA) +
              scale_x_continuous(expand=c(0,0)) +
              scale_y_continuous(expand=c(0,0))



### save sticker  ------------------------

# big
sticker(fig,

        # package name
        package= expression( italic('accessibility')),  p_size= 4, p_y = 1.6, p_color = "gray95", p_family="Roboto",

        # ggplot image size and position
        s_x=1, s_y=.85, s_width=1.4, s_height=1.4,

        # blue hexagon
        h_fill="#440154", h_color="#440154", h_size=1.3,

        ## blackhexagon
        # h_fill="gray20", h_color="gray80", h_size=1.3,

        # url
        url = "github.com/ipeaGIT/accessibility", u_color= "gray95", u_family = "Roboto", u_size = 1.2,

        # save output name and resolution
        filename="./man/figures/logo.png", dpi=300 #
)

#' Minimum travel cost to closest N number of opportunities
#'
#' Calculates the minimum travel cost to the closest N number of opportunities.
#' @template description_generic_cost
#'
#' @template travel_matrix
#' @template land_use_data
#' @template opportunity
#' @template travel_cost
#' @param n A `numeric` vector. The minimum number of opportunities that should
#'   be considered. Defaults to 1. If more than one value is provided, the
#'   output includes an extra column specifying the number of opportunities that
#'   the minimum travel cost refers to.
#' @template group_by
#' @template active
#' @param fill_missing_ids A `logical`. Calculating minimum travel cost to
#'   closest N number of opportunities may result in missing ids in the output
#'   if they cannot reach the specified amount of opportunities across all
#'   destinations they can reach. For example, estimating the minimum travel
#'   time that an origin that can only reach 4 opportunities takes to reach 5
#'   opportunities resulting in such origin not being included in the output.
#'   When `TRUE` (the default), the function identifies which ids would be left
#'   out from the output and fill their respective minimum travel costs with
#'   `Inf`, which incurs in a performance penalty.
#'
#' @template return_accessibility
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' data_dir <- system.file("extdata", package = "accessibility")
#' travel_matrix <- readRDS(file.path(data_dir, "travel_matrix.rds"))
#' land_use_data <- readRDS(file.path(data_dir, "land_use_data.rds"))
#'
#' df <- cost_to_closest(
#'   travel_matrix,
#'   land_use_data,
#'   n = 1,
#'   opportunity = "schools",
#'   travel_cost = "travel_time"
#' )
#' head(df)
#'
#' df <- cost_to_closest(
#'   travel_matrix,
#'   land_use_data,
#'   n = c(1, 2),
#'   opportunity = "schools",
#'   travel_cost = "travel_time"
#' )
#' head(df)
#'
#' @export
cost_to_closest <- function(travel_matrix,
                            land_use_data,
                            opportunity,
                            travel_cost,
                            n = 1,
                            group_by = character(0),
                            active = TRUE,
                            fill_missing_ids = TRUE) {
  checkmate::assert_numeric(
    n,
    lower = 1,
    finite = TRUE,
    any.missing = FALSE,
    min.len = 1,
    unique = TRUE
  )
  checkmate::assert_string(opportunity)
  checkmate::assert_string(travel_cost)
  checkmate::assert_logical(active, len = 1, any.missing = FALSE)
  checkmate::assert_logical(fill_missing_ids, len = 1, any.missing = FALSE)
  assert_group_by(group_by)
  assert_travel_matrix(travel_matrix, travel_cost, group_by)
  assert_land_use_data(
    land_use_data,
    travel_matrix,
    opportunity,
    active = active
  )

  # if not a dt, keep original class to assign later when returning result

  if (!inherits(travel_matrix, "data.table")) {
    original_class <- class(travel_matrix)
    data <- data.table::as.data.table(travel_matrix)
  } else {
    data <- data.table::copy(travel_matrix)
  }

  if (!inherits(land_use_data, "data.table")) {
    land_use_data <- data.table::as.data.table(land_use_data)
  }

  merge_by_reference(
    data,
    land_use_data,
    opportunity,
    left_df_idcol = ifelse(active, "to_id", "from_id")
  )

  group_id <- ifelse(active, "from_id", "to_id")
  groups <- c(group_id, group_by)
  warn_extra_cols(travel_matrix, travel_cost, group_id, groups)

  access <- lapply(
    n,
    function(.x) calculate_cost(data, groups, opportunity, travel_cost, .x)
  )
  names(access) <- n
  access <- data.table::rbindlist(access, idcol = "n")
  access[, n := as.numeric(n)]

  if (fill_missing_ids) {
    unique_values <- lapply(groups, function(x) unique(travel_matrix[[x]]))
    unique_values <- append(unique_values, list(n))
    names(unique_values) <- c(groups, "n")

    possible_combinations <- do.call(data.table::CJ, unique_values)

    if (nrow(access) < nrow(possible_combinations)) {
      access[, min_cost := as.numeric(min_cost)]

      access <- do_fill_missing_ids(
        access,
        possible_combinations,
        groups = c(groups, "n"),
        access_col = "min_cost",
        fill_value = Inf
      )
    }
  }

  data.table::setnames(
    access,
    c(group_id, "min_cost"),
    c("id", travel_cost)
  )
  data.table::setcolorder(
    access,
    c("id", setdiff(groups, group_id), "n", travel_cost)
  )
  if (length(n) == 1) access[, n := NULL]

  if (exists("original_class")) class(access) <- original_class

  return(access[])
}


calculate_cost <- function(data, groups, opportunity, travel_cost, .x) {
  env <- environment()
  .opportunity_colname <- opportunity
  .cost_colname <- travel_cost

  if (.x == 1) {
    cost <- data[
      get(.opportunity_colname) > 0,
      .(min_cost = suppressWarnings(min(get(.cost_colname)))),
      by = eval(groups, envir = env)
    ]
  } else {
    opport_cumsum <- data[get(.opportunity_colname) > 0, ]
    data.table::setorderv(opport_cumsum, c(groups, travel_cost))
    opport_cumsum[
      ,
      cum_opport := cumsum(get(.opportunity_colname)),
      by = eval(groups, envir = env)
    ]

    cost <- opport_cumsum[
      cum_opport >= .x,
      .(min_cost = suppressWarnings(min(get(.cost_colname)))),
      by = eval(groups, envir = env)
    ]
  }

  return(cost)
}

# if running manually, please run the following line first:
# source("tests/testthat/setup.R")

tester <- function(
  travel_matrix = smaller_matrix,
  land_use_data = get("land_use_data", envir = parent.frame()),
  opportunity = "schools",
  travel_cost = "travel_time",
  n = 1,
  group_by = "mode",
  active = TRUE,
  fill_missing_ids = TRUE
) {
  cost_to_closest(
    travel_matrix,
    land_use_data,
    opportunity,
    travel_cost,
    n,
    group_by,
    active,
    fill_missing_ids
  )
}

test_that("raises errors due to incorrect input", {
  expect_error(tester(n = "banana"))
  expect_error(tester(n = -3))
  expect_error(tester(n = Inf))
  expect_error(tester(n = NA))
  expect_error(tester(n = numeric(0)))
  expect_error(tester(n = c(1, 1)))

  expect_error(tester(opportunity = 1))
  expect_error(tester(opportunity = c("schools", "jobs")))

  expect_error(tester(travel_cost = 1))
  expect_error(tester(travel_cost = c("travel_time", "monetary_cost")))

  expect_error(tester(group_by = 1))
  expect_error(tester(group_by = NA))
  expect_error(tester(group_by = "from_id"))
  expect_error(tester(group_by = c("mode", "mode")))

  expect_error(tester(active = 1))
  expect_error(tester(active = c(TRUE, TRUE)))
  expect_error(tester(active = NA))

  expect_error(tester(fill_missing_ids = 1))
  expect_error(tester(fill_missing_ids = c(TRUE, TRUE)))
  expect_error(tester(fill_missing_ids = NA))

  expect_error(tester(as.list(travel_matrix)))
  expect_error(tester(travel_matrix[, .(oi = from_id, to_id, travel_time)]))
  expect_error(tester(travel_matrix[, .(from_id, oi = to_id, travel_time)]))
  expect_error(
    tester(
      travel_matrix[, .(from_id, to_id, oi = travel_time)],
      travel_cost = "travel_time"
    )
  )
  expect_error(
    tester(
      travel_matrix[, .(from_id, to_id, travel_time, oi = mode)],
      group_by = "mode"
    )
  )

  expect_error(tester(as.list(land_use_data)))
  expect_error(tester(land_use_data = land_use_data[, .(oi = id, jobs)]))
  expect_error(
    tester(
      land_use_data = land_use_data[, .(id, oi = jobs)],
      opportunity = "jobs"
    )
  )
})

test_that("throws warning if travel_matrix has an extra col", {
  # i.e. col not listed in travel_cost and by_col
  expect_warning(tester(group_by = character(0)))
})

test_that("returns a dataframe whose class is the same as travel_matrix's", {
  result <- tester()
  expect_is(result, "data.table")
  result <- tester(land_use_data = as.data.frame(land_use_data))
  expect_is(result, "data.table")

  result <- tester(as.data.frame(smaller_matrix))
  expect_false(inherits(result, "data.table"))
  expect_is(result, "data.frame")
  result <- tester(
    as.data.frame(smaller_matrix),
    land_use_data = as.data.frame(land_use_data)
  )
  expect_false(inherits(result, "data.table"))
  expect_is(result, "data.frame")
})

test_that("result has correct structure", {
  result <- tester()
  expect_true(ncol(result) == 3)
  expect_is(result$id, "character")
  expect_is(result$mode, "character")
  expect_is(result$travel_time, "numeric")

  result <- tester(n = c(1, 2))
  expect_true(ncol(result) == 4)
  expect_is(result$id, "character")
  expect_is(result$mode, "character")
  expect_is(result$n, "numeric")
  expect_is(result$travel_time, "numeric")

  suppressWarnings(result <- tester(group_by = character(0)))
  expect_true(ncol(result) == 2)
  expect_is(result$id, "character")
  expect_is(result$travel_time, "numeric")

  result <- tester(
    data.table::data.table(
      mode = character(),
      from_id = character(),
      to_id = character(),
      travel_time = numeric()
    )
  )
  expect_true(ncol(result) == 3)
  expect_true(nrow(result) == 0)
  expect_is(result$id, "character")
  expect_is(result$mode, "character")
  expect_is(result$travel_time, "numeric")

  result <- tester(
    data.table::data.table(
      mode = character(),
      from_id = character(),
      to_id = character(),
      travel_time = numeric()
    ),
    n = c(1, 2)
  )
  expect_true(ncol(result) == 4)
  expect_true(nrow(result) == 0)
  expect_is(result$id, "character")
  expect_is(result$mode, "character")
  expect_is(result$n, "numeric")
  expect_is(result$travel_time, "numeric")
})

test_that("result columns are ordered correctly when length(n)>1", {
  result <- tester(n = c(1, 3))
  expect_identical(names(result), c("id", "mode", "n", "travel_time"))

  result <- tester(n = c(1, 3), fill_missing_ids = FALSE)
  expect_identical(names(result), c("id", "mode", "n", "travel_time"))
})

test_that("input data sets remain unchanged", {
  original_smaller_matrix <- travel_matrix[1:10]
  original_land_use_data <- readRDS(file.path(data_dir, "land_use_data.rds"))

  result <- tester()

  # subsets in other functions tests set smaller_matrix index
  data.table::setindex(smaller_matrix, NULL)

  expect_equal(original_smaller_matrix, smaller_matrix)
  expect_equal(original_land_use_data, land_use_data)
})

test_that("active and passive accessibility is correctly calculated", {
  selected_ids <- c(
    "89a88cdb57bffff",
    "89a88cdb597ffff",
    "89a88cdb5b3ffff",
    "89a88cdb5cfffff",
    "89a88cd909bffff"
  )
  smaller_travel_matrix <- travel_matrix[
    from_id %in% selected_ids & to_id %in% selected_ids
  ]

  result <- tester(smaller_travel_matrix, group_by = "mode")
  expected_result <- data.table::data.table(
    id = rep(selected_ids, 2),
    mode = rep(c("transit", "transit2"), each = 5),
    travel_time = rep(c(47, 5.8, 13, 43, 70), 2)
  )
  expect_identical(result, expected_result)

  result <- tester(
    smaller_travel_matrix,
    opportunity = "population",
    group_by = "mode",
    active = FALSE
  )
  expected_result <- data.table::data.table(
    id = rep(selected_ids, 2),
    mode = rep(c("transit", "transit2"), each = 5),
    travel_time = rep(c(5.8, 5.8, 5.8, 5.8, 58), 2)
  )
  expect_identical(result, expected_result)
})

test_that("output is as expected when ids are missing - n=1", {
  selected_ids <- c(
    "89a88cdb57bffff",
    "89a88cdb597ffff",
    "89a88cdb5b3ffff",
    "89a88cdb5cfffff",
    "89a88cd909bffff"
  )
  smaller_travel_matrix <- travel_matrix[
    from_id %in% selected_ids & to_id %in% selected_ids
  ]
  smaller_travel_matrix <- rbind(
    smaller_travel_matrix,
    data.table::data.table(
      mode = "transit",
      from_id = "fake_id",
      to_id = "89a88cdb57bffff",
      travel_time = 50
    )
  )

  result <- tester(smaller_travel_matrix, n = 1)
  fake_id_result <- result[id == "fake_id"]
  data.table::setkey(fake_id_result, NULL)
  expected_result <- data.table::data.table(
    id = "fake_id",
    mode = c("transit", "transit2"),
    travel_time = Inf
  )
  expect_identical(fake_id_result, expected_result)

  result <- tester(smaller_travel_matrix, n = 1, fill_missing_ids = FALSE)
  fake_id_result <- result[id == "fake_id"]
  data.table::setkey(fake_id_result, NULL)
  expected_result <- data.table::data.table(
    id = character(0),
    mode = character(0),
    travel_time = numeric(0)
  )
  expect_identical(fake_id_result, expected_result)
})

test_that("output is as expected when ids are missing - n>1", {
  selected_ids <- c(
    "89a88cdb57bffff",
    "89a88cdb597ffff",
    "89a88cdb5b3ffff",
    "89a88cdb5cfffff",
    "89a88cd909bffff"
  )
  smaller_travel_matrix <- travel_matrix[
    from_id %in% selected_ids & to_id %in% selected_ids
  ]

  result <- tester(smaller_travel_matrix, n = c(1, 3))
  data.table::setkey(result, NULL)
  expected_result_subset <- data.table::data.table(
    id = rep(selected_ids[order(selected_ids)], each = 2),
    mode = rep(c("transit", "transit2"), times = 5),
    n = 3,
    travel_time = Inf
  )
  expect_identical(result[is.infinite(travel_time)], expected_result_subset)

  result <- tester(smaller_travel_matrix, n = c(1, 3), fill_missing_ids = FALSE)
  expected_result_subset <- data.table::data.table(
    id = character(0),
    mode = character(0),
    n = numeric(0),
    travel_time = numeric(0)
  )
  expect_identical(result[is.infinite(travel_time)], expected_result_subset)
})

test_that("output is as expected when ids are missing - length(n)>1", {
  selected_ids <- c(
    "89a88cdb57bffff",
    "89a88cdb597ffff",
    "89a88cdb5b3ffff",
    "89a88cdb5cfffff",
    "89a88cd909bffff"
  )
  smaller_travel_matrix <- travel_matrix[
    from_id %in% selected_ids & to_id %in% selected_ids
  ]

  result <- tester(smaller_travel_matrix, n = 3)
  data.table::setkey(result, NULL)
  expected_result <- data.table::data.table(
    id = rep(selected_ids[order(selected_ids)], each = 2),
    mode = rep(c("transit", "transit2"), times = 5),
    travel_time = Inf
  )
  expect_identical(result, expected_result)

  result <- tester(smaller_travel_matrix, n = 3, fill_missing_ids = FALSE)
  expected_result <- data.table::data.table(
    id = character(0),
    mode = character(0),
    travel_time = numeric(0)
  )
  expect_identical(result, expected_result)
})

test_that("works even if travel_matrix and land_use has specific colnames", {
  expected_result <- tester()

  smaller_matrix[, opportunity := "oi"]
  result <- suppressWarnings(tester(smaller_matrix))
  expect_identical(expected_result, result)

  smaller_matrix[, opportunity := NULL]
  smaller_matrix[, travel_cost := "oi"]
  result <- suppressWarnings(tester(smaller_matrix))
  expect_identical(expected_result, result)

  smaller_matrix[, travel_cost := NULL]
  smaller_matrix[, groups := "oi"]
  result <- suppressWarnings(tester(smaller_matrix))
  expect_identical(expected_result, result)

  smaller_matrix[, groups := NULL]
  land_use_data[, opportunity := "oi"]
  result <- suppressWarnings(tester(land_use_data = land_use_data))
  expect_identical(expected_result, result)

  land_use_data[, opportunity := NULL]
})

#' Gravity-based accessibility measures
#'
#' Calculates gravity-based accessibility using a decay function specified by
#' the user.
#' @template description_generic_cost
#'
#' @template travel_matrix
#' @template land_use_data
#' @template opportunity
#' @template travel_cost
#' @template decay_function
#' @template group_by
#' @template active
#' @template fill_missing_ids_combinations
#'
#' @template return_accessibility
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' data_dir <- system.file("extdata", package = "accessibility")
#' travel_matrix <- readRDS(file.path(data_dir, "travel_matrix.rds"))
#' land_use_data <- readRDS(file.path(data_dir, "land_use_data.rds"))
#'
#' df_linear <- gravity(
#'   travel_matrix,
#'   land_use_data,
#'   decay_function = decay_linear(cutoff = 50),
#'   opportunity = "schools",
#'   travel_cost = "travel_time"
#' )
#' head(df_linear)
#'
#' df_exp <- gravity(
#'   travel_matrix,
#'   land_use_data,
#'   decay_function = decay_exponential(decay_value = 0.5),
#'   opportunity = "schools",
#'   travel_cost = "travel_time"
#' )
#' head(df_exp)
#'
#' @export
gravity <- function(travel_matrix,
                    land_use_data,
                    opportunity,
                    travel_cost,
                    decay_function,
                    group_by = character(0),
                    active = TRUE,
                    fill_missing_ids = TRUE) {
  checkmate::assert_string(opportunity)
  checkmate::assert_string(travel_cost)
  checkmate::assert_logical(active, len = 1, any.missing = FALSE)
  checkmate::assert_logical(fill_missing_ids, len = 1, any.missing = FALSE)
  assert_decay_function(decay_function)
  assert_group_by(group_by)
  assert_travel_matrix(travel_matrix, travel_cost, group_by)
  assert_land_use_data(
    land_use_data,
    travel_matrix,
    opportunity,
    active = active
  )

  # if not a dt, keep original class to assign later when returning result

  if (!inherits(travel_matrix, "data.table")) {
    original_class <- class(travel_matrix)
    data <- data.table::as.data.table(travel_matrix)
  } else {
    data <- data.table::copy(travel_matrix)
  }

  if (!inherits(land_use_data, "data.table")) {
    land_use_data <- data.table::as.data.table(land_use_data)
  }

  merge_by_reference(
    data,
    land_use_data,
    opportunity,
    left_df_idcol = ifelse(active, "to_id", "from_id")
  )

  data <- apply_gravity_measure(data, decay_function, travel_cost)

  group_id <- ifelse(active, "from_id", "to_id")
  groups <- c(group_id, group_by)
  if ("decay_function_arg" %in% names(data)) {
    groups <- c(groups, "decay_function_arg")
  }
  env <- environment()

  warn_extra_cols(travel_matrix, travel_cost, group_id, groups)

  .opp_colname <- opportunity
  access <- data[
    ,
    .(access = sum(get(.opp_colname) * opp_weight)),
    by = eval(groups, envir = env)
  ]

  if (fill_missing_ids) {
    access <- fill_missing_ids(access, travel_matrix, groups)
  }

  data.table::setnames(access, c(group_id, "access"), c("id", opportunity))

  if (exists("original_class")) class(access) <- original_class

  return(access)
}

tentando a formula dos papers
#' Concentration index
#'
#' Calculates the Concentration Index to measure the socioeconomic inequality of
#' access to opportunities. It measures the extent to which inequalities in
#' access to opportunities are systematically associated with individuals'
#' socioeconomic levels.
#'
#'
#' @template accessibility_data
#' @template sociodem_data_without_income
#' @template opportunity_access
#' @template population
#' @template ses
#' @template type
#' @template group_by_access
#'
#' @template return_inequality
#'
#' @section The concentration index can be defined as a normalized sum of
#' weighted accessibility levels, with the weights being determined by the
#' socioeconomic rank of each individual. This means the CI can theoretically
#' vary between -1 and +1 (when all accessibility is concentrated in the most or
#' in the least disadvantaged person, respectively). Negative values indicate
#' that inequalities favor the poor, while positive values indicate a pro-rich
#' bias. As such, one advantage of CI is that the sign and the magnitude of the
#' index indicate, respectively, the direction and the strength of the
#' relationship between accessibility and socioeconomic position.
#'
#' @family inequality
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' data_dir <- system.file("extdata", package = "accessibility")
#' travel_matrix <- readRDS(file.path(data_dir, "travel_matrix.rds"))
#' land_use_data <- readRDS(file.path(data_dir, "land_use_data.rds"))
#'
#' access <- cumulative_cutoff(
#'   travel_matrix,
#'   land_use_data,
#'   cutoff = 30,
#'   opportunity = "jobs",
#'   travel_cost = "travel_time"
#' )
#'
#' ci <- concentration_index(
#'   accessibility_data = access,
#'   sociodemographic_data = land_use_data,
#'   opportunity = "jobs",
#'   population = "population",
#'   income = "income_per_capita",
#'   type = "CIc"
#' )
#' ci
#'
#' @export
concentration_index <- function(
                         accessibility_data,
                         sociodemographic_data,
                         opportunity,
                         population,
                         income,
                         type,
                         group_by = character(0)) {
  checkmate::assert_string(opportunity)
  checkmate::assert_string(population)
  checkmate::assert_string(income)
  assert_access_group_by(group_by)
  assert_accessibility_data(accessibility_data, opportunity, group_by)
  assert_sociodemographic_data(sociodemographic_data, c(population, income))


  if (!inherits(accessibility_data, "data.table")) {
    original_class <- class(accessibility_data)
    data <- data.table::as.data.table(accessibility_data)
  } else {
    data <- data.table::copy(accessibility_data)
  }

  if (!inherits(sociodemographic_data, "data.table")) {
    sociodemographic_data <- data.table::as.data.table(sociodemographic_data)
  }


  merge_by_reference(
    data,
    sociodemographic_data,
    population,
    left_df_idcol = "id"
  )
  merge_by_reference(data, sociodemographic_data, income, left_df_idcol = "id")

  warn_extra_access_cols(accessibility_data, opportunity, group_by)

  .opp_colname <- opportunity
  .pop_colname <- population
  .inc_colname <- income
  .groups <- group_by


  # remove obs with no pop na missing income
  data <- data[ get(.pop_colname) >0 ]
  data <- data[ ! is.na(get(.inc_colname)) ]
  data <- data[ ! is.na(get(.opp_colname)) ]

  ## order observations by SES
  data.table::setorderv(data, cols = c(group_by, income, population))

  # rank individuals
  data[, rank := 1:.N, by = .groups]

  ### fractional rank (for weighted data)
  # Eddy van Doorslaer and Xander Koolman (4)
  # https://doi.org/10.1002/hec.918
  data[, pop_share := get(.pop_colname)/sum(get(.pop_colname)) , by = .groups]
  data[, frank := (c(0, cumsum(pop_share)[-max(rank)]) + pop_share/2), by = .groups]

  data[, fw := frank - 0.5, by = .groups] # fractional weight

  # weight for access level of each individual
  data[, w := rank - ((.N + 1)/2), by = .groups]


  ### calcuate CI

    # standard CI
    if (type == 'CI') {

      ### unweighted ok
      #> Erreygers (2009) (24)
      #> https://doi.org/10.1016/j.jhealeco.2008.02.003
      data[, avg_access := mean(x = get(.opp_colname)), by = .groups]
      ci <- data[, .(CI = (2 / (.N^2 * (avg_access))) *
                      sum(get(.opp_colname) * w)
                    ), by = .groups]
      # ok 0.2411739

      ### weighted ok
      #> Eddy van Doorslaer and Xander Koolman (2004) (2)
      #> https://doi.org/10.1002/hec.918
      data[, avg_access := weighted.mean(x = get(.opp_colname),
                                         w = get(.pop_colname)
                                         ), by = .groups]

      ci <-  data[, .(CI = (2 / (sum(get(.pop_colname)) * (avg_access))) *
                        sum(get(.opp_colname) * get(.pop_colname) * frank) - 1
                      ), by = .groups]
      # ok 0.2864975

    }

    # # corrected CI
    if (type == 'CIc') {

      # upper and lower bounds
      data[, upper := max(get(.opp_colname), na.rm=TRUE), by = .groups] # ? by = .groups
      data[, lower := min(get(.opp_colname), na.rm=TRUE), by = .groups] # ? by = .groups
      # ou seriam max e min teoricos, e nao os observados

      ### unweighted ok
      #> Erreygers (2009) (27)
      #> https://doi.org/10.1016/j.jhealeco.2008.02.003

      # 0.262953
      ci <- data[, .(CI =
                       (8 / (.N ^ 2 * (upper - lower))) *
                       sum(w * get(.opp_colname))
                     ), by = .groups]


      ### weighted (unweighted + complemento do rineq)
      # 0.3346494
      ci <- data[, .(CI = ((2 / (sum(get(.pop_colname)) * (avg_access))) *
                 sum(get(.opp_colname) * get(.pop_colname) * frank) - 1) *
                   4 * ( avg_access / (upper -lower)) # complemento do ineq
                 ), by = .groups]
    }


  # return output
  ci <- unique(ci)
  return(ci)
}


library(data.table)
df <- merge(access,
            land_use_data[, .(id, population, income_per_capita)]
            , by = 'id') |> setDT()

# https://github.com/kdevkdev/rineq
rineq::ci( ineqvar = df$income_per_capita,
           outcome = df$jobs ,
           method='direct',
           weights = df$population,
           type= 'CI')
#> 0.241175
#> WWW 0.2865013


 concentration_index(
   accessibility_data = access,
   sociodemographic_data = land_use_data,
   opportunity = "jobs",
   population = "population",
   income = "income_per_capita",
   type = "CI")

# 0.2411739  unweighted OK


 #### --------------------------


 rineq::ci( ineqvar = df$income_per_capita,
            outcome = df$jobs ,
            method='direct',
            weights = df$population,
            type= 'CIc')
 #> 0.262953
 #> WWW 0.3346494



 concentration_index(
   accessibility_data = access,
   sociodemographic_data = land_use_data,
   opportunity = "jobs",
   population = "population",
   income = "income_per_capita",
   type = "CIc")

 0.2629539

#' Palma Ratio
#'
#' Calculates the Palma Ratio of a given accessibility distribution. Originally
#' defined as the income share of the richest 10% of a population divided by the
#' income share of the poorest 40%, this measure has been adapted in transport
#' planning as the average accessibility of the richest 10% divided by the
#' average accessibility of the poorest 40%.
#'
#' @template accessibility_data
#' @template sociodem_data_with_income
#' @template opportunity_access
#' @template population
#' @param income A string. The name of the column in `sociodemographic_data`
#'   with the income variable that should be used to classify the population in
#'   socioeconomic groups. Please note that this variable should describe income
#'   per capita (e.g. mean income per capita, household income per capita, etc),
#'   instead of the total amount of income in each cell.
#' @template group_by_access
#'
#' @template return_inequality
#'
#' @family inequality
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' data_dir <- system.file("extdata", package = "accessibility")
#' travel_matrix <- readRDS(file.path(data_dir, "travel_matrix.rds"))
#' land_use_data <- readRDS(file.path(data_dir, "land_use_data.rds"))
#'
#' access <- cumulative_cutoff(
#'   travel_matrix,
#'   land_use_data,
#'   cutoff = 30,
#'   opportunity = "jobs",
#'   travel_cost = "travel_time"
#' )
#'
#' palma <- palma_ratio(
#'   access,
#'   sociodemographic_data = land_use_data,
#'   opportunity = "jobs",
#'   population = "population",
#'   income = "income_per_capita"
#' )
#' palma
#'
#' @export
palma_ratio <- function(accessibility_data,
                        sociodemographic_data,
                        opportunity,
                        population,
                        income,
                        group_by = character(0)) {
  checkmate::assert_string(opportunity)
  checkmate::assert_string(population)
  checkmate::assert_string(income)
  assert_access_group_by(group_by)
  assert_accessibility_data(accessibility_data, opportunity, group_by)
  assert_sociodemographic_data(
    sociodemographic_data,
    accessibility_data,
    population = population,
    income = income
  )

  if (!inherits(accessibility_data, "data.table")) {
    original_class <- class(accessibility_data)
    data <- data.table::as.data.table(accessibility_data)
  } else {
    data <- data.table::copy(accessibility_data)
  }

  if (!inherits(sociodemographic_data, "data.table")) {
    sociodemographic_data <- data.table::as.data.table(sociodemographic_data)
  }

  merge_by_reference(
    data,
    sociodemographic_data,
    population,
    left_df_idcol = "id"
  )
  merge_by_reference(data, sociodemographic_data, income, left_df_idcol = "id")

  warn_extra_access_cols(accessibility_data, opportunity, group_by)

  income_quantiles <- Hmisc::wtd.quantile(
    sociodemographic_data[[income]],
    weights = sociodemographic_data[[population]],
    probs = c(0.4, 0.9)
  )

  .income_colname <- income
  data[
    get(.income_colname) <= income_quantiles["40%"],
    .palma_group := "poorest"
  ]
  data[
    get(.income_colname) > income_quantiles["90%"],
    .palma_group := "wealthiest"
  ]

  .opportunity_colname <- opportunity
  .population_colname <- population
  .groups <- group_by

  wealthiest_access <- data[
    .palma_group == "wealthiest",
    .(
      avg_access = stats::weighted.mean(
        get(.opportunity_colname),
        w = get(.population_colname)
      )
    ),
    by = .groups
  ]
  poorest_access <- data[
    .palma_group == "poorest",
    .(
      avg_access = stats::weighted.mean(
        get(.opportunity_colname),
        w = get(.population_colname)
      )
    ),
    by = .groups
  ]

  if (identical(group_by, character(0))) {
    pr <- if (
      (nrow(wealthiest_access) == 0 || nrow(poorest_access) == 0) &&
        nrow(data) > 0
    ) {
      NA_real_
    } else {
      wealthiest_access$avg_access / poorest_access$avg_access
    }

    palma <- data.table::data.table(palma_ratio = pr)
  } else {
    # if wealthiest_access is missing data on any of the groups, the result is
    # that this group will be missing from the final output, even though its
    # palma should be NA. thus, we have to fill wealthiest_access with missing
    # groups

    wealthiest_access <- fill_groups(wealthiest_access, data, group_by)

    palma <- wealthiest_access[
      poorest_access,
      on = group_by,
      poorest_avg_access := i.avg_access
    ]

    palma[, palma_ratio := avg_access / poorest_avg_access]
    palma[, c("avg_access", "poorest_avg_access") := NULL]
  }

  if (exists("original_class")) class(palma) <- original_class

  return(palma[])
}

fill_groups <- function(wealthiest_access, data, group_by) {
  unique_values <- lapply(group_by, function(x) unique(data[[x]]))
  names(unique_values) <- group_by

  possible_combinations <- do.call(data.table::CJ, unique_values)

  if (nrow(possible_combinations) > nrow(wealthiest_access)) {
    wealthiest_access <- possible_combinations[
      wealthiest_access,
      on = group_by,
      avg_access := i.avg_access
    ]
  }

  return(wealthiest_access[])
}

#' Concentration Index
#'
#' Calculates the Concentration Index (CI) of a given accessibility
#' distribution. This measures estimates the extent to which accessibility
#' inequalities are systematically associated with individuals' socioeconomic
#' levels. CI values can theoretically vary between -1 and +1 (when all
#' accessibility is concentrated in the most or in the least disadvantaged
#' person, respectively). Negative values indicate that inequalities favor the
#' poor, while positive values indicate a pro-rich bias. The function supports
#' calculating the standard relative CI and the corrected CI, as proposed by
#' \insertCite{erreygers2009correcting;textual}{accessibility}.
#'
#' @template accessibility_data
#' @template sociodem_data_with_income
#' @template opportunity_access
#' @template population
#' @param income A string. The name of the column in `sociodemographic_data`
#'   with the income variable that should be used to sort the population from
#'   the least to the most privileged. Please note that this variable should
#'   describe income per capita (e.g. mean income per capita, household income
#'   per capita, etc), instead of the total amount of income in each cell. Also
#'   note that, while income is generally used to rank population groups, any
#'   variable that can be used to describe one's socioeconomic status, such as
#'   education level, can be passed to this argument, as long as it can be
#'   numerically ordered (in which higher values denote higher socioeconomic
#'   status).
#' @param type A string. Which type of Concentration Index to calculate. Current
#'   available options are `"standard"` and `"corrected"`.
#' @template group_by_access
#'
#' @template return_inequality
#'
#' @family inequality
#'
#' @references
#' \insertAllCited{}
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' data_dir <- system.file("extdata", package = "accessibility")
#' travel_matrix <- readRDS(file.path(data_dir, "travel_matrix.rds"))
#' land_use_data <- readRDS(file.path(data_dir, "land_use_data.rds"))
#'
#' access <- cumulative_cutoff(
#'   travel_matrix,
#'   land_use_data,
#'   cutoff = 30,
#'   opportunity = "jobs",
#'   travel_cost = "travel_time"
#' )
#'
#' ci <- concentration_index(
#'   access,
#'   sociodemographic_data = land_use_data,
#'   opportunity = "jobs",
#'   population = "population",
#'   income = "income_per_capita",
#'   type = "corrected"
#' )
#' ci
#'
#' @export
concentration_index <- function(accessibility_data,
                                sociodemographic_data,
                                opportunity,
                                population,
                                income,
                                type,
                                group_by = character(0)) {
  checkmate::assert_string(opportunity)
  checkmate::assert_string(population)
  checkmate::assert_string(income)
  checkmate::assert(
    checkmate::check_string(type),
    checkmate::check_names(type, subset.of = c("standard", "corrected")),
    combine = "and"
  )
  assert_access_group_by(group_by)
  assert_accessibility_data(accessibility_data, opportunity, group_by)
  assert_sociodemographic_data(
    sociodemographic_data,
    accessibility_data,
    population = population,
    income = income
  )

  if (!inherits(accessibility_data, "data.table")) {
    original_class <- class(accessibility_data)
    data <- data.table::as.data.table(accessibility_data)
  } else {
    data <- data.table::copy(accessibility_data)
  }

  if (!inherits(sociodemographic_data, "data.table")) {
    sociodemographic_data <- data.table::as.data.table(sociodemographic_data)
  }

  merge_by_reference(
    data,
    sociodemographic_data,
    population,
    left_df_idcol = "id"
  )
  merge_by_reference(data, sociodemographic_data, income, left_df_idcol = "id")

  warn_extra_access_cols(accessibility_data, opportunity, group_by)

  .opp_colname <- opportunity
  .pop_colname <- population
  .groups <- group_by

  data.table::setorderv(data, cols = c(income, group_by))

  # fractional rank calculation based on equation 4 of this paper:
  # Eddy van Doorslaer and Xander Koolman, https://doi.org/10.1002/hec.918

  data[, pop_share := get(.pop_colname) / sum(get(.pop_colname)), by = .groups]
  data[
    ,
    fractional_rank := c(0, cumsum(pop_share)[-.N]) + pop_share / 2,
    by = .groups
  ]

  data[
    ,
    avg_access := stats::weighted.mean(
      get(.opp_colname),
      w = get(.pop_colname)
    ),
    by = .groups
  ]
  data[, diff_from_avg := get(.opp_colname) - avg_access]

  data[
    ,
    cont_to_total := (fractional_rank - 0.5) * diff_from_avg * pop_share /
      avg_access
  ]

  if (nrow(data) == 0 && identical(group_by, character(0))) {
    ci <- data.table::data.table(concentration_index = numeric(0))
  } else if (type == "standard") {
    ci <- data[
      ,
      .(concentration_index = 2 * sum(cont_to_total)),
      by = .groups
    ]
  } else if (type == "corrected") {
    # max()/min() may result in warnings if nrow(data) == 0
    data[, upper := suppressWarnings(max(get(.opp_colname))), by = .groups]
    data[, lower := suppressWarnings(min(get(.opp_colname))), by = .groups]

    # from Erreygers (2009), comparing equations 27 and 24, we have the
    # following correction factor
    # https://doi.org/10.1016/j.jhealeco.2008.02.003
    data[, correction_factor := 4 * avg_access / (upper - lower)]

    ci <- data[
      ,
      .(concentration_index = 2 * sum(cont_to_total * correction_factor)),
      by = .groups
    ]
  }

  if (exists("original_class")) class(ci) <- original_class

  return(ci)
}

#' @param sociodemographic_data A data frame. The distribution of
#'   sociodemographic characteristics of the population in the study area cells.
#'   Must contain the columns `id` and any others specified in `population` and
#'   `income`.

tester <- function(decay_value = 0.5) decay_power(decay_value)

test_that("adequately raises errors", {
  expect_error(tester("banana"))
  expect_error(tester(-0.1))
  expect_error(tester(Inf))
  expect_error(tester(NA))
  expect_error(tester(numeric(0)))
  expect_error(tester(c(0.1, 0.1)))
})

test_that("output is a decay function that returns a numeric", {
  expect_is(tester(), "function")

  output_fn <- tester()
  output_list <- output_fn(0)
  expect_is(output_list, "list")
  expect_length(output_list, 1)
  expect_named(output_list, "0.5")

  expect_equal(
    output_fn(c(0, 0.5, 1, 16, 64))[["0.5"]],
    c(1, 1, 1, 0.25, 0.125)
  )

  output_fn <- tester(0.25)
  output_list <- output_fn(0)
  expect_is(output_list, "list")
  expect_length(output_list, 1)
  expect_named(output_list, "0.25")

  expect_equal(
    output_fn(c(0, 0.5, 1, 16, 256))[["0.25"]],
    c(1, 1, 1, 0.5, 0.25)
  )

  output_fn <- tester(c(0.5, 0.25))
  output_list <- output_fn(0)
  expect_is(output_list, "list")
  expect_length(output_list, 2)
  expect_named(output_list, c("0.5", "0.25"))

  expect_equal(
    output_fn(c(0, 0.5, 1, 16, 64))[["0.5"]],
    c(1, 1, 1, 0.25, 0.125)
  )
  expect_equal(
    output_fn(c(0, 0.5, 1, 16, 256))[["0.25"]],
    c(1, 1, 1, 0.5, 0.25)
  )
})

test_that("output fn returns empty numeric if receives empty numeric/integer", {
  output_fn <- tester(c(0.5, 0.25))

  expect_identical(output_fn(integer())[["0.5"]], numeric())
  expect_identical(output_fn(integer())[["0.25"]], numeric())

  expect_identical(output_fn(numeric())[["0.5"]], numeric())
  expect_identical(output_fn(numeric())[["0.25"]], numeric())
})

---
title: "Introduction to accessibility: calculating accessibility measures"
output: rmarkdown::html_vignette
bibliography: '`r system.file("REFERENCES.bib", package = "accessibility")`'
vignette: >
  %\VignetteIndexEntry{Introduction to accessibility: calculating accessibility measures} 
  %\VignetteEngine{knitr::rmarkdown} 
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
Sys.setenv(OMP_THREAD_LIMIT = 2)

knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

**accessibility** is an R package that offers a set of fast and convenient
functions to calculate multiple transport accessibility measures. Given a
pre-computed travel cost matrix and a land use dataset, the package allows one
to calculate active and passive accessibility levels using multiple
accessibility metrics, such as cumulative opportunities (using either a travel
cost cutoff or a travel cost interval), minimum travel cost to closest N
number of activities, gravitational measures and different floating catchment
area methods. This vignette briefly overviews the package with a few
reproducible examples.

# Installation

Before using **accessibility** please make sure that you have it installed in
your computer. You can download either the most stable version from CRAN...

```{r, eval = FALSE}
install.packages("accessibility")
```

...or the development version from GitHub.

```{r, eval = FALSE}
# install.packages("remotes")
remotes::install_github("ipeaGIT/accessibility")
```

# Overview of the package

As of the time of writing this vignette, the **accessibility** package includes
five different functions to calculate accessibility:

1. `cost_to_closest()` - calculates the minimum travel cost to the closest
   *n* number of opportunities.
2. `cumulative_cutoff()` - calculates the frequently used threshold-based
   cumulative opportunities measure.
3. `cumulative_interval()` - calculates accessibility as the median/mean (or
   any other summary measure, really) number of opportunities that can be
   reached within a cost interval. 
4. `gravity()` - calculates gravity-based accessibility, taking a decay
   function specified by the user (more on that in a few paragraphs).
5. `floating_catchment_area()` - calculates accessibility taking into account
   the effects of competition for opportunities with different floating
   catchment area measures.

You may have noticed that we've mentioned a few times that the functions
calculate accessibility using travel *cost*, and not travel *time*. That's
because we're treating *costs* here in its generic sense: anything that
increases the impedance from an origin to a destination, such as travel time,
monetary costs, distances, risk perception, etc., can be considered a generic
*cost*.

The `gravity()` and `floating_catchment_area()` functions can use
different decay functions when estimating accessibility levels. These decay
functions effectively weigh the number of opportunities in a destination by a
factor that depends on the travel cost between the origin and the destination.
For convenience, the package currently includes the following decay functions:

1. `decay_binary()` - binary decay function (the one used in cumulative
   opportunities measures).
2. `decay_exponential()` - negative exponential decay function.
3. `decay_linear()` - linear decay function (weights decay linearly from 1 to 0
   until a specific travel cost cutoff is reached).
4. `decay_power()` - inverse power decay function.
5. `decay_stepped()` - stepped decay function (similar to `decay_binary()`, but
   can take an arbitrary number of steps, instead of a single one).

The users can also specify their own custom decay functions, if they need to
use functions currently not included in the package. For more details on this,
please read the [decay functions vignette](decay_functions.html).

# Demonstration on sample data

Enough talking. Let's demonstrate some of the key features of the package.
First we'll need to load the libraries we'll be using:

```{r, message = FALSE, warning = FALSE}
library(accessibility)
library(data.table)
library(ggplot2)
library(sf)
```

## Data requirements

To use **accessibility**, you will need a pre-computed travel cost matrix and
some land use data (e.g. location of jobs, healthcare, population, etc.). As
mentioned before, travel costs can be presented in terms of travel times,
distances or monetary costs, for example. This dataset must be structured in a
`data.frame` containing, at least, the columns `from_id`, `to_id` and the
travel cost between each origin-destination pair.

Your data should look similar to this sample dataset with public transport
travel times for the city of Belo Horizonte, Brazil, included in the package
for demonstration purposes[^1].

[^1]: If you would like to calculate such travel cost matrices yourself, there
are several computational packages to do that in R, such as
[r5r](https://github.com/ipeaGIT/r5r),
[dodgr](https://github.com/UrbanAnalyst/dodgr),
[gtfsrouter](https://github.com/UrbanAnalyst/gtfsrouter),
[hereR](https://github.com/munterfi/hereR) and
[opentripplanner](https://github.com/ropensci/opentripplanner).

```{r}
data_dir <- system.file("extdata", package = "accessibility")

travel_matrix <- readRDS(file.path(data_dir, "travel_matrix.rds"))
head(travel_matrix)
```

The land use data must also be structured in a `data.frame` and must contain an
`id` column, referring to the ids listed in the travel matrix, and the number
of opportunities/facilities/services in each spatial unit. The sample dataset
we'll be using looks like this:

```{r}
land_use_data <- readRDS(file.path(data_dir, "land_use_data.rds"))
head(land_use_data)
```

## Minimum travel cost

`cost_to_closest()` calculates the minimum travel cost to a given number of
opportunities. Much like the other functions we'll be demonstrating in this
section, it takes as inputs the travel matrix and land use datasets, the name
of the column in the latter with the opportunities to be considered and the
name of the column in the former with the travel cost to be considered.
Additionally, it takes the minimum number of opportunities to be considered.
Here's how calculating the time from each origin in Belo Horizonte to the
closest school looks like:

```{r, message = FALSE}
mtc <- cost_to_closest(
  travel_matrix,
  land_use_data,
  opportunity = "schools",
  travel_cost = "travel_time",
  n = 1
)
head(mtc)
```

## Cutoff-based cumulative opportunities

`cumulative_cutoff()` calculates the traditional cumulative opportunities
measure, indicating the number of opportunities that are accessible within a
given travel cost threshold. In this example, we estimate how many jobs can be
reached from each origin with trips taking up to 30 minutes of travel time.

```{r, message = FALSE}
cum_cutoff <- cumulative_cutoff(
  travel_matrix,
  land_use_data,
  opportunity = "jobs",
  travel_cost = "travel_time",
  cutoff = 30
)
head(cum_cutoff)
```

Let's say that we wanted to, instead, calculate *passive* accessibility - i.e.
by how many people each destination can be reached within a given travel cost.
Doing so requires very few changes to the call: just change the "opportunity"
column to `"population"` and set `active` (`TRUE` by default) to `FALSE`.

```{r, message = FALSE}
passive_cum_cutoff <- cumulative_cutoff(
  travel_matrix,
  land_use_data,
  opportunity = "population",
  travel_cost = "travel_time",
  cutoff = 30,
  active = FALSE
)
head(passive_cum_cutoff)
```

The `active` parameter is available in most other accessibility functions as
well (with the exception of `floating_catchment_area()`), making it very easy
to calculate both active and passive accessibility.

## Interval-based cumulative opportunities

`cumulative_time_interval()` calculates the interval-based cumulative
opportunities measure. This measure, developed by @tomasiello2023time,
mitigates the impacts of arbitrary choices of cost cutoffs, one of the main
disadvantages of the traditional threshold-based cumulative opportunities
measure. Given a cost interval, it calculates several accessibility estimates
within the interval and summarizes it using a user-specified function. In the
example below, we calculate the average number of accessible jobs considering
multiple minute-by-minute time thresholds between 40 and 60 minutes.

```{r, message = FALSE}
cum_interval <- cumulative_interval(
  travel_matrix = travel_matrix,
  land_use_data = land_use_data,
  opportunity = "jobs",
  travel_cost = "travel_time",
  interval = c(40, 60),
  summary_function = base::mean
)
head(cum_interval)
```

## Gravity measures

`gravity()` calculates gravity-based measures - i.e. measures in which the
weight of opportunities is gradually discounted as the travel cost increases.
Of course, several different decay functions can be used to so, each one of
them with a range of possible different parameters. In order to accommodate such
generalization, the function takes the decay function to be used as a
parameter.

In the example below, we calculate accessibility using a negative exponential
function with a `decay_value` (usually referred as the \eqn{\beta} in its
formulation) of 0.2. Please see [the vignette on decay
functions](decay_functions.html) for more information on the decay functions
shipped with the package and how to use custom functions.

```{r, message = FALSE}
negative_exp <- gravity(
  travel_matrix,
  land_use_data,
  opportunity = "schools",
  travel_cost = "travel_time",
  decay_function = decay_exponential(decay_value = 0.2)
)
head(negative_exp)
```

## Floating catchment area

`floating_catchment_area()` calculates accessibility accounting for competition
of resources using different floating catchment area (FCA) methods. The FCA
family includes several different methods, which can be specified using the
`method` parameter. As of the time of writing this vignette, the package
supports two different methods:

- 2-Step Floating Catchment Area (`"2sfca"`) - the first metric in the FCA
family, originally proposed by @luo2003measures.
- Balanced Floating Catchment Area (`"bfca"`) - takes competition affects into
account while correcting for issues of inflation of demand and service levels.
Originally proposed by @paez2019demand and named in @pereira2021geographic.

Please note that, since FCA measures consider competition effects, we have to
specify which column in the land use dataset represents the population
competing for opportunities with the `demand` parameter. The function
also supports different decay functions. In the example below, we calculate
accessibility to jobs using the BFCA method, considering that the entire
population of the city compete for these jobs and using a negative exponential
decay function.

```{r, message = FALSE}
bfca <- floating_catchment_area(
  travel_matrix,
  land_use_data,
  opportunity = "jobs",
  travel_cost = "travel_time",
  demand = "population",
  method = "bfca",
  decay_function = decay_exponential(decay_value = 0.5)
)
head(bfca)
```

## Spatial availability

`spatial_availability()` also calculates accessibility considering competition
effects, though using the spatial availability measure proposed by
@soukhov2023introducing. The results from this metric are proportional both to
the demand in each origin and the travel cost it takes to reach the
destinations. As with the FCA function, we have to specify the column in the
land use dataset that contains the population competing for opportunities and we
can use different decay functions to calculate the impedance between
origin-destination pairs.

```{r}
spatial_avlblt <- spatial_availability(
  travel_matrix,
  land_use_data,
  opportunity = "jobs",
  travel_cost = "travel_time",
  demand = "population",
  decay_function = decay_exponential(decay_value = 0.1)
)
head(spatial_avlblt)
```

## Balancing cost

`balancing_cost()` calculates the balancing cost accessibility measure.
Originally proposed by @barboza2021balancing under the name "balancing time",
this metric is defined as the travel cost required to reach as many
opportunities as the number of people in a given origin. Just like the
previous two functions, `balancing_cost()` also includes a
parameter to specify the population competing for opportunities.

The function also includes a `cost_increment` parameter, that should be used to
specify the increment that defines the travel cost distribution from which the
potential balancing costs will be picked. For example, an increment of 1 (the
default) tends to suit travel time distributions, meaning that the function will
first check if any origins reach their balancing cost with a travel time of 0
minutes, then 1 minute, 2 minutes, 3, 4, ..., etc. On the other hand, an
increment of 1 might be too big for a distribution of monetary costs, which
could possibly benefit from a smaller increment of 0.05 (5 cents), for example.
Such increment results in the function looking for balancing costs first at a
monetary cost of 0, then 0.05, 0.10, ..., etc. In the example below, we use the
default cost increment of 1.

```{r}
bal_cost <- balancing_cost(
  travel_matrix,
  land_use_data,
  opportunity = "jobs",
  travel_cost = "travel_time",
  demand = "population"
)
head(bal_cost)
```

## Visualize results

If you have the spatial data of your origins/destinations, you can easily merge
it with the accessibility to create spatial visualizations of the results. The
example below quickly shows how to create a simple map using `{ggplot2}`.

```{r, eval = requireNamespace(c("sf", "ggplot2"), quietly = TRUE), out.width = "80%", fig.width = 6, fig.height = 6}
grid <- system.file("extdata/grid_bho.rds", package = "accessibility")
grid <- readRDS(grid)

spatial_data <- merge(grid, cum_cutoff, by = "id")

ggplot() +
  geom_sf(data = spatial_data, aes(fill = jobs), color = NA) +
  labs(
    title = "Job accessibility by transit in under 30 min.",
    fill = "Accessible jobs"
  ) +
  scale_fill_viridis_c() +
  theme_void()
```

# References

---
title: "Decay functions"
output: rmarkdown::html_vignette
bibliography: '`r system.file("REFERENCES.bib", package = "accessibility")`'
vignette: >
  %\VignetteIndexEntry{Decay functions}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
Sys.setenv(OMP_THREAD_LIMIT = 2)

knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

Some functions of the **accessibility** package, such as
`floating_catchment_area()` and `gravity()`, use decay functions to
continuously discount the weight of opportunities as travel costs become
larger. For convenience, the package ships with some of the most frequently
used functions in the accessibility literature, which are discussed below.
Additionally, users can pass custom functions to convert travel costs into
weights to be applied to the opportunities, which is discussed further down in
this the vignette.

# Currently available decay functions

## Binary

Also known as the step decay function, it's most commonly used in cumulative
opportunities measures.

$$
\begin{aligned}
  f(t_{ij})= \left\{
      \begin{array}{ll}
          1 & \quad \text{for }t_{ij} \leq T \\
          0 & \quad \text{for }t_{ij} > T
      \end{array}
    \right.\\
\end{aligned}
$$

Where $t_{ij}$ is the travel cost between origin *i* and destination *j*, and
$T$ is the travel cost cutoff.

## Linear

Weights decay linearly until the travel cost cutoff is reached. From this point
onward weights assume the value of 0.

$$
\begin{aligned}
  f(t_{ij})= \left\{
      \begin{array}{ll}
          (1 - t_{ij}/ T) & \quad \text{for }t_{ij} \leq T \\
          0 & \quad \text{for }t_{ij} > T
      \end{array}
    \right.\\
\end{aligned}
$$

Where $t_{ij}$ is the travel cost between origin *i* and destination *j*, and
$T$ is the travel cost cutoff.

## Negative exponential

$$
\begin{aligned}
  f(t_{ij})= e^{(-\beta t_{ij})}
\end{aligned}
$$

Where $t_{ij}$ is the travel cost between origin *i* and destination *j*, and
$\beta$ is the parameter that tells the speed of decay.

## Inverse power

$$
\begin{aligned}
  f(t_{ij})= \left\{
      \begin{array}{ll}
          1 & \quad \text{for } t_{ij}\leq 1 \\
          t_{ij}^{-\beta} & \quad \text{for }t_{ij} > 1
      \end{array}
    \right.\\
\end{aligned}
$$

Where $t_{ij}$ is the travel cost between origin *i* and destination *j*, and
$\beta$ is the parameter that tells the speed of decay.

## Stepped

Similar to the binary function, but can take an arbitrary number of steps. The
current implementation assumes that values changes at each step, instead of
right after it.

$$
\begin{aligned}
  f(t_{ij})= \left\{
      \begin{array}{ll}
          1 & \quad \text{for } t_{ij} \lt S_{1} \\
          v_{1} & \quad \text{for } t_{ij} \lt S_{2} \\
          v_{2} & \quad \text{for } t_{ij} \lt S_{3} \\
          ... \\
          v_{n-1} & \quad \text{for } t_{ij} \lt S_{n} \\
          v_{n} & \quad \text{otherwise} \\
      \end{array}
    \right.\\
\end{aligned}
$$

Where $t_{ij}$ is the travel cost between origin *i* and destination *j*, $n$
is the total number of steps, $S_{k}$ is the travel cost cutoff that delimits
the $k^{th}$ step, and $v_{k}$ is the value that the decay function assumes at
the $k^{th}$ step.

## Logistic

Weights decay sigmoidally, according to a reversed cumulative logistic curve.
Currently, the function implements a logistic distribution parameterized with
the cutoff that sets its inflection point and the standard deviation that sets
its steepness, according to the logistic decay curve proposed by
@bauer2016measuring. Standard deviations values near 0 result in weighting
curves that approximate binary decay, while higher values tend to linearize the
curve.

$$
\begin{aligned}
  f(t_{ij}) = \frac{
    1 + e^\frac{-IP \times \pi}{SD \times \sqrt{3}}}{
    1 + e^\frac{(t_{ij} - IP) \times \pi}{SD \times \sqrt{3}}}
\end{aligned}
$$

Where $t_{ij}$ is the travel cost between origin *i* and destination *j*, $IP$
is the distribution inflection point, and $SD$ is the distribution standard
deviation.

### Quick demonstration

All decay functions (`decay_*()`) take decay parameters as input and return a
function as output. This output function, in turn, takes a vector of numeric
values as input, the travel costs, and returns a list of numeric vectors as
output. Since the decay function can take multiple decay parameters as input,
each element of the output list refers to the opportunities weights calculated
with one of these decay parameters.

Let's check this behavior with an example. With the code snippet below, we
calculate the opportunities weights for the same travel costs, but using
different negative exponential decay values (0.2 and 0.3):

```{r}
library(accessibility)

output_fn <- decay_exponential(c(0.2, 0.3))

output_fn(c(10, 15, 20))
```

`decay_stepped()`, as the only decay function that takes more than one argument
(both `steps` and `weights`) names the output elements after the combination of
steps and weights:

```{r}
stepped_output <- decay_stepped(
  steps = list(c(10, 20, 30), c(10, 20, 30, 40)),
  weights = list(c(0.67, 0.33, 0), c(0.75, 0.5, 0.25, 0))
)

stepped_output(c(15, 25, 35, 45))
```

With the code below, we demonstrate each decay function with travel costs
ranging from 1 to 100:

```{r, eval = requireNamespace("ggplot2", quietly = TRUE), out.width = "80%", fig.width = 6, fig.height = 6}
library(data.table)
library(ggplot2)

binary <- decay_binary(cutoff = 50)
linear <- decay_linear(cutoff = 50)
negative_exp <- decay_exponential(decay_value = 0.2)
inverse_power <- decay_power(decay_value = 0.2)
stepped <- decay_stepped(steps = c(30, 60, 90), weights = c(0.67, 0.33, 0))
logistic <- decay_logistic(cutoff = 50, sd = 10)

travel_costs <- seq(1, 100, 0.1)

weights <- data.table(
  minutes = travel_costs,
  binary = as.numeric(binary(travel_costs)[["50"]]),
  linear = linear(travel_costs)[["50"]],
  negative_exp = negative_exp(travel_costs)[["0.2"]],
  inverse_power = inverse_power(travel_costs)[["0.2"]],
  stepped = stepped(travel_costs)[["s(30,60,90);w(0.67,0.33,0)"]],
  logistic = logistic(travel_costs)[["c50;sd10"]]
)

# reshape data to long format
weights <- melt(
  weights,
  id.vars = "minutes",
  variable.name = "decay_function",
  value.name = "weights"
)

ggplot(weights) +
  geom_line(
    aes(minutes, weights, color = decay_function),
    show.legend = FALSE
  ) +
  facet_wrap(. ~ decay_function, ncol = 2) +
  theme_minimal()
```

# Using a custom decay function

**accessibility** also allows you to use a custom decay function, instead of one
of the functions shipped with the package. A valid decay function is one that
takes a `numeric` vector of travel costs as input and returns either:

- A `numeric` vector of weights, with the same length of input, or;
- A named `list` of `numeric` vectors to be used as weights, each one with the
  same length of input.

Let's check the difference between each case with an example. Suppose we want to
use a very simple decay function that defines the weights as the multiplicative
inverse of travel cost - i.e. $travel\_cost^{-1}$. We just have to take care of
the case when travel cost is less than 1, in which case the function would
return values greater than 1 and which we will replace with 1, but otherwise the
implementation is pretty simple:

```{r}
my_decay <- function(travel_cost) {
  weights <- 1 / travel_cost
  weights[weights > 1] <- 1
  return(weights)
}
```

Given a `numeric` vector of travel costs, the function returns a `numeric`
vector of weights:

```{r}
my_decay(c(0, 0.5, 1, 2, 5, 10))
```

Using this function to calculate accessibility is as easy as any of the built-in
decay functions:

```{r}
data_dir <- system.file("extdata", package = "accessibility")

travel_matrix <- readRDS(file.path(data_dir, "travel_matrix.rds"))
land_use_data <- readRDS(file.path(data_dir, "land_use_data.rds"))

custom_gravity <- gravity(
  travel_matrix,
  land_use_data,
  opportunity = "jobs",
  travel_cost = "travel_time",
  decay_function = my_decay
)
head(custom_gravity)
```

Great! But now suppose we want to change this function a bit. Instead of the
simple multiplicative inverse of the travel cost, we want to multiply this
inverse by a given decay parameter. Now we face a "problem": our function would
need to take two inputs (the travel cost, as above, and the decay parameter to
multiply the travel cost), but `gravity()` can only take functions that receive
a single input (the travel cost). In this case, we resort to the same strategy
used in the decay functions shipped with the package, we create a function that
takes a decay parameter as input and returns a function that takes travel cost
input as output:

```{r}
my_second_decay <- function(decay_parameter) {
  function(travel_cost) {
    weights <- 1 / (decay_parameter * travel_cost)
    weights[weights > 1] <- 1
    return(weights)
  }
}

output_fn <- my_second_decay(2)
output_fn(c(0, 0.5, 1, 2, 5, 10))

# compare to the first custom decay function
my_decay(c(0, 0.5, 1, 2, 5, 10))
```

Great, it works! In fact, we can achieve the exact same `gravity()` result shown
above if we use `my_second_decay(1)`, instead of `my_decay`:

```{r}
second_custom_gravity <- gravity(
  travel_matrix,
  land_use_data,
  opportunity = "jobs",
  travel_cost = "travel_time",
  decay_function = my_second_decay(1)
)
head(second_custom_gravity)
```

A small difference is that in the first example we passed `my_decay` (the
function *object*) to the `decay_function` parameter, whereas in the second we
passed the function *call* `my_second_decay(1)`. That's because
`my_second_decay()` (and the built-in decay functions as well) is actually a
function "factory": it's a function that returns a function.

```{r}
decay_power(1)
```

What if we want our custom function to take many decay parameters as inputs,
just like the built-in decay functions? In this case, our output function has to
return a `list` of `numeric` weights named after the decay parameters we have
set:

```{r}
my_third_decay <- function(decay_parameter) {
  function(travel_cost) {
    weighting_list <- lapply(
      decay_parameter,
      function(x) {
        weights <- 1 / (x * travel_cost)
        weights[weights > 1] <- 1
        return(weights)
      }
    )

    names(weighting_list) <- decay_parameter
    weighting_list
  }
}

output_fn <- my_third_decay(c(1, 2))
output_fn(c(0, 0.5, 1, 2, 5, 10))

# compare to the first and second custom decay functions

my_decay(c(0, 0.5, 1, 2, 5, 10))

my_second_decay(2)(c(0, 0.5, 1, 2, 5, 10))
```

This new custom decay function is a bit more complex than the previous two, yet
much more powerful. We can use it in conjunction with `gravity()` (and the other
accessibility functions that take decay functions as input as well) to calculate
the accessibility with multiple decay parameters in a single call:

```{r}
third_custom_gravity <- gravity(
  travel_matrix,
  land_use_data,
  opportunity = "jobs",
  travel_cost = "travel_time",
  decay_function = my_third_decay(c(1, 2))
)
third_custom_gravity
```

When we create our own custom decay function, we don't have to think about
general cases, just the cases that matter for us. Creating a simple function
that converts one vector into another, therefore, might be enough, given our
needs. If, however, you're testing a new decay function and you want to check
its sensitivity to different decay parameters, for example, you might be better
off developing a decay function that can take many decay values as inputs and
returns a list of weight vectors as output.

---
title: "Calculating accessibility inequality and poverty"
output: rmarkdown::html_vignette
bibliography: '`r system.file("REFERENCES.bib", package = "accessibility")`'
vignette: >
  %\VignetteIndexEntry{Calculating accessibility inequality and poverty} 
  %\VignetteEngine{knitr::rmarkdown} 
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
Sys.setenv(OMP_THREAD_LIMIT = 2)

knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

**accessibility** not only includes functions to calculate accessibility levels
from travel time matrices and land use datasets, but also to calculate
accessibility inequality and poverty from an accessibility distribution. This
vignette briefly presents these functions.

# Demonstration on sample data

We first need to calculate the accessibility levels that we are going to use in
this demonstration. To do so, we use the `cumulative_cutoff()` function and the
sample data included in the package.

```{r}
library(accessibility)

data_dir <- system.file("extdata", package = "accessibility")
travel_matrix <- readRDS(file.path(data_dir, "travel_matrix.rds"))
land_use_data <- readRDS(file.path(data_dir, "land_use_data.rds"))

access <- cumulative_cutoff(
  travel_matrix,
  land_use_data,
  opportunity = "jobs",
  travel_cost = "travel_time",
  cutoff = 30
)
head(access)
```

The functions we’ll be demonstrating in this section take much of the same
information as input, including:

- the accessibility distribution;
- a sociodemographic dataset containing information such as the distribution of
  people in the study area and their income;
- the name of the column in the accessibility dataset with the opportunities
  whose accessibility should be considered in the calculation; and
- the name of the column in the sociodemographic dataset with the number of
  people in each cell, used to weigh accessibility levels when calculating
  inequality/poverty.

## Palma Ratio

`palma_ratio()` calculates the Palma Ratio of a given accessibility
distribution. Originally defined as the income share of the wealthiest 10% of a
population divided by the income share of the poorest 40%, this measure has been
adapted in transport planning as the average accessibility of the wealthiest 10%
divided by the average accessibility of the poorest 40%. Palma Ratio values
higher than 1 indicate a scenario in which the wealthiest population has higher
accessibility levels than the poorest, whereas values lower than 1 indicate the
opposite situation.

This function includes an additional `income` parameter, used to list the column
in the sociodemographic dataset with the income variable that should be used to
classify the population in socioeconomic groups. Please note that this variable
should describe income per capita (e.g. mean income per capita, household income
per capita, etc), instead of the total amount of income in each cell.

```{r}
palma <- palma_ratio(
  access,
  sociodemographic_data = land_use_data,
  opportunity = "jobs",
  population = "population",
  income = "income_per_capita"
)
palma
```

## Gini Index

`gini_index()` calculates the Gini Index of a given accessibility distribution.
Probably the most frequently used inequality measure in transport planning, this
index estimates how much a distribution deviates from a hypothetical situation
in which everyone has the exact same accessibility conditions. Gini Index values
range from 0 to 1. A value of 0 indicates a scenario of perfect
equality, in which everyone has the same accessibility levels, whereas a value
of 1 indicates a scenario of perfect inequality, in which the accessibility
levels in a study area are all concentrated into a single cell.

```{r}
gini <- gini_index(
  access,
  sociodemographic_data = land_use_data,
  opportunity = "jobs",
  population = "population"
)
gini
```

## Concentration Index

`concentration_index()` calculates the Concentration Index (CI) of a given
accessibility distribution. This indicator estimates the extent to which
accessibility inequalities are systematically associated with individuals'
socioeconomic levels. CI values can theoretically vary between -1 and 1 (when
all accessibility is concentrated in the most or in the least disadvantaged
cell, respectively). Negative values indicate that inequalities favor the poor,
while positive values indicate a pro-rich bias.

Just like `palma_ratio()`, this function includes an `income` parameter to
indicate which variable from the sociodemographic dataset should be used to rank
the population from the least to the most privileged groups. Unlike the Palma
Ratio function, however, any variable that can be used to describe one's
socioeconomic status, such as education level, for example, can be passed to
this argument, as long as it can be numerically ordered (in which higher values
denote higher socioeconomic status).

`concentration_index()` also includes a `type` parameter, used to indicate which
Concentration Index to calculate. This parameter currently supports two values,
`"standard"` and `"corrected"`, which respectively identify the standard
relative CI and the corrected CI, proposed by @erreygers2009correcting.

```{r}
ci <- concentration_index(
  access,
  sociodemographic_data = land_use_data,
  opportunity = "jobs",
  population = "population",
  income = "income_per_capita",
  type = "corrected"
)
ci
```

## Theil T

`theil_t()` calculates the Theil T Index of a given accessibility distribution.
Values range from 0 (when all individuals have exactly the same accessibility
levels) to the natural log of *n*, in which *n* is the number of individuals in
the accessibility dataset.

If the individuals can be classified into mutually exclusive and completely
exhaustive groups (i.e. into groups that do not overlap and cover the entire
population), the index can be decomposed into a between- and a within-groups
inequality component. The function includes a `socioeconomic_groups` parameter
to indicate which variable from the sociodemographic dataset should be used
identify the socioeconomic groups used to calculate these components.

Please note that the output `theil_t()` varies based on the value of
`socioeconomic_groups`. If `NULL` (the default), the between- and within-groups
components are not calculated, and the function returns a data frame containing
only the total aggregate inequality for the returned area. If
`socioeconomic_groups` is not `NULL`, however, the function returns a list
containing three dataframes: one summarizing the total inequality and the
between- and within-groups components, one listing the contribution of each
group to the between-groups component and another listing the contribution of
each group to the within-groups component. Both behaviors are shown below.

```{r}
theil_without_groups <- theil_t(
  access,
  sociodemographic_data = land_use_data,
  opportunity = "jobs",
  population = "population"
)
theil_without_groups

# some cells are classified as in the decile NA because their income per capita
# is NaN, as they don't have any population. we filter these cells from our
# accessibility data, otherwise the output would include NA values (note that
# subsetting the data like this doesn't affect the assumption that groups are
# completely exhaustive, because cells with NA income decile don't have any
# population)

na_decile_ids <- land_use_data[is.na(land_use_data$income_decile), ]$id
no_na_access <- access[! access$id %in% na_decile_ids, ]
sociodem_data <- land_use_data[! land_use_data$id %in% na_decile_ids, ]

theil_with_groups <- theil_t(
  no_na_access,
  sociodemographic_data = sociodem_data,
  opportunity = "jobs",
  population = "population",
  socioeconomic_groups = "income_decile"
)
theil_with_groups
```


## Foster-Greer-Thorbecke (FGT) poverty measures

`fgt_poverty()` calculates the FGT metrics, a family of poverty measures
originally proposed by @foster1984class that capture the extent and severity of
poverty within an accessibility distribution. The FGT family is composed of
three measures that differ based on the $\alpha$ parameter used to calculate
them (either 0, 1 or 2) and which also changes their interpretation:

- with $\alpha = 0$ (FGT0) the measure captures the extent of poverty as a
  simple headcount - i.e. the proportion of people below the poverty line;
- with $\alpha = 1$ (FGT1) the measure, also know as the "poverty gap index",
  captures the severity of poverty as the average percentage distance between
  the poverty line and the accessibility of individuals below the poverty line;
- with $\alpha = 2$ (FGT2) the measure simultaneously captures the extent and
  the severity of poverty by calculating the number of people below the poverty
  line weighted by the size of the accessibility shortfall relative to the
  poverty line.

This function includes an additional `poverty_line` parameter, used to define
the poverty line below which individuals are considered to be in accessibility
poverty.

```{r}
poverty <- fgt_poverty(
  access,
  sociodemographic_data = land_use_data,
  opportunity = "jobs",
  population = "population",
  poverty_line = 50000
)

poverty
```

# References

---
title: "testing_constrained"
output: html_document
---

```{r}
library(data.table)
library(dplyr)
devtools::load_all(".")
```

# Test 1. Toy data from the spaital availability (2023) paper.
## TOT CONST
```{r}
# Toy travel matrix and land use data
travel_matrix <- expand.grid(from_id = c("A", "B", "C"),
                             to_id = c("1", "2", "3")) |>
  dplyr::mutate(travel_time = c(15, 30, 100, 30, 15, 100, 100, 100, 15)) |>
  data.table::data.table()

land_use_data <- data.table::data.table(
  id = c("A", "B", "C","1", "2", "3"),
  population = c(50000, 150000, 10000),   # supply
  jobs = c(100000, 100000, 10000)     # demand
)

# Decay function
decay <- decay_exponential(decay_value = 0.1)
```


```{r}
active_det_false <- constrained_accessibility(
  constraint = "total",
  travel_matrix, land_use_data,
  travel_cost = "travel_time",
  decay_function = decay,
  demand = NULL,
  supply = "jobs",
  detailed_results = FALSE,
  active = TRUE)

active_det_false
```


```{r}
passive_det_false <- constrained_accessibility(
  constraint = "total", 
  travel_matrix = travel_matrix, 
  land_use_data = land_use_data,
  travel_cost = "travel_time",
  decay_function = decay,
  demand = "population",
  supply = NULL,
  detailed_results = FALSE,
  active = FALSE
  )

passive_det_false
```


```{r}
active_det_true <- constrained_accessibility(
  constraint = "total", 
  travel_matrix, 
  land_use_data,
                                travel_cost = "travel_time",
                                decay_function = decay,
                                demand = NULL,
                                supply = "jobs",
                                detailed_results = TRUE,
                                active = TRUE
  )

active_det_true
```


```{r}
passive_det_true <- constrained_accessibility(constraint = "total", travel_matrix, land_use_data,
                               travel_cost = "travel_time",
                               decay_function = decay,
                               demand = "population",
                               supply = NULL,
                               detailed_results = TRUE,
                               active = FALSE)
passive_det_true
```

```{r}
passive_det_true$demand |> sum() #POPULATION 210,000
active_det_true$supply |> sum() #jobs 210,000
passive_det_false$demand |> sum() #POPULATION 210,000
active_det_false$supply |> sum() #jobs 210,000

active_det_true$kappa_total |> sum()
passive_det_true$hatkappa_total |> sum()
```

## SINGLE CONST
```{r}
active_det_false <- constrained_accessibility(constraint = "singly", travel_matrix, land_use_data,
                                travel_cost = "travel_time",
                                decay_function = decay,
                                demand = "population",
                                supply = "jobs",
                                detailed_results = FALSE,
                                active = TRUE)
active_det_false
```
Should be: 
A 66833.466			
B	133203.363			
C	9963.171	

```{r}
passive_det_false <- constrained_accessibility(constraint = "singly", travel_matrix, land_use_data,
                                travel_cost = "travel_time",
                                decay_function = decay,
                                demand = "population",
                                supply = "jobs",
                                detailed_results = FALSE,
                                active = FALSE)
passive_det_false
```
Should be:
1	68261.682			
2	131775.520			
3	9962.798	

```{r}
active_det_true <- constrained_accessibility(constraint = "singly", travel_matrix, land_use_data,
                                travel_cost = "travel_time",
                                decay_function = decay,
                                demand = "population",
                                supply = "jobs",
                                detailed_results = TRUE,
                                active = TRUE)
active_det_true
```
e.g., for summing As... 59900.642536+6922.691048+10.132187 = 66833.47

```{r}
passive_det_true <- constrained_accessibility(constraint = "singly", travel_matrix, land_use_data,
                               travel_cost = "travel_time",
                               decay_function = decay,
                               demand = "population",
                                supply = "jobs",
                               detailed_results = TRUE,
                               active = FALSE)
passive_det_true
```

Checks!
```{r}
passive_det_true$demand |> sum() #POPULATION 210,000
active_det_true$supply |> sum() #jobs 210,000
passive_det_false$demand |> sum() #POPULATION 210,000
active_det_false$supply |> sum() #jobs 210,000

active_det_true$kappa_singly |> sum()
passive_det_true$hatkappa_singly |> sum()
```


```{r}
active_det_true$kappa_singly |> sum() #should be the number of destinations
active_det_true[, sum(kappa_singly), by = to_id]  #should all be 1

passive_det_true$hatkappa_singly |> sum() #should be the number of origins
passive_det_true[, sum(hatkappa_singly), by = from_id]  #should all be 1
```

# Test 2. Testing results with total/singly constrained toy example in 2025 paper
Data:
```{r test-data}
travel_matrix <- expand.grid(from_id = c("1", "2", "3"),
                  to_id = c("1", "2", "3")) |>
  mutate(travel_time = c(10, 30, 15, 30, 10, 25, 15, 25, 10)) |> data.table()

land_use_data <- data.table(
  id = c("1", "2", "3"),
  population = c(4, 10, 6),
  jobs = c(160, 150, 180)
)
```

## Total constrained--ACTIVE
Checking for decay_power scenario 3:  
```{r}
active_det_false <- constrained_accessibility(constraint="total",
                          travel_matrix, land_use_data,
                                travel_cost = "travel_time",
                                decay_function = decay_power(decay_value = 3),
                                demand = NULL,
                                supply = "jobs",
                                detailed_results = FALSE,
                                active = TRUE)

active_det_true <- constrained_accessibility(constraint="total",
                          travel_matrix, land_use_data,
                                travel_cost = "travel_time",
                                decay_function = decay_power(decay_value = 3),
                                demand = NULL,
                                supply = "jobs",
                                detailed_results = TRUE,
                                active = TRUE)
active_det_false
active_det_true

active_det_false$supply |> sum()
active_det_true$supply |> sum()
active_det_true$kappa_total |> sum()
```
Should be:
1	172.0653			
2	131.6267			
3	186.3080

Checking for decay_power scenario 2:  
```{r}
active_det_false <- constrained_accessibility(constraint="total",
                          travel_matrix, land_use_data,
                                travel_cost = "travel_time",
                                decay_function = decay_power(decay_value = 2),
                                demand = NULL,
                                supply = "jobs",
                                detailed_results = FALSE,
                                active = TRUE)

active_det_true <- constrained_accessibility(constraint="total",
                          travel_matrix, land_use_data,
                                travel_cost = "travel_time",
                                decay_function = decay_power(decay_value = 2),
                                demand = NULL,
                                supply = "jobs",
                                detailed_results = TRUE,
                                active = TRUE)
active_det_false
active_det_true

active_det_false$supply |> sum()
active_det_true$supply |> sum()
active_det_true$kappa_total |> sum()
```
Should be:
1	172.6721			
2	132.2474			
3	185.0805	

Checking for decay_power scenario 0.1:  
```{r}
active_det_false <- constrained_accessibility(constraint="total",
                          travel_matrix, land_use_data,
                                travel_cost = "travel_time",
                                decay_function = decay_power(decay_value = 0.1),
                                demand = NULL,
                                supply = "jobs",
                                detailed_results = FALSE,
                                active = TRUE)

active_det_true <- constrained_accessibility(constraint="total",
                          travel_matrix, land_use_data,
                                travel_cost = "travel_time",
                                decay_function = decay_power(decay_value = 0.1),
                                demand = NULL,
                                supply = "jobs",
                                detailed_results = TRUE,
                                active = TRUE)
active_det_false
active_det_true

active_det_false$supply |> sum()
active_det_true$supply |> sum()
active_det_true$kappa_total |> sum()
```
Should be:
1	164.0802			
2	160.6921			
3	165.2277

## Total constrained--PASSIVE
Checking for decay_power scenario 3:  
```{r}
passive_det_false <- constrained_accessibility(constraint="total",
                          travel_matrix, land_use_data,
                                travel_cost = "travel_time",
                                decay_function = decay_power(decay_value = 3),
                                demand = "population",
                                supply = NULL,
                                detailed_results = FALSE,
                                active = FALSE)

passive_det_true <- constrained_accessibility(constraint="total",
                          travel_matrix, land_use_data,
                                travel_cost = "travel_time",
                                decay_function = decay_power(decay_value = 3),
                                demand = "population",
                                supply = NULL,
                                detailed_results = TRUE,
                                active = FALSE)
passive_det_false
passive_det_true

passive_det_false$demand |> sum()
passive_det_true$demand |> sum()
passive_det_true$hatkappa_total |> sum()
```
Should be:
1	5.017774			
2	8.595749			
3	6.386477	

Checking for decay_power scenario 2:  
```{r}
passive_det_false <- constrained_accessibility(constraint="total",
                          travel_matrix, land_use_data,
                                travel_cost = "travel_time",
                                decay_function = decay_power(decay_value = 2),
                                demand = "population",
                                supply = NULL,
                                detailed_results = FALSE,
                                active = FALSE)

passive_det_true <- constrained_accessibility(constraint="total",
                          travel_matrix, land_use_data,
                                travel_cost = "travel_time",
                                decay_function = decay_power(decay_value = 2),
                                demand = "population",
                                supply = NULL,
                                detailed_results = TRUE,
                                active = FALSE)
passive_det_false
passive_det_true

passive_det_false$demand |> sum()
passive_det_true$demand |> sum()
passive_det_true$hatkappa_total |> sum()
```
Should be:
1	5.446623			
2	7.986306			
3	6.567071	

Checking for decay_power scenario 0.1:  
```{r}
passive_det_false <- constrained_accessibility(constraint="total",
                          travel_matrix, land_use_data,
                                travel_cost = "travel_time",
                                decay_function = decay_power(decay_value = 0.1),
                                demand = "population",
                                supply = NULL,
                                detailed_results = FALSE,
                                active = FALSE)

passive_det_true <- constrained_accessibility(constraint="total",
                          travel_matrix, land_use_data,
                                travel_cost = "travel_time",
                                decay_function = decay_power(decay_value = 0.1),
                                demand = "population",
                                supply = NULL,
                                detailed_results = TRUE,
                                active = FALSE)
passive_det_false
passive_det_true

passive_det_false$demand |> sum()
passive_det_true$demand |> sum()
passive_det_true$hatkappa_total |> sum()
```
Should be:
1	6.598332			
2	6.717223			
3	6.684444	

## Singly constrained--ACTIVE
Checking for decay_power scenario 3:  
```{r}
active_det_false <- constrained_accessibility(constraint="singly",
                          travel_matrix, land_use_data,
                                travel_cost = "travel_time",
                                decay_function = decay_power(decay_value = 3),
                                demand = "population",
                                supply = "jobs",
                                detailed_results = FALSE,
                                active = TRUE)

active_det_true <- constrained_accessibility(constraint="singly",
                          travel_matrix, land_use_data,
                                travel_cost = "travel_time",
                                decay_function = decay_power(decay_value = 3),
                                demand = "population",
                                supply = "jobs",
                                detailed_results = TRUE,
                                active = TRUE)
active_det_false
active_det_true

active_det_false$supply |> sum()
active_det_true$supply |> sum()
active_det_true$kappa_singly |> sum()  #should be the number of destinations or origins

round(active_det_false$supply / land_use_data$population, 6)
```
Should be:
1	133.4687			
2	166.7813			
3	189.7499	

And the per capita should be:
33.36718 16.67813 31.62499

Checking for decay_power scenario 2:  
```{r}
active_det_false <- constrained_accessibility(constraint="singly",
                          travel_matrix, land_use_data,
                                travel_cost = "travel_time",
                                decay_function = decay_power(decay_value = 2),
                                demand = "population",
                                supply = "jobs",
                                detailed_results = FALSE,
                                active = TRUE)

active_det_true <- constrained_accessibility(constraint="singly",
                          travel_matrix, land_use_data,
                                travel_cost = "travel_time",
                                decay_function = decay_power(decay_value = 2),
                                demand = "population",
                                supply = "jobs",
                                detailed_results = TRUE,
                                active = TRUE)
active_det_false
active_det_true

active_det_false$supply |> sum()
active_det_true$supply |> sum()
active_det_true$kappa_singly|> sum()  #should be the number of destinations or origins

round(active_det_false$supply / land_use_data$population, 6)
```
Should be:
1	122.2546			
2	185.0957			
3	182.6497	

And the per capita should be:
30.56365 18.50957 30.44161

Checking for decay_power scenario 0.1:  
```{r}
active_det_false <- constrained_accessibility(constraint="singly",
                          travel_matrix, land_use_data,
                                travel_cost = "travel_time",
                                decay_function = decay_power(decay_value = 0.1),
                                demand = "population",
                                supply = "jobs",
                                detailed_results = FALSE,
                                active = TRUE)

active_det_true <- constrained_accessibility(constraint="singly",
                          travel_matrix, land_use_data,
                                travel_cost = "travel_time",
                                decay_function = decay_power(decay_value = 0.1),
                                demand = "population",
                                supply = "jobs",
                                detailed_results = TRUE,
                                active = TRUE)
active_det_false
active_det_true

active_det_false$supply |> sum()
active_det_true$supply |> sum()
active_det_true$kappa_singly |> sum()  #should be the number of destinations or origins

round(active_det_false$supply / land_use_data$population, 6)
```
Should be:
1	98.84766			
2	241.87721			
3	149.27513	

And the per capita should be:
24.71192 24.18772 24.87919

## Singly constrained--PASSIVE
Checking for decay_power scenario 3:  
```{r}
passive_det_false <- constrained_accessibility(constraint="singly",
                          travel_matrix, land_use_data,
                                travel_cost = "travel_time",
                                decay_function = decay_power(decay_value = 3),
                                demand = "population",
                                supply = "jobs",
                                detailed_results = FALSE,
                                active = FALSE)

passive_det_true <- constrained_accessibility(constraint="singly",
                          travel_matrix, land_use_data,
                                travel_cost = "travel_time",
                                decay_function = decay_power(decay_value = 3),
                                demand = "population",
                                supply = "jobs",
                                detailed_results = TRUE,
                                active = FALSE)
passive_det_false
passive_det_true

passive_det_false$demand |> sum()
passive_det_true$demand |> sum()
passive_det_true$hatkappa_singly |> sum()
```


Checking for decay_power scenario 2:  
```{r}
passive_det_false <- constrained_accessibility(constraint="singly",
                          travel_matrix, land_use_data,
                                travel_cost = "travel_time",
                                decay_function = decay_power(decay_value = 2),
                                demand = "population",
                                supply = "jobs",
                                detailed_results = FALSE,
                                active = FALSE)

passive_det_true <- constrained_accessibility(constraint="singly",
                          travel_matrix, land_use_data,
                                travel_cost = "travel_time",
                                decay_function = decay_power(decay_value = 2),
                                demand = "population",
                                supply = "jobs",
                                detailed_results = TRUE,
                                active = FALSE)
passive_det_false
passive_det_true

passive_det_false$demand |> sum()
passive_det_true$demand |> sum()
passive_det_true$hatkappa_singly |> sum()
```

Checking for decay_power scenario 0.1:  
```{r}
passive_det_false <- constrained_accessibility(constraint="singly",
                          travel_matrix, land_use_data,
                                travel_cost = "travel_time",
                                decay_function = decay_power(decay_value = 0.1),
                                demand = "population",
                                supply = "jobs",
                                detailed_results = FALSE,
                                active = FALSE)

passive_det_true <- constrained_accessibility(constraint="singly",
                          travel_matrix, land_use_data,
                                travel_cost = "travel_time",
                                decay_function = decay_power(decay_value = 0.1),
                                demand = "population",
                                supply = "jobs",
                                detailed_results = TRUE,
                                active = FALSE)
passive_det_false
passive_det_true

passive_det_false$demand |> sum()
passive_det_true$demand |> sum()
passive_det_true$hatkappa_singly |> sum()
```

# Test 3. Testing results with doubly constrained toy example in 2025 paper
```{r}
# Toy dataset where totals match (sum demand = sum supply = 20)
travel_matrix <- expand.grid(from_id = c("1", "2", "3"),
                             to_id = c("1", "2", "3")) |>
  dplyr::mutate(travel_time = c(10, 30, 15, 30, 10, 25, 15, 25, 10)) |>
  data.table()

land_use_data <- data.table(
  id = c("1", "2", "3"),
  population = c(4, 10, 6),
  jobs = c(7, 5, 8)
)
```


Checking for decay_power scenario 3:  
```{r}
active_det_true <- constrained_accessibility(constraint="doubly",
                          travel_matrix, land_use_data,
                                travel_cost = "travel_time",
                                decay_function = decay_power(decay_value = 3),
                                demand = "population",
                                supply = "jobs",
                                detailed_results = TRUE,
                                active = TRUE)

active_det_true

active_det_true$flow |> sum()
active_det_true$kappa_doubly |> sum()  #should be the number of destinations or origins
```
Should be:
1	1	3.23585899	
1	2	0.01032226	
1	3	0.75565683	
2	1	2.13260167	
2	2	4.95932483	
2	3	2.90443915	
3	1	1.63153933	
3	2	0.03035291	
3	3	4.33990401	

So... this matches table 10, but I accidentally flipped the columns. like 2->1 in the table should be 1->2. 
 

Checking for decay_power scenario 2:  
```{r}
active_det_true <- constrained_accessibility(constraint="doubly",
                          travel_matrix, land_use_data,
                                travel_cost = "travel_time",
                                decay_function = decay_power(decay_value = 2),
                                demand = "population",
                                supply = "jobs",
                                detailed_results = TRUE,
                                active = TRUE)

active_det_true

active_det_true$flow |> sum()
active_det_true$kappa_doubly |> sum()  #should be the number of destinations or origins
```

Checking for decay_power scenario 0.1:  
```{r}
active_det_true <- constrained_accessibility(constraint="doubly",
                          travel_matrix, land_use_data,
                                travel_cost = "travel_time",
                                decay_function = decay_power(decay_value = 0.1),
                                demand = "population",
                                supply = "jobs",
                                detailed_results = TRUE,
                                active = TRUE)

active_det_true

active_det_true$flow |> sum()
active_det_true$kappa_doubly |> sum()  #should be the number of destinations or origins
```


# Test 4. Testing singly constrained's equality to spatial_availbility() and 2SFCA(). 
```{r}
travel_matrix <- expand.grid(from_id = c("1", "2", "3"),
                  to_id = c("1", "2", "3")) |>
  mutate(travel_time = c(10, 30, 15, 30, 10, 25, 15, 25, 10)) |> data.table()

land_use_data <- data.table(
  id = c("1", "2", "3"),
  population = c(4, 10, 6),
  jobs = c(160, 150, 180)
)
```

Spatial_availability() equality:
```{r}
constrained_accessibility(constraint = "singly",
                          travel_matrix, land_use_data,
                                travel_cost = "travel_time",
                                decay_function = decay_power(decay_value = 2),
                                demand = "population",
                                supply = "jobs",
                                detailed_results = FALSE,
                                active = TRUE)

spatial_availability(travel_matrix,
    land_use_data,
    opportunity = "jobs",
    travel_cost = "travel_time",
    demand = "population",
    decay_function = decay_power(decay_value = 2),
    alpha = 1
  )
```

floating_catchment_area(method="2sfca") equality:
```{r}
result_singly <- constrained_accessibility(constraint = "singly",
                          travel_matrix, land_use_data,
                                travel_cost = "travel_time",
                                decay_function = decay_power(decay_value = 2),
                                demand = "population",
                                supply = "jobs",
                                detailed_results = FALSE,
                                active = TRUE)

round(result_singly$supply / land_use_data$population, 6)
  
floating_catchment_area(
    travel_matrix,
    land_use_data,
    opportunity = "jobs",
    travel_cost = "travel_time",
    demand = "population",
    method = "2sfca",
    decay_function = decay_power(decay_value = 2)
  )

```
yay!

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/decay_logistic.R
\name{decay_logistic}
\alias{decay_logistic}
\title{Logistic decay function}
\usage{
decay_logistic(cutoff, sd)
}
\arguments{
\item{cutoff}{A \code{numeric} vector. The cost value that serves as the
inflection point of the cumulative logistic curve.}

\item{sd}{A \code{numeric} vector with same length as \code{cutoff}. The standard
deviation of the logistic curve. Values near 0 result in weighting curves
that approximate binary decay, while higher values tend to linearize the
decay.}
}
\value{
A \code{function} that takes a generic travel cost vector (\code{numeric}) as
input and returns a vector of weights (\code{numeric}).
}
\description{
Returns a logistic weighting function (in which the weights
follow the distribution of a reversed cumulative logistic curve) to be
used inside accessibility calculating functions. The logistic curve is
parameterized with the cutoff that sets its inflection point and the standard
deviation that sets its steepness.

This function is generic over any kind of numeric travel cost,
such as distance, time and money.
}
\details{
When using a function created with \code{decay_logistic()}, the output is
named after the combination of cutoffs (\code{"c"}) and standard deviations
(\code{"sd"}) - e.g. given the cutoffs \code{c(30, 40)} and the standard deviations
\code{c(10, 20)}, the first element of the output will be named \code{"c30;sd10"} and
the second will be named \code{"c40;sd20"}. This function uses the adjusted
logistic decay curve proposed by
\insertCite{bauer2016measuring;textual}{accessibility}, in which the
condition f(0) = 1 is met (i.e. the weight of an opportunity whose cost to
reach is 0 is 1).
}
\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}
weighting_function <- decay_logistic(cutoff = 30, sd = 5)

weighting_function(c(0, 30, 45, 60))

weighting_function <- decay_logistic(cutoff = c(30, 45), sd = c(5, 10))

weighting_function(c(0, 30, 45, 60))
\dontshow{\}) # examplesIf}
}
\references{
\insertAllCited{}
}
\seealso{
Other decay functions: 
\code{\link{decay_binary}()},
\code{\link{decay_exponential}()},
\code{\link{decay_linear}()},
\code{\link{decay_power}()},
\code{\link{decay_stepped}()}
}
\concept{decay functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cost_to_closest.R
\name{cost_to_closest}
\alias{cost_to_closest}
\title{Minimum travel cost to closest N number of opportunities}
\usage{
cost_to_closest(
  travel_matrix,
  land_use_data,
  opportunity,
  travel_cost,
  n = 1,
  group_by = character(0),
  active = TRUE,
  fill_missing_ids = TRUE
)
}
\arguments{
\item{travel_matrix}{A data frame. The travel matrix describing the costs
(i.e. travel time, distance, monetary cost, etc.) between the origins and
destinations in the study area. Must contain the columns \code{from_id}, \code{to_id}
and any others specified in \code{travel_cost}.}

\item{land_use_data}{A data frame. The distribution of opportunities within
the study area cells. Must contain the columns \code{id} and any others
specified in \code{opportunity}.}

\item{opportunity}{A string. The name of the column in \code{land_use_data}
with the number of opportunities/resources/services to be considered when
calculating accessibility levels.}

\item{travel_cost}{A string. The name of the column in \code{travel_matrix} with
the travel cost between origins and destinations. The notion of cost
here is generic over any kind of numeric travel cost, such as distance,
time and money.}

\item{n}{A \code{numeric} vector. The minimum number of opportunities that should
be considered. Defaults to 1. If more than one value is provided, the
output includes an extra column specifying the number of opportunities that
the minimum travel cost refers to.}

\item{group_by}{A \code{character} vector. When not \code{character(0)} (the default),
indicates the \code{travel_matrix} columns that should be used to group the
accessibility estimates by. For example, if \code{travel_matrix} includes a
departure time column, that specifies the departure time of each entry in
the data frame, passing \code{"departure_time"} to this parameter results in
accessibility estimates grouped by origin and by departure time.}

\item{active}{A logical. When \code{TRUE}, the function calculates active
accessibility (the quantity of opportunities that can be reached from
a given origin). when \code{FALSE}, it calculates passive accessibility (by
how many people each destination can be reached), which is equivalent
to the notion of market potential.}

\item{fill_missing_ids}{A \code{logical}. Calculating minimum travel cost to
closest N number of opportunities may result in missing ids in the output
if they cannot reach the specified amount of opportunities across all
destinations they can reach. For example, estimating the minimum travel
time that an origin that can only reach 4 opportunities takes to reach 5
opportunities resulting in such origin not being included in the output.
When \code{TRUE} (the default), the function identifies which ids would be left
out from the output and fill their respective minimum travel costs with
\code{Inf}, which incurs in a performance penalty.}
}
\value{
A data frame containing the accessibility estimates for each
origin/destination (depending if \code{active} is \code{TRUE} or \code{FALSE}) in the
travel matrix.
}
\description{
Calculates the minimum travel cost to the closest N number of opportunities.

This function is generic over any kind of numeric travel cost,
such as distance, time and money.
}
\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}
data_dir <- system.file("extdata", package = "accessibility")
travel_matrix <- readRDS(file.path(data_dir, "travel_matrix.rds"))
land_use_data <- readRDS(file.path(data_dir, "land_use_data.rds"))

df <- cost_to_closest(
  travel_matrix,
  land_use_data,
  n = 1,
  opportunity = "schools",
  travel_cost = "travel_time"
)
head(df)

df <- cost_to_closest(
  travel_matrix,
  land_use_data,
  n = c(1, 2),
  opportunity = "schools",
  travel_cost = "travel_time"
)
head(df)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gravity.R
\name{gravity}
\alias{gravity}
\title{Gravity-based accessibility measures}
\usage{
gravity(
  travel_matrix,
  land_use_data,
  opportunity,
  travel_cost,
  decay_function,
  group_by = character(0),
  active = TRUE,
  fill_missing_ids = TRUE
)
}
\arguments{
\item{travel_matrix}{A data frame. The travel matrix describing the costs
(i.e. travel time, distance, monetary cost, etc.) between the origins and
destinations in the study area. Must contain the columns \code{from_id}, \code{to_id}
and any others specified in \code{travel_cost}.}

\item{land_use_data}{A data frame. The distribution of opportunities within
the study area cells. Must contain the columns \code{id} and any others
specified in \code{opportunity}.}

\item{opportunity}{A string. The name of the column in \code{land_use_data}
with the number of opportunities/resources/services to be considered when
calculating accessibility levels.}

\item{travel_cost}{A string. The name of the column in \code{travel_matrix} with
the travel cost between origins and destinations. The notion of cost
here is generic over any kind of numeric travel cost, such as distance,
time and money.}

\item{decay_function}{A \code{fuction} that converts travel cost into an
impedance factor used to weight opportunities. This function should take a
\code{numeric} vector and also return a \code{numeric} vector as output, with the
same length as the input. For convenience, the package currently includes
the following functions: \code{\link[=decay_binary]{decay_binary()}}, \code{\link[=decay_exponential]{decay_exponential()}},
\code{\link[=decay_power]{decay_power()}} and \code{\link[=decay_stepped]{decay_stepped()}}. See the documentation of each decay
function for more details.}

\item{group_by}{A \code{character} vector. When not \code{character(0)} (the default),
indicates the \code{travel_matrix} columns that should be used to group the
accessibility estimates by. For example, if \code{travel_matrix} includes a
departure time column, that specifies the departure time of each entry in
the data frame, passing \code{"departure_time"} to this parameter results in
accessibility estimates grouped by origin and by departure time.}

\item{active}{A logical. When \code{TRUE}, the function calculates active
accessibility (the quantity of opportunities that can be reached from
a given origin). when \code{FALSE}, it calculates passive accessibility (by
how many people each destination can be reached), which is equivalent
to the notion of market potential.}

\item{fill_missing_ids}{A \code{logical}. When calculating grouped accessibility
estimates (i.e. when \code{by_col} is not \code{NULL}), some combinations of groups
and origins may be missing. For example, if a single trip can depart from
origin \code{A} at 7:15am and reach destination \code{B} within 55 minutes, but no
trips departing from \code{A} at 7:30am can be completed at all, this second
combination will not be included in the output. When \code{TRUE} (the default),
the function identifies which combinations would be left out and fills
their respective accessibility values with 0, which incurs in a
performance penalty.}
}
\value{
A data frame containing the accessibility estimates for each
origin/destination (depending if \code{active} is \code{TRUE} or \code{FALSE}) in the
travel matrix.
}
\description{
Calculates gravity-based accessibility using a decay function specified by
the user.

This function is generic over any kind of numeric travel cost,
such as distance, time and money.
}
\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}
data_dir <- system.file("extdata", package = "accessibility")
travel_matrix <- readRDS(file.path(data_dir, "travel_matrix.rds"))
land_use_data <- readRDS(file.path(data_dir, "land_use_data.rds"))

df_linear <- gravity(
  travel_matrix,
  land_use_data,
  decay_function = decay_linear(cutoff = 50),
  opportunity = "schools",
  travel_cost = "travel_time"
)
head(df_linear)

df_exp <- gravity(
  travel_matrix,
  land_use_data,
  decay_function = decay_exponential(decay_value = 0.5),
  opportunity = "schools",
  travel_cost = "travel_time"
)
head(df_exp)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/singly_constrained.R
\name{singly_constrained}
\alias{singly_constrained}
\title{Singly constrained accessibility}
\value{
A \code{data.table}/\code{data.frame} with results (structure mirrors the wrapper).
}
\description{
Allocates opportunities at each destination proportionally based on travel
impedance and population at the origin. Uses the logic of Wilson's single
constraint. Returns values in the unit of 'demand'.
Internal helper function used by \code{\link[=constrained_accessibility]{constrained_accessibility()}} when \code{constraint = "singly"}.
}
\examples{
NULL
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cumulative_cutoff.R
\name{cumulative_cutoff}
\alias{cumulative_cutoff}
\title{Cumulative access based on a travel cost cutoff}
\usage{
cumulative_cutoff(
  travel_matrix,
  land_use_data,
  opportunity,
  travel_cost,
  cutoff,
  group_by = character(0),
  active = TRUE,
  fill_missing_ids = TRUE
)
}
\arguments{
\item{travel_matrix}{A data frame. The travel matrix describing the costs
(i.e. travel time, distance, monetary cost, etc.) between the origins and
destinations in the study area. Must contain the columns \code{from_id}, \code{to_id}
and any others specified in \code{travel_cost}.}

\item{land_use_data}{A data frame. The distribution of opportunities within
the study area cells. Must contain the columns \code{id} and any others
specified in \code{opportunity}.}

\item{opportunity}{A string. The name of the column in \code{land_use_data}
with the number of opportunities/resources/services to be considered when
calculating accessibility levels.}

\item{travel_cost}{A \code{character} vector. The name of the columns in
\code{travel_matrix} with the travel costs between origins and destinations to
be considered in the calculation.}

\item{cutoff}{Either a \code{numeric} vector or a list of \code{numeric} vectors, one
for each cost specified in \code{travel_cost}. The travel cost cutoffs to
consider when calculating accessibility levels. If a list, the function
finds every single possible cutoff combination and use them to calculate
accessibility (e.g. if one specifies that travel time cutoffs should be 30
and 60 minutes and that monetary cost cutoffs should be 5 and 10 dollars,
the output includes accessibility estimates limited at 30 min & 5 dollars,
30 min & 10 dollars, 60 min & 5 dollars and 60 min & 10 dollars). In these
cases, cost constraints are considered simultaneously - i.e. only trips
that take 30 minutes or less AND 5 dollars or less to be completed, for
example, are included in the accessibility output. The cutoff parameter is
not included in the final output if the input includes only a single cutoff
for a single travel cost.}

\item{group_by}{A \code{character} vector. When not \code{character(0)} (the default),
indicates the \code{travel_matrix} columns that should be used to group the
accessibility estimates by. For example, if \code{travel_matrix} includes a
departure time column, that specifies the departure time of each entry in
the data frame, passing \code{"departure_time"} to this parameter results in
accessibility estimates grouped by origin and by departure time.}

\item{active}{A logical. When \code{TRUE}, the function calculates active
accessibility (the quantity of opportunities that can be reached from
a given origin). when \code{FALSE}, it calculates passive accessibility (by
how many people each destination can be reached), which is equivalent
to the notion of market potential.}

\item{fill_missing_ids}{A \code{logical}. Calculating cumulative accessibility may
result in missing ids if the they cannot reach any of the destinations
within the specified travel cost cutoff. For example, using a travel time
cutoff of 20 minutes, when estimating the accessibility of origin \code{A} that
can only reach destinations with more than 40 minutes results in id \code{A}
not being included in the output. When \code{TRUE} (the default), the function
identifies which origins would be left out and fills their respective
accessibility values with 0, which incurs in a performance penalty.}
}
\value{
A data frame containing the accessibility estimates for each
origin/destination (depending if \code{active} is \code{TRUE} or \code{FALSE}) in the
travel matrix.
}
\description{
Calculates the number of opportunities accessible under a given specified
travel cost cutoff.

This function is generic over any kind of numeric travel cost,
such as distance, time and money.
}
\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}
data_dir <- system.file("extdata", package = "accessibility")
travel_matrix <- readRDS(file.path(data_dir, "travel_matrix.rds"))
land_use_data <- readRDS(file.path(data_dir, "land_use_data.rds"))

# active accessibility: number of schools accessible from each origin
df <- cumulative_cutoff(
  travel_matrix = travel_matrix,
  land_use_data = land_use_data,
  cutoff = 30,
  opportunity = "schools",
  travel_cost = "travel_time"
)
head(df)

df <- cumulative_cutoff(
  travel_matrix = travel_matrix,
  land_use_data = land_use_data,
  cutoff = c(30, 60),
  opportunity = "schools",
  travel_cost = "travel_time"
)
head(df)

# passive accessibility: number of people that can reach each destination
df <- cumulative_cutoff(
  travel_matrix = travel_matrix,
  land_use_data = land_use_data,
  cutoff = 30,
  opportunity = "population",
  travel_cost = "travel_time",
  active = FALSE
)
head(df)

# using multiple travel costs
pareto_frontier <- readRDS(file.path(data_dir, "pareto_frontier.rds"))

df <- cumulative_cutoff(
  pareto_frontier,
  land_use_data = land_use_data,
  opportunity = "jobs",
  travel_cost = c("travel_time", "monetary_cost"),
  cutoff = list(c(20, 30), c(0, 5, 10))
)
head(df)
\dontshow{\}) # examplesIf}
}
\seealso{
Other cumulative access: 
\code{\link{cumulative_interval}()}
}
\concept{cumulative access}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/concentration_index.R
\name{concentration_index}
\alias{concentration_index}
\title{Concentration Index}
\usage{
concentration_index(
  accessibility_data,
  sociodemographic_data,
  opportunity,
  population,
  income,
  type,
  group_by = character(0)
)
}
\arguments{
\item{accessibility_data}{A data frame. The accessibility levels whose
inequality should be calculated. Must contain the columns \code{id} and any
others specified in \code{opportunity}.}

\item{sociodemographic_data}{A data frame. The distribution of
sociodemographic characteristics of the population in the study area cells.
Must contain the columns \code{id} and any others specified in \code{population} and
\code{income}.}

\item{opportunity}{A string. The name of the column in \code{accessibility_data}
with the accessibility levels to be considerend when calculating inequality
levels.}

\item{population}{A string. The name of the column in \code{sociodemographic_data}
with the number of people in each cell. Used to weigh accessibility levels
when calculating inequality.}

\item{income}{A string. The name of the column in \code{sociodemographic_data}
with the income variable that should be used to sort the population from
the least to the most privileged. Please note that this variable should
describe income per capita (e.g. mean income per capita, household income
per capita, etc), instead of the total amount of income in each cell. Also
note that, while income is generally used to rank population groups, any
variable that can be used to describe one's socioeconomic status, such as
education level, can be passed to this argument, as long as it can be
numerically ordered (in which higher values denote higher socioeconomic
status).}

\item{type}{A string. Which type of Concentration Index to calculate. Current
available options are \code{"standard"} and \code{"corrected"}.}

\item{group_by}{A \code{character} vector. When not \code{character(0)} (the default),
indicates the \code{accessibility_data} columns that should be used to group the
inequality estimates by. For example, if \code{accessibility_data} includes a
\code{scenario} column that identifies distinct scenarios that each
accessibility estimates refer to (e.g. before and after a transport policy
intervention), passing \code{"scenario"} to this parameter results in inequality
estimates grouped by scenario.}
}
\value{
A data frame containing the inequality estimates for the study area.
}
\description{
Calculates the Concentration Index (CI) of a given accessibility
distribution. This measures estimates the extent to which accessibility
inequalities are systematically associated with individuals' socioeconomic
levels. CI values can theoretically vary between -1 and +1 (when all
accessibility is concentrated in the most or in the least disadvantaged
person, respectively). Negative values indicate that inequalities favor the
poor, while positive values indicate a pro-rich bias. The function supports
calculating the standard relative CI and the corrected CI, as proposed by
\insertCite{erreygers2009correcting;textual}{accessibility}.
}
\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}
data_dir <- system.file("extdata", package = "accessibility")
travel_matrix <- readRDS(file.path(data_dir, "travel_matrix.rds"))
land_use_data <- readRDS(file.path(data_dir, "land_use_data.rds"))

access <- cumulative_cutoff(
  travel_matrix,
  land_use_data,
  cutoff = 30,
  opportunity = "jobs",
  travel_cost = "travel_time"
)

ci <- concentration_index(
  access,
  sociodemographic_data = land_use_data,
  opportunity = "jobs",
  population = "population",
  income = "income_per_capita",
  type = "corrected"
)
ci
\dontshow{\}) # examplesIf}
}
\references{
\insertAllCited{}
}
\seealso{
Other inequality: 
\code{\link{gini_index}()},
\code{\link{palma_ratio}()},
\code{\link{theil_t}()}
}
\concept{inequality}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/decay_stepped.R
\name{decay_stepped}
\alias{decay_stepped}
\title{Stepped decay function}
\usage{
decay_stepped(steps, weights)
}
\arguments{
\item{steps}{A \code{numeric} vector or a list of \code{numeric} vectors. The travel
cost steps, in ascending order. Please do not include travel cost 0 as a
step: this is already handled by the function.}

\item{weights}{A \code{numeric} vector with same length as \code{steps} or a list of
\code{numeric} vectors whose lengths are equal to the lengths of the elements of
same index in \code{steps}. The values, between 0 and 1, that the function
assumes at each step. Please do not include weight 1 as the first value:
this is already handled by the function. The function considers the steps'
intervals "open on the right", meaning that the function assumes the step
value at the actual step, not right after it. Please see the illustrative
examples for effects of this assumption on the results.}
}
\value{
A \code{function} that takes a generic travel cost vector (\code{numeric}) as
an input and returns a vector of weights (\code{numeric}).
}
\description{
Returns a stepped weighting function to be used inside accessibility
calculating functions.

This function is generic over any kind of numeric travel cost,
such as distance, time and money.
}
\details{
When both \code{steps} and \code{weights} parameters are given \code{list}s, their
content are matched element-wise to define each stepped weighting function
\itemize{
\item i.e. the first element of \code{steps} is matched to the first element of
\code{weights}, the second element of \code{steps} is matched to the second of
\code{weights}, etc. When using a function created with \code{decay_stepped()}, the
output is named after the combination of steps (\code{"s"}) and weights (\code{"w"})
\item e.g. given the steps \code{c(10, 20, 30)} and the weights \code{c(0.66, 0.33, 0)},
the output will be named \code{"s(10,20,30);w(0.66,0.33,0)"}.
}
}
\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}
weighting_function <- decay_stepped(
  c(10, 20, 30, 40),
  weights = c(0.75, 0.5, 0.25, 0)
)

weighting_function(c(5, 25, 35, 45))

weighting_function <- decay_stepped(
  list(c(10, 20, 30, 40), c(10, 20, 30, 40)),
  weights = list(c(0.75, 0.5, 0.25, 0), c(0.8, 0.6, 0.4, 0.2))
)

weighting_function(c(5, 25, 35, 45))

# intervals are open on the right, so the values change exactly at each step
weighting_function(c(0, 10, 20, 30, 40))
\dontshow{\}) # examplesIf}
}
\seealso{
Other decay functions: 
\code{\link{decay_binary}()},
\code{\link{decay_exponential}()},
\code{\link{decay_linear}()},
\code{\link{decay_logistic}()},
\code{\link{decay_power}()}
}
\concept{decay functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/balancing_cost.R
\name{balancing_cost}
\alias{balancing_cost}
\title{Balancing cost accessibility measure}
\usage{
balancing_cost(
  travel_matrix,
  land_use_data,
  opportunity,
  travel_cost,
  demand,
  cost_increment = 1,
  group_by = character(0),
  fill_missing_ids = TRUE
)
}
\arguments{
\item{travel_matrix}{A data frame. The travel matrix describing the costs
(i.e. travel time, distance, monetary cost, etc.) between the origins and
destinations in the study area. Must contain the columns \code{from_id}, \code{to_id}
and any others specified in \code{travel_cost}.}

\item{land_use_data}{A data frame. The distribution of opportunities within
the study area cells. Must contain the columns \code{id} and any others
specified in \code{opportunity}.}

\item{opportunity}{A string. The name of the column in \code{land_use_data}
with the number of opportunities/resources/services to be considered when
calculating accessibility levels.}

\item{travel_cost}{A string. The name of the column in \code{travel_matrix} with
the travel cost between origins and destinations. The notion of cost
here is generic over any kind of numeric travel cost, such as distance,
time and money.}

\item{demand}{A string. The name of the column in \code{land_use_data} with the
number of opportunity-demanders at each origin (e.g., people) that will be
considered.}

\item{cost_increment}{A number. The cost increment that should be used when
defining the travel cost distribution from which the potential balancing
costs will be picked. For example, an increment of 1 tends to suitable for
travel time distributions, meaning that the function will first check if
any origins reach their balancing cost with a travel time of 0 minutes,
then 1 minute, 2 minutes, 3, 4, ..., etc. A increment of 1 might be too big
for a distribution of monetary costs, on the other hand, which could
possibly benefit from a smaller increment of 0.05, for example, resulting
in the function looking for balancing costs first at a cost of 0, then
0.05, 0.10, ..., etc. Defaults to 1.}

\item{group_by}{A \code{character} vector. When not \code{character(0)} (the default),
indicates the \code{travel_matrix} columns that should be used to group the
accessibility estimates by. For example, if \code{travel_matrix} includes a
departure time column, that specifies the departure time of each entry in
the data frame, passing \code{"departure_time"} to this parameter results in
accessibility estimates grouped by origin and by departure time.}

\item{fill_missing_ids}{A \code{logical}. When calculating grouped accessibility
estimates (i.e. when \code{by_col} is not \code{NULL}), some combinations of groups
and origins may be missing. For example, if a single trip can depart from
origin \code{A} at 7:15am and reach destination \code{B} within 55 minutes, but no
trips departing from \code{A} at 7:30am can be completed at all, this second
combination will not be included in the output. When \code{TRUE} (the default),
the function identifies which combinations would be left out and fills
their respective accessibility values with 0, which incurs in a
performance penalty.}
}
\value{
A data frame containing the accessibility estimates for each
origin/destination (depending if \code{active} is \code{TRUE} or \code{FALSE}) in the
travel matrix.

A data frame containing the accessibility estimates for each origin
in the travel matrix. Origins marked with a \code{NA} balancing cost never reach
as many opportunities as there is people residing in them, given the
specified travel matrix.
}
\description{
Calculates the balancing cost measure, which is defined as the travel cost
required to reach as many opportunities as the number of people in a given
origin. Originally proposed by
\insertCite{barboza2021balancing;textual}{accessibility}, under the name
"balancing time".

This function is generic over any kind of numeric travel cost,
such as distance, time and money.
}
\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}
data_dir <- system.file("extdata", package = "accessibility")
travel_matrix <- readRDS(file.path(data_dir, "travel_matrix.rds"))
land_use_data <- readRDS(file.path(data_dir, "land_use_data.rds"))

bc <- balancing_cost(
  travel_matrix,
  land_use_data,
  opportunity = "jobs",
  travel_cost = "travel_time",
  demand = "population"
)
head(bc)
\dontshow{\}) # examplesIf}
}
\references{
\insertAllCited{}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fgt_poverty.R
\name{fgt_poverty}
\alias{fgt_poverty}
\title{Foster-Greer-Thorbecke (FGT) poverty measures}
\usage{
fgt_poverty(
  accessibility_data,
  sociodemographic_data,
  opportunity,
  population,
  poverty_line,
  poor_below_threshold = TRUE,
  group_by = character(0)
)
}
\arguments{
\item{accessibility_data}{A data frame. The accessibility levels whose
poverty levels should be calculated. Must contain the columns \code{id} and any
others specified in \code{opportunity}.}

\item{sociodemographic_data}{A data frame. The distribution of
sociodemographic characteristics of the population in the study area cells.
Must contain the columns \code{id} and any others specified in \code{population}.}

\item{opportunity}{A string. The name of the column in \code{accessibility_data}
with the accessibility levels to be considerend when calculating
accessibility poverty.}

\item{population}{A string. The name of the column in \code{sociodemographic_data}
with the number of people in each cell. Used to weigh accessibility levels
when calculating poverty.}

\item{poverty_line}{A \code{numeric}. The poverty line below which individuals are
considered to be in accessibility poverty.}

\item{poor_below_threshold}{Logic. If \code{TRUE} (default,) the observations below
the poverty line are considered to be poor. This is the correct approach
for primal accessibility measures (e.g.cumulative accessibility). If
\code{FALSE}, then observations above the poverty line are considered to be
poor. This is the correct approach for dual accessibility measures (e.g.
travel time to the closest facility). When set to \code{FALSE}, FGT 1 and 2
do not have an upper bound.}

\item{group_by}{A \code{character} vector. When not \code{character(0)} (the default),
indicates the \code{accessibility_data} columns that should be used to group the
poverty estimates by. For example, if \code{accessibility_data} includes a
\code{race} column that specifies the racial category of the population (e.g.
\code{"black"} and \code{"white"}) that each entry refers to, passing \code{"race"} to
this parameter results in poverty estimates grouped by race.}
}
\value{
A data frame containing the three poverty estimates (FGT0, FGT1 and
FGT2) for the study area.
}
\description{
Calculates the FGT metrics, a family of poverty measures originally proposed
by \insertCite{foster1984class;textual}{accessibility} that capture the
extent and severity of poverty within an accessibility distribution. The FGT
family is composed of three measures that differ based on the \eqn{\alpha}
parameter used to calculate them (either 0, 1 or 2) and which also changes
their interpretation. Please see the details section for more information on
the interpretation of the measures.
}
\section{Interpretation of FGT measures}{

The interpretation of each FGT measure depends on the \eqn{\alpha} parameter
used to calculate it:
\itemize{
\item with \eqn{\alpha = 0} (FGT0) the measure captures the \emph{extent} of poverty
as a simple headcount - i.e. the proportion of people below the poverty line;
\item with \eqn{\alpha = 1} (FGT1) the measure, also know as the "poverty gap
index", captures the \emph{severity} of poverty as the average percentage distance
between the poverty line and the accessibility of individuals below the
poverty line;
\item with \eqn{\alpha = 2} (FGT2) the measure simultaneously captures the
\emph{extent} and the \emph{severity} of poverty by calculating the number of people
below the poverty line weighted by the size of the accessibility shortfall
relative to the poverty line.
}

FGT values range from 0 to 1. A value of 0 indicates that every individual is
above the poverty line. When every individual is below the poverty line,
however, FGT0 value is 1 and FGT1 and FGT2 values approach 1.
}

\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}
data_dir <- system.file("extdata", package = "accessibility")
travel_matrix <- readRDS(file.path(data_dir, "travel_matrix.rds"))
land_use_data <- readRDS(file.path(data_dir, "land_use_data.rds"))

access <- cumulative_cutoff(
  travel_matrix,
  land_use_data,
  cutoff = 30,
  opportunity = "jobs",
  travel_cost = "travel_time"
)

poverty <- fgt_poverty(
  accessibility_data = access,
  opportunity = "jobs",
  sociodemographic_data = land_use_data,
  population = "population",
  poverty_line = 50000
)
poverty
\dontshow{\}) # examplesIf}
}
\references{
\insertAllCited{}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/doubly_constrained.R
\name{doubly_constrained}
\alias{doubly_constrained}
\title{Doubly constrained accessibility}
\value{
A \code{data.table}/\code{data.frame} with either OD-level flows (\code{detailed_results = TRUE})
or marginals; see Details.
}
\description{
Calculates accessibility using Wilson's doubly-constrained gravity model.
This measure allocates flows between origins and destinations such that
origin totals equal demand and destination totals equal supply. Note: this is
an internal helper function used by \code{\link[=constrained_accessibility]{constrained_accessibility()}} when
\code{constraint = "doubly"}.
}
\examples{
NULL
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constrained_accessibility.R
\name{constrained_accessibility}
\alias{constrained_accessibility}
\title{Constrained accessibility}
\usage{
constrained_accessibility(
  travel_matrix,
  land_use_data,
  travel_cost,
  demand = NULL,
  supply = NULL,
  constraint,
  decay_function,
  active = NULL,
  error_threshold = 0.001,
  improvement_threshold = 1e-06,
  max_iterations = 1000,
  group_by = character(0),
  fill_missing_ids = TRUE,
  detailed_results = FALSE
)
}
\arguments{
\item{travel_matrix}{A data frame. The travel matrix describing the costs
(i.e. travel time, distance, monetary cost, etc.) between the origins and
destinations in the study area. Must contain the columns \code{from_id}, \code{to_id}
and any others specified in \code{travel_cost}.}

\item{land_use_data}{A data frame. The distribution of opportunities within
the study area cells. Must contain the columns \code{id} and any others
specified in \code{opportunity}.}

\item{travel_cost}{A string. The name of the column in \code{travel_matrix} with
the travel cost between origins and destinations. The notion of cost
here is generic over any kind of numeric travel cost, such as distance,
time and money.}

\item{demand}{A string. The name of the column in \code{land_use_data} with the
number of opportunity-demanders at each origin (e.g., people) that will be
considered.}

\item{supply}{A string. The name of the column in \code{land_use_data} with the
number of opportunity supply at each destination (e.g., jobs, school-seats)
that will be considered.}

\item{constraint}{A string. One of \code{"total"}, \code{"singly"}, or \code{"doubly"}. See
Details section for more information.}

\item{decay_function}{A \code{fuction} that converts travel cost into an
impedance factor used to weight opportunities. This function should take a
\code{numeric} vector and also return a \code{numeric} vector as output, with the
same length as the input. For convenience, the package currently includes
the following functions: \code{\link[=decay_binary]{decay_binary()}}, \code{\link[=decay_exponential]{decay_exponential()}},
\code{\link[=decay_power]{decay_power()}} and \code{\link[=decay_stepped]{decay_stepped()}}. See the documentation of each decay
function for more details.}

\item{active}{A logical. When \code{TRUE}, the function calculates active
accessibility (the quantity of opportunities that can be reached from
a given origin). When \code{FALSE}, it calculates passive accessibility (by
how many people each destination can be reached), which is equivalent
to the notion of market potential. This parameter only works for
\code{constraint} types \code{"total"} and \code{"singly"}. Ignored for
\code{constraint = "doubly"}.}

\item{error_threshold}{Numeric. Convergence criterion used only for
calibration in the doubly-constrained case (\code{constraint = "doubly"}).}

\item{improvement_threshold}{Numeric. Convergence criterion for improvement
used only for calibration in the doubly-constrained case
(\code{constraint = "doubly"}).}

\item{max_iterations}{Integer. Maximum iterations used only for calibration
in the doubly-constrained case (\code{constraint = "doubly"}).}

\item{group_by}{A \code{character} vector. When not \code{character(0)} (the default),
indicates the \code{travel_matrix} columns that should be used to group the
accessibility estimates by. For example, if \code{travel_matrix} includes a
departure time column, that specifies the departure time of each entry in
the data frame, passing \code{"departure_time"} to this parameter results in
accessibility estimates grouped by origin and by departure time.}

\item{fill_missing_ids}{A \code{logical}. When calculating grouped accessibility
estimates (i.e. when \code{by_col} is not \code{NULL}), some combinations of groups
and origins may be missing. For example, if a single trip can depart from
origin \code{A} at 7:15am and reach destination \code{B} within 55 minutes, but no
trips departing from \code{A} at 7:30am can be completed at all, this second
combination will not be included in the output. When \code{TRUE} (the default),
the function identifies which combinations would be left out and fills
their respective accessibility values with 0, which incurs in a
performance penalty.}

\item{detailed_results}{Logical. Whether to return detailed OD-level results.}
}
\description{
Calculates accessibility using constraints, as proposed in
\insertCite{soukhov2025family;textual}{accessibility}. Accessibility is
conceptualised as potential spatial interaction. This function covers three
constraint cases. Please see the Details section for more information.
}
\section{Details}{

This function covers the family of constrained accessibility measures
proposed in \insertCite{soukhov2025family;textual}{accessibility}.
\subsection{Total Constrained Accessibility}{

Allocates the system-wide total proportionally based on travel impedance
between origins and destinations. This measure uses the logic of a
total ~(or 'unconstrained' by Wilson's terms)~ constraint.

Use this measure when the total quantity of \strong{supply} OR \strong{demand} in the
system is known and representing accessibility as a proportion of this total
is meaningful.
\subsection{Requirement:}{
\itemize{
\item Either \code{demand} or \code{supply} must be provided (cannot provide both).
}
}

\subsection{Interpretation:}{
\itemize{
\item \code{active = TRUE} (\emph{active accessibility}): Results represent the total
number of  \strong{opportunities} (supply) accessible from each origin based on
region-relative travel impedance. The units are in 'supply' (e.g., jobs,
school seats).
\itemize{
\item If \code{detailed_results = FALSE}, outputs are aggregated and returned by
origin.
\item If \code{detailed_results = TRUE}, OD-level flows are returned. Summing flows
by origin equals the aggregated result.
}
\item \code{active = FALSE} (\emph{passive accessibility}, the notion of market potential):
Results represent the total number of \strong{population} (demand) that can reach
each destination based on region-relative travel impedance. The units are in
'demand' (e.g., population).
\itemize{
\item If \code{detailed_results = FALSE}, outputs are aggregated by destination.
\item If \code{detailed_results = TRUE}, OD-level flows are returned. Summing flows
by destination equals the aggregated result.
}
}
}

\subsection{Use cases:}{
\itemize{
\item Active accessibility (aggregated):
"How many jobs can be reached from origin zone A given its region-relative
travel impedance?"
\item Active accessibility (flow-level):
"How many jobs can be reached by flow A->1 given A->1's region-relative
travel impedance?"
\item Passive accessibility (aggregated):
"How many people can reach destination zone 1 given its region-relative
travel impedance?"
\item Passive accessibility (flow-level):
"How many people are reached by flow 1->A given 1->A's region-relative
travel impedance?"
}
}

}

\subsection{Singly Constrained Accessibility}{

Allocates opportunities at each destination (or population at each origin)
proportionally based on travel impedance and the opposite marginal. This
measure uses the logic of single constraint from
\insertCite{wilson1971family;textual}{accessibility}.

Use this measure when modeling \strong{competition}, where both demand and supply
are conceptualised to influence accessibility but only one side is fixed.
The measure distributes flows so that totals match the constrained side
while weighting by travel impedance and the unconstrained side.
\subsection{Requirements:}{
\itemize{
\item Both \code{demand} and \code{supply} must be provided (the logical for \code{active}
determines if either demand or supply is constrained).
}
}

\subsection{Interpretation:}{
\itemize{
\item \code{active = TRUE} (\emph{active accessibility}): constrains supply. Results
represent the total number of \strong{opportunities} (supply) accessible from each
origin based on region-relative travel impedance and population at the origin.
The units are in 'supply' (e.g., jobs, school seats).
\itemize{
\item If \code{detailed_results = FALSE}, outputs are aggregated and returned by origin.
\item If \code{detailed_results = TRUE}, OD-level flows are returned. Summing flows
by origin equals the aggregated result.
}
\item \code{active = FALSE} (\emph{passive accessibility}, the notion of market potential):
constrains demand. Results represent the total number of
\strong{population} (demand) that can reach each destination based on
region-relative travel impedance and opportunities at the destination. The
units are in 'demand' (e.g., population).
\itemize{
\item If \code{detailed_results = FALSE}, outputs are aggregated by destination.
\item If \code{detailed_results = TRUE}, OD-level flows are returned. Summing flows
by destination equals the aggregated result.
}
}
}

\subsection{Use cases:}{
\itemize{
\item Active accessibility (aggregated):
"How many jobs can be reached from origin zone A given its region-relative
travel impedance and demand?"
\item Active accessibility (flow-level):
"How many jobs can be reached by flow A->1 given A->1's region-relative
travel impedance and demand?"
\item Passive accessibility (aggregated):
"How many people can reach destination zone 1 given its region-relative
travel impedance and supply?"
\item Passive accessibility (flow-level):
"How many people are reached by flow 1->A given 1->A's region-relative
travel impedance and supply?"
}

\strong{NOTE:} the active form of this measure yields equivalent results to the
\code{spatial_availability()} function, through different logic.
}

}

\subsection{Doubly Constrained Accessibility}{

Allocates flows so supply at each destination matches demand at each origin.
It uses Wilson's \emph{doubly-constrained} gravity model \insertCite{wilson1971family;textual}{accessibility}.

The model uses iterative proportional fitting to update balancing factors in
order to calibrate OD flows on both margins (\code{A_i} for origins and \code{B_j} for
destinations) until convergence (i.e. the sum of \code{demand} and \code{supply} match).
This guarantees that flows satisfy both marginals while being weighted by
travel  impedance.
\subsection{Requirements:}{
\itemize{
\item Both \code{demand} and \code{supply} must be provided.
\item Unlike \code{total} and \code{singly}, \code{doubly} requires the sum of demand and supply
to match; otherwise, the model will not converge.
\item \code{active} must be \code{NULL}. Since supply must match demand, their units are
the same and there is no distinction between 'active' and 'passive' notions.
\item Only accepts\code{detailed_results = TRUE}.
}
}

\subsection{Interpretation:}{
\itemize{
\item Results include OD-level flows (\code{flow}) along with balancing factors (\code{A_i},
\code{B_j}) and travel impedance weights. The resulting flows represent the
distribution of demand and supply across all origin-destination pairs. NOTE:
OD flows are in flow units (jointly determined by demand and supply).
}
}

\subsection{Use cases:}{
\itemize{
\item flow-level:
"What is the count of A->1 flows given A->1's region-relative travel impedance,
demand and supply?"
}
}

}
}

\examples{
# Load demo data shipped with the package
data_dir <- system.file("extdata", package = "accessibility")
travel_matrix <- readRDS(file.path(data_dir, "travel_matrix.rds"))
land_use_data <- readRDS(file.path(data_dir, "land_use_data.rds"))

# Total-constrained (active accessibility, aggregated): returns units of
# accessible supply by origin (requires supply)
constrained_accessibility(
  travel_matrix   = travel_matrix,
  land_use_data   = land_use_data,
  travel_cost     = "travel_time",
  constraint      = "total",
  decay_function  = decay_exponential(0.1),
  demand          = NULL,
  supply          = "jobs",
  active          = TRUE,
  detailed_results = FALSE
)

# Total-constrained (passive accessibility, aggregated): returns units of
# accessible demand by destination  (requires demand)
constrained_accessibility(
  travel_matrix  = travel_matrix,
  land_use_data  = land_use_data,
  travel_cost    = "travel_time",
  constraint     = "total",
  decay_function = decay_exponential(0.1),
  demand         = "population",
  supply         = NULL,
  active         = FALSE,
  detailed_results = FALSE
)

# Singly-constrained (active accessibility, aggregated): returns units of
# accessible supply by origin (requires supply and demand)
constrained_accessibility(
  travel_matrix   = travel_matrix,
  land_use_data   = land_use_data,
  travel_cost     = "travel_time",
  constraint      = "singly",
  decay_function  = decay_exponential(0.1),
  demand          = "population",
  supply          = "jobs",
  active          = TRUE,
  detailed_results = FALSE
)

# Doubly-constrained: returns units of flow (requires both demand and supply
# (totals that match) and `detailed_results = TRUE`)

# Using a small toy dataset with matching totals.
tm_small <- data.table::data.table(
  expand.grid(from_id = c("1","2","3"), to_id = c("1","2","3"))
)
tm_small[, travel_time := c(10, 30, 15, 30, 10, 25, 15, 25, 10)]
lu_small <- data.table::data.table(
  id         = c("1","2","3"),
  population = c(4, 10, 6),   # sum = 20
  jobs       = c(7,  5,  8)   # sum = 20
)

constrained_accessibility(
  travel_matrix   = tm_small,
  land_use_data   = lu_small,
  travel_cost     = "travel_time",
  constraint      = "doubly",
  decay_function  = decay_exponential(0.1),
  demand          = "population",
  supply          = "jobs",
  detailed_results = TRUE
)

}
\references{
\insertAllCited{}
}
\seealso{
Other Constrained accessibility: 
\code{\link{spatial_availability}()}
}
\concept{Constrained accessibility}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/palma_ratio.R
\name{palma_ratio}
\alias{palma_ratio}
\title{Palma Ratio}
\usage{
palma_ratio(
  accessibility_data,
  sociodemographic_data,
  opportunity,
  population,
  income,
  group_by = character(0)
)
}
\arguments{
\item{accessibility_data}{A data frame. The accessibility levels whose
inequality should be calculated. Must contain the columns \code{id} and any
others specified in \code{opportunity}.}

\item{sociodemographic_data}{A data frame. The distribution of
sociodemographic characteristics of the population in the study area cells.
Must contain the columns \code{id} and any others specified in \code{population} and
\code{income}.}

\item{opportunity}{A string. The name of the column in \code{accessibility_data}
with the accessibility levels to be considerend when calculating inequality
levels.}

\item{population}{A string. The name of the column in \code{sociodemographic_data}
with the number of people in each cell. Used to weigh accessibility levels
when calculating inequality.}

\item{income}{A string. The name of the column in \code{sociodemographic_data}
with the income variable that should be used to classify the population in
socioeconomic groups. Please note that this variable should describe income
per capita (e.g. mean income per capita, household income per capita, etc),
instead of the total amount of income in each cell.}

\item{group_by}{A \code{character} vector. When not \code{character(0)} (the default),
indicates the \code{accessibility_data} columns that should be used to group the
inequality estimates by. For example, if \code{accessibility_data} includes a
\code{scenario} column that identifies distinct scenarios that each
accessibility estimates refer to (e.g. before and after a transport policy
intervention), passing \code{"scenario"} to this parameter results in inequality
estimates grouped by scenario.}
}
\value{
A data frame containing the inequality estimates for the study area.
}
\description{
Calculates the Palma Ratio of a given accessibility distribution. Originally
defined as the income share of the richest 10\% of a population divided by the
income share of the poorest 40\%, this measure has been adapted in transport
planning as the average accessibility of the richest 10\% divided by the
average accessibility of the poorest 40\%.
}
\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}
data_dir <- system.file("extdata", package = "accessibility")
travel_matrix <- readRDS(file.path(data_dir, "travel_matrix.rds"))
land_use_data <- readRDS(file.path(data_dir, "land_use_data.rds"))

access <- cumulative_cutoff(
  travel_matrix,
  land_use_data,
  cutoff = 30,
  opportunity = "jobs",
  travel_cost = "travel_time"
)

palma <- palma_ratio(
  access,
  sociodemographic_data = land_use_data,
  opportunity = "jobs",
  population = "population",
  income = "income_per_capita"
)
palma
\dontshow{\}) # examplesIf}
}
\seealso{
Other inequality: 
\code{\link{concentration_index}()},
\code{\link{gini_index}()},
\code{\link{theil_t}()}
}
\concept{inequality}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessibility-package.R
\docType{package}
\name{accessibility}
\alias{accessibility}
\alias{accessibility-package}
\title{accessibility: Transport accessibility measures}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

A set of fast and convenient functions to help conducting accessibility analyses. Given a pre-computed travel cost matrix and a land use dataset (containing for example the location of jobs, healthcare and population), the package allows one to calculate accessibility levels, and accessibility poverty and inequality. The package covers the majority of the most commonly used accessibility measures (such as cumulative opportunities, gravity-based and floating catchment areas methods), some cutting edge measures proposed in the literature (e.g. balancing cost and constrained accessibility) as well as the most frequently used inequality and poverty metrics (such as the Palma ratio, the concentration and Theil indices and the FGT family of measures).
}
\section{Usage}{

Please check the vignettes for more on the package usage:
\itemize{
\item Introduction to accessibility: calculating accessibility measures. Run
\code{vignette("accessibility")} or check it on the \href{https://ipeagit.github.io/accessibility/articles/accessibility.html}{website}.
\item Decay functions. Run \code{vignette("decay_functions", package = "accessibility")} or check it on the \href{https://ipeagit.github.io/accessibility/articles/decay_functions.html}{website}.
\item Calculating accessibility inequality and poverty. Run
\code{vignette("inequality_and_poverty", package = "accessibility")} or check it
on the
\href{https://ipeagit.github.io/accessibility/articles/inequality_and_poverty.html}{website}.
}
}

\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/ipeaGIT/accessibility}
  \item \url{https://ipeagit.github.io/accessibility/}
  \item Report bugs at \url{https://github.com/ipeaGIT/accessibility/issues}
}

}
\author{
\strong{Maintainer}: Daniel Herszenhut \email{dhersz@gmail.com} (\href{https://orcid.org/0000-0001-8066-1105}{ORCID})

Authors:
\itemize{
  \item Rafael H. M. Pereira \email{rafa.pereira.br@gmail.com} (\href{https://orcid.org/0000-0003-2125-7465}{ORCID})
  \item Anastasia Soukhov
}

Other contributors:
\itemize{
  \item Christopher Higgins [contributor]
  \item Joey Reid [contributor]
  \item Ipea - Institute for Applied Economic Research [copyright holder, funder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/floating_catchment_area.R
\name{floating_catchment_area}
\alias{floating_catchment_area}
\title{Floating catchment area accessibility}
\usage{
floating_catchment_area(
  travel_matrix,
  land_use_data,
  opportunity,
  travel_cost,
  demand,
  method,
  decay_function,
  group_by = character(0),
  fill_missing_ids = TRUE
)
}
\arguments{
\item{travel_matrix}{A data frame. The travel matrix describing the costs
(i.e. travel time, distance, monetary cost, etc.) between the origins and
destinations in the study area. Must contain the columns \code{from_id}, \code{to_id}
and any others specified in \code{travel_cost}.}

\item{land_use_data}{A data frame. The distribution of opportunities within
the study area cells. Must contain the columns \code{id} and any others
specified in \code{opportunity}.}

\item{opportunity}{A string. The name of the column in \code{land_use_data}
with the number of opportunities/resources/services to be considered when
calculating accessibility levels.}

\item{travel_cost}{A string. The name of the column in \code{travel_matrix} with
the travel cost between origins and destinations. The notion of cost
here is generic over any kind of numeric travel cost, such as distance,
time and money.}

\item{demand}{A string. The name of the column in \code{land_use_data} with the
number of opportunity-demanders at each origin (e.g., people) that will be
considered.}

\item{method}{A string. Which floating catchment area measure to use.
Current available options are \code{"2sfca"} and \code{"bfca"}. More info in the
details.}

\item{decay_function}{A \code{fuction} that converts travel cost into an
impedance factor used to weight opportunities. This function should take a
\code{numeric} vector and also return a \code{numeric} vector as output, with the
same length as the input. For convenience, the package currently includes
the following functions: \code{\link[=decay_binary]{decay_binary()}}, \code{\link[=decay_exponential]{decay_exponential()}},
\code{\link[=decay_power]{decay_power()}} and \code{\link[=decay_stepped]{decay_stepped()}}. See the documentation of each decay
function for more details.}

\item{group_by}{A \code{character} vector. When not \code{character(0)} (the default),
indicates the \code{travel_matrix} columns that should be used to group the
accessibility estimates by. For example, if \code{travel_matrix} includes a
departure time column, that specifies the departure time of each entry in
the data frame, passing \code{"departure_time"} to this parameter results in
accessibility estimates grouped by origin and by departure time.}

\item{fill_missing_ids}{A \code{logical}. When calculating grouped accessibility
estimates (i.e. when \code{by_col} is not \code{NULL}), some combinations of groups
and origins may be missing. For example, if a single trip can depart from
origin \code{A} at 7:15am and reach destination \code{B} within 55 minutes, but no
trips departing from \code{A} at 7:30am can be completed at all, this second
combination will not be included in the output. When \code{TRUE} (the default),
the function identifies which combinations would be left out and fills
their respective accessibility values with 0, which incurs in a
performance penalty.}
}
\value{
A data frame containing the accessibility estimates for each
origin/destination (depending if \code{active} is \code{TRUE} or \code{FALSE}) in the
travel matrix.
}
\description{
Calculates accessibility accounting for the competition of resources using a
measure from the floating catchment area (FCA) family. Please see the details
for the available FCA measures.

This function is generic over any kind of numeric travel cost,
such as distance, time and money.
}
\section{Details}{

The package currently includes two built-in FCA measures:
\itemize{
\item 2SFCA - the 2-Step Floating Catchment Area measure was the first
accessibility metric in the FCA family. It was originally proposed by
\insertCite{luo2003measures;textual}{accessibility}.
\item BFCA - the Balanced Floating Catchment Area measure calculates
accessibility accounting for competition effects while simultaneously
correcting for issues of inflation of demand and service levels that are
present in other FCA measures. It was originally proposed by
\insertCite{paez2019demand;textual}{accessibility} and named in
\insertCite{pereira2021geographic;textual}{accessibility}.
}
}

\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}
data_dir <- system.file("extdata", package = "accessibility")
travel_matrix <- readRDS(file.path(data_dir, "travel_matrix.rds"))
land_use_data <- readRDS(file.path(data_dir, "land_use_data.rds"))

# 2SFCA with a step decay function
df <- floating_catchment_area(
  travel_matrix,
  land_use_data,
  method = "2sfca",
  decay_function = decay_binary(cutoff = 50),
  opportunity = "jobs",
  travel_cost = "travel_time",
  demand = "population"
)
head(df)


# BFCA with an exponential decay function
df <- floating_catchment_area(
  travel_matrix,
  land_use_data,
  method = "bfca",
  decay_function = decay_exponential(decay_value = 0.5),
  opportunity = "jobs",
  travel_cost = "travel_time",
  demand = "population"
)
head(df)
\dontshow{\}) # examplesIf}
}
\references{
\insertAllCited{}
}
\concept{Floating catchment area}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/total_constrained.R
\name{total_constrained}
\alias{total_constrained}
\title{Total constrained accessibility}
\value{
A \code{data.table}/\code{data.frame} with results (structure mirrors the wrapper).
}
\description{
Allocates total opportunities in the region proportionally based on travel
impedance. Uses the logic of a total (or unconstrained by Wilson's terms)
constraint. Returns values in units of 'supply' (i.e., opportunities) if
\code{active = TRUE} and returns values in units of demand' (i.e., population)
if \code{active = FALSE}.
}
\details{
Internal helper used by \code{\link[=constrained_accessibility]{constrained_accessibility()}} when \code{constraint = "total"}.
}
\examples{
NULL

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/decay_exponential.R
\name{decay_exponential}
\alias{decay_exponential}
\title{Negative exponential decay function}
\usage{
decay_exponential(decay_value)
}
\arguments{
\item{decay_value}{A \code{numeric} vector. The calibration parameters that, when
multiplied by the travel cost, are used as the exponent of \code{e} in the
negative exponential function.}
}
\value{
A \code{function} that takes a generic travel cost vector (\code{numeric}) as
an input and returns a \code{list} of weight vectors (a list of \code{numeric}
vectors, named after the arguments passed to the decay function).
}
\description{
Returns a negative exponential weighting function to be used inside
accessibility calculating functions.

This function is generic over any kind of numeric travel cost,
such as distance, time and money.
}
\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}
weighting_function <- decay_exponential(decay_value = 0.1)

weighting_function(c(20, 30))

weighting_function <- decay_exponential(decay_value = c(0.1, 0.2))

weighting_function(c(20, 30))
\dontshow{\}) # examplesIf}
}
\seealso{
Other decay functions: 
\code{\link{decay_binary}()},
\code{\link{decay_linear}()},
\code{\link{decay_logistic}()},
\code{\link{decay_power}()},
\code{\link{decay_stepped}()}
}
\concept{decay functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gini_index.R
\name{gini_index}
\alias{gini_index}
\title{Gini Index}
\usage{
gini_index(
  accessibility_data,
  sociodemographic_data,
  opportunity,
  population,
  group_by = character(0)
)
}
\arguments{
\item{accessibility_data}{A data frame. The accessibility levels whose
inequality should be calculated. Must contain the columns \code{id} and any
others specified in \code{opportunity}.}

\item{sociodemographic_data}{A data frame. The distribution of
sociodemographic characteristics of the population in the study area cells.
Must contain the columns \code{id} and any others specified in \code{population}.}

\item{opportunity}{A string. The name of the column in \code{accessibility_data}
with the accessibility levels to be considerend when calculating inequality
levels.}

\item{population}{A string. The name of the column in \code{sociodemographic_data}
with the number of people in each cell. Used to weigh accessibility levels
when calculating inequality.}

\item{group_by}{A \code{character} vector. When not \code{character(0)} (the default),
indicates the \code{accessibility_data} columns that should be used to group the
inequality estimates by. For example, if \code{accessibility_data} includes a
\code{scenario} column that identifies distinct scenarios that each
accessibility estimates refer to (e.g. before and after a transport policy
intervention), passing \code{"scenario"} to this parameter results in inequality
estimates grouped by scenario.}
}
\value{
A data frame containing the inequality estimates for the study area.
}
\description{
Calculates the Gini Index of a given accessibility distribution.
}
\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}
data_dir <- system.file("extdata", package = "accessibility")
travel_matrix <- readRDS(file.path(data_dir, "travel_matrix.rds"))
land_use_data <- readRDS(file.path(data_dir, "land_use_data.rds"))

access <- cumulative_cutoff(
  travel_matrix,
  land_use_data,
  cutoff = 30,
  opportunity = "jobs",
  travel_cost = "travel_time"
)

gini <- gini_index(
  access,
  sociodemographic_data = land_use_data,
  opportunity = "jobs",
  population = "population"
)
gini
\dontshow{\}) # examplesIf}
}
\seealso{
Other inequality: 
\code{\link{concentration_index}()},
\code{\link{palma_ratio}()},
\code{\link{theil_t}()}
}
\concept{inequality}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/decay_binary.R
\name{decay_binary}
\alias{decay_binary}
\title{Binary (a.k.a. step) decay function}
\usage{
decay_binary(cutoff)
}
\arguments{
\item{cutoff}{A \code{numeric} vector. The numbers indicating the travel cost
cutoffs.}
}
\value{
A \code{function} that takes a generic travel cost vector (\code{numeric}) as
an input and returns a \code{list} of weight vectors (a list of \code{numeric}
vectors, named after the arguments passed to the decay function).
}
\description{
Returns a binary weighting function (frequently used to calculate cumulative
opportunities measures) to be used inside accessibility calculating
functions.

This function is generic over any kind of numeric travel cost,
such as distance, time and money.
}
\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}
weighting_function <- decay_binary(cutoff = 30)

weighting_function(c(20, 35))

weighting_function <- decay_binary(cutoff = c(30, 45))

weighting_function(c(20, 35))
\dontshow{\}) # examplesIf}
}
\seealso{
Other decay functions: 
\code{\link{decay_exponential}()},
\code{\link{decay_linear}()},
\code{\link{decay_logistic}()},
\code{\link{decay_power}()},
\code{\link{decay_stepped}()}
}
\concept{decay functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cumulative_interval.R
\name{cumulative_interval}
\alias{cumulative_interval}
\title{Cumulative access based on maximum travel time interval}
\usage{
cumulative_interval(
  travel_matrix,
  land_use_data,
  opportunity,
  travel_cost,
  interval,
  interval_increment = 1,
  summary_function = stats::median,
  group_by = character(0),
  active = TRUE
)
}
\arguments{
\item{travel_matrix}{A data frame. The travel matrix describing the costs
(i.e. travel time, distance, monetary cost, etc.) between the origins and
destinations in the study area. Must contain the columns \code{from_id}, \code{to_id}
and any others specified in \code{travel_cost}.}

\item{land_use_data}{A data frame. The distribution of opportunities within
the study area cells. Must contain the columns \code{id} and any others
specified in \code{opportunity}.}

\item{opportunity}{A string. The name of the column in \code{land_use_data}
with the number of opportunities/resources/services to be considered when
calculating accessibility levels.}

\item{travel_cost}{A string. The name of the column in \code{travel_matrix} with
the travel cost between origins and destinations. The notion of cost
here is generic over any kind of numeric travel cost, such as distance,
time and money.}

\item{interval}{A \code{numeric} vector of length 2. Indicates the start and end
points of the interval of travel cost thresholds to be used. The first
entry must be lower than the second.}

\item{interval_increment}{A \code{numeric}. How many travel cost units separate
the cutoffs used to calculate the accessibility estimates which will be
used to calculate the summary estimate within the specified interval.
Should be thought as the resolution of the distribution of travel costs
within the interval. Defaults to 1.}

\item{summary_function}{A function. This function is used to summarize a
distribution of accessibility estimates within a travel cost interval as a
single value. Can be any function that takes an arbitrary number of
numeric values as as input and returns a single number as output. Defaults
to \code{\link[stats:median]{stats::median()}}.}

\item{group_by}{A \code{character} vector. When not \code{character(0)} (the default),
indicates the \code{travel_matrix} columns that should be used to group the
accessibility estimates by. For example, if \code{travel_matrix} includes a
departure time column, that specifies the departure time of each entry in
the data frame, passing \code{"departure_time"} to this parameter results in
accessibility estimates grouped by origin and by departure time.}

\item{active}{A logical. When \code{TRUE}, the function calculates active
accessibility (the quantity of opportunities that can be reached from
a given origin). when \code{FALSE}, it calculates passive accessibility (by
how many people each destination can be reached), which is equivalent
to the notion of market potential.}
}
\value{
A data frame containing the accessibility estimates for each
origin/destination (depending if \code{active} is \code{TRUE} or \code{FALSE}) in the
travel matrix.
}
\description{
Calculates the average or median number of opportunities that can be reached
considering multiple maximum travel cost thresholds within a given travel
cost interval specified by the user. The time interval cumulative
accessibility measures was originally proposed by
\insertCite{tomasiello2023time;textual}{accessibility}.

This function is generic over any kind of numeric travel cost,
such as distance, time and money.
}
\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}
data_dir <- system.file("extdata", package = "accessibility")
travel_matrix <- readRDS(file.path(data_dir, "travel_matrix.rds"))
land_use_data <- readRDS(file.path(data_dir, "land_use_data.rds"))

df <- cumulative_interval(
  travel_matrix = travel_matrix,
  land_use_data = land_use_data,
  interval = c(20, 30),
  opportunity = "schools",
  travel_cost = "travel_time"
)
head(df)

df <- cumulative_interval(
  travel_matrix = travel_matrix,
  land_use_data = land_use_data,
  interval = c(40, 80),
  opportunity = "jobs",
  travel_cost = "travel_time"
)
head(df)
\dontshow{\}) # examplesIf}
}
\references{
\insertAllCited{}
}
\seealso{
Other cumulative access: 
\code{\link{cumulative_cutoff}()}
}
\concept{cumulative access}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/theil_t.R
\name{theil_t}
\alias{theil_t}
\title{Theil T Index}
\usage{
theil_t(
  accessibility_data,
  sociodemographic_data,
  opportunity,
  population,
  socioeconomic_groups = NULL,
  group_by = character(0)
)
}
\arguments{
\item{accessibility_data}{A data frame. The accessibility levels whose
inequality should be calculated. Must contain the columns \code{id} and any
others specified in \code{opportunity}.}

\item{sociodemographic_data}{A data frame. The distribution of
sociodemographic characteristics of the population in the study area cells.
Must contain the columns \code{id} and any others specified in \code{population} and
\code{socioeconomic_groups}.}

\item{opportunity}{A string. The name of the column in \code{accessibility_data}
with the accessibility levels to be considerend when calculating inequality
levels.}

\item{population}{A string. The name of the column in \code{sociodemographic_data}
with the number of people in each cell. Used to weigh accessibility levels
when calculating inequality.}

\item{socioeconomic_groups}{A string. The name of the column in
\code{sociodemographic_data} whose values identify the socioeconomic groups that
should be used to calculate the between- and within-groups inequality
levels. If \code{NULL} (the default), between- and within-groups components are
not calculated and only the total aggregate inequality is returned.}

\item{group_by}{A \code{character} vector. When not \code{character(0)} (the default),
indicates the \code{accessibility_data} columns that should be used to group the
inequality estimates by. For example, if \code{accessibility_data} includes a
\code{scenario} column that identifies distinct scenarios that each
accessibility estimates refer to (e.g. before and after a transport policy
intervention), passing \code{"scenario"} to this parameter results in inequality
estimates grouped by scenario.}
}
\value{
If \code{socioeconomic_groups} is \code{NULL}, a data frame containing the
total Theil T estimates for the study area. If not, a list containing three
dataframes: one summarizing the total inequality and the between- and
within-groups components, one listing the contribution of each group to the
between-groups component and another listing the contribution of each group
to the within-groups component.
}
\description{
Calculates the Theil T Index of a given accessibility distribution. Values
range from 0 (when all individuals have exactly the same accessibility
levels) to the natural log of \emph{n}, in which \emph{n} is the number of individuals
in the accessibility dataset. If the individuals can be classified into
mutually exclusive and completely exhaustive groups, the index can be
decomposed into a between-groups inequaliy component and a within-groups
component.
}
\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}
data_dir <- system.file("extdata", package = "accessibility")
travel_matrix <- readRDS(file.path(data_dir, "travel_matrix.rds"))
land_use_data <- readRDS(file.path(data_dir, "land_use_data.rds"))

access <- cumulative_cutoff(
  travel_matrix,
  land_use_data,
  cutoff = 30,
  opportunity = "jobs",
  travel_cost = "travel_time"
)

ti <- theil_t(
  access,
  sociodemographic_data = land_use_data,
  opportunity = "jobs",
  population = "population"
)
ti

# to calculate inequality between and within income deciles, we pass
# "income_decile" to socioeconomic_groups.
# some cells, however, are classified as in the decile NA because their
# income per capita is NaN, as they don't have any population. we filter
# these cells from our accessibility data, otherwise the output would include
# NA values (note that subsetting the data like this doesn't affect the
# assumption that groups are completely exhaustive, because cells with NA
# income decile don't have any population)

na_decile_ids <- land_use_data[is.na(land_use_data$income_decile), ]$id
access <- access[! access$id \%in\% na_decile_ids, ]
sociodem_data <- land_use_data[! land_use_data$id \%in\% na_decile_ids, ]

ti <- theil_t(
  access,
  sociodemographic_data = sociodem_data,
  opportunity = "jobs",
  population = "population",
  socioeconomic_groups = "income_decile"
)
ti
\dontshow{\}) # examplesIf}
}
\seealso{
Other inequality: 
\code{\link{concentration_index}()},
\code{\link{gini_index}()},
\code{\link{palma_ratio}()}
}
\concept{inequality}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial_availability.R
\name{spatial_availability}
\alias{spatial_availability}
\title{Spatial availability}
\usage{
spatial_availability(
  travel_matrix,
  land_use_data,
  opportunity,
  travel_cost,
  demand,
  decay_function,
  alpha = 1,
  group_by = character(0),
  fill_missing_ids = TRUE,
  detailed_results = FALSE
)
}
\arguments{
\item{travel_matrix}{A data frame. The travel matrix describing the costs
(i.e. travel time, distance, monetary cost, etc.) between the origins and
destinations in the study area. Must contain the columns \code{from_id}, \code{to_id}
and any others specified in \code{travel_cost}.}

\item{land_use_data}{A data frame. The distribution of opportunities within
the study area cells. Must contain the columns \code{id} and any others
specified in \code{opportunity}.}

\item{opportunity}{A string. The name of the column in \code{land_use_data}
with the number of opportunities/resources/services to be considered when
calculating accessibility levels.}

\item{travel_cost}{A string. The name of the column in \code{travel_matrix} with
the travel cost between origins and destinations. The notion of cost
here is generic over any kind of numeric travel cost, such as distance,
time and money.}

\item{demand}{A string. The name of the column in \code{land_use_data} with the
number of opportunity-demanders at each origin (e.g., people) that will be
considered.}

\item{decay_function}{A \code{fuction} that converts travel cost into an
impedance factor used to weight opportunities. This function should take a
\code{numeric} vector and also return a \code{numeric} vector as output, with the
same length as the input. For convenience, the package currently includes
the following functions: \code{\link[=decay_binary]{decay_binary()}}, \code{\link[=decay_exponential]{decay_exponential()}},
\code{\link[=decay_power]{decay_power()}} and \code{\link[=decay_stepped]{decay_stepped()}}. See the documentation of each decay
function for more details.}

\item{alpha}{A \code{numeric}. A parameter used to modulate the effect of demand
by population. When less than 1, opportunities are allocated more rapidly
to smaller centers relative to larger ones; values higher than 1 achieve
the opposite effect.}

\item{group_by}{A \code{character} vector. When not \code{character(0)} (the default),
indicates the \code{travel_matrix} columns that should be used to group the
accessibility estimates by. For example, if \code{travel_matrix} includes a
departure time column, that specifies the departure time of each entry in
the data frame, passing \code{"departure_time"} to this parameter results in
accessibility estimates grouped by origin and by departure time.}

\item{fill_missing_ids}{A \code{logical}. When calculating grouped accessibility
estimates (i.e. when \code{by_col} is not \code{NULL}), some combinations of groups
and origins may be missing. For example, if a single trip can depart from
origin \code{A} at 7:15am and reach destination \code{B} within 55 minutes, but no
trips departing from \code{A} at 7:30am can be completed at all, this second
combination will not be included in the output. When \code{TRUE} (the default),
the function identifies which combinations would be left out and fills
their respective accessibility values with 0, which incurs in a
performance penalty.}

\item{detailed_results}{A \code{logical}. Whether to return spatial availability
results aggregated by origin-destination pair (\code{TRUE}) or by origin
(\code{FALSE}, the default). When \code{TRUE}, the output also includes the demand,
impedance and combined balancing factors used to calculate spatial
availability. Please note that the argument \code{fill_missing_ids}
does not affect the output when \code{detailed_results} is \code{TRUE}.}
}
\value{
A data frame containing the accessibility estimates for each
origin/destination (depending if \code{active} is \code{TRUE} or \code{FALSE}) in the
travel matrix.
}
\description{
Calculates spatial availability, an accessibility measured proposed by
\insertCite{soukhov2023introducing;textual}{accessibility} that takes into
account competition effects. The accessibility levels that result from using
this measure are proportional both to the demand in each origin and to the
travel cost it takes to reach the destinations. The spatial availability is
a particular case of \code{constrained_accessibility(constraint = "singly")}.

This function is generic over any kind of numeric travel cost,
such as distance, time and money.
}
\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}
# the example below is based on Soukhov et al. (2023) paper

travel_matrix <- data.table::data.table(
  from_id = rep(c("A", "B", "C"), each = 3),
  to_id = as.character(rep(1:3, 3)),
  travel_time = c(15, 30, 100, 30, 15, 100, 100, 100, 15)
)
land_use_data <- data.table::data.table(
  id = c("A", "B", "C", "1", "2", "3"),
  population = c(50000, 150000, 10000, 0, 0, 0),
  jobs = c(0, 0, 0, 100000, 100000, 10000)
)

df <- spatial_availability(
  travel_matrix,
  land_use_data,
  opportunity = "jobs",
  travel_cost = "travel_time",
  demand = "population",
  decay_function = decay_exponential(decay_value = 0.1)
)
df

detailed_df <- spatial_availability(
  travel_matrix,
  land_use_data,
  opportunity = "jobs",
  travel_cost = "travel_time",
  demand = "population",
  decay_function = decay_exponential(decay_value = 0.1),
  detailed_results = TRUE
)
detailed_df
\dontshow{\}) # examplesIf}
}
\references{
\insertAllCited{}
}
\seealso{
Other Constrained accessibility: 
\code{\link{constrained_accessibility}()}
}
\concept{Constrained accessibility}
